/*

  Copyright (C) 2001 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.associations;

import jac.core.AspectComponent;
import jac.core.Wrappee;
import jac.core.rtti.*;
import jac.util.Log;

import java.util.Collection;
import java.util.Iterator;

/**
 * <p>This aspect manages associations. If two relations are declared to
 * constitue an association, modifications on one relation are
 * refelected in the other.</p>
 *
 * <p>For instance, if you have a Customer class and an Order class :</p>
 *
 * <pre>
 *    ,----------. 1    n ,-------.
 *    | Customer |--------| Order |
 *    `----------'        `-------'
 * </pre>
 * 
 * <p>You can declare an association made of Cutomer.orders and
 * Order.customer, so that setting the cutomer of an order will
 * automatically add this order in the customer's list of orders. And
 * vice-versa.</p>
 *
 * <p>Associations are declared within the RTTI aspect.</p>
 *
 * @see jac.core.rtti.RttiAC.declareAssociation(ClassItem,String,ClassItem,String)
 */
public class AssociationsAC extends AspectComponent 
{
   static RelationWrapper relationWrapper = new RelationWrapper();
   
   public void whenUsingNewInstance() 
   {
      Wrappee wrappee = wrappee();
      Log.trace("associations","whenUsingNewInstance"+wrappee.getClass());
      ClassItem cli = ClassRepository.get().getClass( wrappee.getClass() );
      Collection methods = cli.getMethods();
      Iterator it = methods.iterator();
      while ( it.hasNext() ) {
         MethodItem[] method_items = (MethodItem[]) it.next();
         if (method_items!=null) {
            for (int i=0; i<method_items.length; i++) {
               // wrap references' setters
               MethodItem method = method_items[i];
               FieldItem[] fields = method.getWrittenFields();
               if ( fields != null) {
                  for (int j=0; j<fields.length; j++) {
                     if (fields[j].isReference() && 
                         fields[j].getAttribute("rtti.oppositeRelation") != null) {
                        wrappee.wrap(relationWrapper,"updateOppositeRelations",
                                     method.getName());
                        break;
                     }
                  }
               }
               
               // wrap adders
               CollectionItem[] collections = method.getAddedCollections();
               if (collections!=null) {
                  for (int j=0; j<collections.length; j++) {
                     if (collections[j].getAttribute("rtti.oppositeRelation")!=null)
                        wrappee.wrap(relationWrapper,"updateOppositeRelations",
                                     method.getName());
                  }
               }
               // wrap removers
               collections = method.getRemovedCollections();
               if (collections!=null) {
                  for (int j=0; j<collections.length; j++) {
                     if (collections[j].getAttribute("rtti.oppositeRelation")!=null)
                        wrappee.wrap(relationWrapper,"updateOppositeRelations",
                                     method.getName());
                  }
               }
            }
         }
      }      
   }
}
