/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.distribution;

import jac.core.*;
import jac.core.dist.*;
import jac.core.rtti.ClassRepository;
import jac.core.rtti.ClassItem;
import jac.core.rtti.FieldItem;
import jac.util.*;
import jac.aspects.naming.*;

import java.util.*;

/**
 * This aspect component implements a generic deployment aspect.
 * 
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 *
 * @see DeploymentConf
 * @see DeploymentRule */

public class DeploymentAC extends AspectComponent implements DeploymentConf {

   transient Hashtable treated = new Hashtable();

    /**
     * Force the treatment of the objects that are being
     * serialized. This allows the deployment aspect to be sure that
     * an object that is passed as a parameter of a remote invocation
     * is already deployed when arriving on the new site. If none of
     * the deployment rule can be applied on the serialized object,
     * then the object will not be a forwarder but a copy of the
     * original object.
     *
     * @param finalObject the structure to store the serialized infos
     */

   public void whenSerialized ( SerializedJacObject finalObject ) {
	
      Log.trace("serialization","DeploymentAC.whenSerialize");
      NameRepository nr = (NameRepository) NameRepository.get();
      Wrappee orgObject = (Wrappee)attr( "orgObject" );
      String name = nr.getName( orgObject );
      if ( name != null ) {
         Log.trace("serialization","DeploymentAC forces the treatment when "+
                   orgObject+" serialized");
         forceTreatment( orgObject );
      }
      //if ( ! isMatchingARule( orgObject ) ) {
      Log.trace("serialization",name + " is not matching any rule");
      finalObject.disableForwarding();
      Object[] state = orgObject.getState();
      String[] fieldsName = (String[]) state[0];
      Object[] fieldsValue = (Object[]) state[1];
      
      ClassItem cl = ClassRepository.get().getClass(orgObject.getClass());
      for ( int i = 0; i < fieldsName.length; i++ ) {
         Log.trace("serialization"," serializing field "+fieldsName[i]+
                   " - "+fieldsValue[i]); 
         FieldItem fi = cl.getField(fieldsName[i]);
         if( fi != null && fi.getAttribute("deployment.transient") == null ) {
            finalObject.addField( fieldsName[i], fieldsValue[i] );
         }
      }
         
      /*} else {
        DeploymentRule r = getMatchingRule( orgObject );
        if( r != null && r.getType().equals( "dynamic client-server" ) ) {
        finalObject.disableForwarding();
        finalObject.setACInfos( 
        ACManager.get().getName( this ),
        RemoteRef.create( NameRepository.get().getName( orgObject ), orgObject ) );
        }
        }*/
   }

   /**
    * Fill the field values when the forwarding is disabled for this
    * object..
    *
    * @param orgObject the JAC object that is being deserialized. */
   
   public void whenDeserialized ( SerializedJacObject orgObject ) {
      Log.trace("serialization","DeploymentAC.whenDeserialize");
      if ( ! orgObject.isForwarder() ) {
         Log.trace("serialization","not a forwarder");
         Wrappee finalObject = (Wrappee)attr("finalObject");
         RemoteRef server;
         if( (server = (RemoteRef) orgObject.getACInfos( ACManager.get().getName( this ) ) ) != null ) {
            finalObject.wrapAll( new StubWrapper( server ), "invoke" );
         } else {
            Iterator it = orgObject.getFields().keySet().iterator();
            while ( it.hasNext() ) {
               String fn = (String) it.next();
               Log.trace("serialization","setting field "+fn+" <- "+
                         orgObject.getField( fn ));
               finalObject.setFieldValue( fn, orgObject.getField( fn ) );
            }
         }
      }
   }   

   /**
    * Actually deploys an object with a target host expression. */

   public void deployTo( String hostExpr, boolean state ) {
      if (treated == null) treated = new Hashtable();
      if (treated.get( wrappee() ) != null) return;
      treated.put(wrappee(),"");
      System.out.println( "DEPLOYED UPCALLED WITH "+hostExpr+" wrappee="+wrappee()+", topology="+Topology.get());
      Topology topology = Topology.getPartialTopology( hostExpr );
      Deployment dep = new Deployment( topology );
      if( state ) {
         dep.replicate( wrappee() );
      } else {
         dep.replicateStruct( wrappee() );
      }
   }

   public void deploy( String deploymentHost,
                       String nameRegexp, String containerName ) {

      pointcut( nameRegexp, ".*", "deployTo", 
                new Object[] { containerName, new Boolean(true) }, 
                deploymentHost );

      //      addDeploymentRule( new DeploymentRule( 
      //   "deployment", nameRegexp, containerName, true ) ); 
      
   }

   public void replicate ( String deploymentHost, String nameRegexp,
                           String contRegexp ) {

      pointcut( nameRegexp, ".*", "deployTo", 
                new Object[] { contRegexp, new Boolean(true) }, 
                deploymentHost );

      //addDeploymentRule( new DeploymentRule( 
      //   "replicate", nameRegexp, contRegexp, true ) );
   }

   public void createTypedStubsFor( String name, String serverHost, 
                                    String hosts, 
                                    String stubType ) {
      pointcut( name, ".*", "deployTo", 
                new Object[] { hosts, new Boolean(false) }, serverHost );
      //addDeploymentRule( new DeploymentRule( 
      //   "replicate", name, hosts, true ) );
      try {
         pointcut( name, ".*", ".*", 
                   (Wrapper)Class.forName(stubType)
                   .getConstructor(new Class[]{String.class})
                   .newInstance(new Object[]{serverHost}), 
                   "invoke", hosts+" && !"+serverHost );
      } catch ( Exception e ) {
         Log.error("pointcut creation failed");
         e.printStackTrace();
      }
   }

   public void createStubsFor( String name, String serverHost, String hosts ) {
      createTypedStubsFor( name, serverHost, hosts,
                           "jac.core.dist.StubWrapper" );
   }

   public void createAsynchronousStubsFor( String name, String serverHost, 
                                           String hosts ) {
      createTypedStubsFor( name, serverHost, hosts,
                           "jac.core.dist.NonBlockingStubWrapper" );
   }

   public void setTransient(ClassItem classItem, String fieldName) {
      FieldItem fi = classItem.getField(fieldName);
      fi.setAttribute("deployment.transient","true");
   }


}


