/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.distribution;

import jac.core.*;
import jac.core.dist.*;

import gnu.regexp.*;
import java.util.*;

import jac.util.Log;

/**
 * This Aspect Component allows the programmer to easily implement
 * load-balancing features for its application when JAC is running in
 * distributed mode.
 *
 * @see LoadBalancingConf
 * @author Renaud Pawlak
 * @version 0.8.1 */

public class LoadBalancingAC extends AspectComponent implements LoadBalancingConf {

   public void addRoundTripLoadBalancer( String wrappeeName,
                                         String methods,
                                         String hostName, 
                                         String replicaExpr ) {

      pointcut( wrappeeName, ".*", methods,
                new LoadBalancingWrapper( replicaExpr ),
                "roundTripBalance", hostName );
   }

   public void addRandomLoadBalancer( String wrappeeName,
                                      String methods,
                                      String hostName, 
                                      String replicaExpr ) {

      pointcut( wrappeeName, ".*", methods,
                new LoadBalancingWrapper( replicaExpr ),
                "randomBalance", hostName );
   }
   
   /**
    * This inner-wrapper handles the load-balancing wrapping methods that
    * actually implement the load-balancing algorithms. */

   public class LoadBalancingWrapper extends Wrapper {

      int count = 0;
      Vector replicas = null;
      Random random = new Random();
      String hostExpr;
      boolean doFill = true;

      public LoadBalancingWrapper( String hostExpr ) {
         this.hostExpr = hostExpr;
      }

      public void invalidate() {
         doFill = true;
      }

      /**
       * Performs a round-trip load-balancing. */

      public Object roundTripBalance() {
         if( doFill ) {
            replicas = Topology.getPartialTopology(hostExpr)
               .getReplicas( this.wrappee() );
            System.out.println("filled partial topo with "+hostExpr+" on "+Topology.get()+": "+replicas);
            doFill = false;
         }
         if( replicas.size() == 0 ) {
            // none replicas where found, we perform a local call and 
            // will try to get them again on the next call
            doFill = true;
            Log.warning("load-balancing: no replica found, local call performed");
            return proceed();
         }
         if( count >= replicas.size() ) {
            count = 0;
         }
         return ((RemoteRef)replicas.get(count++)).invoke(this.method(),this.args());
      }

      /**
       * Performs a random load-balancing. */

      public Object randomBalance() {
         if( doFill ) {
            replicas = Topology.getPartialTopology(hostExpr)
               .getReplicas( this.wrappee() );
            doFill = false;
         }
         if( replicas.size() == 0 ) {
            // none replicas where found, we perform a local call and 
            // will try to get them again on the next call
            doFill = true;
            Log.warning("load-balancing: no replica found, local call performed");
            return proceed();
         }
         return ((RemoteRef)replicas.get(random.nextInt(replicas.size())))
            .invoke(this.method(),this.args());
      }
   }      

}
