/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.distribution.consistency;

import jac.util.*;
import jac.core.dist.*;
import jac.aspects.naming.*;
import jac.core.*;

import java.util.Arrays;
import java.util.Vector;
import java.io.Serializable;
import java.lang.reflect.*;

import gnu.regexp.*;

/**
 * This class is a high-level class that allows the programer to
 * easily create replication groups.
 *
 * <p>In the JAC terminology, a replication group is a set of objects
 * of the same class and name that are located on several remote
 * containers (defined by the replication group topology).<p>
 *
 * The set of replicas are put on consistency with a consistency
 * protocol that can be of various forms. A consistency protocol is
 * defined by its class (wich is a subclass of
 * <code>ConsistencyWrapper</code> and that implements the core
 * protocol interactions) and by its knowledge graph (i.e. the set of
 * replicas known by a given replica of the replication group).<p>
 *
 * This programming framework allows many kinds of distribution
 * schemes that are not limited to simple replication as known in the
 * litterature. For instance, the programmer can implement
 * client/server shemes, mobile-agents, caches, and so on.<p>
 *
 * For the programmers that just want to use this feature in
 * distribution aspects, use the higher-level abstraction called
 * "distribution rules" that implement pre-defined replication
 * groups.<p>
 *
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 *
 * @see Consistency
 * @see ConsistencyWrapper
 * @see jac.aspects.distribution.DeploymentAC
 * @see jac.core.dist.Topology */

public class ReplicationGroup implements Serializable {

   /** The topology says where the replicas are located. */
   Topology topology = null;
   /** The local replica. */
   transient Object localReplica = null;
   /** The class of the replicas. */
   Class type = null;
   /** The consistency of this group. */
   Consistency consistency = null;
   /** The deployement of this group. */
   Deployment deployment = null;
   /** The read method names array. */
   String[] readMethods = null;
   /** The write method names array. */
   String[] writeMethods = null;
   /** The call method names array. */
   String[] callMethods = null;
   
   /**
    * Creates a new replication group.<p>
    *
    * This constructor needs an existing object to be replicated (the
    * initatial replica).<p>
    *
    * In some cases, you need to create a replication group from some
    * already deployed object. In this case, use the other
    * constructor.<p>
    *
    * @param localReplica the local replica (from which the
    * replication group is created)
    * @param type the class of the replicas
    * @param topology the containers where the replicas are located
    * @param consistencyType the type of the consistency wrapper
    * (subclass of <code>ConsistencyWrapper</code>); can be null if no
    * consistency is needed for the servers
    * @param readMethods an array containing the names of the methods
    * that read the object state (can be empty or null)
    * @param writeMethods an array containing the names of the methods
    * that write the object state (can be empty or null)
    * @param callMethods an array containing the names of the methods
    * that neither read or write the object state (can be empty or
    * null)
    * @param knowledgeStyle the knowledge style for the consistency
    * protocol
    * @param knowledgeGraph the knowledge graph for the consistency
    * protocol
    *
    * @see Consistency
    * @see ConsistencyWrapper */

   public ReplicationGroup ( Object localReplica, 
                             Class type,
                             Topology topology,
                             Class consistencyType,
                             String[] readMethods,
                             String[] writeMethods,
                             String[] callMethods,
                             int knowledgeStyle,
                             int[] knowledgeGraph ) {
      this.localReplica = localReplica;
      this.topology = topology;
      this.type = type;
      this.readMethods = readMethods;
      this.writeMethods = writeMethods;
      this.callMethods = callMethods;
      if ( consistencyType != null ) {
         consistency = new Consistency (
            consistencyType, knowledgeStyle, knowledgeGraph );
      }
      deployment = new Deployment ( topology );
      if ( localReplica != null ) {
         RemoteRef[] replicas = deploy();
         deployConsistency( replicas );
      }
   }

   /**
    * Creates a new replication group from an existing set of remote
    * objects.<p>
    *
    * The set of remote objects can be obtained by replicating them
    * with the deployment features.<p>
    *
    * @param localReplica the local replica (from which the
    * replication group is created)
    * @param replicas the non-local replicas (from which the
    * replication group is created)
    * @param type the class of the replicas
    * @param topology the containers where the replicas are located
    * @param consistencyType the type of the consistency wrapper
    * (subclass of <code>ConsistencyWrapper</code>); can be null if no
    * consistency is needed for the servers
    * @param readMethods an array containing the names of the methods
    * that read the object state (can be empty or null)
    * @param writeMethods an array containing the names of the methods
    * that write the object state (can be empty or null)
    * @param callMethods an array containing the names of the methods
    * that neither read or write the object state (can be empty or
    * null)
    * @param knowledgeStyle the knowledge style for the consistency
    * protocol
    * @param knowledgeGraph the knowledge graph for the consistency
    * protocol
    *
    * @see Consistency
    * @see ConsistencyWrapper
    * @see Deployment#deploy(Object[])
    * @see Deployment#deployStruct(Object[])
    * @see Deployment#replicate(Object)
    * @see Deployment#replicateStruct(Object)
    */

   public ReplicationGroup ( Object localReplica,
                             RemoteRef[] replicas, 
                             Class type,
                             Topology topology,
                             Class consistencyType,
                             String[] readMethods,
                             String[] writeMethods,
                             String[] callMethods,
                             int knowledgeStyle,
                             int[] knowledgeGraph ) {
      this.localReplica = localReplica;
      this.topology = topology;
      this.type = type;
      this.readMethods = readMethods;
      this.writeMethods = writeMethods;
      this.callMethods = callMethods;
      if ( consistencyType != null ) {
         consistency = new Consistency (
            consistencyType, knowledgeStyle, knowledgeGraph );
      }
      deployment = new Deployment ( topology );
      if ( localReplica != null ) {
         deployConsistency( replicas );
      }
   }

   /**
    * Deploy the group by remotely instanciating the members on the
    * group hosts.<p>
    *
    * This method do not deploy the consistency protocol on the
    * replicas.<p>
    *
    * @return a set of remote references on the created replicas
    *
    * @see #deployConsistency(RemoteRef[]) */

   public RemoteRef[] deploy () {

      if ( localReplica == null ) return null;

      RemoteRef[] rreplicas = null;
      RemoteRef[] replicas = null;

      try {
         rreplicas = deployment.replicate ( localReplica );
         if( topology.containsLocal() ) {
            replicas = new RemoteRef[rreplicas.length + 1];
            System.arraycopy( rreplicas, 0, replicas, 0, rreplicas.length );
            replicas[rreplicas.length] = RemoteRef.create( 
               NameRepository.get().getName( localReplica ), (Wrappee) localReplica );
         } else {
            replicas = rreplicas;
         }
      } catch (Exception e) {
         System.out.println ("Cannot deploy the replication group.");
         e.printStackTrace();
      }      
      return replicas;
   }            

   /**
    * Deploy the consistency protocol on the set of replicas.<p>
    *
    * @param replicas the replica to be consistent
    * 
    * @see #deploy() */

   public void deployConsistency( RemoteRef[] replicas ) {
      if ( consistency != null ) {
         try {
            Log.trace ( "distribution.topology", "deploying the consistency on the replicas " + 
                        Arrays.asList( replicas ) );
            consistency.deploy ( replicas, localReplica.getClass(),
                                 readMethods,
                                 writeMethods,
                                 callMethods );
         } catch (Exception e) {
            System.out.println ("Cannot deploy the replication group.");
            e.printStackTrace();
         }      
         /** Force the replicated objects to support all the aspects
             of the remote hosts */
         for ( int i = 0; i < replicas.length; i++ ) {
            replicas[i].invoke( "forceTreatment", new Object[0] );
         }  
         //         replicas[0].invoke( "invokeRoleMethod", new Object[] {
         //   "addMember", new Object[] {
         //      RemoteRef.create ( NameRepository.get().getName( localReplica ), 
         //                         (Wrappee)localReplica )
         //   }
         //} );
      }   
   }   

   /**
    * Get the hosts where the group is deployed.<p>
    * 
    * @return the replication group topology */

   public Topology getTopology() {
      return topology;
   }
   
   /**
    * Get the remote reference of a given replica.<p>
    * 
    * @param container the container where the replica is located
    * @return a remote reference on this replica
    *
    * @see #getReplicas() */

   public RemoteRef getReplica( RemoteContainer container ) {
      if ( ! topology.isContainer( container ) ) return null;
      if ( localReplica == null ) return null;
      return container.bindTo( NameRepository.get().getName( localReplica ));
   }

   /**
    * Gets a set of remote references on all the replicas of this
    * replication group.<p>
    *
    * @return the replicas references
    * @see #getReplica(RemoteContainer) */

   public RemoteRef[] getReplicas() {
      RemoteRef[] res = new RemoteRef[topology.countContainers()];
      for ( int i = 0; i < res.length; i++ ) {
         res[i] = getReplica( topology.getContainer( i ) );
      }  
      return res;
   }

   /**
    * Gets a set of references on all the replicas located on the
    * given hosts.
    *
    * @param name the name of the replica (the object's name as given
    * by the naming aspect
    * @param hostExpr a regular expression that denote the hosts where
    * the replicas have to be looked up
    * @return the replicas references */

   public static RemoteRef[] getReplicas( String name, String hostExpr ) {

      // apply the aspects to the local replica if exist
      // so that the distribution aspect can replicate it if needed
      Wrappee localReplica = (Wrappee)NameRepository.get().getObject( name );
      localReplica.forceTreatment();

      // construct the host regexp
      RE regexp = null;
      try {
         regexp = new RE( hostExpr );
      } catch( Exception e ) {
         e.printStackTrace();
      }  

      // resolve and bind to the remote replica
      RemoteContainer[] containers = 
         Topology.get().getContainers( regexp );
      RemoteRef[] replicas = new RemoteRef[containers.length];
      for( int i=0; i<replicas.length; i++ ) {
         replicas[i] = containers[i].bindTo( name );
      }

      return replicas;
   }         


   /**
    * Get the type of the group (the class of its members) <p>
    *
    * @return the class of the members */
   
   public Class getType() {
      return type;
   }

}





