/*

  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.distribution.consistency;

import jac.util.*;
import jac.core.*;
import jac.core.dist.*;
import java.util.*;
import jac.aspects.naming.*;

/**
 * This wrapper class implements the core protocol for a strong
 * consistency that is based on a pull strategy.
 *
 * <p>On contrary to the <code>StrongPushConsistencyWrapper</code>,
 * this protocol pulls the data from the other replicas. Indeed, each
 * time a data is read and is not locally available, it is fetched
 * from the known replicas.<p>
 *
 * @see StrongPushConsistencyWrapper
 * 
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 */

public class StrongPullConsistencyWrapper extends ConsistencyWrapper {

   String hosts = null;

   /**
    * A friendly constructor for a pull consistency wrapper.
    *
    * @param hosts a regular expression that defines the host where
    * the consistency protocol is installed */

   public StrongPullConsistencyWrapper( String hosts ) {
      this.hosts = hosts;
      knownReplicas = null;
   } 

   /** An empty constructor for the Consistency class. */
   public StrongPullConsistencyWrapper() {} 

   /**
    * This wrapping method first try to use the wrappee to get the
    * result of the read method.<p>
    * 
    * If this result is null or if an exception occurs, it considers
    * that the read information was not locally present and tries to
    * fetch it from the replicas it knows (recursivelly all the
    * replicas are finally asked). If none of the replica returns a
    * value, it returns null or throws an exception.
    *
    * @return the value returned by the wrapped read method */

   public Object whenRead () {

      Object ret = null;
      
      ret = proceed();

      //System.out.println("<<< Strong consistency: kr = " + Arrays.asList ( knownReplicas ) + " >>>");

      if ( ret == null ) {

         if(knownReplicas != null) {
            Collaboration c = JacObject.getCollaboration();
            Vector asked_replicas = (Vector)c.getAttribute( visitedReplicas );
            if (asked_replicas == null) {
               asked_replicas = (Vector)c.addTransientAttribute( 
                  visitedReplicas, new Vector() );
            }
            //System.out.println ( "########### " + notified_replicas );
            
            Vector new_ar = new Vector ();
            RemoteRef cur_replica = RemoteRef.create( 
               NameRepository.get().getName( wrappee() ), wrappee() );
            
            for (int i = 0; i < knownReplicas.size(); i++) {
               if ( (! asked_replicas.contains( knownReplicas.get(i) ) ) &&
                    (! ((RemoteRef)knownReplicas.get(i)).getRemCont().isLocal()) ) {
                  
                  Vector kr = new Vector(knownReplicas);
                  kr.remove( knownReplicas.get(i) );
                  new_ar.clear();
                  new_ar.addAll ( asked_replicas );
                  new_ar.addAll ( kr );
                  new_ar.add ( cur_replica );
                  c.addTransientAttribute( visitedReplicas, new_ar );
                  
                  Log.trace("consistency", "(strong pull) read event on " + 
                            wrappee() + ":" + method() + ":" + 
                            ((RemoteRef)knownReplicas.get(i)).getRemCont().getName());
                  ret = ((RemoteRef)knownReplicas.get(i)).invoke(
                     "invokeRoleMethod",
                     new Object[] {
                        StrongPullConsistencyWrapper.class, "acceptRemoteRead",
                        new Class[] { RemoteRef.class, Object[].class },
                        new Object[] { null, new Object[] { method(), args() } }
                     }
                  );
                  if ( ret != null ) {
                     break;
                  }
               }
            }
         }
      }   
      return ret;
   }
      
   /**
    * Try to read the method asked by <code>whenRead</code>.<p>
    *
    * The data is :<p>
    * <ul><pre>
    * data[0] = the read method name string
    * data[1] = an array that contains the arguments of the read method 
    * </pre></ul>
    *
    * @param remoteReplica the replica that received the read event
    * @param data the read event data
    * @return the return value of the <code>data[0]</code> method*/

   public Object acceptRemoteRead ( RemoteRef remoteReplica,
                                    Object[] data) {
      Log.trace("consistency", "(strong pull) remote read event on " + 
                wrappee() + ":" + (String)data[0]);
      return wrappee().invoke ( (String)data[0], (Object[])data[1] );
   }
   
      
}






