/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.distribution.consistency;

import jac.util.Log;
import jac.core.*;
import jac.core.dist.Topology;
import jac.core.dist.RemoteRef;
import java.util.*;
import jac.aspects.naming.*;

/**
 * This wrapper implements a consistency protocol that forwards all
 * the writing calls to all the replicas that are known by the
 * wrapper.
 *
 * <p>It is called "push" since the wrapper pushes the data to the
 * other replicas. Despite this strategy is the most curently used,
 * other strong or weak consistency strategies can be implemented by
 * other consistency wrappers.
 * 
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 *
 * @see #whenWrite()
 * @see StrongPullConsistencyWrapper
 * @see WeakConsistencyWrapper */

public class StrongPushConsistencyWrapper extends ConsistencyWrapper {

   /** a false that is true during notification. */
   boolean inNotification = false;

   /**
    * A friendly constructor for a push consistency wrapper.
    *
    * @param hosts a regular expression that defines the host where
    * the consistency protocol is installed */

   public StrongPushConsistencyWrapper( String hosts ) {
      knownReplicas = null;
      this.hosts = hosts;
   } 

   /** An empty constructor for the Consistency class. */
   public StrongPushConsistencyWrapper() {} 

   /**
    * Forwards the call to all the replicas and then call the
    * replica.<p>
    *
    * The pushing mecanism is stopped by using the collaboration
    * attribute value defined by <code>visitedReplicas</code>.
    *
    * @return the value returned by the wrapped method.
    * @see ConsistencyWrapper#getVisitedReplicas() */

   public Object whenWrite () {

      Object ret = null;

      if( knownReplicas == null ) {
         calculateKnownReplicas();
      } 

      attrdef( "Persistence.disabled", "true" );

      if(knownReplicas != null) {
         if( inNotification ) return proceed();
         inNotification = true;
         Collaboration c = JacObject.getCollaboration();
         Vector notified_replicas = (Vector)c.getAttribute( visitedReplicas );
         if (notified_replicas == null) {
            notified_replicas = (Vector)c.addTransientAttribute( 
               visitedReplicas, new Vector() );
         }
         
         Vector new_nr = new Vector ();
         RemoteRef cur_replica = RemoteRef.create( 
            NameRepository.get().getName( wrappee() ), wrappee() );

         for (int i = 0; i < knownReplicas.size(); i++) {
            if ( (! notified_replicas.contains( knownReplicas.get(i) ) ) && 
                 (! ((RemoteRef)knownReplicas.get(i)).getRemCont().isLocal()) ) {

               Vector kr = new Vector ( knownReplicas );
               kr.remove( knownReplicas.get(i) );
               new_nr.clear();
               new_nr.addAll ( notified_replicas );
               new_nr.addAll ( kr );
               new_nr.add ( cur_replica );
               c.addTransientAttribute( visitedReplicas, new_nr );
               
               Log.trace("consistency", "(strong) write event on " + 
                         NameRepository.get().getName(wrappee()) + 
                         ":" + method() + ":" + 
                         ((RemoteRef)knownReplicas.get(i)).getRemCont().getName());
               try {
                  ((RemoteRef)knownReplicas.get(i)).invoke(
                     "invokeRoleMethod",
                     new Object[] {
                        "acceptRemoteWrite",
                        new Object[] { null, new Object[] { method(), args() } }
                     }
                  );
               } catch ( Exception e ) {
                  Log.error( "strong consistency error: "+
                             "failed to remotely invoke "+
                             "acceptRemoteWrite for "+
                             wrappee()+"."+method());
                  e.printStackTrace();
                  break;
               }
            }
         }
      }
      
      ret = proceed();

      inNotification = false;

      return ret;
   }

   /**
    * This method is called by <code>whenWrite</code> to push the
    * needed data when a state is writen in a remote replica.
    *
    * The data is :<p>
    * <ul><pre>
    * data[0] = the write method name string
    * data[1] = an array that contains the arguments of the write method 
    * </pre></ul>
    *
    * @param remoteReplica expected to be a reference on the remote
    * replica that recieved the write event
    * @param data the data transmittedd by <code>whenWrite</code>
    * @return null by default
    * @see #whenWrite() */

   public Object acceptRemoteWrite ( RemoteRef remoteReplica,
                                     Object[] data) {
      Log.trace("consistency", "(strong) remote write event on " + 
                NameRepository.get().getName(wrappee()) + 
                ":" + (String)data[0]);
      Object ret=null;
      try {
         ret = wrappee().invoke ( (String)data[0], (Object[])data[1] );
      } catch (Exception e) {
         e.printStackTrace();
      }
      return ret;
   }
   
   /**
    * Push the current replica state to the binding new replica.
    *
    * @param newReplica the replica that is currently binding
    */

   public void whenBindingNewReplica( RemoteRef newReplica ) {
      Log.trace("consistency", "(strong) initialized " + 
                newReplica + " with " + 
                NameRepository.get().getName(wrappee()));
      newReplica.remoteCopy( wrappee() );
   }
      
}



