/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.distribution.thinclient;

import java.io.Reader;
import java.util.Map;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Arrays;
import java.util.Collection;
import java.lang.reflect.InvocationTargetException;

import jac.core.Display;
import jac.core.NameRepository;
import jac.core.InvokeThread;
import jac.core.Wrappee;
import jac.core.JacObject;
import jac.util.Log;
import jac.core.rtti.ClassRepository;
import jac.core.rtti.ClassItem;
import jac.core.rtti.FieldItem;
import jac.core.rtti.CollectionItem;
import jac.core.rtti.MethodItem;
import jac.core.rtti.ConstructorItem;
import jac.core.rtti.AbstractMethodItem;
import jac.aspects.naming.NoSuchObjectError;

/**
 * This abstract class provides the bases to build a server for
 * clients that implement the Jac Thinclient Protocol (JTP).
 *
 * <p>The JTP defines 6 requests. All of them takes as a first
 * parameter the session ID (sid) that must be unique for each client.
 *
 * <ul>
 * <li><b>viewRequest(sid,jid)</b>: allows the client to open a view 
 * on an object called "jid" within the accessed JAC container
 * <li><b>updateRequest(sid,jid,fieldValues)</b>: allows the client 
 * to updates an object state 
 * <li><b>refreshRequest(sid)</b>: allows the client to refresh the 
 * current  view
 * <li><b>newRequest(sid,className)</b>: allows the client to 
 * instantiate a new object of the given class and opens a view on 
 * this new object 
 * <li><b>closeRequest(sid)</b>: allows the client to close the current
 *  wiew (the previous view is re-opened) 
 * <li><b>invokeRequest(sid,jid,methodName)</b>: allows the client to 
 * invoke a method called "method" on an object called "jid" within the 
 * accesed JAC container; if some parameters are needed by the invoked
 * method, the server will automatically open an input form to allow 
 * the user to fill their values and that invocation process is 
 * suspended until an input request occurs
 * <li><b>inputRequest(sid,methodName,params)</b>: allows the client 
 * to give to the server the parameters values that correspond to the 
 * given method invocation (this request should be send when the 
 * invoke request ask the client for the parameters); it then restart
 * the waiting invocation request
 * <li><b>cancelInputRequest(sid,methodName)</b>: allows the client 
 * to cancel the current invocation during the parameters imput
 * process 
 * </ul>
 *
 * <p>This class also provides some support for sessions allowing
 * multiple clients to access the same server.
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@bearteam.org">Laurent Martelli</a> */

public abstract class AbstractServer implements Display {

   /**
    * Handles a "view" request from a thin client.
    *
    * @param sid the session id
    * @param name the name of the object to open the view on
    * @return the data that will be used by the thin client to refresh
    * its state */

   public byte[] viewRequest( String sid, String name ) {
      initSession( sid );
      Log.trace("AbstractServer","viewRequest("+name+")"+ " " + getSession());
      Object substance = NameRepository.get().getObject( name ) ;
      if ( substance == null ) {
         errorResponse( "no such object (" + name + ")" );
      } else {
         getSession().newRequest( 
            new Request( Request.VIEW, 
                         substance.getClass().getName(), 
                         name, null, null ) );

         viewResponse( substance );
      }
      return waitForResponse();
   }  

   /**
    * Handles an "update" request from a thin client.
    *
    * <p>It uses the <code>fieldValues</code> hashtable to set the
    * fields of the substance object.
    *
    * @param sid the session id
    * @param name the name of the object to be updated
    * @param fieldValues the fields values (fieldName -> fieldValue);
    * all the fields of the object do not have to be in the hashtable;
    * if a field name does not correspond to an existing field name,
    * an error message is printed but the updating process is not
    * interrupted
    * @return the response (a view on the substance object) */

   public byte[] updateRequest( String sid, String name, 
                                Hashtable fieldValues ) {
      initSession( sid );
      Log.trace("AbstractServer","updateRequest("+name+","+
                fieldValues+")"+ " " + getSession());
      Object substance = NameRepository.get().getObject( name ) ;
      if ( substance == null ) {
         errorResponse( "no such object (" + name + ")" );
      } else { 
         ClassItem cli = ClassRepository.get().getClass(substance.getClass());
         Enumeration fieldNames = fieldValues.keys();
         while(fieldNames.hasMoreElements()) {
            String fieldName = (String) fieldNames.nextElement();
            FieldItem f = cli.getField( fieldName );
            if ( f != null ) {
               MethodItem setter = f.getSetter();
               Object[] v = buildParameterValues(
                  setter,new String[] { (String)fieldValues.get( fieldName ) });
               Log.trace("AbstractServer","setting state of "+substance+
                         "."+f+" to "+v[0]);
               f.setThroughWriter( substance, v[0] );
            } else {
               Log.error("update request contains a non-existing field");
            }            
         }
         Request currentRequest = getSession().getCurrentRequest();
         if ( getSession().getRequestCount() > 1 ) {
            getSession().endCurrentRequest();
         }
         // release the waiting thread in case of a modal view
         if ( currentRequest.isModalView() ) {
            // the request is valid
            currentRequest.setState(true); 
            currentRequest.getSemaphore().release();
            // do not dispatch the request since the release thread
            // will do it
         } else {
            dispatch( getSession().getCurrentRequest() );
         }
      }
      return waitForResponse();
   }  

   /**
    * Asks the server to be in frame-mode if supported by the client.
    *
    * <p>A frame is a sub-window of the client GUI.
    *
    * @param sid the session id
    * @param jid1 the object to open in frame 1
    * @param jid2 the object to open in frame 2
    * @param jid3 the object to open in frame 3
    * @return the response */

   /*public byte[] frameModeRequest( String sid, String jid1, String jid2, String jid3 ) {
      initSession( sid );
      Log.trace("AbstractServer","frameModeRequest()"+ " " + getSession());

      frameModeResponse( jid1, jid2, jid3 );
      return waitForResponse();

      }*/

   /**
    * Handles a "refresh" request from a thin client.
    *
    * <p>The refresh operation takes the object that is on the top of
    * the opened objects stack, creates its view again, and returns it
    * as a result.
    *
    * @param sid the session id
    * @return the data that will be used by the thin client to refresh
    * its state */

   public byte[] refreshRequest( String sid ) {
      initSession( sid );
      Log.trace("AbstractServer","refreshRequest()"+ " " + getSession());
      Object substance = getSession().getCurrentRequest().getObject();
      if ( substance == null ) {
         errorResponse( "no current object" );
      } else {
         //getSession().newRequest( new Request( name ) );
         viewResponse( substance );
      }
      return waitForResponse();
   }

   /**
    * Handles an "invoke" request from a thin client.
    *
    * <p>The invoke operation calls the method to invoke within a new
    * thread and waits for a result to be put into the session
    * result. The session id is passed to the created thread as a new
    * collaboration attribute.
    *
    * @param sid the session id
    * @param jid the JAC object ID to invoke
    * @param methodName the name of the method to invoke on jid
    * @return the data that will be used by the thin client to refresh
    * its state */

   public byte[] invokeRequest( String sid, String jid, String methodName ) 
      throws IllegalAccessException, InvocationTargetException, 
      NoSuchMethodException
   {
      initSession( sid );
      Log.trace("AbstractServer","invokeRequest("+jid+","+
                methodName+")"+ " " + getSession());
      Object substance = NameRepository.get().getObject( jid ) ;
      if ( substance != null ) {
         MethodItem method = ClassRepository.get()
            .getClass(substance.getClass())
            .getMethod( methodName );
         Object container = null;
         if( method.getAttribute("Rtti.containerOperation" ) != null ) {
            container = substance;
         }

         CollectionItem[] collections = method.getAddedCollections();
         CollectionItem collection = null;
         if( collections != null && collections.length>0 ) {
            collection = collections[0];
         }

         if ( collection != null && method.isAdder() &&
              collection.getAttribute("Gui.autoCreate") != null ) {
            Log.trace("AbstractServer","autoCreate...");

            ClassItem type = ClassRepository.get().getClass(
               method.getParameterTypes()[0]);

            InvokeThread.run( 
               substance, method, new Object[method.getParameterTypes().length], 
               new String[] {
                  "WebGui.sid", "Session.sid" },
               new Object[] { sid, getSession().getId() },
               new String[] {
                  "Gui.display", "Thinclient.session", "Gui.askForParameters", 
                  "Gui.autoCreate", "Rtti.container" },
               new Object[] { this, getSession(), method, type, container }
            );

         } else if( method.getAttribute("gui.openView" )!=null) {
            Log.trace("AbstractServer","openView...");

            InvokeThread.run( 
               substance, method, 
               new Object[method.getParameterTypes().length], 
               new String[] {
                  "WebGui.sid", "Session.sid" },
               new Object[] { sid, getSession().getId() },
               new String[] {
                  "Gui.display", "Thinclient.session", "Gui.askForParameters",
                  "Rtti.container", "Gui.openView" },
               new Object[] { this, getSession(), method, container, "true" }
            );

         } else {
            Log.trace("AbstractServer","other...");

            InvokeThread.run( 
               substance,
               method, new Object[method.getParameterTypes().length], 
               new String[] {
                  "WebGui.sid", "Session.sid" },
               new Object[] { sid, getSession().getId() },
               new String[] {
                  "Gui.display", "Thinclient.session", "Gui.askForParameters",
                  "Rtti.container" },
               new Object[] { this, getSession(), method, container }
            );
         }
      } else {
         throw new NoSuchObjectError(jid);
      }
      return waitForResponse();
   }

   /**
    * Handles a "new" request from a thin client.
    *
    * <p>It tries to instantiate the class that corresponds to the
    * given name and creates a view on the new object.
    *
    * @param sid the session id
    * @param className the name of the class that must be instantiated
    * @return the data that will be used by the thin client to refresh
    * its state */

   public byte[] newInstanceRequest(String sid, String className) {
      initSession( sid );
      Log.trace("AbstractServer","newInstanceRequest("+className+")"+ " " + getSession() );
      ClassItem classItem = ClassRepository.get().getClass(className);
      if (classItem==null) {
         errorResponse( "no such class \"" + className + "\"" );
      } else {
         try {
            Vector constructors = classItem.getConstructors();
            ConstructorItem c;
            if (constructors.size()==1) {
               c = (ConstructorItem)classItem.getConstructors().get(0);
            } else {
               c = (ConstructorItem)classItem.getConstructors().get(1);
            }
            if ( c.getParameterTypes().length > 0 ) {
               Object[] parameters = new Object[c.getParameterTypes().length];

               InvokeThread.quietRun( 
                  this, ClassRepository.get().getClass( this.getClass() )
                  .getMethod( "newInstanceInput" ),
                  new Object[] { c, parameters },
                  new String[] { "WebGui.sid",
                                 "Session.sid" },
                  new Object[] { sid, getSession().getId() },
                  new String[] { "Gui.display", "Thinclient.session" }, 
                  new Object[] { this, getSession() }
               );

            } else {
               Object newInstance = c.newInstance();
               // This is a HACK because the persistence aspect may have
               // unnamed the object
               ((Wrappee)newInstance).forceTreatment();
               getSession().newRequest( 
                  new Request( Request.VIEW, classItem.getName(), 
                               NameRepository.get().getName(newInstance), 
                               null, null ) );
               viewResponse( newInstance );
            }
         } catch (Exception e) {
            errorResponse( "instanciation of class " + className + " failed : "+e);
            e.printStackTrace();
         }
      }
      Log.trace("AbstractServer","<<< thread 1 waiting for result >>>" );
      return waitForResponse();
   }

   public void newInstanceInput(ConstructorItem c, Object[] parameters) {

      Log.trace("AbstractServer","<<< thread 2 showing input for "+c+" >>>" );
      String inputSequenceName =  (String) c.getAttribute("Gui.inputSequence");
      Log.trace("AbstractServer","input sequence name = "+inputSequenceName );
      if( inputSequenceName == null ) {
         showInput ( null, c, parameters );
      } else {
         Object inputSeq = null;
         Object ret = null;
         try {
            inputSeq = Class.forName( inputSequenceName )
               .getConstructor(new Class[] { Display.class, AbstractMethodItem.class, Object[].class })
               .newInstance(new Object[] {this, c, parameters});
            ret = inputSeq.getClass().getMethod("proceedInputs",new Class[0]).invoke(inputSeq,new Object[0]);
         } catch ( Exception e ) { e.printStackTrace(); return; }
         /*            while( ((Boolean)inputSeq.getClass().getMethod("hasNextStep",new Class[0]).invoke(inputSeq,new Object[0])).booleanValue() ) {
                       inputSeq.getClass().getMethod("nextStep",new Class[0]).invoke(inputSeq,new Object[0]);
                       }
                       inputSeq.getClass().getMethod("validate",new Class[0]).invoke(inputSeq,new Object[0]);*/
         if( ((Boolean)ret).booleanValue() ) {
            throw new RuntimeException();
         }
      }

      Log.trace("AbstractServer","<<< thread 2 creating instance >>>" );
      Object newInstance = null;
      try {
         newInstance = c.newInstance(parameters);
         ((Wrappee)newInstance).forceTreatment();
      } catch ( Exception e ) {
         e.printStackTrace();
      }
      refresh();
   } 
   
   /**
    * Handles an "input" request from a thin client.
    *
    * <p>This request is called by a client when it has filled an
    * input form that corresponds to the parameters values needed by
    * the given method. Generally, this request releases a semaphore
    * that triggers the actual method invocation.
    *
    * @param sid the session id
    * @param methodName the name of the method the parameters must be inputted
    * @param params the default values of the inputs in the forms
    * @return the data that will be used by the thin client to refresh
    * its state
    * @see #cancelInputRequest(String,String) */

   public byte[] inputRequest( String sid, String methodName, String[] params ) {

      initSession( sid );
      Log.trace("AbstractServer","inputRequest("+methodName+","+Arrays.asList(params)+")"+ " " + getSession());
      Request request = getSession().endCurrentRequest();
      Object substance = request.getObject();
      Log.trace("AbstractServer","inputRequest -> substance = "+substance);

      Class substance_type = null;
      ClassItem cli = null;
      AbstractMethodItem method = null;
      
      if ( substance != null ) {
         substance_type = substance.getClass();
         cli = ClassRepository.get().getClass(substance_type);
         method = cli.getMethod( methodName );
      } else {
         Log.trace("AbstractServer","inputRequest -> static invoke = "+request.getMethod());
         method = request.getMethod();
         cli = (ClassItem)method.getParent();
         substance_type = cli.getActualClass();
      }
      Object[] v = buildParameterValues(method,params);

      /** fill the inputs */
      request.setInputResult(v);
      /** releases the semaphore so that the waiting method call
          performs */
      request.getSemaphore().release();

      return waitForResponse();
   }

   public abstract Object[] buildParameterValues(AbstractMethodItem method, String[] params);

   /**
    * Handles an "input" request that has been cancelled from a thin
    * client.
    *
    * <p>Notifies the server that an input form has been
    * cancelled. The waiting method is also released by the input
    * result is null so that the program knows that it has been
    * cancelled.
    *
    * @param sid the session id
    * @param methodName the name of the method the parameters must be
    * inputted
    * @return the data that will be used by the thin client to refresh
    * its state
    * @see #inputRequest(String,String,String[]) */

   public byte[] cancelInputRequest( String sid, String methodName ) {
      initSession( sid );
      Log.trace("AbstractServer","cancelInputRequest("+methodName+")"+ " " + getSession());
      Request request = getSession().endCurrentRequest();
      /** Release the blocked thread but do not set the result */
      request.getSemaphore().release();
      return waitForResponse();
   }

   /**
    * Handles a "close" request from a thin client.
    *
    * <p>Pops the current request from the session's requests
    * stack. Then peeks the preview one and re-dispatch it.
    *
    * @param sid the session id
    * @return the data that will be used by the thin client to refresh
    * its state */

   public byte[] closeRequest( String sid ) {      
      initSession( sid );
      Log.trace("AbstractServer","closeRequest()"+ " " + getSession());
      Request currentRequest = getSession().getCurrentRequest();
      if ( getSession().getRequestCount() > 1 ) {
         getSession().endCurrentRequest();
      }
      // release the waiting thread in case of a modal view
      if ( currentRequest.isModalView() ) {
         // the request is not valid (cancelled by the user)
         currentRequest.setState(false); 
         currentRequest.getSemaphore().release();
         // do not dispatch the current request since the released 
         // thread will do it
      } else {
         dispatch( getSession().getCurrentRequest() );
      }
      return waitForResponse();
   }

   /**
    * Dispatches a given request (handles it regarding its type).
    *
    * @param request the request to dispatch
    */

   public void dispatch( Request request ) {
      if (request.isView()) {
         viewResponse( NameRepository.get().getObject( request.get("jid") ) );
      } else if (request.isModalView()) {
         viewResponse( NameRepository.get().getObject( request.get("jid") ) );
      } else if (request.isInvoke()) {
         inputResponse( request.getObject(),request.getMethod() );
      }
   }

   /** Stores the active sessions. */
   protected Hashtable sessions = new Hashtable();

   /**
    * Initializes/creates the current session with the given ID.
    *
    * <p>If the ID corresponds to a currently active session, then the
    * session object is returned, otherwise a new session is created
    * and activated.
    *
    * <p>A request handling method implementation must first call this
    * method to set the current collaboration session attribute. If
    * the session is not set, other methods using sessions will raise
    * a runtime error.
    *
    * @param sid the session's ID
    * @return the session that has been initialized
    *
    * @see #getSession()
    * @see #waitForResponse()
    * @see #setResponse(byte[]) */

   protected final Session initSession( String sid ) {
      Session result = null;
      if( ! sessions.containsKey( sid ) ) {
         sessions.put( sid, result = new Session( sid ) );
      } else {
         result = (Session) sessions.get( sid );
      }
      JacObject.getCollaboration().addTransientAttribute( 
         "Thinclient.session", result );
      JacObject.getCollaboration().addTransientAttribute( 
         "Session.sid", sid );
      return result;
   }

   /**
    * Gets the current session.
    *
    * <p>This method works only if <code>initSession</code> has
    * previously been called in the current collaboration.
    *
    * @return the current session
    * @see #initSession(String) */

   protected final Session getSession() {
      return (Session) JacObject.getCollaboration()
         . getAttribute( "Thinclient.session" );
   }
   
   /**
    * Makes the current requesting thread block and wait until a
    * response from the same session is available.
    *
    * <p>The thread that call this method waits until a call to
    * <code>setResponse</code> occurs in the same session. Then the
    * thread wakes up and the value that was set by
    * <code>setResponse</code> is returned.
    *
    * <p>This method works only if <code>initSession</code> has
    * previously been called in the current collaboration.
    *
    * @return the response value
    * @see #initSession(String)
    * @see #setResponse(byte[]) */
   
   public byte[] waitForResponse() {
      Log.trace("AbstractServer","Waiting for response...");
      return getSession().waitForResponse();
   }

   /**
    * Sets the result of the response to a given string value.
    *
    * <p>If another requesting thread was blocked by a
    * <code>waitForResult</code> call, then this thread is unlocked.
    *
    * <p>This method works only if <code>initSession</code> has
    * previously been called in the current collaboration.
    *
    * @param sid the session of the current thread 
    * @param response the response value
    * @see #initSession(String)
    * @see #waitForResponse() */

   public void setResponse( byte[] response ) {
      Log.trace("AbstractServer","setResult");
      getSession().setResponse( response );
   }

   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to show a view on a given object.
    *
    * <p>For instance, if the thin client is a terminal, the view
    * response can build a textual representation of the object by
    * using the <code>toString()</code> method:
    *
    * <ul><pre>
    * protected void viewResponse( Object substance ) {
    *     setResponse( substance.toString().getBytes() );
    * }
    * </pre></ul>
    *
    * @param substance the object to build the view of */

   protected abstract void viewResponse( Object substance );

   //   protected abstract void frameModeResponse( String jid1, String jid2, String jid3 );

   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to show a message.
    *
    * <p>For instance, if the thin client is a terminal:
    *
    * <ul><pre>
    * protected void messageResponse( String text ) {
    *     setResponse( text.getBytes() );
    * }
    * </pre></ul>
    *
    * @param text the text of the message */

   protected abstract void messageResponse( String text );
   
   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to show a file
    *
    * @param reader the file's stream
    * @param mimeType the file's mime type
    */
   protected abstract void fileResponse( Reader reader, String mimeType );

   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to show an error.
    *
    * <p>For instance, if the thin client is a terminal:
    *
    * <ul><pre>
    * protected void stringResponse( String description ) {
    *     setResponse( "Error: " + description.getBytes() );
    * }
    * </pre></ul>
    *
    * @param description the error description */
   
   protected abstract void errorResponse( String description );
   
   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to show a collection of objects.
    *
    * <p>For instance, each item within the collection may corresponds
    * to one row of a table view drawn by the thin client. Headers and
    * collums can be formed with the fields of the objects within the
    * collection.
    *
    * @param c the collection to show */

   protected abstract void collectionResponse( Collection c );
   
   /**
    * Must be defined by the programmer to set a response that will
    * allow the thin client to ask to the user the input of all the
    * parameters values of the given method.
    *
    * @param object the object being invoked
    * @param method the method being invoked (can be a constructor or
    * a static method - object is null in these cases) */

   protected abstract void inputResponse( Object object, 
                                          AbstractMethodItem method );

   /** A counter to name non-JAC object that need to be accessed by
       web clients. */
   static int lastName = 0;

   /**
    * Generates a locally unique name for an object.
    *
    * @param object the object to name
    * @return the proposed name */

   static String getNextName(Object object) {
      return object.getClass().getName()+(lastName++);
   }

   /***********************************************************
    * The following methods implement the Display interface
    ***********************************************************/
   
   public void show( Object object ) {
      Log.trace("AbstractServer","show "+object+","+getSession());
      if (object == null) {
         messageResponse( "null" );
      } else if ( object instanceof Wrappee ) {
         viewResponse( object );
      } else if ( object.getClass().isArray() || 
                  object instanceof Collection ) {
         Vector c;
         if( object instanceof Collection )
            c = new Vector( (Collection) object );
         else
            c = new Vector( Arrays.asList( (Object[])object ) );
         
         if ( c.size() == 0 ) {
            messageResponse( "O item returned" );
         } else {
            collectionResponse( c );
         }

      } else if ( object instanceof Throwable ) {
         messageResponse( object.toString() );
         ((Throwable)object).printStackTrace();
      } else if ( object instanceof Reader ) {
         String mimeType  = "text/plain";
         fileResponse( (Reader)object, mimeType );
      } else  {
         messageResponse( object.toString() );
      }
      
   }

   public boolean showMessage(String message, String title,
                              boolean canValidate, boolean canCancel, 
                              boolean canClose) {
      return showModal(null,title,message,canValidate,canCancel,canClose);
   }

   public boolean showModal( Object object, String title, String header,  
                             boolean canValidate, boolean canCancel, 
                             boolean canClose ) {
      Log.trace("AbstractServer","show modal "+object+","+getSession());
      Request request = new Request( 
         Request.MODAL_VIEW, null, 
         NameRepository.get().getName(object), null, null );
      getSession().newRequest( request );
      request.title = title;
      request.header = header;
      request.canValidate = canValidate;
      request.canCancel = canCancel;
      request.canClose = canClose;
      Log.trace("AbstractServer","session "+getSession());
      show(object);
      //Request request = getSession().getCurrentRequest();
      // we block the current thread until the view is closed
      request.getSemaphore().acquire();
      Log.trace("AbstractServer","showModal thread unlocked");
      // if getState() == false, then the invocation has been cancelled
      // by the user
      return request.getState();
   }

   public void openView( Object object ) {
      Log.trace("AbstractServer","open view "+object+","+getSession());
      // put a view request on the stack
      getSession().newRequest( 
         new Request( Request.VIEW, 
                      object.getClass().getName(), 
                      NameRepository.get().getName(object), null, null ) );
      show(object);
   }

   public Object showInvocation( Object object, AbstractMethodItem method ) { 
      return null;
   }
   
   public boolean showInput( Object object, AbstractMethodItem method, 
                             Object[] parameters ) {      

      String jid = NameRepository.get().getName(object);
      if (object!=null && jid==null) {
         // create a local name for non wrappee objects
         jid = getNextName(object);
         NameRepository.get().register(jid,object);
      }

      /*      if ( ! getSession().getCurrentRequest().isInvoke() )
         throw new InvalidActionException();
      */

      Log.trace("AbstractServer",
                "showInput("+object+","+method.getName()+","+
                Arrays.asList(parameters)+")");

      Request request = new Request( 
         Request.INVOKE, ((ClassItem)method.getParent()).getName(), 
         jid, method.getFullName(), null );

      getSession().newRequest( request );
      Log.trace("AbstractServer", getSession().toString() );

      inputResponse( object, method );
         
      request.getSemaphore().acquire();

      if( request.getInputResult() == null ) {
         return false;
      } else {
         for ( int i = 0; i < parameters.length; i ++ ) {
            parameters[i] = request.getInputResult(i);
         }
         return true;
      }
   }
   
   /**
    * Does nothing by default. */

   public void showStatus(String message) {};

   public void refresh() {
      dispatch( getSession().getCurrentRequest() );
   }

   public void newObject( Object newObject ) {}

   public void applicationStarted() {}

}
