/*
  Copyright (C) 2001 Laurent Martelli, Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.distribution.thinclient;

import java.util.*;
import jac.util.*;
import jac.core.*;
import jac.core.rtti.*;

/**
 * This class defines requests that come from thin clients.
 *
 * <p>When a request is performed to a <code>AbstractServer</code>,
 * a new instance of this class is created and is pushed on the
 * request stack of the current session.
 *
 * @see AbstractServer */

public class Request implements java.io.Serializable {

   /** The unknown request type id. */
   public static String UNKNOWN = "unknown";
   /** The view request type id. */
   public static String VIEW = "view";
   /** The modal view request type id. */
   public static String MODAL_VIEW = "modal_view";
   /** The invoke request type id. */
   public static String INVOKE = "invoke";
   /** The input request type id. */
   public static String INPUT = "input";
   /** The cancel request type id. */
   public static String CANCEL = "cancel";
   /** The close request type id. */
   public static String CLOSE = "close";
   /** The refresh request type id. */
   public static String REFRESH = "refresh";
   /** The new request type id. */
   public static String NEW = "new";

   /** Information on the current window: the title. */
   public String title;
   /** Information on the current window: the body header. */
   public String header;
   /** Information on the current window: validation feature. */
   public boolean canValidate=false;
   /** Information on the current window: cancellation feature. */
   public boolean canCancel=false;
   /** Information on the current window: close feature. */
   public boolean canClose=false;
   
   /** Stores the current request type id. */
   protected String requestType;

   /** Stores the parameters of the requests. */
   protected Hashtable parameters = new Hashtable();

   /** The semaphore that is used to make the thread that asked for
       an input operation wait that the user actually inputs the
       data (only used in case of input requests). */
   protected Semaphore semaphore = new Semaphore();

   /** The request state (true = validated, false = cancelled) */
   protected boolean state = true;

   /** The input result (only used in case of input requests). */
   protected Object[] inputResult;

   /**
    * A constuctor that initializes the request with all the needed
    * information.
    *
    * <p>Some of these parameters must be null to be consistent with
    * the request type.
    *
    * @param type the request's type ID (UNKNOWN, VIEW, INVOKE, INPUT,
    * CANCEL, CLOSE, or REFRESH)
    * @param className the name of the object's class
    * @param jid the object's ID
    * @param method the method name
    * @param parameters the method parameters string values */

   public Request( String type, String className, String jid, 
                   String method, Object[] params ) {
      if ( type == null ) requestType = UNKNOWN;
      else requestType = type;
      if( className != null ) parameters.put("class",className);
      if( jid != null ) parameters.put("jid",jid);
      if( method != null ) parameters.put("method",method);
      if( params != null ) parameters.put("params",params);      
   }   

   /**
    * Sets the parameters of a "view" request.
    *
    * @param jid the wiewed object's JAC id */

   public void setViewPage(String jid) {
      parameters.clear();
      parameters.put("jid",jid);
   }

   /**
    * Gets the semaphore of the request.
    *
    * It is used to make the thread that asked for an input operation
    * wait that the user actually inputs the data (only used in case
    * of input requests).
    *
    * @return the semaphore */

   public Semaphore getSemaphore() {
      return semaphore;
   }

   /**
    * Tells if the request is valid (if not, it may have been
    * cancelled by the user).
    *
    * @return the request state */

   public boolean getState() {
      return state;
   }

   /**
    * Sets the request's state.
    *
    * @param state the new state (false is cancelled by the user) */

   public void setState( boolean state ) {
      this.state = state;
   }

   /**
    * Sets the input result of this request (only used in case of
    * input requests).
    *
    * @param result the input result
    * @see #getInputResult() */

   public void setInputResult(Object[] result) {
      inputResult = result;
   }

   /**
    * Gets the input result (only used in case of input requests).
    *
    * @return the input result
    * @see #setInputResult(Object[]) */

   public Object[] getInputResult() {
      return inputResult;
   }

   /**
    * Gets the i-th input result (only used in case of input requests).
    *
    * @return the i-th input result
    * @see #getInputResult() */
   
   public Object getInputResult(int i) {
      return inputResult[i];
   }

   /**
    * Set the parameters of an "invoke" request.
    *
    * @param jid the invoked object's JAC id
    * @param method the method that is invoked by this
    * request */

   public void setInvokePage(String className, String jid, String method) {
      parameters.clear();
      parameters.put("class",className);
      parameters.put("jid",jid);
      parameters.put("method",method);
   }

   /**
    * Set the parameters of a "new" request.
    *
    * @param className the name of the class that must be created by
    * this request */

   public void setNewInstancePage(String className) {
      parameters.clear();
      parameters.put("class",className);
   }

   /**
    * Tells if this request is a "view" request.
    *
    * @return true if the current request is a "view" request */

   public boolean isView() {
      return requestType == VIEW;
   }

   /**
    * Tells if this request is a "modal view" request.
    *
    * @return true if the current request is a "modal view" request */

   public boolean isModalView() {
      return requestType == MODAL_VIEW;
   }

   /**
    * Tells if this request is an "invoke" request.
    *
    * @return true if the current request is an "invoke" request */
    
   public boolean isInvoke() {
      return requestType == INVOKE;
   }
   
   /**
    * Tells if this request is a "new" request.
    *
    * @return true if the current request is a "new" request */

   public boolean isNew() {
      return requestType == NEW;
   }

   /**
    * Gets a parameter of this request.
    *
    * @param name the name of the parameter
    * @return the parameter value
    */

   public String get(String name) {
      return (String)parameters.get(name);
   }

   /**
    * Gets the invoked method in case of an "invoke" request.
    *
    * @return the invoked method */

   public AbstractMethodItem getMethod() {
      if ( get("jid") != null ) {
         Object substance = NameRepository.get().getObject(get("jid"));
         return ClassRepository.get().getClass(substance.getClass()).getMethod(get("method"));
      } else {
         Log.trace("webgui","getMethod "+get("class") + ":" + get("method"));
         AbstractMethodItem res = null;
         try {
            res = ClassRepository.get().getClass(get("class")).getAbstractMethod(get("method"));
         } catch ( Exception e ) {
            e.printStackTrace();
         }
         return res;
      }
   }

   /**
    * Returns the object that is involved by this request.
    *
    * <p>It can be a viewed object in case of a "view" request or an
    * invoked object in case of a "invoke" request.
    *
    * @return the object that is involved by the request */

   public Object getObject() {
      return NameRepository.get().getObject(get("jid"));
   }


   /** Contextual info of the session */
   Hashtable context = new Hashtable();

   public Map getContext() {
      return context;
   }

   /**
    * Returns a printable string representation of the current
    * request.
    *
    * @return a printable string */ 

   public String toString() {
      return requestType + parameters.toString();
   }
}
