/*
  Copyright (C) 2002 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import jac.core.Display;
import jac.core.InvokeThread;
import jac.core.NameRepository;
import jac.core.rtti.AbstractMethodItem;
import jac.core.rtti.MethodItem;
import jac.core.rtti.CollectionItem;
import jac.util.Log;

import java.util.Arrays;
import java.util.Enumeration;
import java.util.Stack;

import java.io.Reader;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.IOException;

import java.awt.Window;
import java.awt.Dialog;
import java.awt.Frame;
import javax.swing.JFrame;
import javax.swing.JDialog;
import javax.swing.JFileChooser;

/**
 * This class provides the base impllementation of a display for the
 * Swing toolkit.
 *
 * <p>It is a frame (top-level window) that implements default
 * behaviors for the <code>Display</code> insterface except the
 * <code>show</code> method.
 *
 * @see jac.core.Display */

public abstract class AbstractSwingDisplay extends JFrame 
   implements Display {

   Stack dialogs = new Stack();

   public void addDialog(JDialog dialog) {
      dialogs.push(dialog);
   }

   public void removeDialog(JDialog dialog) {
      dialogs.remove(dialog);
   }

   String displayID;

   public String getDisplayID() {
      return displayID;
   }

   public void setDisplayID(String displayID) {
      this.displayID = displayID;
   }

   public boolean showMessage(String message, String title,
                              boolean okButton, 
                              boolean cancelButton, 
                              boolean closeButton ) {
      return showModal(null,title,message,okButton,cancelButton,closeButton);
   }

   // implements the display interface
 
   public boolean showModal( Object object, 
                             String title, String header, 
                             boolean okButton, 
                             boolean cancelButton, 
                             boolean closeButton ) {
      Window owner = this;
      if( dialogs.size() > 0 ) {
         owner = (Window)dialogs.peek();
      }
      if (object == null) {
         return ObjectView.addViewFor( 
            null, this, owner,
            title, header, 
            okButton, cancelButton, closeButton );
      } else if (object instanceof Throwable) {
         ((Throwable)object).printStackTrace();
      } else if (object.getClass().isArray()) {
         return ObjectView.addViewFor(
            Arrays.asList((Object[])object), this, owner,
            title, header, 
            okButton, cancelButton, closeButton );
      } else {
         return ObjectView.addViewFor(
            object, this, owner,
            title, header, 
            okButton, cancelButton, closeButton );
      }
      return false;
   }

   public void openView( Object object ) {
      show(object);
   }

   public Object showInvocation( Object object, AbstractMethodItem method ) {
      Window owner = this;
      CallingBox popup = null;
      if( dialogs.size() > 0 ) {
         owner = (Window)dialogs.peek();
         popup = new CallingBox( this, (Dialog)owner, object, (MethodItem)method );
      } else {
         popup = new CallingBox( this, (Frame)owner, object, (MethodItem)method );
      }
      return popup.getResult();
   }

   public boolean showInput( Object object, AbstractMethodItem method, 
                             Object[] parameters ) {
      Window owner = this;
      InputBox popup = null;
      if( dialogs.size() > 0 ) {
         owner = (Window)dialogs.peek();
         popup = new InputBox( this, (Dialog)owner, object, method );
      } else {
         popup = new InputBox( this, (Frame)owner, object, method );
      }
      Object[] inputResult = popup.getResult();
      if( inputResult == null ) {
         return false;
      } else {
         for ( int i = 0; i < parameters.length; i ++ ) {
            parameters[i] = inputResult[i];
         }
         return true;
      }
   }

   public void showStatus(String message) {
      System.out.println(message);
   }

   public void refresh() {}

   public void newObject( Object newObject ) {}   

   public void applicationStarted() {}

   /**
    * A GUI level method to invoke a method on an object.
    *
    * @param method the invoked method
    * @param object the invoked object
    * @param collection the corresponding collection item if the
    * invoked object is a collection */

   public void invoke(AbstractMethodItem method, Object object, CollectionItem collection) {
      if ( method.getParameterTypes().length == 0 ) {
         Log.trace("gui", "Invoking " + method + " on " + object );
         InvokeThread.run( object, method, new Object[0],
                           null, null,
                           new String[] { "Gui.display", "Session.sid" }, 
                           new Object[] { this, "Swing"+jac.core.dist.Distd.getLocalContainerName() } );
      } else {
         //new CallingBox( this, object, method );
         Log.trace("gui","Invoking "+method+" on "+object+" (ask for parameters is on)");
         if (collection != null) {
            InvokeThread.run( object, method, new Object[method.getParameterTypes().length], 
                              null, null, 
                              new String[] { "Gui.display", "Session.sid", "Gui.askForParameters" , "Gui.collectionItem"}, 
                              new Object[] { this, "Swing"+jac.core.dist.Distd.getLocalContainerName(), method, collection } );
         } else {
            InvokeThread.run( object, method, new Object[method.getParameterTypes().length], 
                              null, null, 
                              new String[] { "Gui.display", "Session.sid", "Gui.askForParameters" }, 
                              new Object[] { this, "Swing"+jac.core.dist.Distd.getLocalContainerName(), method } );
         }
      }
   }


   /** show a save dialog and save the stream into the selected file 
    * @param reader the stream to save
    */
   public void saveStreamToFile(Reader reader) 
   {
      JFileChooser chooser = new JFileChooser();
      if (JFileChooser.APPROVE_OPTION == chooser.showSaveDialog(this)) {
         try {
            File file = chooser.getSelectedFile();
            Log.trace("gui.stream","saving stream to "+file.getAbsolutePath());
            OutputStreamWriter writer = new OutputStreamWriter(
               new FileOutputStream(file),"UTF-8");
            int b = reader.read();
            while (b != -1) {
               writer.write(b);
               b = reader.read();
            }
            writer.close();
         } catch (IOException e) {
            Log.error("saveStreamToFile() failed : "+e);
            e.printStackTrace();
         }
      }
   }

}


