/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.List;
import java.lang.reflect.InvocationTargetException;

import jac.core.Display;
import jac.core.JacObject;
import jac.util.Log;
import jac.core.rtti.MethodItem;

/**
 * This dialog is used to ask the parameters values when a method is
 * called on a viewed JAC object.<p>
 *
 * @see View
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a>
 */

public class CallingBox extends JDialog implements ActionListener /*, TextListener, ListSelectionListener*/ {

   private JButton okButton;
   private JButton cancelButton;

   private Object substance = null;
   private MethodItem method = null;
   
   private Object result = null;

   /**
    * Returns the result of the invoked method.
    *
    * <p>If the invocation is canceled by the user, then this method
    * is meaningless. The client should catch and handle the
    * <code>InputCanceledException</code> to treat cancelation.
    *
    * @return the value returned by the invoked method
    * @see InputCanceledException */

   public Object getResult() {
      return result;
   }

   /**
    * Construct a new calling box to get the parameters of the called
    * method.<p>
    *
    * @param display the owning display
    * @param parent the owning window (here a frame)
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

   public CallingBox( Window display, Frame parent, Object substance, MethodItem method ) {
      super(parent,"Call " + method.getName(),true);
      init( display, substance, method );
   }

   /**
    * Construct a new calling box to get the parameters of the called
    * method.<p>
    *
    * @param display the owning display
    * @param parent the owning window (here a dialog)
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

   public CallingBox( Window display, Dialog parent, Object substance, MethodItem method ) {
      super(parent,"Call " + method.getName(),true);
      init( display, substance, method );
   }

   /**
    * Initializes the dialog.
    *
    * @param display the owning display
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

   public void init( Window display, Object substance, MethodItem method ) {
      ((AbstractSwingDisplay)display).addDialog(this);
      this.substance = substance;
      this.method = method;

      Container contentPane = getContentPane();

      Class substance_type = substance.getClass();

      String tn = substance_type.getName();

      addWindowListener ( new WindowAdapter () {
            public void windowClosing( WindowEvent e ) {
               e.getWindow().dispose();
            }
         }
      );
      Component panel = RenderFactory.createParameters(display,method,substance);
      contentPane.add("Center", panel);
      JPanel p2 = new JPanel();
      p2.setBorder( BorderFactory.createEtchedBorder() );
      okButton = addButton( p2, "Ok" );
      cancelButton = addButton( p2, "Cancel" );
      contentPane.add( "South", p2 );
      pack();

      // open the box centerd in the screen...
      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      Rectangle rect = getBounds();
      double left = (screenDim.getWidth()-rect.getWidth())/2;
      double top = (screenDim.getHeight()-rect.getHeight())/2;
      Rectangle newRect = new Rectangle(
         (int)left,(int)top,
         (int)rect.getWidth(),(int)rect.getHeight());
      setBounds(newRect);

      setVisible ( true );
   }

   /**
    * For internal use.<p>
    */
   JButton addButton(Container c, String name)  {
      JButton button = new JButton(name);
      button.addActionListener(this);
      c.add(button);
      return button;
   }

   /**
    * Implements what is done when a button is pressed (may be either
    * OK or CANCEL).<p>
    *
    * @param evt tell what button was pressed */

   public void actionPerformed(ActionEvent evt) {

      try {

         Object source = evt.getSource();
         if(source == okButton) {
            
            Object[] v = RenderFactory.buildParameterValues(method,getContentPane());
            JacObject.getCollaboration().addPersistentAttribute(
               "Gui.display", (Display)getParent() );
            //JacObject.getCollaboration().addPersistentAttribute(
            //   "Gui.askForParameters", "true" );
            try {
               //InvokeThread.run( substance, method, v.toArray(), (Display)getParent() );
               result = method.invoke( substance, v );
            } catch ( Exception e ) {
               dispose();
               if ( e instanceof InvocationTargetException ) { 
                  new ResultBox( this, substance, method.getActualMethod(), 
                                 ((InvocationTargetException)e).getTargetException().toString() );
                  ((InvocationTargetException)e).printStackTrace();
               } else {
                  new ResultBox( this, substance, method.getActualMethod(), e.toString() );
                  e.printStackTrace();
               }  
            }
            dispose();
            if ( method.getType() != void.class ) {
               System.out.println( (result==null?"null":result.toString()) );
            }
         }
         else if (source == cancelButton)
            dispose();
         
      } catch ( Exception e ) {
         e.printStackTrace ();
      }
   }

}
