/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

import jac.core.rtti.*;
import jac.util.*;

import jac.core.*;
import jac.wrappers.*;
import jac.aspects.naming.*;

/**
 * This class defines a Swing component view for collections of objects.
 *
 * <p>By default this view constructs an embedded <code>JList</code>
 * containing the string representation of the contained
 * object. However, depending on the attributes defined on the
 * collection by the GUI aspect component, the view can be
 * customized.
 *
 * @see GuiAC */

public class CollectionView extends JPanel
   implements FieldItemView, ActionListener, ListSelectionListener, MouseListener {

   JTable table;
   Vector headers;

   Object[] objects;
   Window parent;
   CollectionItem collection;
   JList wordList;

   Object substance = null;
   private View parentView = null;

   String[] sortCriteria = null;

   class MyComparator implements Comparator {
      public int compare(Object o1, Object o2) {
         if( sortCriteria == null ) {
            return o1.toString().compareToIgnoreCase(o2.toString());
         } else {
            FieldItem fi = ClassRepository.get().getClass(o1.getClass())
               .getField(sortCriteria[0]);
            Object v1 = fi.getThroughAccessor(o1);
            Object v2 = fi.getThroughAccessor(o2);
            if( v1 != null && v2 != null ) {
               if( v1 instanceof Number ) {
                  return (int) (((Number)v1).doubleValue() - ((Number)v2).doubleValue());
               } else {
                  return v1.toString().compareToIgnoreCase(v2.toString());
               }
            }
         }
         return -1;
      }
      public boolean equals(Object o1, Object o2) {
         if( sortCriteria == null ) {
            return o1.toString().equals(o2.toString());
         } else {
            return ((Wrappee)o1).getFieldValue(sortCriteria[0]) ==
               ((Wrappee)o2).getFieldValue(sortCriteria[0]);
         }
      }
   }

   /**
    * Constructs a new collection view.
    *
    * @param parent the Swing dialog parent
    * @param parentView a view that aggregates this view if any
    * @param substance the object the viewed collection belongs to
    * @param collection the collection that is showed by this view */

   public CollectionView(Window parent, 
                         View parentView, Object substance,
                         CollectionItem collection ) {
      this.parent = parent;
      this.substance = substance;
      this.parentView = parentView;
      this.collection = collection;

      setLayout( new BoxLayout( this, BoxLayout.Y_AXIS ) );
      if( collection.getAttribute("Gui.border") == null || 
          collection.getAttribute("Gui.border").equals("true")) {
         setBorder( 
            BorderFactory.createTitledBorder( 
               BorderFactory.createLineBorder( Color.black ),
               NamingConventions.textForName(" "+collection.getName()+" ")));
      }

      JPanel upperCont = new JPanel();
      upperCont.setLayout( new BoxLayout( upperCont, BoxLayout.Y_AXIS ) );
      JPanel downCont = new JPanel();
      //downCont.setLayout( new BoxLayout( downCont, BoxLayout.X_AXIS ) );

      sortCriteria = (String[])collection.getAttribute( "Gui.sortCriteria" );

      Integer height = (Integer)collection.getAttribute( "Gui.viewHeight" );
      Integer width = (Integer)collection.getAttribute( "Gui.viewWidth" );

      if ( height == null ) height = new Integer(70);
      if ( width == null ) width = new Integer(500);
      
      if ( GuiAC.isTableView(collection) ) 
      {
         objects = null;
         table = new JTable();

         JScrollPane scrollPane = new JScrollPane(table);
         table.setPreferredScrollableViewportSize(
            new Dimension(width.intValue(), height.intValue()));
         //      table.setAutoResizeMode( JTable.AUTO_RESIZE_ALL_COLUMNS );
         upperCont.add( scrollPane, BorderLayout.CENTER );
         table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
         BetterListSelectionModel selModel = new BetterListSelectionModel( table );
         table.setSelectionModel( selModel );         
         table.setName( collection.getName() );
         selModel.addListSelectionListener( this );
         table.addMouseListener(this);
      } else {
         wordList = new JList();
         JScrollPane scrollPane = new JScrollPane(wordList);
         wordList.setName( collection.getName() );
         int rowCount = height.intValue()/10;
         wordList.setVisibleRowCount( rowCount );
         upperCont.add( scrollPane );
         wordList.addListSelectionListener( this );
         wordList.addMouseListener(this);
      }

      // add, remove and view buttons
      JButton b = null;
      if ( !GuiAC.getViewOnSelection(collection) ) {
          b = new JButton ( ResourceManager.getViewIcon() );
          b.setToolTipText( "View" );
          b.setActionCommand( "open" );
          b.addActionListener( this );
          downCont.add( b );
      }

      if ( collection.isList() || collection.isSet() ) {
         b = new JButton ( ResourceManager.getNewIcon() );
         b.setToolTipText( "Add" );
         b.setActionCommand( "add" );
         b.addActionListener( this );
         downCont.add( b );

         b = new JButton ( ResourceManager.getRemoveIcon() );
         b.setToolTipText( "Remove" );
         b.setActionCommand( "remove" );
         b.addActionListener( this );
         downCont.add( b );
      } else {
         MethodItem[] addingMeths = collection.getAddingMethods();
         if ( addingMeths != null ) {
            b = new JButton ( ResourceManager.getNewIcon() );
            b.setToolTipText( "Add" );
            b.setActionCommand( "add" );
            b.addActionListener( this );
            downCont.add( b );
         }

         MethodItem[] removingMeths = collection.getRemovingMethods();
         if ( removingMeths != null ) {
            b = new JButton ( ResourceManager.getRemoveIcon() );
            b.setToolTipText( "Remove" );
            b.setActionCommand( "remove" );
            b.addActionListener( this );
            downCont.add( b );
         }

      }

      String[] directMeths = (String[])
         collection.getAttribute("Gui.directCollectionMethods");
      if( directMeths != null ) {
         for( int i=0; i < directMeths.length; i++ ) {
            b = new JButton ( NamingConventions.textForName( directMeths[i] ) );
            b.setActionCommand( "DIRECT:"+ directMeths[i] );
            b.addActionListener( this );
            downCont.add( b );
         }
      }

      add( "North", upperCont );
      add( "South", downCont );

      refreshView();
   }
   
   protected void setCellRenderers(JTable table, Vector headers) {
      for (int i=0; i<headers.size(); i++) {
         FieldItem field = (FieldItem)headers.get(i);
         MetaItem type = (MetaItem)field.getAttribute("rtti.fieldType");
         if (type == null) 
            type = field.getTypeItem();
         ClassItem viewerClass = GuiAC.getTypeViewer(type);
         if (viewerClass != null)
            try {
               table.getColumnModel().getColumn(i).setCellRenderer(
                  (TableCellRenderer)viewerClass.newInstance());
            } catch (Exception e) {
               Log.error("Failed to instanciate TableCellRenderer "+
                         viewerClass.getName()+" for type "+type);
            }
      }
   }

   /**
    * Returns an array of displayable values from a collection of
    * objects.<p>
    *
    * @param collection the collection
    * @return an array of printable values */

   public static String[] toPrintableArray ( Collection collection ) {

      String[] names = null;

      if ( collection == null ) {
         return new String[0];
      }

      names = new String[collection.size()];
      Iterator it = collection.iterator();
      for ( int j = 0; j < names.length; j++ ) {
         try {
            Object cur = it.next();
            if ( cur != null ) {
               if ( cur instanceof Wrappee ) {
                  if ( ((Wrappee)cur).isExtendedBy( ForwardingWrapper.class ) ) {
                     cur = ((ForwardingWrapper)((Wrappee)cur).getWrappers( 
                        ForwardingWrapper.class )[0]). getForwardee();
                  }
                  if ( ((Wrappee)cur).isExtendedBy( BindingWrapper.class ) ) {
                     names[j] = ((BindingWrapper)((Wrappee)cur).getWrappers( 
                        BindingWrapper.class )[0]). getLogicalName() 
                        + ": " + "<not bound yet>";
                     continue;
                  }
               }
               MetaItem showItem = (MetaItem)ClassRepository.get().getClass( cur.getClass() )
                  .getAttribute( "Gui.List.showItem" );
               Object tmp = null;
               if ( showItem != null ) {
                  if ( showItem instanceof MethodItem ) {
                     tmp = ((MethodItem)showItem).invoke( cur, new Object[0] );
                     names[j] = (tmp==null?"null":tmp.toString());
                  } else if ( showItem instanceof FieldItem ) {
                     tmp = ((FieldItem)showItem).get( cur );
                     names[j] = (tmp==null?"null":tmp.toString());
                  } else {
                     names[j] = "<invalid show item>";
                  }
               } else {
                  names[j]= cur.toString();
                  /*if ( (names[j] = NameRepository.get().getName( cur )) == null ) {
                     names[j] = 
                  } else {
                     if ( NamingConventions.getDeclaredMethodByName( cur.getClass(), "toString" ) != null ) {
                        String ts = cur.toString();
                        if ( ts != null ) {
                           names[j] = names[j] + ": " + cur.toString();
                        }
                     }
                     }*/
               }
               if( names[j].length() > 102 ) {
                  names[j] = names[j].substring( 0, 100 ) + "...";
               }
            } else {
               names[j] = "null";
            }
         } catch (Exception e) { 
            Log.error("Failed do build string for index "+j+" : "+e);
            e.printStackTrace();
            names[j] = "null"; 
         }
      }
      return names;      
   }

   /**
    * Builds the headers and and the data in case of a view that
    * displays a table.
    *
    * <p>For details on tables, see the Swing's JTable class.
    *
    * @return a vector that is filled with the headers
    */

   protected Vector buildTableHeader() {
      FieldItem[] fields = null;
      Vector headers = new Vector();
      for (int j=0; j<objects.length; j++) {
         Object o = objects[j];
         if(o==null) {j++; continue;}
         ClassItem clit = ClassRepository.get().getClass(o.getClass());
         if (fields==null) {
            fields = (FieldItem[])collection.getAttribute("Gui.attributesOrder");
            if (fields==null)
               fields = (FieldItem[])clit.getAttribute("Gui.tableAttributesOrder");
            if (fields==null)
               fields = (FieldItem[])clit.getAttribute("Gui.attributesOrder");
            if (fields==null)
               fields = clit.getFields();
         }
         if (fields!=null) {
            FieldItem opposite = (FieldItem)collection.getAttribute("rtti.oppositeRelation");
            Log.trace("associations","opposite of "+collection+" is "+opposite);
            for ( int i = 0; i < fields.length; i++ ) {
               if ( ( fields[i].getAttribute("Gui.visible") == null || 
                      fields[i].getAttribute("Gui.visible").equals("true") ) &&
                    opposite!=fields[i] && 
                    ! (fields[i] instanceof CollectionItem) )
               {
                     headers.add(fields[i]);
               }
            }
            return headers;
         }
      }
      return headers;
   }

   /**
    * Returns an array of the selected objects. The array is empty if
    * no object is selected, but not null.
    **/

   public Object[] getSelected() {
      Object[] selected = null;
      int[] indices = null;
      if ( GuiAC.isTableView(collection) ) {
         indices = table.getSelectedRows();
      } else {
         indices = wordList.getSelectedIndices();
      }
      if (indices != null) {
         selected = new Object[indices.length];
         for (int i=0; i<indices.length; i++)
         {
            selected[i] = objects[indices[i]];
         }
      } else {
         selected = new Object[0];
      }
      return selected;
   }

   /**
    * Handles the actions on this view.
    *
    * <p>On a collection view, the three default possible actions are
    * to open a new view on the selected item, to add a new item to
    * the collection, and to remove the selected item from the
    * collection.
    *
    * @param evt the user event */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("open")) {
         Object[] to_open = getSelected();
         JPanel target = (JPanel) collection.getAttribute( "Gui.Swing.targetContainer" );
         if ( target == null ) {
            for (int i=0; i<to_open.length; i++) {
               ObjectView.addViewFor( to_open[i], parent, parent );
            }
         } else {
            // sending the close event on the views that where opened
            // in the container
            Component[] comps = target.getComponents();
            for(int i=0; i<comps.length; i++) {
               if( comps[i] instanceof View ) {
                  ((View)comps[i]).close();
               }
            }
            // removing the views
            target.removeAll();
            // creating and adding the new view
            for (int i=0; i<to_open.length; i++) {
               target.add( new ObjectView(parent, null, to_open[i]) );
            }
            target.validate();
         }
      } else if (evt.getActionCommand().equals("add")) {
         MethodItem[] methods = collection.getAddingMethods();

         Collaboration collab = JacObject.getCollaboration();
         FieldItem oppositeRole = 
            (FieldItem)collection.getAttribute("rtti.oppositeRelation");
         if (oppositeRole!=null) {
            Log.trace("associations","Setting opposite role for add : "+oppositeRole);
            collab.addTransientAttribute("Gui.oppositeRole", oppositeRole);
         }
         
         try {
         if (methods!=null && methods.length>=1) {
            MethodItem addMethod = methods[0];
            Log.trace("gui","Invoking add method "+addMethod.getName()+" on collection owner");

            if ( collection.getAttribute("Gui.autoCreate") != null ) {
               
               ClassItem type = ClassRepository.get().getClass(
                  addMethod.getParameterTypes()[0]);

               collab.addTransientAttribute("Gui.askForParameters", addMethod);
               collab.addTransientAttribute("Gui.autoCreate", type);
               collab.addTransientAttribute("Gui.view", this);
               collab.addTransientAttribute("Gui.display", parent);
               try {
                  addMethod.invoke( substance, new Object[] { null } );
               } catch (Exception e) {
                  //        Log.error("Instanciation of "+type.getName()+" failed : "+e);
                  e.printStackTrace();
               }
               
            } else {
               GuiAC.invoke((Display)parent,addMethod,substance,null);
            }
         } else {
            methods = collection.getAccessingMethods();
            if (methods!=null && methods.length==1) {
               MethodItem getMethod = methods[0];
               MethodItem method = null;
               Object c;
               try {
                  c = getMethod.invoke(substance,new Object[0]);
               } catch (Exception e) {
                  Log.error("Getting collection with "+getMethod.getName()+
                            "failed : "+e);
                  return;
               }
               ClassItem cl = ClassRepository.get().getClass(c.getClass());
               if (collection.isMap()) 
                  method = cl.getMethod("put(Object,Object)");
               else
                  method = cl.getMethod("add(Object)");
               Log.trace("gui","Invoking "+method.getName()+" on collection itself");
               GuiAC.invoke((Display)parent,method,c,collection);
            }
         }
         } finally {
            collab.addTransientAttribute("Gui.oppositeRole",null);
         }
      } else if (evt.getActionCommand().equals("remove")) {
         Object[] selected = getSelected();
         MethodItem[] methods = collection.getRemovingMethods();
         if (methods!=null && methods.length>=1) {
            // try with the removing method
            MethodItem removeMethod = methods[0];            
            for (int i=0; i<selected.length; i++) {
               try {
                  Log.trace("gui","Invoking remove method "+removeMethod.getName()+" on collection owner");
                  removeMethod.invoke(substance,new Object[] {selected[i]});
               } catch (Exception e) {
                  e.printStackTrace();
               }
            }
         } else {
            // call the collection's method directly
            methods = collection.getAccessingMethods();
            if (methods!=null && methods.length==1) {
               MethodItem getMethod = methods[0];
               MethodItem method = null;
               Object c;
               try {
                  c = getMethod.invoke(substance,new Object[0]);
               } catch (Exception e) {
                  Log.error("Getting collection with "+getMethod.getName()+
                            " failed : "+e);
                  return;
               }
               ClassItem cl = ClassRepository.get().getClass(c.getClass());
               if (!collection.isMap()) {
                  method = cl.getMethod("remove(Object)");
                  Log.trace("gui","Invoking "+method.getName()+" on collection itself");
                  for (int i=0; i<selected.length; i++) {
                     try {
                        method.invoke(c,new Object[] {selected[i]});
                     } catch (Exception e) {
                        Log.error("Removing from collection with "+method.getName()+
                                  " failed : "+e);
                     }
                  }
               }
            }
            
         }
      } else if ( evt.getActionCommand().startsWith("DIRECT:") ) {

         String directMethodName = evt.getActionCommand().substring(7);
         
         int selected = -1;
         if ( GuiAC.isTableView(collection) ) {
            selected = table.getSelectedRow();
         } else {
            selected = wordList.getSelectedIndex();
         }            
         if (selected!=-1) {
            Object to_invoke = objects[selected];
            MethodItem directMethod =
               ClassRepository.get().getClass( to_invoke.getClass() )
               . getMethod( directMethodName );
            if( directMethod != null )
               GuiAC.invoke((Display)parent,directMethod,to_invoke,null);
         }
         
      }
   }

   /**
    * Defines what happens when the selection changes. */

   public void valueChanged(ListSelectionEvent event) {
      if (event.getValueIsAdjusting())
         return;

      Log.trace("gui.events","valueChanged");

      if ( event.getSource() instanceof JList ) {
         JList theList = (JList)event.getSource();
         if (! theList.isSelectionEmpty()) {
            ((CustomizedGUI)parent).onSelection(
               collection,getSelected()[0],null,null);
         }
         
      } else {

         ListSelectionModel lsm = (ListSelectionModel)event.getSource();
         
         int firstIndex = event.getFirstIndex();
         int lastIndex = event.getLastIndex();
         boolean isAdjusting = event.getValueIsAdjusting();
         
         if (lsm.isSelectionEmpty() || isAdjusting) {
            return;
         } else {
            MethodItem handler  = null;
            ((CustomizedGUI)parent).onSelection(
               collection,getSelected()[0],null,null);
         }
      }
   }

   /**
    * Shows a popup if needed.
    *
    * @param e mouse event descriptor */

   public void mousePressed(MouseEvent e) {
      Log.trace("gui.events","CollectionView.mousePressed "+e);
      //      maybeShowPopup(e);
   }

   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseClicked(MouseEvent e) {
      Log.trace("gui.events","CollectionView.mouseClicked "+e);
      if (e.getClickCount() == 2) {
         if ( e.getSource() instanceof JList ) {
            JList theList = (JList)e.getSource();
            if (! theList.isSelectionEmpty()) {
               if ( collection.getAttribute( "Gui.viewOnDoubleClick" ) != null ) {
                  actionPerformed( new ActionEvent(this, 0, "open") );
               }
            }
         }
      }
   }

   public void mouseEntered(MouseEvent e) {
   }

   public void mouseExited(MouseEvent e) {
   }

   public void mouseReleased(MouseEvent e) {
   }
     
   private void maybeShowPopup(MouseEvent e) {
   }
   
 
   // interface View

   public Object getSubstance() {
      return substance;
   }

   public View getParentView() {
      return parentView;
   }

   public void refreshView() {

      Collection c = collection.getActualCollectionThroughAccessor(substance);
      if(c==null) return;
      Log.trace("gui","CollectionView.refreshView "+c.size());
      if ( GuiAC.isTableView(collection) ) {
         if ( c != null) { 
            objects = c.toArray();
            try {
               Arrays.sort(objects, new MyComparator() );
            } catch (ClassCastException e) {
               Log.error("ChoiceFieldEditor : failed to sort array");
               e.printStackTrace();
            }
            headers = buildTableHeader();
            Log.trace("gui.refresh","refreshing collection view with headers = " + headers);
            if( table == null ) return;
            table.setModel(new MyTableModel( objects, headers ) );
            setCellRenderers(table,headers);
         } 

      } else {
         if ( c != null) { 
            objects = c.toArray();
         }
         String[] cvalue = toPrintableArray(c);
         wordList.setListData( cvalue );
      }
      RegisterViews();
   }

   /**
    * Register ourself as a view on all objects of the collection
    */
   protected void RegisterViews() {
      if (headers==null)
         return;
      Log.trace("gui","CollectionView.registerViews "+objects.length);
      for (int i=0; i<objects.length; i++) {
         if (objects[i] instanceof Wrappee) {
            Wrappee wrappee = (Wrappee)objects[i];
            Object[] viewWrappers = (wrappee).getWrappers(ViewControlWrapper.class);
            if (viewWrappers.length>0) {
               ((ViewControlWrapper)viewWrappers[0]).addView( this );
            } else {
               Log.warning("wrappee not wrapped by ViewControlWrapper : "+wrappee);
            }
            for (int j=0; j<headers.size(); j++) {
               FieldItem field = (FieldItem)headers.get(j);
               if (field.isReference()) {
                  wrappee = (Wrappee)field.getThroughAccessor(objects[i]);
                  if (wrappee!=null) {
                     viewWrappers = (wrappee).getWrappers(ViewControlWrapper.class);
                     if (viewWrappers.length>0) {
                        ((ViewControlWrapper)viewWrappers[0]).addView( this );
                     } else {
                        Log.warning("wrappee not wrapped by ViewControlWrapper : "+wrappee);
                     }
                  }
               }
            }
         }
      }
   }

   /**
    * Unregister ourself as a view on all objects of the collection
    */
   protected void unregisterViews() {
      Log.trace("gui","CollectionView.unRegisterViews "+objects.length);
      for (int i=0; i<objects.length; i++) {
         if (objects[i] instanceof Wrappee) {
            Wrappee wrappee = (Wrappee)objects[i];
            Object[] viewWrappers = (wrappee).getWrappers(ViewControlWrapper.class);
            if (viewWrappers.length>0) {
               ((ViewControlWrapper)viewWrappers[0]).removeView( this );
            } else {
               Log.warning("wrappee not wrapped by ViewControlWrapper : "+wrappee);
            }
            for (int j=0; j<headers.size(); j++) {
               FieldItem field = (FieldItem)headers.get(j);
               if (field.isReference()) {
                  wrappee = (Wrappee)field.getThroughAccessor(objects[i]);
                  viewWrappers = (wrappee).getWrappers(ViewControlWrapper.class);
                  if (viewWrappers.length>0) {
                     ((ViewControlWrapper)viewWrappers[0]).removeView( this );
                  } else {
                     Log.warning("wrappee not wrapped by ViewControlWrapper : "+wrappee);
                  }
               }
            }
         }
      }
   }

   public void close() {
      unregisterViews();
   }

   public void refreshViewItem( FieldItem item, MethodItem method, Object[] args ) {
      refreshView();
   }

   class BetterListSelectionModel extends DefaultListSelectionModel {
      JComponent list = null;
      public BetterListSelectionModel( JComponent list ) {
         super();
         this.list = list;
      }
      public JComponent getList() {
         return list;
      }
   }

   class MyTableModel extends AbstractTableModel 
   {
      Object[] objects;
      Vector fields;

      public int getRowCount() {
         return objects.length;
      }
      public int getColumnCount() {
         return fields.size();
      }
      public String getColumnName(int column) {
         return NamingConventions.textForName(getField(column).getName());
      }
      public Class getColumnClass(int column) {
         Log.trace("gui.table","getColumnClass("+column+") -> "+getField(column).getType());
         return Classes.getPrimitiveTypeWrapper(getField(column).getType());
      }
      public Object getValueAt(int row, int column) {
         Log.trace("gui.table","getValueAt("+row+","+column+") -> "+
                   getField(column).getThroughAccessor(objects[row]));
         Class type = getField(column).getType();
         Object value = getField(column).getThroughAccessor(objects[row]);
         return value;
      }
      public boolean isCellEditable(int row, int column) {
         return false;
      }
      public MyTableModel(Object[] objects, Vector fields) {
         this.objects = objects;
         this.fields = fields;
      }
      protected final FieldItem getField(int column) {
         return (FieldItem)fields.get(column);
      }
   }
   public void setFocus(FieldItem field, Object extraOption) {}

   public FieldItem getFieldItem() {
      return collection;
   }
}

