/*
  Copyright (C) 2002 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import jac.util.Log;
import jac.core.Display;
import jac.core.NameRepository;
import jac.core.ACManager;
import jac.core.JacObject;
import jac.core.rtti.AbstractMethodItem;
import jac.core.rtti.ClassRepository;
import jac.core.rtti.MethodItem;
import jac.core.rtti.CollectionItem;
import jac.core.rtti.FieldItem;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Arrays;
import java.util.Enumeration;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;

import java.io.Reader;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.BoxLayout;

/**
 * This class allows the programmer to create simple customized GUIs.
 *
 * <p>The idea is to define sub-panels that can contain views on the
 * objects of the application. The geometric placement of the
 * sub-panel is defined by the <code>setSubPanesGeometry</code> method
 * that has to be called at contruction-time.
 *
 * <p>Once the geometry is chosen, you can tell a pane to contain a
 * view on a given object by using the <code>setObjectForPane</code>
 * or the <code>addReferenceToPane</code> methods.
 *
 * <p>This class must be subclassed by the porgrammer of a JAC
 * application to provide a customized GUI.
 *
 * @version 0.8.1
 * @author <a href="mailto:pawlak@cnam.fr">Renaud Pawlak</a>
 *
 * @see GuiAC
 * @see #setSubPanesGeometry(int,int,String[])
 * @see #setObjectForPane(String,int)
 * @see #addReferenceToPane(String,String,int) */

public class CustomizedGUI extends AbstractSwingDisplay
   implements ActionListener, MouseListener { 

   //GuiAC guiAC = null;

   Vector subPanes = new Vector();
   Hashtable subPaneObjects = new Hashtable();
   Vector splitters = new Vector();
   String displayID;

   Hashtable invalidPanes = new Hashtable(); 

   // WARNING: the following order is very important
   /** The subpanes are separated by an horizontal splitter */
   public static int HORIZONTAL=0;
   /** The upper subpanes are separated by a vertical splitter */
   public static int HORIZONTAL_UP=1;
   /** The down subpanes are separated by a vertical splitter */
   public static int HORIZONTAL_DOWN=2;
   /** The subpanes are separated by a vertical splitter */
   public static int VERTICAL=3;
   /** The left subpanes are separated by an horizontal splitter */
   public static int VERTICAL_LEFT=4;
   /** The right subpanes are separated by an horizontal splitter */
   public static int VERTICAL_RIGHT=5;

   public void setInvalidPane(int changedPane, int invalidPane) {
      invalidPanes.put(subPanes.get(changedPane),subPanes.get(invalidPane));
   }

   //void invokeSelectionHandler(MethodItem handler, CollectionObject object) {
   //}

   /**
    * Called when an object is selected in the GUI
    *
    * @param container the collection which contains the selected object
    * @param selectedObject the selected object
    * @param field if not null, this field will be made visible, and
    * it will get the focus if it has an embedded editor
    */
   public void onSelection(CollectionItem container, 
                           Object selectedObject, 
                           FieldItem field, Object extraOption) {
      Log.trace("gui.events","onSelection("+container+","+selectedObject+")");

      MethodItem handler  = null;
      if ( (handler = GuiAC.getSelectionHandler(container)) != null) {
         try {
            Log.trace("gui.events","invoking selectionHandler "+handler);
            Object result = 
               handler.invokeStatic(new Object[] { this,
                                                   container,
                                                   selectedObject});
            if (handler.getType()!=void.class && result!=null) {
               if (handler.getType()==HandlerResult.class) {
                  HandlerResult hres = (HandlerResult)result;
                  onSelection((CollectionItem)hres.container,hres.object,hres.field,hres.extraOption);
               } else {
                  onSelection(null,result,null,null);
               }
            }
         } catch(Exception e) {
            Log.error("gui: invocation of event handler "+handler+
                      " for onSelection failed with "+e);
            e.printStackTrace();
         }
      }

      JPanel target = (JPanel) container.getAttribute("Gui.Swing.targetContainer");

      Log.trace("gui.events","target="+target);
            
      if( target != null ) {

         if( selectedObject != null ) {
         
            JacObject.getCollaboration()
               .addTransientAttribute(
                  "Gui.view.small",
                  container.getAttribute("Gui.Swing.smallTargetContainer"));
            // creating and adding the new view
            View view=invalidatePane(target,selectedObject);
            if (field!=null && view!=null)
               view.setFocus(field,extraOption);

            JacObject.getCollaboration()
               .addTransientAttribute(
                  "Gui.view.small",null);
         }
         target.validate();
         JPanel invalidPane = (JPanel)invalidPanes.get(target);
         if( invalidPane != null ) {
            invalidatePane(invalidPane,null);
            invalidPane.paint(invalidPane.getGraphics());
         }
      }
   }

   View invalidatePane(JPanel panel,Object selectedObject) {
      Component[] comps = panel.getComponents();
      View view=null;
      if(selectedObject==null || 
         (selectedObject!=null && comps.length>0 && (comps[0] instanceof View) && 
          selectedObject!=((View)comps[0]).getSubstance())) {
         for(int i=0; i<comps.length; i++) {
            if( comps[i] instanceof View ) {
               ((View)comps[i]).close();
            }
         }
         // removing the views
         panel.removeAll();
      }
      comps = panel.getComponents();
      if(comps.length==0 && selectedObject!=null) {
         view = new ObjectView(this, null, selectedObject);
         panel.add( (Component)view );
      }
      return view;
   }

   /**
    * Gets the number of sub-panes in the main window of the GUI.
    *
    * @return the number of panels */

   public int getSubPanesCount() {
      return subPanes.size();
   }

   /**
    * Sets the geometric arrangement of the panes.
    *
    * <p><ul><li><code>subPanesCount == 4 && geometry == VERTICAL</code>:
    * <pre>
    *    +-------+
    *    | 0 | 2 |
    *    +---|---+
    *    | 1 | 3 |
    *    +-------+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 4 && geometry == HORIZONTAL</code>:
    * <pre>
    *    +-------+
    *    | 0 | 1 |
    *    +-------+
    *    | 2 | 3 |
    *    +-------+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 3 && geometry == HORIZONTAL_UP</code>:
    * <pre>
    *    +-------+
    *    | 0 | 1 |
    *    +-------+
    *    |   2   |
    *    +-------+
    * </pre>
    * </li>
    * </li>
    * <li><code>subPanesCount == 3 && geometry == HORIZONTAL_DOWN</code>:
    * <pre>
    *    +-------+
    *    |   2   |
    *    +-------+
    *    | 0 | 1 |
    *    +-------+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 3 && geometry == VERTICAL_LEFT</code>:
    * <pre>
    *    +---+---+
    *    | 0 |   |
    *    +---| 2 |
    *    | 1 |   |
    *    +---+---+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 3 && geometry == VERTICAL_RIGHT</code>:
    * <pre>
    *    +---+---+
    *    |   | 0 |
    *    | 2 |---+
    *    |   | 1 |
    *    +---+---+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 2 && geometry == VERTICAL</code>:
    * <pre>
    *    +---+---+
    *    |   |   |
    *    | 0 | 1 |
    *    |   |   |
    *    +---+---+
    * </pre>
    * </li>
    * <li><code>subPanesCount == 2 && geometry == HORIZONTAL</code>:
    * <pre>
    *    +-------+
    *    |   0   |
    *    +-------+
    *    |   1   |
    *    +-------+
    * </pre>
    * </li>
    *
    * @param subPanesCount the number of subpanes in the window
    * @param geometry the geometry = <code>VERTICAL || HORIZONTAL ||
    * VERTICAL_LEFT || VERTICAL_RIGHT || HORIZONTAL_UP ||
    * HORIZONTAL_DOWN</code> (see above)
    * @param scrollings a set of string that tells if the sub-panes
    * must be srollable or not */

   public void setSubPanesGeometry(int subPanesCount, int geometry, 
                                   String[] scrollings) {
      splitters.clear();
      subPanes.clear();
      getContentPane().removeAll();
      Vector subScrollPanes = new Vector();
      for(int i=0;i<subPanesCount;i++) {
         JPanel subPane = new JPanel();
         JScrollPane subScrollPane = null;
         if( scrollings[i].equals("true") ) {
             subScrollPane = new JScrollPane(subPane=new JPanel());
         }         
         subPane.setLayout( new BoxLayout( subPane, BoxLayout.Y_AXIS ) );
         subPanes.add( subPane );
         if( subScrollPane == null ) {
            subScrollPanes.add(subPane);
         } else {
            subScrollPanes.add( subScrollPane );
         }
      }
      if( subPanesCount == 1 ) {
         getContentPane().add((Component)subScrollPanes.get(0));
      } else if( subPanesCount == 2 ) {
         // create a split panel
         JSplitPane splitPane  = new JSplitPane(
            geometry==HORIZONTAL?JSplitPane.VERTICAL_SPLIT:JSplitPane.HORIZONTAL_SPLIT,
            (Component)subScrollPanes.get(0), (Component)subScrollPanes.get(1));
         splitPane.setOneTouchExpandable(true);
         splitters.add(splitPane);
         getContentPane().add(splitPane);
      } else if( subPanesCount == 4 ) {
         // create two sub-split panels
         JSplitPane subSplitPane1  = new JSplitPane(
            geometry==HORIZONTAL?JSplitPane.HORIZONTAL_SPLIT:JSplitPane.VERTICAL_SPLIT,
            (Component)subScrollPanes.get(0), (Component)subScrollPanes.get(1));
         subSplitPane1.setOneTouchExpandable(true);
         JSplitPane subSplitPane2  = new JSplitPane(
            geometry==HORIZONTAL?JSplitPane.HORIZONTAL_SPLIT:JSplitPane.VERTICAL_SPLIT,
            (Component)subScrollPanes.get(2), (Component)subScrollPanes.get(3));
         subSplitPane2.setOneTouchExpandable(true);
         // create a split panel
         JSplitPane splitPane  = new JSplitPane(
            geometry==HORIZONTAL?JSplitPane.VERTICAL_SPLIT:JSplitPane.HORIZONTAL_SPLIT,
            subSplitPane1, subSplitPane2);
         splitPane.setOneTouchExpandable(true);
         splitters.add(splitPane);
         splitters.add(subSplitPane1);
         splitters.add(subSplitPane2);
         getContentPane().add(splitPane);
      } else if( subPanesCount == 3 ) {
         // create one sub-split panels
         JSplitPane subSplitPane  = new JSplitPane(
            geometry<VERTICAL?JSplitPane.HORIZONTAL_SPLIT:JSplitPane.VERTICAL_SPLIT,
            (Component)subScrollPanes.get(0), (Component)subScrollPanes.get(1));
         subSplitPane.setOneTouchExpandable(true);
         // create a split panel
         int type;
         int splitType;
         if(geometry<VERTICAL) {
            type = geometry - HORIZONTAL;
            splitType = JSplitPane.VERTICAL_SPLIT;
         } else {
            type = geometry - VERTICAL;
            splitType = JSplitPane.HORIZONTAL_SPLIT;
         }
         JSplitPane splitPane = new JSplitPane(
            splitType,
            type==1?subSplitPane:(Component)subScrollPanes.get(2),
            type==1?(Component)subScrollPanes.get(2):subSplitPane);
         splitPane.setOneTouchExpandable(true);
         splitters.add(splitPane);
         splitters.add(subSplitPane);
         getContentPane().add(splitPane);
      }
   }

   /**
    * Gets a sub-panel from its ID.
    *
    * @param paneId the ID */

   public JPanel getPane( int paneId ) {
      return (JPanel) subPanes.get(paneId);
   }

   /**
    * Sets the object that should be opened in the given panel id.
    *
    * @param name the object's name as registered in the name
    * repository
    * @param paneId the panel id (see the geometry to know its
    * placement)
    *
    * @see #setSubPanesGeometry(int,int,String[]) */

   public void setObjectForPane( String name, int paneId ) {
      subPaneObjects.put(name,subPanes.get(paneId));
   }

   /**
    * Tells a referenced object to open in a given panel when a view
    * is asked by the user (instead of opening in a popup).
    *
    * @param className the class the reference belongs to
    * @param referenceName the name of the reference
    * @param paneId the panel id where the view must be opened */

   public void addReferenceToPane( String className, 
                                   String referenceName,
                                   int paneId ) {
      GuiAC.setTargetContainer(className,referenceName,subPanes.get(paneId));
   }

   /**
    * Sets the dimensions and position of the window regarding to the
    * main screen.
    *
    * @param left left-border pixel
    * @param up upper-border pixel
    * @param width in percentage regarding the screen
    * @param height in percentage regarding the screen */

   public void setPosition(int left, int up, int width, int height) {
      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      setBounds(new Rectangle(
         left,up,
         (int)(((float)screenDim.getWidth()*width)/100),
         (int)(((float)screenDim.getHeight()*height)/100)));
   }

   /**
    * Sets a splitter location.
    *
    * <p>The splitter is referenced by its index going from the
    * front-end splitter to the back-end splitters. For instance, in
    * the case of a 3 sub-panel window, the 0 index references the
    * splitter that splits the main window in two, the 1 index, the
    * one that splits the half-window in two other smaller parts.
    *
    * @param splitterId the splitter's index
    * @param location the position in pixel, regarding to the top/left
    * component, a negative value means that the splitter should be
    * set at the preferred sized of the inner components */

   public void setSplitterLocation( int splitterId, int location ) {
      JSplitPane splitter = (JSplitPane)splitters.get(splitterId);
      splitter.setDividerLocation( location );
   }

   JPanel statusBar; 
   
   /**
    * Adds a status bar to the GUI.
    */

   public void addStatusBar() {
      statusBar = new JPanel();
      statusBar.add(new JLabel(" "));
      getContentPane().add("South",statusBar);
   }
    
   public void showStatus(String message) {
      if( statusBar == null ) {
         System.out.println(message);
      } else {
         ((JLabel)statusBar.getComponent(0)).setText(message);
      }
   }

   boolean subDisplay=false;

   /**
    * Creates a new customized GUI.
    *
    * <p>When subclassing, a typical implementation of the constructor is:
    *
    * <pre class=code>
    * super()
    * // initialization calls such as setSubPanesGeometry, setObjectForPane
    * // addReferenceToPane, setPosition
    * ...
    * show();
    * </pre>
    *
    * @see #setSubPanesGeometry(int,int,String[])
    * @see #setObjectForPane(String,int)
    * @see #addReferenceToPane(String,String,int)
    * @see #setPosition(int,int,int,int)
    * @see #setSplitterLocation(int,int) */
   
   public CustomizedGUI() {
      super();
      unregisterOnClose=true;
      init();
      pack();
   }
   public CustomizedGUI(boolean unregisterOnClose) {
      super();
      this.unregisterOnClose=unregisterOnClose;
      init();
      pack();  
   }

   boolean hidden=false;
   Point actualLocation;
   boolean unregisterOnClose;
   protected void init() {
      //if (unregisterOnClose && !subDisplay) {
      addWindowListener ( new WindowAdapter () {
            public void windowClosing( WindowEvent e ) {
               System.out.println("closing window");
               close();
               // notify the system that this display has been closed
               if( unregisterOnClose ) {
                  ((ACManager)ACManager.get()).whenCloseDisplay((Display)e.getWindow());
               }
               if( subDisplay ) {
                  // a dirty hack since Swing is not able to hide the
                  // window properly (with Window Maker as a window
                  // manager
                  hidden=true;
                  actualLocation = getLocation(new Point());
                  setLocation(new Point(2000,0));
                  throw new RuntimeException();
               } else {
                  e.getWindow().dispose();
               }
            }
         }
      );
   }

   public void showAsSubDisplay() {
      if( hidden ) {
         setLocation(actualLocation);
         toFront();
      } else {
         subDisplay = true;
         unregisterOnClose = false;
         super.show();
         applicationStarted();
      }
   }

   Hashtable menus = new Hashtable();
   Hashtable menuItems = new Hashtable();

   public void setMenuBar(String[] smenus) {
      JMenuBar menuBar = new JMenuBar();
      for(int i=0;i<smenus.length;i++) {
         JMenu menu = new JMenu(smenus[i]);
         menus.put(smenus[i],menu);
         menuBar.add(menu);
      }
      setJMenuBar(menuBar);
   }

   public void addMenuItem(String menu,String name,
                           String className,String methodName) {
      JMenu parent = (JMenu)menus.get(menu);
      if( parent != null ) {
         JMenuItem menuItem = new JMenuItem(name);
         menuItems.put(name,new String[] {className,methodName});
         parent.add(menuItem);
         menuItem.addActionListener(this);
      }
   }

   public void addMenuSeparator(String menu) {
      JMenu parent = (JMenu)menus.get(menu);
      if( parent != null ) {
         //JMenuItem menuItem = new JMenuItem(name);
         //menuItems.put(name,new String[] {className,methodName});
         parent.addSeparator();
         //menuItem.addActionListener(this);
      }
   }

   /**
    * Shows the window on the screen.
    */

   public void show() {
      validate();
      super.show();
   }
   
   /** Do nothing. */
   public void mouseClicked(MouseEvent me){}
   /** Do nothing. */
   public void mousePressed(MouseEvent e) {}
   /** Do nothing. */
   public void mouseReleased(MouseEvent e) {}
   /** Do nothing. */
   public void mouseExited(MouseEvent me){}
   /** Do nothing. */
   public void mouseEntered(MouseEvent me){}   
   /** Handles menus actions. */
   public void actionPerformed(ActionEvent ae){
      String command = ae.getActionCommand();
      if( menuItems.get(command) != null ) {
         String className = ((String[])menuItems.get(command))[0];
         String methodName = ((String[])menuItems.get(command))[1];
         try {
            //JacObject.getCollaboration().addTransientAttribute("Gui.display",this);
            //Object ret = Class.forName(className).getMethod(methodName,new Class[0])
            //   .invoke(null,new Object[0]);
            invoke(ClassRepository.get().getClass(className).getMethod(methodName),null,null);
            //if( ret != null ) {
            //   ret.invokeRoleMethod("
         } catch( Exception e ) {
            e.printStackTrace();
         }
      }
   }
   
   // Implements the Display interface

   public void show( Object object ) {
      Log.trace("display","show("+object.getClass().getName()+")");
      if (object instanceof Throwable) {
         showMessage(((Throwable)object).getMessage(),"Error",false,false,true);
         ((Throwable)object).printStackTrace();
      } else if (object instanceof Reader) {
         saveStreamToFile((Reader)object);
      } else if (object.getClass().isArray()) {
         CustomizedGUI gui = new CustomizedGUI(true);
         gui.setSubPanesGeometry(2, HORIZONTAL, new String[] {"false","true"});
         CollectionWrapper c = new CollectionWrapper(Arrays.asList((Object[])object));
         gui.setObjectForPane(NameRepository.get().getName(c),0);
         gui.addReferenceToPane("jac.aspects.gui.CollectionWrapper","collection",1);
         gui.applicationStarted();
         gui.setPosition(0,0,60,60);
         gui.show();
      } else {
         String name = NameRepository.get().getName(object);
         Log.trace("display","name of "+object.getClass().getName()+" is "+name);
         if( name != null ) {
            JPanel pane = (JPanel)subPaneObjects.get( name );
            if( pane != null ) {
               Log.trace("display","got a pane for "+name);
               if( pane.getComponentCount() > 0 ) {
                  for(int i=0;i<pane.getComponentCount();i++) {
                     ((View)pane.getComponent(i)).close();
                  }
                  pane.removeAll();
               }
               pane.add( new ObjectView( this, null, object ) );
               pane.validate();
               return;
            } else {
               Log.trace("display","Haven't got a pane for "+name);
               ObjectView.addViewFor( object, this, this );
            }
         } else {
            ObjectView.addViewFor( object, this, this );
         }
      }

   }

   // implements the display interface
 
   public void applicationStarted() {
      
      // open the views on the objects in the right subpanes
      for (Enumeration e=subPaneObjects.keys();e.hasMoreElements();) {
         String name = (String)e.nextElement();
         Object o = NameRepository.get().getObject( name );
         if( o != null ) {
            JPanel pane = (JPanel)subPaneObjects.get( name );
            if( pane.getComponentCount() == 0 ) {
               pane.add( new ObjectView( this, null, o ) );
               pane.validate();
            }
         }
      }
      
      // refresh the views that need it
      Iterator it = subPanes.iterator();
      while(it.hasNext()) {
         JPanel p = (JPanel)it.next();
         if( p.getComponentCount()==1 && 
             (p.getComponent(0) instanceof TreeView)) {
            ((View)p.getComponent(0)).refreshView();
         }
      }
   }

   public void close() {
      Iterator it = subPanes.iterator();
      while( it.hasNext() ) {
         // cleanly close all the current views
         try {
            ((View)((JPanel)it.next()).getComponent(0)).close();
         } catch( Exception ex) {}
      }
   }
      
      /*
      // this method can be called several times if another GUI is
      // running... bug??
      if( gm != null ) return;

      // open a view on the contact repository if created and 
      // add it to the upper panel
      Object main=null;
      if( (main=NameRepository.get().getObject(upperPaneObject))!=null ) {
         //((Wrappee)main).forceTreatment();
         gm = new ObjectView( this, null, main );
         upperPane.add(gm);
         splitPane.resetToPreferredSizes();
         upperPane.validate();
      }
      */
}



















