/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.gui;

import java.util.*;
import java.text.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import jac.core.*;
import jac.core.rtti.*;
import jac.util.*;

/**
 * This class defines a Swing component editor for primitive fields in
 * objects.
 *
 * <p>By default this view constructs an embedded <code>JTextField</code>
 * containing the string representation of the field value. However,
 * the field can be attributed to be displayed with a customized
 * rendering by the GUI aspect component.
 *
 * <p>When the field of an object is a collection of a reference, then
 * other types of wiews are used.
 *
 * @see GuiAC
 * @see CollectionView
 * @see ReferenceView */

public class FieldEditor extends JPanel 
   implements FieldItemView, ActionListener, FocusListener {

   Window parent;
   Object substance;
   FieldItem field;
   Object value = null;
   MethodItem setter = null;
   Component component;
   /**
    * Constructs a new field editor.
    *
    * @param parent the Swing dialog parent
    * @param substance the object the viewed field belongs to
    * @param field the field that is showed by this view */
   
   public FieldEditor(Window parent, 
                      Object substance, FieldItem field) {
      this.parent = parent;
      this.substance = substance;
      this.field = field;

      int width = -1;
      int columns = -1;
      int height = -1;
      int rows = -1;

      if(JacObject.getCollaboration().getAttribute("Gui.view.small")!=null) {

         Integer tmp = (Integer)field.getAttribute("Gui.editor.smallWidth");
         if( tmp != null ) {
            width = tmp.intValue();
         }
         if(width == -1) {
            if(field.getAttribute("rtti.fieldType")
               !=ClassRepository.get().getObject("text")) {
               width=8;
            } else {
               width=200;
            }
         }

         tmp = (Integer)field.getAttribute("Gui.editor.smallCols");
         if( tmp != null ) {
            columns = tmp.intValue();
         }
         tmp = (Integer)field.getAttribute("Gui.editor.smallHeight");
         if( tmp != null ) {
            height = tmp.intValue();
         }
         tmp = (Integer)field.getAttribute("Gui.editor.smallRows");
         if( tmp != null ) {
            rows = tmp.intValue();
         }

      } else {

         Integer tmp = (Integer)field.getAttribute("Gui.editor.width");
         if( tmp != null ) {
            width = tmp.intValue();
         }
         tmp = (Integer)field.getAttribute("Gui.editor.cols");
         if( tmp != null ) {
            columns = tmp.intValue();
         }
         tmp = (Integer)field.getAttribute("Gui.editor.height");
         if( tmp != null ) {
            height = tmp.intValue();
         }
         tmp = (Integer)field.getAttribute("Gui.editor.rows");
         if( tmp != null ) {
            rows = tmp.intValue();
         }
      }


      value = field.getThroughAccessor( substance );

      MethodItem writingMethods[] = field.getWritingMethods();
      if (writingMethods!=null && writingMethods.length>0) {
         ValueEditor ve = RenderFactory.getEditorComponent(
            parent,substance,writingMethods[0],0);
         if(width!=-1) ve.setWidth(width);
         if(height!=-1) ve.setHeight(height);
         component = (Component)ve;
         if (GuiAC.getCommitOnFocusLost())
            ve.getValueComponent().addFocusListener(this);
         add(component);
      }
      setLayout(new FlowLayout(FlowLayout.CENTER,0,0));
      refreshView();
   }

   /**
    * Returns the current value in the editor.
    *
    * @return the value */

   Object getValue() {
      if( component instanceof ValueEditor ) {
         return ((ValueEditor)component).getValue();
      } else {
         return ((JTextField)component).getText();
      }
   }

   /**
    * Set the current value in the editor.
    */

   void setValue(Object newValue) {
      if( component instanceof ValueEditor ) {
         ((ValueEditor)component).setValue(newValue);
      } else {
         ((JTextField)component).setText((String)newValue);
      }
   }

   /** do nothing */
   public void focusGained(FocusEvent e) {
      Log.trace("gui.focus","focus gained event");
   }

   /** Update the object's field if needed. */
   public void focusLost(FocusEvent e) {
      Log.trace("gui.focus","focus lost event");
      if( valueHasChanged() ) {
         Log.trace("gui.focus","value changed");
         field.setThroughWriter(substance,getValue());
      }
   }

   /**
    * Tells wether the value in the editor was changed
    */
   boolean valueHasChanged() {
      Object newValue = getValue();
      boolean ret;
      if( value == null &&  newValue != null ) {
         ret = true;
      } else if( value == null ) {
         ret = false;
      } else {
         ret = ! value.equals(getValue());
      }
      Log.trace("gui.editor","valueHasChanged("+field.getName()+") "+
                value+" / "+newValue+" -> "+ret);
      return ret;
   }

   /**
    * Handles the actions on this view.
    *
    * <p>On a field view, the two default possible actions are to open
    * a new view on the field value object, or to edit the field
    * value.
    *
    * @param evt the user event */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("open")) {
         if (value!=null) {
            ObjectView.addViewFor( value, parent, parent );
         }
      } else if (evt.getActionCommand().equals("edit")) {
         GuiAC.invoke((Display)parent,setter,substance,null);
      }
   }

   /**
    * Returns the field edited by the current editor.
    *
    * @return the edited field */
 
   public FieldItem getField() {
      return field;
   }

   // View interface

   public Object getSubstance() {
      return substance;
   }

   public View getParentView() {
      return null;
   }

   public void refreshViewItem( FieldItem item, MethodItem method, Object[] args ) {
      refreshView();
   }

   public void refreshView( ) {
      value = field.getThroughAccessor(substance);
      if (component.getClass() == JTextField.class) {
         ((JTextField)component).setText(value!=null?value.toString():"");
      } else {
         ((ValueEditor)component).setValue(value);
      }
   }

   public void close() {}

   public void setFocus(FieldItem field, Object extraOption) {
      if (component instanceof ValueEditor) {
         component.requestFocus();
         ((ValueEditor)component).onSetFocus(extraOption);
      }
   }

   public FieldItem getFieldItem() {
      return field;
   }
}
