/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.gui;

import java.util.*;
import java.text.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import jac.core.*;
import jac.core.rtti.*;
import jac.util.*;

/**
 * This class defines a Swing component view for primitive fields in
 * objects.
 *
 * <p>By default this view constructs an embedded <code>JLabel</code>
 * containing the string representation of the field value. However,
 * the field can be attributed to be displayed with a customized
 * rendering by the GUI aspect component.
 *
 * <p>When the field of an object is a collection of a reference, then
 * other types of wiews are used.
 *
 * @see GuiAC
 * @see CollectionView
 * @see ReferenceView */

public class FieldView extends JPanel implements FieldItemView, ActionListener {

   Window parent;
   Object substance;
   FieldItem field;
   Object object = null;
   MethodItem setter = null;
   Component component;

   /**
    * Constructs a new field view.
    *
    * @param parent the Swing dialog parent
    * @param substance the object the viewed field belongs to
    * @param field the field that is showed by this view */
   
   public FieldView(Window parent, 
                    Object substance, FieldItem field) {
      this.parent = parent;
      this.substance = substance;
      this.field = field;

      object = field.getThroughAccessor( substance );
      
      MetaItem type = (MetaItem)field.getAttribute("rtti.fieldType");
      if (type==null) {
         if (object!=null)
            type = ClassRepository.get().getClass(object);
         else
            type = ClassRepository.get().getClass(field.getType());
      }
      ClassItem viewerClass = GuiAC.getTypeViewer(type);
      if (viewerClass != null) {
         try {
            component = (Component)viewerClass.newInstance();
            ((ValueViewer)component).setValue(object);
         } catch (Exception e) {
            Log.error("Instanciation of viewer component "+viewerClass.getName()+" failed");
            e.printStackTrace();
         }
      } else {
         JLabel l = new JLabel ( object!=null ? object.toString() : "null");
         component = l;
      }

      add(component);

      JButton b;
      if ( field.getWritingMethods() != null ) {
         // we should be more specific
         setter = field.getWritingMethods()[0];
         if ( setter != null ) {
            add( Box.createHorizontalStrut( 20 ) );
            b = new JButton ( ResourceManager.getEditIcon() );
            b.setToolTipText("Edit");
            b.setActionCommand("edit");
            b.addActionListener(this);
            b.setMargin(new Insets(1,1,1,1));
            add( b );
         }
      }
      setLayout(new FlowLayout(FlowLayout.CENTER,0,0));
      
      refreshView();
   }

   /**
    * Handles the actions on this view.
    *
    * <p>On a field view, the two default possible actions are to open
    * a new view on the field value object, or to edit the field
    * value.
    *
    * @param evt the user event */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("open")) {
         if (object!=null) {
            ObjectView.addViewFor( object, parent, parent );
         }
      } else if (evt.getActionCommand().equals("edit")) {
         GuiAC.invoke((Display)parent,setter,substance,null);
      }
   }

   // View interface

   public Object getSubstance() {
      return substance;
   }

   public View getParentView() {
      return null;
   }

   public void refreshViewItem( FieldItem item, MethodItem method, Object[] args ) {
      Log.trace("gui.viewControl","refreshViewItem("+item+") on "+getClass().getName());
      refreshView();
   }

   public void refreshView( ) {
      Log.trace("gui.viewControl","refreshViewItem on "+getClass().getName());
      object = field.getThroughAccessor(substance);
      if (component.getClass() == JLabel.class) {
         ((JLabel)component).setText(object!=null?object.toString():"null");
      } else {
         ((ValueViewer)component).setValue(object);
      }
   }

   public void close() {}

   public void setFocus(FieldItem field, Object extraOption) {
   }

   public FieldItem getFieldItem() {
      return field;
   }
}
