/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.gui;

import java.lang.reflect.*;
import java.util.*;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.awt.Font;
import javax.swing.UIManager;
import javax.swing.ImageIcon;
import javax.swing.JTextArea;
import javax.swing.JPanel;

import jac.core.*;
import jac.core.rtti.*;
import jac.core.rtti.NoSuchFieldException;
import jac.util.*;
import jac.core.dist.Distd;
import jac.core.dist.RemoteContainer;

/**
 * This aspect component implements a very simple GUI that allows the
 * user to browse the named object of the system and call methods on
 * them.
 *
 * <p>It implements a MVC design pattern where the controllers are
 * implemented by <code>ViewWrapper</code> instances.<p>
 *
 * @see ViewControlWrapper
 * @see View
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a>
 */

public class GuiAC extends AspectComponent implements GuiConf {

   /** Use this constant to tell that a component is displayed with
       its default view. */
   public static String DEFAULT_VIEW = "";

   /** Use this constant to tell that a component is displayed /
       modified with a file chooser box. */
   public static String FILE_CHOOSER_VIEW = "fileChooser";
   
   /** Stores the language of the GUI (0 = english, 1 = french). */
   public static int LANGUAGE = 0;

   static InputWrapper inputWrapper = new InputWrapper();
   static AdminWrapper adminWrapper = new AdminWrapper();

   /**
    * Stores the displays of the GUI (class -> display). */
   
   protected static Hashtable displays = new Hashtable();

   /**
    * Gets a display from its ID.
    *
    * @param displayID the id of the display
    * @see Display  */

   public static final Display getDisplay( String displayID ) {
      return (Display) displays.get( displayID );
   }

   /**
    * This configuration method creates a new display and registers
    * its.
    *
    * @param name the display name
    * @param displayType the display type */

   public final Display createDisplay( String displayID, Class displayType ) {
      Display ret = null;
      Log.trace("display","createDisplay("+displayType.getName()+")");
      try {
         ret = (Display) displayType.newInstance();
         ret.setDisplayID( displayID );
         addDisplay( displayID, ret );
      } catch ( Exception e ) {
         e.printStackTrace();
      }
      return ret;
   }

   /**
    * Adds a new display for this GUI.
    *
    * @param newDisplay the display to add
    * @see Display */

   public static void addDisplay( String displayID, Display newDisplay ) {
      Log.trace("display","addDisplay("+newDisplay.getClass().getName()+")");
      displays.put( displayID, newDisplay );
   }


   /**
    * Removes a display for this GUI.
    *
    * @param display the display to remove
    * @see Display */

   public static void removeDisplay( Display display ) {
      if ( display.getDisplayID() != null )
         displays.remove( display.getDisplayID() );
   }

   /**
    * This configuration method allows the user to register a new
    * display.
    * 
    * <p>It instantiates a new display from its class name and
    * registers its.
    * 
    * @param name the display name
    * @param displayType the display type */

   public void registerDisplay( String name, String displayType ) {
      try { 
         Class cl = Class.forName(displayType);
         if( getDisplay(name) != null ) {
            Log.warning("display "+name+" is already launched");
            return;
         }
         createDisplay( name, cl );
      } catch ( Exception e ) {
         e.printStackTrace();
      }
   }

   public void registerCustomized( String name ) {
      try { 
         //Class cl = Class.forName("jac.aspects.gui.CustomizedGUI");
         if( getDisplay(name) != null ) {
            Log.warning("display "+name+" is already launched");
            return;
         }
         createDisplay( name, CustomizedGUI.class );         
      } catch ( Exception e ) {
         e.printStackTrace();
      }
   }

   public void showCustomized( String gui, String host ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setSubPanesGeometry: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      if( host == null || 
          RemoteContainer.resolve(host).getName().equals(Distd.getLocalContainerName()) ) 
      {
         cgui.show();
      } else {
         displays.remove(gui);
      }
      ((ACManager)ACManager.get()).afterApplicationStarted();
   }

   public void showCustomized( String gui ) {
      showCustomized(gui,null);
   }

   public void addStatusBar( String gui ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("addStatusBar: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.addStatusBar();
   }

   public void setSubPanesGeometry(String gui, int subPanesCount, 
                                   String geometry) {
      String[] scrollings = new String[subPanesCount];
      for(int i=0;i<subPanesCount;i++) {
         scrollings[i]="true";
      }
      setSubPanesGeometry(gui,subPanesCount,geometry,scrollings);
   }

   public void setSubPanesGeometry(String gui, int subPanesCount, 
                                   String geometry,
                                   String[] scrollings) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setSubPanesGeometry: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      int geom = 0;
      if(geometry.equals("VERTICAL")) geom = CustomizedGUI.VERTICAL;
      if(geometry.equals("VERTICAL_LEFT")) geom = CustomizedGUI.VERTICAL_LEFT;
      if(geometry.equals("VERTICAL_RIGHT")) geom = CustomizedGUI.VERTICAL_RIGHT;
      if(geometry.equals("HORIZONTAL")) geom = CustomizedGUI.HORIZONTAL;
      if(geometry.equals("HORIZONTAL_UP")) geom = CustomizedGUI.HORIZONTAL_UP;
      if(geometry.equals("HORIZONTAL_DOWN")) geom = CustomizedGUI.HORIZONTAL_DOWN;
      cgui.setSubPanesGeometry(subPanesCount,geom,scrollings);
   }

   public void setObjectForPane( String gui, String name, int paneId ) {      
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setObjectForPane: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setObjectForPane(name,paneId);
   }

   public void setInvalidPane( String gui,
                               int changedPane, int invalidPane ) {      
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setInvalidPane: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setInvalidPane(changedPane,invalidPane);
   }

   public void addReferenceToPane( String gui,
                                   String className, 
                                   String referenceName,
                                   int paneId ) {
      addReferenceToPane( gui, className, referenceName, null, paneId);
   }

   public void addReferenceToPane( String gui,
                                   String className, 
                                   String referenceName,
                                   String small,
                                   int paneId ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("addReferenceToPane: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      ClassItem cli = ClassRepository.get().getClass( className );
      MetaItem mi = cli.getField( referenceName );
      if ( small != null ) {
         mi.setAttribute( "Gui.Swing.smallTargetContainer", small );
      }
      cgui.addReferenceToPane(className,referenceName,paneId);
      
   }

   public void setPosition(String gui, int left, int up, 
                           int width, int height) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setPosition: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setPosition(left,up,width,height);
   }

   public void setMenuBar(String gui, String[] smenus) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setMenuBar: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setMenuBar(smenus);
   }

   public void addMenuItem(String gui, String menu,String name,
                           String className,String methodName) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setMenuBar: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.addMenuItem(menu,name,className,methodName);
   }

   public void addMenuSeparator(String gui, String menu) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setMenuSeperator: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.addMenuSeparator(menu);
   }

   public void setSplitterLocation( String gui, 
                                    int splitterId, int location ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setSplitterLocation: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setSplitterLocation( splitterId, location );
   }

   public void setTitle( String gui, String title ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setTitle: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      cgui.setTitle( title );
   }

   public void setLoggingMethod( String gui, 
                                 String objects, 
                                 String classes, 
                                 String methods, 
                                 int paneID ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("setLoggingMethod: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      JPanel pane = cgui.getPane(paneID);
      JTextArea textArea = new JTextArea();
      pane.add(textArea);
      pointcut( objects, classes, methods, 
                new LoggingWrapper(textArea), 
                "logIntoConsole");
   }

   /**
    * The wrapper that logs methods into a Swing console. */
   public class LoggingWrapper extends Wrapper {
      JTextArea console;
      public LoggingWrapper(JTextArea console) {
         this.console = console;
      }
      public Object logIntoConsole() {
         if( console != null ) {
            console.append( (String)arg(0) );
            console.setCaretPosition( console.getText().length() );
         }
         return proceed();
      }
   }

   public void addTreeView( String gui, 
                            String pathDef,
                            boolean showRelations,
                            int paneID ) {
      CustomizedGUI cgui = (CustomizedGUI) getDisplay( gui );
      if( cgui == null ) {
         Log.error("addTreeView: "+gui+
                   " has not been declared with registerCustomized");
         return;
      }
      JPanel pane = cgui.getPane(paneID);
      TreeView treeView = new TreeView(cgui,null,null,pathDef,showRelations);
      pane.add(treeView);
   }

   public void defineResource(String type, String name, String path ) {
      if( type.equals("ICON") ) {
         ResourceManager.setIcon(name,path);
      }
   }
   
   public void setIcon(ClassItem classItem, String name) {
      classItem.setAttribute("Gui.attachedIcon", 
                             ResourceManager.getIcon(name));
   }
   
   public void setMenu(ClassItem classItem, String[] menu) {
      classItem.setAttribute("Gui.attachedMenu", 
                             menu);
   }

   /**
    * Catch this event and remove the display from the handled list if
    * exists.
    *
    * @param display the closing display */

   public void whenCloseDisplay(Display display) {
      removeDisplay(display);
   }

   /**
    * The default constructor for the GUI aspect component.
    *
    * <p>By default it defines the visible property to false for some
    * JAC system items.
    *
    */

   public GuiAC() {
      try {
         setDefaults();
      } catch (Exception e) {
         Log.error("Failed to initialize the GuiAC : "+e);
      }
   }

   /**
    * Defines the visible property to false for some JAC system items.  
    */
   protected void setDefaults() {
      setVisible( "jac.core.NameRepository", "names", "false" );
      setVisible( "jac.core.ACManager", new String[] { 
         "registering", "orderedObjects", "orderedNames", "allWoven", "nonWovenACs" }, "false" );

      setVisible( "jac.core.dist.Topology", "nameReps", "false" );
      setVisible( "jac.core.dist.Topology", "getContainerIndex", "false" );
      setVisible( "jac.core.dist.Topology", "getContainerIndexes", "false" );
      setVisible( "jac.core.dist.Topology", "countContainers", "false" );
      setVisible( "jac.core.dist.Topology", "getContainer", "false" );
      setVisible( "jac.core.dist.Topology", "isContainer", "false" );
      setVisible( "jac.core.dist.Topology", "createNameReps", "false" );

      setTableView( "jac.core.dist.Topology", "containers" );

      setDirectCollectionMethods( 
         "jac.core.dist.Topology", "containers", new String[] { "launchRemoteGUI" } );

      setVisible( "jac.core.dist.rmi.RMIRemoteContainerStub", "copy", "false" );
      setVisible( "jac.core.dist.rmi.RMIRemoteContainerStub", "getByteCodeFor", "false" );
      setVisible( "jac.core.dist.rmi.RMIRemoteContainerStub", "instantiates", "false" );

      generateDefaultParameterNames( "jac.core.Application" );
      generateDefaultParameterNames( "jac.core.ApplicationRepository" );

      setVisible( "java.util.Arrays$ArrayList", "indexOf", "false" );
      setVisible( "java.util.Arrays$ArrayList", "removeAll", "false" );
      setVisible( "java.util.Arrays$ArrayList", "containsAll", "false" );
      setVisible( "java.util.Arrays$ArrayList", "contains", "false" );
      setVisible( "java.util.Arrays$ArrayList", "subList", "false" );
      setVisible( "java.util.Arrays$ArrayList", "set", "false" );
      setVisible( "java.util.Arrays$ArrayList", "isEmpty", "false" );
      setVisible( "java.util.Arrays$ArrayList", "add", "false" );
      setVisible( "java.util.Arrays$ArrayList", "size", "false" );
      setVisible( "java.util.Arrays$ArrayList", "iterator", "false" );
      setVisible( "java.util.Arrays$ArrayList", "clear", "false" );
      setVisible( "java.util.Arrays$ArrayList", "listIterator", "false" );
      setVisible( "java.util.Arrays$ArrayList", "retainAll", "false" );
      setVisible( "java.util.Arrays$ArrayList", "toArray", "false" );
      setVisible( "java.util.Arrays$ArrayList", "lastIndexOf", "false" );
      setVisible( "java.util.Arrays$ArrayList", "addAll", "false" );
      setVisible( "java.util.Arrays$ArrayList", "get", "false" );
      setVisible( "java.util.Arrays$ArrayList", "remove", "false" );

      ClassItem applicationClass = ClassRepository.get().getClass( jac.core.Application.class );
      
      setVisible( "jac.core.Application", "props", "false" );
      applicationClass.setAttribute( "Gui.List.showItem", applicationClass.getMethod("getName") );
      ClassItem applicationManagerClass = ClassRepository.get().getClass( jac.core.ApplicationRepository.class );
      
      applicationClass.setAttribute( "Gui.tableAttributesOrder", 
                               new FieldItem[] {
                                  applicationClass.getField("name"),
                                  applicationClass.getField("constructorClass"),
                                  applicationClass.getField("instantiated") } );

      setViewOnSelection( "jac.core.ApplicationRepository", "applications" );
      setAutoCreate( "jac.core.ApplicationRepository", "applications" );

      applicationManagerClass.getMethod( "addApplication" ).setAttribute( "Gui.askingSequence", 
                                                              new Object[] { "autoCreate" } );
      setVisible( "jac.core.ApplicationRepository", "getApplication", "false" );

      setCategories( "jac.core.Application", new String[] { "Aspects", "Properties" } );

      setCategory( "jac.core.Application", "name", "Properties" );
      setCategory( "jac.core.Application", "path", "Properties" );
      setCategory( "jac.core.Application", "instantiated", "Properties" );
      setCategory( "jac.core.Application", "constructorClass", "Properties" );
      setCategory( "jac.core.Application", "arguments", "Properties" );
      setCategory( "jac.core.Application", "start", "Properties" );

      setCategory( "jac.core.Application", "acs", "Aspects" );
      setCategory( "jac.core.Application", "acConfigurations", "Aspects" );

      setEmbeddedEditor( "jac.core.Application", "name" );
      setEmbeddedEditor( "jac.core.Application", "path" );

      setFileChooserView( "jac.core.Application", "setPath", new String[] { "" }, "directories" );

      setViewOnSelection( "jac.core.Application", "acConfigurations" );
      setAutoCreate( "jac.core.Application", "acConfigurations" );

      //setTableView( "jac.aspects.distribution.Topology", "containers" );

      applicationClass.getMethod( "addAcConfiguration" ).setAttribute( "Gui.askingSequence", 
                                                                 new Object[] { "autoCreate" } );
      applicationClass.getMethod( "addAcConfiguration" ).setAttribute( "Rtti.containerOperation", 
                                                                 "" );
      applicationClass.getMethod( "removeAcConfiguration" ).setAttribute( "Rtti.containerOperation", 
                                                                    "" );
      //applicationClass.getMethod( "weave" ).setAttribute( "Gui.category", "Aspects" );
      //applicationClass.getMethod( "unweave" ).setAttribute( "Gui.category", "Aspects" );
      applicationClass.getMethod( "realizes" ).setAttribute( "Gui.category", "Aspects" );

      applicationClass.getMethod( "configures" ).setAttribute( "Gui.visible", "false" );

      applicationClass.getCollection( "acs" ).setAttribute( "Gui.visible", "false" );

      //applicationClass.getCollection( "acs" ).addAddingMethod( applicationClass.getMethod( "weave" ) );
      //applicationClass.getCollection( "acs" ).addRemovingMethod( applicationClass.getMethod( "unweave" ) );

      ClassItem acConfClass = ClassRepository.get().getClass( jac.core.ACConfiguration.class );

      acConfClass.setAttribute( "Gui.tableAttributesOrder", 
                                new FieldItem[] {
                                   acConfClass.getField("name"),
                                   acConfClass.getField("weaveOnDemand"),
                                   acConfClass.getField("instance") } );
      
      acConfClass.setAttribute( "Gui.List.showItem", acConfClass.getMethod("getName") );

      acConfClass.getField( "weaveOnDemand" ).setAttribute( "Gui.embeddedEditor", "true" );

      acConfClass.getAbstractMethod(
         "ACConfiguration(Application,String,String,boolean)").setAttribute( 
            "Rtti.containerParam",
            new Integer(0) );

      /*setDynamicFieldChoice( "jac.core.ACConfiguration", "name", 
                             "jac.aspects.gui.ConsoleAdmin",
                             "getDeclaredACNames" ); */

      generateDefaultParameterNames( "jac.core.ACConfiguration" );
      
      setParameterNames( "jac.core.ACConfiguration",
                         "ACConfiguration(Application,String,String,boolean)",
                         new String[] { "Application", "Aspect name", "Configuration file", "Weave on demand" } );
      
      
      setParameterFields( "jac.core.ACConfiguration",
                          "ACConfiguration(Application,String,String,boolean)",
                          new String[] { "application", "name", "filePath", "weaveOnDemand" } );
      
      setParameterNames( "jac.core.Application", "Application(String,String,String,String[])",
                         new String[] {"Name", "Path", "Constructor class", "Launching arguments"} );

      setParameterNames( "jac.aspects.authentication.AuthenticationAC", 
                         "addTrustedUser(String,String)",
                         new String[] {"Username", "Password"} );

      setSystemListener( true );

      // Trace configuration
      generateDefaultParameterNames( "jac.aspects.gui.WrappableMap" );
      setMethodDynamicParameterChoice("jac.aspects.gui.WrappableMap","put",
                                      new String[] { "jac.aspects.gui.WrappableMap", 
                                                     "jac.aspects.gui.WrappableMap" }, 
                                      new String[] { "getCategories", "getLevels" },
                                      new String[] { "true", "false"});
      setParameterNames("jac.aspects.gui.WrappableMap","put", 
                        new String[] { "trace", "level"} );
      RttiAC rttiac = (RttiAC)ACManager.get().getObject("rtti");
      rttiac.addWrittenFields("jac.aspects.gui.WrappableMap","put",new String[] {"delegate"});
   }

   /**
    * Notifies all the displays that a new object has just been
    * created by upcalling the <code>newObject</code> method of the
    * displays.
    * 
    * @param newObject the newly created object 
    * @see Display */

   protected void notifyDisplaysForNewObject( Object newObject ) {
      Iterator it = displays.values().iterator();
      while( it.hasNext() ) {
         Display display = (Display) it.next();
         display.newObject( newObject );
      }
   }

   /**
    * The default behavior of this <code>BaseProgramListener</code>
    * method is to notify the display with the
    * <code>applicationStarted</code> event.
    *
    * @see jac.core.Display#applicationStarted() */

   public void afterApplicationStarted() {
      Display old = (Display)JacObject.getCollaboration()
         .getAttribute("Gui.display");
      Iterator it = displays.values().iterator();
      while( it.hasNext() ) {
         Display display = (Display) it.next();
         JacObject.getCollaboration()
            .addTransientLocalAttribute("Gui.display",display);
         display.applicationStarted();
      }
      if( old != null ) {
         JacObject.getCollaboration()
            .addTransientLocalAttribute("Gui.display",old);
      }
   }


   /**
    * A GUI level method to invoke a method on an object.
    *
    * @param display the display that must be used to print the
    * result, ask for inputs, or show errors or warnings
    * @param method the invoked method
    * @param object the invoked object
    * @param collection the corresponding collection item if the
    * invoked object is a collection */

   public static void invoke(Display display, AbstractMethodItem method, 
                             Object object, CollectionItem collection ) {
      if ( method.getParameterTypes().length == 0 ) {
         Log.trace("gui","Invoking " + method + " on " + object );
         InvokeThread.run( object, method, new Object[0], null, null,  
                           new String[] { "Gui.display" }, new Object[] { display } );
      } else {
         //new CallingBox( this, object, method );
         Log.trace("gui","Invoking "+method+" on "+object+" (ask for parameters is on)");
         if (collection != null) {
            InvokeThread.run( object, method, new Object[method.getParameterTypes().length],
                              null, null,
                              new String[] { "Gui.display", "Session.sid", "Gui.askForParameters" , "Gui.collectionItem"}, 
                              new Object[] { display, "Swing"+jac.core.dist.Distd.getLocalContainerName(), method, collection } );
         } else {
            InvokeThread.run( object, method, new Object[method.getParameterTypes().length],
                              null, null,
                              new String[] { "Gui.display", "Session.sid", "Gui.askForParameters" }, 
                              new Object[] { display, "Swing"+jac.core.dist.Distd.getLocalContainerName(), method } );
         }
      }
   }

   /**
    * The programmer of a new GUI aspect must overload this method to
    * return the right default program name.<p>
    *
    * By default, it returns the package path name of the current GUI
    * aspect (must be right most of the time).
    *
    * @return the default program name
    */

   protected String getDefaultProgramName() {
      return getClass().getName()
         .substring( 0, getClass().getName().lastIndexOf( '.' ) );
   }

   /** A flag to memorize that the mainWindow is beeing launched (or
       has already been launched). */
   static boolean launched = false;

   /**
    * This method updates the objects browser view when a new
    * instance is created.<p>
    *
    * If it does not exist yet, this method also creates a main GUI
    * window when the first drawable instance is created.<p>
    *
    * @param newInstance the new instance
    */

   public void whenNewInstance( Wrappee newInstance ) {
      notifyDisplaysForNewObject( newInstance );
   }

   /**
    * This method updates the objects browser view when a new
    * instance is created.<p>
    *
    * If it does not exist yet, this method also creates a main GUI
    * window when the first drawable instance is created.<p>
    */


   public void whenUsingNewInstance() {
      super.whenUsingNewInstance();
      if( (wrappee() instanceof ACManager) || 
          (wrappee() instanceof NameRepository) ||
          (wrappee() instanceof Display) ) return;
      if( wrappee().isExtendedBy(InputWrapper.class) ) return;
      Log.trace("gui","When using new instance" + wrappee().getClass().getName() );
      wrappee().wrapAllEx( inputWrapper, "askForParameters", new Class[] {} );
      wrappee().wrapModifiers ( new ViewControlWrapper(), "updateView" );
      wrappee().addExceptionHandler(inputWrapper,"catchInputCanceled");
      if( (wrappee() instanceof ApplicationRepository) && 
          ProgramView.get() != null ) {
         wrappee().wrap( adminWrapper, "whenNewProgram", 
                         "addApplication" );
      }
   }

   /**
    * This method updates the objects browser view when a new
    * instance is created from a remote host.<p>
    *
    * If it does not exist yet, this method also creates a main GUI
    * window when the first drawable instance is created.<p>
    *
    * @param newInstance the instance that have been created by a
    * remote host
    * @param the name of the new instance
    */


   public void whenRemoteInstantiation( Wrappee newInstance,
                                        String name ) {
      notifyDisplaysForNewObject( newInstance );
   }

   public void setVisible( String cl, String itemName, String visible ) {
      Log.trace("gui","setVisible("+cl+","+itemName+","+visible+")");
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem field = cli.getFieldNoError( itemName );
      if ( field != null ) {
         field.setAttribute( "Gui.visible", visible );
         return;
      }
      MethodItem[] methods = cli.getMethods( itemName );
      if ( methods != null ) {
         for (int i=0; i< methods.length; i++) {
            methods[i].setAttribute( "Gui.visible", visible );
         }
         return;
      }
   }

   /**
    * Tells if an item is visible (displayed by the GUI).
    *
    * @param metaItem the meta item to check 
    * @return true if visible
    *
    * @see #setVisible(String,String,String) */

   public static boolean isVisible( MetaItem metaItem ) {
      Object value = metaItem.getAttribute( "Gui.visible" );

      if ( value == null )
         return true;
      else
         return ((String)value).equals( "true" );
   }

   public void setVisible( String cl, String itemNames[], 
                           String visible ) {
      for( int i = 0; i < itemNames.length; i++ ) {
         setVisible( cl, itemNames[i], visible );
      }
   }

   public void setEmbeddedEditors(String cl) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem[] fields = cli.getFields();
      if( fields!=null) {
         for(int i=0;i<fields.length;i++) {
            setEmbeddedEditor(cl,fields[i].getName());
         }
      }
   }

   public void setEmbeddedEditor( String cl, String itemName ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem field = cli.getField( itemName );
      if ( field != null ) {
         if (field.getWritingMethods()==null || 
             field.getWritingMethods().length==0)
         {
            Log.warning("setEmbeddedEditor: no setter for "+cl+"."+itemName+" , ignored");
         } else {
            field.setAttribute( "Gui.embeddedEditor", "true" );
         }
         return;
      }
      MethodItem[] methods = cli.getMethods( itemName );
      if ( methods != null ) {
         for (int i=0; i< methods.length; i++) {
            methods[i].setAttribute( "Gui.embeddedEditor", "true" );
         }
         return;
      }
   }

   /**
    * @return true is the item has the property Gui.embeddedEditor set
    * to "true"
    * @see #hasEmbeddedEditors(ClassItem)
    * @see #setEmbeddedEditor(String,String)
    */
   public static boolean isEmbeddedEditor( MetaItem item ) {
      return 
         item.getAttribute("Gui.embeddedEditor" ) != null && 
         item.getAttribute("Gui.embeddedEditor" ).equals("true");
   }

   /**
    * @return true if at least one field of the class has the property
    * Gui.embeddedEditor set to "true" 
    * @see #isEmbeddedEditor(MetaItem)
    * @see #setEmbeddedEditor(String,String)
    */
   public static boolean hasEmbeddedEditors( ClassItem cli )
   {
      FieldItem fields[] = cli.getFields( );
      String[] cats = (String[])cli.getAttribute("Gui.categories");
      if (fields != null) {
         for (int i=0; i<fields.length; i++) {
            if(fields[i] instanceof CollectionItem) continue;
            if( cats != null && 
                fields[i].getAttribute("Gui.category") == null ) continue; 
            //(Arrays.asList(cats).contains(fields[i].getAttribute("Gui.category"))) 
            //  continue;
            if ( isEmbeddedEditor(fields[i]) && isVisible(fields[i]) ) {
               return true;
            }
         }
      }
      return false;
   }

   public void setBorder( String className, 
                          String fieldName,
                          String border ) {
   
      ClassItem cli = ClassRepository.get().getClass( className );
      FieldItem field = cli.getField( fieldName );
      field.setAttribute( "Gui.border", border );
   }

   public void setEditorWidth( String className, 
                               String fieldName, int width, int columns ) {
   
      ClassItem cli = ClassRepository.get().getClass( className );
      FieldItem field = cli.getField( fieldName );
      field.setAttribute( "Gui.editor.width", new Integer(width) );
      field.setAttribute( "Gui.editor.cols", new Integer(columns) );
   }

   public void setSmallEditorWidth( String className, 
                                    String fieldName, int width, int columns ) {
   
      ClassItem cli = ClassRepository.get().getClass( className );
      FieldItem field = cli.getField( fieldName );
      field.setAttribute( "Gui.editor.smallWidth", new Integer(width) );
      field.setAttribute( "Gui.editor.smallCols", new Integer(columns) );
   }

   public void setEditorHeight( String className, 
                                String fieldName, int height, 
                                int rows ) {
   
      ClassItem cli = ClassRepository.get().getClass( className );
      FieldItem field = cli.getField( fieldName );
      field.setAttribute( "Gui.editor.height", new Integer(height) );
      field.setAttribute( "Gui.editor.rows", new Integer(rows) );
   }

   public void setSmallEditorHeight( String className, 
                                     String fieldName, int height, 
                                     int rows ) {
   
      ClassItem cli = ClassRepository.get().getClass( className );
      FieldItem field = cli.getField( fieldName );
      field.setAttribute( "Gui.editor.smallHeight", new Integer(height) );
      field.setAttribute( "Gui.editor.smallRows", new Integer(rows) );
   }

   public void setFont( String fontName, String fontType, int fontSize ) {
      int type = Font.PLAIN;
      if( fontType.equals("BOLD") ) {
         type = Font.BOLD;
      }
      else if( fontType.equals("PLAIN") ) {
         type = Font.PLAIN;
      }
      else if( fontType.equals("ITALIC") ) {
         type = Font.ITALIC;
      }
      UIManager.put( "Label.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "ComboBox.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "Menu.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "MenuBar.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "Panel.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "Border.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "MenuItem.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "Button.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "RadioButton.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "RadioButtonMenuItem.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "CheckBox.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "CheckBoxMenuItem.font", new Font( fontName, type, fontSize ) );
      UIManager.put( "TabbedPane.font", new Font( fontName, type, fontSize ) );

   }

   public void setCategories( String cl, String[] categories ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      cli.setAttribute( "Gui.categories", categories );
   }

   public void setCategory( String cl, String itemName, String category ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem field = cli.getFieldNoError( itemName );
      if ( field != null ) {
         field.setAttribute( "Gui.category", category );
         return;
      }
      MethodItem[] methods = cli.getMethods( itemName );
      if ( methods != null ) {
         for (int i=0; i< methods.length; i++) {
            methods[i].setAttribute( "Gui.category", category );
         }
         return;
      }
   }

   /**
    * Get the category of a field as configured with setCategory().
    *
    * @param field the FieldItem
    * @return the category of the field or null.
    */
   public static String getCategory(FieldItem field) {
      return (String)field.getAttribute("Gui.category");
   }

   public void setEmbeddedView(String cl, String itemName) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getField( itemName );
      mi.setAttribute( "Gui.embeddedView", "true" );
   }

   public void setCustomizedView( String className, String type ) {
      ClassItem cli = ClassRepository.get().getClass( className );
      cli.setAttribute( "Gui.customizedView", type );
   }

   public void setTableView( String cl, String fieldName ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getField( fieldName );
      mi.setAttribute( "Gui.tableView", "true" );
   }

   public void setCollectionSortCriteria(String cl,
                                         String fieldName,
                                         String[] criteria) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getField( fieldName );
      mi.setAttribute( "Gui.sortCriteria", criteria );
   }

   public void setAutoCreate( String cl, String fieldName ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getField( fieldName );
      mi.setAttribute( "Gui.autoCreate", "true" );
   }

   public void setAutoCreateParameters( String cl, String methodName ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getMethod( methodName );
      mi.setAttribute( "Gui.autoCreate", "true" );
   }

   public void setAutoCreateParameters(String cl) {
      setAutoCreateParametersEx(cl,new String[0]);
   }

   public void setAutoCreateParametersEx(String cl, String[] excludedMethods) {
      Class clazz=null;
      boolean error = false;
      try {
         clazz = Class.forName(cl);
         Iterator it = ClassRepository.get().getClass(clazz)
            .getAllMethods().iterator();
         List l = Arrays.asList(excludedMethods);
         while(it.hasNext()) {
            MethodItem cur = (MethodItem)it.next();
            if( (!l.contains(cur.getName())) && 
                cur.getActualMethod().getDeclaringClass() == clazz )
            {
               try {
                  setAutoCreateParameters(
                     cl,cur.toString().substring(0,cur.toString().indexOf(":")));
               } catch(Exception e) {
                  e.printStackTrace();
               }
            }
         }
      } catch(Exception e) {
         e.printStackTrace();
      }
   }

   public void setAutoCreateState( String cl,
                                   String[] fields ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      cli.setAttribute( "Gui.autoCreatedState", fields );
   }

   public void setOpenViewMethod( String cl, String methodName ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getMethod( methodName );
      mi.setAttribute( "Gui.openView", "true" );
   }

   public void setAttributesOrder( String cl, String[] attributeNames ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem[] fields = new FieldItem[attributeNames.length];
      for(int i=0;i<attributeNames.length;i++) {
         fields[i] = cli.getField(attributeNames[i]);
      }
      cli.setAttribute("Gui.attributesOrder",fields);
   }

   public void setGroupAttributes( String cl, int[] groupIndexes) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      cli.setAttribute("Gui.groups",groupIndexes);
   } 

   public void setMethodsOrder( String cl, String[] methodNames ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MethodItem[] methods = new MethodItem[methodNames.length];
      for(int i=0;i<methodNames.length;i++) {
         methods[i] = cli.getMethod(methodNames[i]);
      }
      cli.setAttribute("Gui.methodsOrder",methods);
   }

   public void setCreationAttributesOrder( String cl, String[] attributeNames ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem[] fields = new FieldItem[attributeNames.length];
      for(int i=0;i<attributeNames.length;i++) {
         fields[i] = cli.getField(attributeNames[i]);
      }
      cli.setAttribute("Gui.creationAttributesOrder",fields);
   }

   public void setAttributesOrder( ClassItem cl, String fieldName, 
                                   ClassItem targetClass, String[] attributeNames)
   {
      FieldItem field = cl.getField( fieldName );
      FieldItem[] fields = new FieldItem[attributeNames.length];
      for(int i=0;i<attributeNames.length;i++) {
         fields[i] = targetClass.getField(attributeNames[i]);
      }
      field.setAttribute("Gui.attributesOrder",fields);
   }

   public void setTableAttributesOrder( String cl, String[] attributeNames ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem[] fields = new FieldItem[attributeNames.length];
      for(int i=0;i<attributeNames.length;i++) {
         fields[i] = cli.getField(attributeNames[i]);
      }
      cli.setAttribute("Gui.tableAttributesOrder",fields);
   }

   public void setFieldChoice( String cl, String fieldName, String editable,
                               String[] choice) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem field = cli.getField( fieldName );
      Object[] choice2 = new Object[choice.length];
      try {
         Log.trace("gui","setFieldChoice "+field.getType());
         for(int i=0; i<choice.length;i++) {
            choice2[i] = ClassRepository.instantiate(field.getType(),choice[i]);
         }
      } catch (Exception e) {
         e.printStackTrace();
      }
      field.setAttribute( "Gui.fieldChoice", choice2 );
      field.setAttribute( "Gui.editableChoice", editable );
   }

   public static boolean isEditableChoice(FieldItem field) {
      String tmp = (String)field.getAttribute("Gui.editableChoice");
      return tmp != null && tmp.equals("true");
   }

   public void setDynamicFieldChoice( String cl, String fieldName, String editable,
                                      String targetClass, String targetMethod) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MetaItem mi = cli.getField( fieldName );
      mi.setAttribute( "Gui.fieldChoice", 
                       ClassRepository.get().getClass(targetClass).getMethod(targetMethod) );
      mi.setAttribute( "Gui.editableChoice", editable );
      return;
   }

   public void setMethodDynamicParameterChoice( String cl, String methodName, 
                                                String[] targetClasses, 
                                                String[] targetMethods,
                                                String[] editable) 
   {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MethodItem mi = cli.getMethod( methodName );
      MethodItem[] methods = new MethodItem[mi.getParameterTypes().length];
      boolean[] editableChoices = new boolean[editable.length];
      for (int i=0; i < methods.length;i++) {
         if (targetClasses[i] != null && !targetClasses[i].equals("")) {
            ClassItem targetClass = ClassRepository.get().getClass(targetClasses[i]);
            methods[i] = targetClass.getMethod(targetMethods[i]);
         } else {
            methods[i] = null;
         }
         editableChoices[i] = editable[i].equals("true");
      }
      mi.setAttribute( "Gui.parameterChoices", methods);
      mi.setAttribute( "Gui.editableChoices", editableChoices);
   }

   public void setCollectionType( String cl, String collectionName, 
                                  String type) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      CollectionItem collection = (CollectionItem)cli.getField(collectionName);
      collection.setAttribute("Gui.collectionType",ClassRepository.get().getClass(type));
   }

   /**
    * Sets the selection mode for an URL.
    *
    * @param cl the class name
    * @param fieldName the field within this class (URL type) 
    * @param mode the selection mode (DIRECTORIES_ONLY ||
    * FILES_AND_DIRECTORIES), default is FILES_ONLY
    * @param extensions the selection extensions (gif, ps, etc.)
    * @param extensionsDecription the description of the extensions
    * (can be null) */

   public void setURLSelectionMode( String cl, String fieldName, 
                                    String mode, String[] extensions, 
                                    String extensionsDescription) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      FieldItem field = (FieldItem)cli.getField(fieldName);
      field.setAttribute("Gui.fileSelectionMode",mode);
      field.setAttribute("Gui.fileExtensions",extensions);
      field.setAttribute("Gui.fileExtensionsDescription",extensionsDescription);
   }

   public void setViewOnSelection( String cl, String collectionName ) 
   {
      ClassItem cli = ClassRepository.get().getClass( cl );
      CollectionItem collection = (CollectionItem)cli.getField(collectionName);
      collection.setAttribute("Gui.viewOnSelection","true");
   }

   public static boolean getViewOnSelection(FieldItem field) {
      return "true".equals(field.getAttribute("Gui.viewOnSelection"));
   }

   public void setOnSelection( ClassItem cl, String collectionName, 
                               MethodItem eventHandler) 
   {
      CollectionItem collection = (CollectionItem)cl.getField(collectionName);
      collection.setAttribute("Gui.viewOnSelection",eventHandler);
   }

   public static MethodItem getSelectionHandler(FieldItem field) {
      Object handler = field.getAttribute("Gui.viewOnSelection");
      return (handler instanceof MethodItem) ? (MethodItem)handler : null;
   }
     
   public static void setTargetContainer( String className, 
                                          String fieldName, 
                                          Object container ) {
      ClassItem cli = ClassRepository.get().getClass( className );
      cli.getField(fieldName).setAttribute( 
         "Gui.Swing.targetContainer", container);
   }

   public void setContainerOperation( String className, 
                                      String methodName ) {
      ClassItem cli = ClassRepository.get().getClass( className );
      cli.getMethod(methodName).setAttribute( 
         "Rtti.containerOperation", "");
   }

   public void setPreferredHeight( String className, 
                                   String collectionName, 
                                   int height ) {
      ClassItem cli = ClassRepository.get().getClass( className );
      cli.getField(collectionName)
         .setAttribute("Gui.viewHeight",new Integer(height));
   }

   public void setPreferredWidth( String className, 
                                  String collectionName, 
                                  int width ) {
      ClassItem cli = ClassRepository.get().getClass( className );
      cli.getField(collectionName)
         .setAttribute("Gui.viewWidth",new Integer(width));
   }

   public void setFileChooserView( String cl, String setterName, 
                                   String[] fileExtensions,
                                   String fileDescription ) {
      ClassItem cli = ClassRepository.get().getClass( cl );
      MethodItem mi = cli.getMethod( setterName );
      mi.setAttribute( "Gui.fileChooserView", 
                       new Object[] { fileExtensions, fileDescription } );
      return;
   }

   /**
    * Tells if the given method call by the GUI opens a file chooser
    * dialog.<p>
    *
    * @param setter the setter for the field
    * @return true if a file chooser is oppened
    * @see #getFileExtensions(MethodItem)
    * @see #getFileDescription(MethodItem) */

   public boolean isFileChooserView( MethodItem setter ) { 
      if ( setter.getAttribute( "Gui.fileChooserView" ) == null ) {
         return false;
      } else {
         return true;
      }
   }

   /**
    * Gets the file extensions for a given field that is view as a
    * file chooser (more precisely, the field that is set by the given
    * setter).<p>
    *
    * For instance, the file extensions can be a strings array like
    * {"gif", "jpg", "png"} or {"txt", "doc"}.<p>
    *
    * @param setter the setter for the field
    * @return the description of the files that are associated to the
    * field
    * @see #isFileChooserView(MethodItem) */

   public String[] getFileExtensions( MethodItem setter ) { 
      Object attr = null;
      if ( ( attr = setter.getAttribute( "Gui.fileChooserView" ) ) != null ) {
         return (String[]) ((Object[])attr)[0];
      }
      return null;
   }

   /**
    * Gets the file description for a given field that is view as a
    * file chooser (more precisely, the field that is set by the given
    * setter).<p>
    *
    * For instance, the file description can be a string like "Image
    * files" or "Text files".<p>
    *
    * @param setter the setter for the field
    * @return the description of the files that are associated to the
    * field
    * @see #isFileChooserView(MethodItem) */

   public String getFileDescription( MethodItem setter ) { 
      Object attr = null;
      if ( ( attr = setter.getAttribute( "Gui.fileChooserView" ) ) != null ) {
         return (String) ((Object[])attr)[1];
      }
      return null;
   }

   /**
    * Returns true if the given item must be displayed by the GUI as a
    * table.<p>
    *
    * @param metaItem the item to check */

   public static boolean isTableView( MetaItem metaItem ) {
      Object value = metaItem.getAttribute( "Gui.tableView" );
      if ( value == null )
         return false;
      else
         return ((String)value).equals( "true" );
   }

   public void setParameterNames( String cl, String methodName, 
                                  String[] parameterNames ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      setParameterNames( classItem.getAbstractMethod( methodName ), parameterNames );
   }

   public void setParameterNames( AbstractMethodItem method,
                                  String[] parameterNames ) {
      method.setAttribute( "Gui.parameterNames", parameterNames );
   }

   public void setParameterFields( String cl, String methodName, 
                                   String[] parameterFields ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      AbstractMethodItem method = classItem.getAbstractMethod( methodName );
      method.setAttribute( "Gui.parameterFields", parameterFields );
   }

   public void setPasswordParameters( String cl, String methodName, 
                                      String[] params ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      AbstractMethodItem method = classItem.getAbstractMethod( methodName );
      method.setAttribute( "Gui.passwordParameters", params );
   }

   public void setDirectCollectionMethods( String cl, String collectionName, 
                                           String[] methods ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      CollectionItem collection = classItem.getCollection( collectionName );
      collection.setAttribute( "Gui.directCollectionMethods", methods );
   }
    
   public void generateDefaultParameterNames( String cl ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      Vector methods = classItem.getMethods();
      for (int i=0; i<methods.size(); i++) {
         MethodItem[] meths = (MethodItem[])methods.get(i);
         for (int j=0; j<meths.length; j++) {
            MethodItem method = meths[j];
            if (method.isSetter() || method.isAdder() || method.isRemover()) {
               String coreName = NamingConventions.removePrefixFrom( method.getName() );
               if (method.getParameterCount()==1) {
                  method.setAttribute( "Gui.parameterNames", 
                                       new String[] { NamingConventions.textForName( coreName ) } );
               } else if (method.getParameterCount()==2) {
                  method.setAttribute( "Gui.parameterNames", 
                                       new String[] { 
                                          "key",
                                          NamingConventions.textForName( coreName ) } );
               } 
            } 
         }
      }
   }

   public void setDefaultValues( ClassItem classItem, String methodName, 
                                 Object[] values ) {
      AbstractMethodItem method = classItem.getAbstractMethod( methodName );
      method.setAttribute( "Gui.defaultValues", values );
   }

   public void setClosingMethod( String cl, String methodName ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      AbstractMethodItem method = classItem.getAbstractMethod( methodName );
      method.setAttribute( "Gui.closingMethod", "true" );
   }

   public void setModifyingBoxes( String cl, 
                                  String methodName, 
                                  String[] modifyingBoxes ) {
      ClassItem classItem = ClassRepository.get().getClass( cl );
      AbstractMethodItem method = classItem.getAbstractMethod( methodName );
      method.setAttribute( "Gui.modifyingBoxes", modifyingBoxes );
   }

   /**
    * Get the modifying boxes for the given abstract method item (a
    * method or a constructor).<p>
    *
    * @return the mofifying boxes names for each parameter */

   public String[] getModifyingBoxes( AbstractMethodItem method ) {
      return (String[]) method.getAttribute( "Gui.modifyingBoxes" );
   }

   public void setStyleSheetURL( String url ) {
      WebGui webgui = (WebGui)getDisplay("web");
      if (webgui != null)
         webgui.setStyleSheetURL(url);
      else
         Log.warning("gui","setStyleSheetURL("+url+"): no WebGui found");
   }

   /**
    * Tells the swing gui to capture System.out so that it appears a
    * in tab.
    */

   public void captureSystemOut( ) {
      ProgramView pv = (ProgramView)getDisplay("admin");
      if (pv != null)
         pv.captureSystemOut();
      else
         Log.warning("gui","captureSystemOut(): no ProgramView found");
   }

   public void captureSystemErr( ) {
      ProgramView pv = (ProgramView)getDisplay("admin");
      if (pv != null)
         pv.captureSystemErr();
      else
         Log.warning("gui","captureSystemErr(): no ProgramView found");
   }
   
   String defaultCurrency;

   public void setDefaultCurrency(String currencyName, int precision) {
      MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass("currency");
      classItem.setAttribute("gui.defaultCurrency",currencyName);
      declareCurrency(currencyName,precision,1);
   }

   /**
    * Returns the default currency
    */
   public static String getDefaultCurrency() {
      MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass("currency");
      return (String)classItem.getAttribute("gui.defaultCurrency");
   }

   public void declareCurrency( String currencyName, int precision, double rate) {
      MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass("currency");
      Log.trace("gui.currency","currency classItem =  "+classItem);
      Hashtable currencies = (Hashtable)classItem.getAttribute("gui.currencies");
      if (currencies==null) {
         Log.trace("gui.currency","building currencies");
         currencies = new Hashtable();
         classItem.setAttribute("gui.currencies",currencies);
      }
      currencies.put(currencyName, new Currency(currencyName, precision, rate));
      Log.trace("gui.currency","declareCurrency "+currencyName+" "+rate);
   }
   /**
    * Returns an enumeration of all declared currencies, including the
    * default currency. 
    * 
    * @param return the declared currencies as an Enumeration of
    * Currency.
    */
   public static Enumeration getCurrencies() {
      MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass("currency");
      Hashtable currencies = (Hashtable)classItem.getAttribute("gui.currencies");
      Log.trace("gui.currency",currencies.size()+" currencies");
      return currencies.keys();
   }
   /**
    * Returns the Currency object associated with a currency name.
    */
   public static Currency getCurrency(String currencyName) {
      MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass("currency");
      Hashtable currencies = (Hashtable)classItem.getAttribute("gui.currencies");
      return (Currency)currencies.get(currencyName);
   }

   public void setTypeEditor(String type, ClassItem editorClass) {
      if ( !java.awt.Component.class.isAssignableFrom(editorClass.getActualClass()) ) {
         Log.error("Editor class "+editorClass+" does not inherits java.awt.Component");
      } else if (!jac.aspects.gui.ValueEditor.class.isAssignableFrom(editorClass.getActualClass())) {
         Log.error("Editor class "+editorClass+" does not implement jac.aspects.gui.ValueEditorclass");
      } else {
         MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass(type);
         classItem.setAttribute("gui.swing.typeEditor",editorClass);
      }
   }
   
   public void setTypeEditor(String type, String editorClass) {
      setTypeEditor(type, ClassRepository.get().getClass(editorClass));
   }

   public void setHTMLTypeEditor(String type, ClassItem editorClass) {
      if ( !jac.aspects.gui.HTMLEditor.class.isAssignableFrom(editorClass.getActualClass()) ) {
         Log.error("Editor class "+editorClass+" does not implement jac.aspects.gui.HTMLEditor");
      } else {
         MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass(type);
         classItem.setAttribute("gui.html.typeEditor",editorClass);
      }
   }

   /**
    * Get the editor component class for a type
    *
    * @param type the type your the editor component class for
    * @param return the ClassItem of the type editor you asked
    */
   public static ClassItem getTypeEditor(MetaItem item) {
      ClassItem editorType = (ClassItem)item.getAttribute("gui.swing.typeEditor");
      if (editorType == null ) {
         editorType = ClassRepository.get().getClass(PrimitiveFieldEditor.class);
      }
      return editorType;
   }

   /**
    * Get the editor component class for a type
    *
    * @param type the type your the editor component class for
    * @param return the ClassItem of the type editor you asked
    */
   public static ClassItem getHTMLTypeEditor(MetaItem item) {
      ClassItem editorType = (ClassItem)item.getAttribute("gui.html.typeEditor");
      if (editorType == null ) {
         editorType = ClassRepository.get().getClass(HTMLPrimitiveFieldEditor.class);
      }
      return editorType;
   }


   Hashtable typeViewers = new Hashtable();

   public void setTypeViewer(String type, ClassItem viewerClass) {
      if ( !java.awt.Component.class.isAssignableFrom(viewerClass.getActualClass()) ) {
         Log.error("Viewer class "+viewerClass+" does not inherits java.awt.Component");
      } else {
         MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass(type);
         classItem.setAttribute("gui.swing.typeViewer",viewerClass);
      }
   }

   public void setTypeViewer(String type, String viewerClass) {
      setTypeViewer(type, ClassRepository.get().getClass(viewerClass));
   }

   public void setHTMLTypeViewer(String type, ClassItem viewerClass) {
      if ( !jac.aspects.gui.HTMLViewer.class.isAssignableFrom(viewerClass.getActualClass()) ) {
         Log.error("Viewer class "+viewerClass+" does not implement jac.aspects.gui.HTMLViewer");
      } else {
         MetaItem classItem = (MetaItem)ClassRepository.get().getVirtualClass(type);
         classItem.setAttribute("gui.html.typeViewer",viewerClass);
      }
   }

   public static ClassItem getTypeViewer(MetaItem item) {
      return (ClassItem)item.getAttribute("gui.swing.typeViewer");
   }

   public static ClassItem getHTMLTypeViewer(MetaItem item) {
      return (ClassItem)item.getAttribute("gui.html.typeViewer");
   }

   /**
    * Get the editor component class for a type
    *
    * @param type the type your the editor component class for
    */
   public ClassItem getTypeViewer(String type) {
      return (ClassItem)typeViewers.get(type);
   }

   public void setDateFormat(String dateFormat) {
      ClassRepository.get().getClass("java.util.Date").setAttribute(
         "gui.format",dateFormat);
   }

   /**
    * Returns the date format as set by setDateFormat.
    *
    * @return the date format
    */
   public static String getDateFormat() {
      String format = (String)ClassRepository.get()
         .getClass("java.util.Date")
         .getAttribute("gui.format");
      if( format==null ) format = "dd/MM/yyyy";
      return format;
   }

   public void setCommitOnFocusLost(boolean value) {
      ClassRepository.get().getClass("jac.aspects.gui.ObjectView").setAttribute(
         "gui.commitOnFocusLost",
         new Boolean(value)
            );
   }

   public static boolean getCommitOnFocusLost() {
      Boolean val = (Boolean)ClassRepository.get().getClass("jac.aspects.gui.ObjectView").
         getAttribute("gui.commitOnFocusLost");
      if (val == null)
         return true;
      else 
         return val.booleanValue();
   }
}
