/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.gui;

import jac.core.rtti.*;

/**
 * This is the configuration interface of the Gui aspect. 
 *
 * @see GuiAC
 * @see View
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a>
 */

public interface GuiConf {

   /**
    * This configuration method allows the user to create and register
    * a new customized GUI.
    * 
    * <p>A configuration file should first use this method, then
    * configure the customized GUI, and then call the
    * <code>showCustomized</code> method when the GUI configuration is
    * finished.
    *
    * @param name the new customized GUI name
    */

   void registerCustomized( String name );

   /**
    * This configuration method allows the programmer to actually show
    * a personal UI that extends the CustomizedGUI class on a
    * specified host.
    *
    * <p>Note that the customized GUI should have been configured with
    * some available configuration methods.
    *
    * @param gui the name of the customized GUI
    * @param host the host where the GUI must be launched. If null,
    * the GUI is launched on the local host */

   void showCustomized( String gui, String host );

   /**
    * This configuration method allows the programmer to actually show
    * a personal UI that extends the CustomizedGUI class on a
    * specified host.
    *
    * <p>Note that the customized GUI should have been configured with
    * some available configuration methods.
    *
    * @param gui the name of the customized GUI
    * @param host the host where the GUI must be launched */

   void showCustomized( String gui );

   /**
    * This configuration method allows the programmer to actually show
    * a status bar in a declared personal UI
    *
    * @param gui the name of the customized GUI */

   void addStatusBar( String gui );

   /**
    * This configuration method delegates to
    * <code>setSubPanesGeometry(int,int,String[])</code> with all the
    * panes scrollable.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param subPanesCount the number of subpanes in the window
    * @param geometry the geometry = <code>VERTICAL || HORIZONTAL ||
    * VERTICAL_LEFT || VERTICAL_RIGHT || HORIZONTAL_UP ||
    * HORIZONTAL_DOWN</code>
    *
    * @see #registerCustomized(String)
    * @see #setSubPanesGeometry(String,int,String,String[]) */

   void setSubPanesGeometry(String gui, int subPanesCount, String geometry);

   /**
    * This configuration method delegates to the corresponding
    * customized GUI.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param subPanesCount the number of subpanes in the window
    * @param geometry the geometry = <code>VERTICAL || HORIZONTAL ||
    * VERTICAL_LEFT || VERTICAL_RIGHT || HORIZONTAL_UP ||
    * HORIZONTAL_DOWN</code>
    * @param scrollings a set of string that tells if the sub-panes
    * must be srollable or not (use setSubPanesGeometry(String,int,String) to
    * make all the sub-panes scrollable
    *
    * @see #registerCustomized(String)
    * @see #setSubPanesGeometry(String,int,String) 
    */

   void setSubPanesGeometry(String gui, int subPanesCount, 
                            String geometry,
                            String[] scrollings);

   /**
    * This configuration method delegates to the corresponding
    * customized GUI.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param name the object's name as registered in the name
    * repository
    * @param paneId the panel id (see the geometry to know its
    * placement)
    * 
    * @see #registerCustomized(String)
    */ 

   void setObjectForPane( String gui, String name, int paneId );

   void setInvalidPane( String gui, int changedPane, int invalidPane );

   void addReferenceToPane( String gui,
                            String className, 
                            String referenceName,
                            int paneId );

   /**
    * This configuration method delegates to the corresponding
    * customized GUI.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param className the class the reference belongs to
    * @param referenceName the name of the reference
    * @param small tell if the viewed object in this pane should be
    * small
    * @param paneId the panel id where the view must be opened
    * @see #registerCustomized(String)
    * @see CustomizedGUI#addReferenceToPane(String,String,int) */ 

   void addReferenceToPane( String gui,
                            String className, 
                            String referenceName,
                            String small, 
                            int paneId );

   /**
    * This configuration method delegates to the corresponding
    * customized GUI.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param left left-border pixel
    * @param up upper-border pixel
    * @param width in percentage regarding the screen
    * @param height in percentage regarding the screen
    * @see #registerCustomized(String)
    * @see CustomizedGUI#setPosition(int,int,int,int) */ 

   void setPosition(String gui, int left, int up, 
                    int width, int height);

   void setMenuBar(String gui, String[] smenus);

   void addMenuItem(String gui, String menu,String name,
                    String className,String methodName);

   void addMenuSeparator(String gui, String menu);

   /**
    * This configuration method delegates to the corresponding
    * customized GUI.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param splitterId the splitter's index
    * @param location the position in pixel, regarding to the top/left
    * component, a negative value means that the splitter should be
    * set at the preferred sized of the inner components
    * @see #registerCustomized(String)
    * @see CustomizedGUI#setSplitterLocation(int,int) */ 

   void setSplitterLocation( String gui, int splitterId, int location );

   /**
    * This configuration method sets the title of the GUI main window.
    *
    * <p>The GUI must have been declared.
    *
    * @param gui the GUI name
    * @param title the window title
    * @see #registerCustomized(String) */

   void setTitle( String gui, String title );

   /**
    * Sets the given methods to be logging.
    *
    * <p>When it is called, the argument --- that must be a string ---
    * is written into a text area added at the end of a subpanel
    * defined by a customized GUI.
    *
    * @param gui the customized gui (must be registered and
    * configured with at least <code>setSubPanesGeometry</code>)
    * @param objects the objects that contain the method (pointcut
    * expression)
    * @param classes the class the contains the method (pointcut
    * expression)
    * @param methods the method item names (pointcut expression)
    * @param paneId the subpanel id
    * @param componentNumber the component to tog into
    * @see #registerCustomized(String)
    * @see #setSubPanesGeometry(String,int,String) */


   void setLoggingMethod( String gui, 
                          String objects, 
                          String classes, 
                          String methods, 
                          int paneID );

   /**
    * Adds a tree view in one of the panels of a customized GUI.
    *
    * @param gui the customized gui (must be registered and
    * configured with at least <code>setSubPanesGeometry</code>)
    * @param pathDef the path definition of the tree view (an 
    * <a href="http://jac.aopsys.com/doc/programmer_guide.html#3.2.5.2">
    * OPath expr</a> as explained in the programmer's guide of JAC), this 
    * path definition matches a set of objects and relations that are used
    * to create the tree
    * @param showRelations if true, shows the relations as tree nodes,
    * else hide them
    * @param paneID the subpanel id
    * @see #registerCustomized(String)
    * @see #setSubPanesGeometry(String,int,String) */

   void addTreeView( String gui, 
                     String pathDef,
                     boolean showRelations,
                     int paneID );

   /**
    * This configuration method allows the user to define new
    * resources that can be used by the GUI later-on.
    *
    * @param type the resource type (ICON) -- other types should be
    * supported soon
    * @param name the identifier of the resource (should be unique)
    * @param path the path where the resource is located (can be
    * classpath relative)
    * @see #setIcon(ClassItem,String) */

   void defineResource(String type, String name, String path );

   /**
    * This configuration method attaches an icon to a given class so
    * that the iconized instances of this class will be represented by
    * this icon (for instance in a treeview).
    *
    * @param classItem the class
    * @param name the icon's resource name */
   
   void setIcon(ClassItem classItem, String name);

   /**
    * This configuration method attaches a contextual menu to a given
    * class so that the instances of this class will show the menu
    * when the user performs a right click on it (by default the menu
    * shows all the methods).
    *
    * @param classItem the class
    * @param menu an array containing the names of the methods that
    * form the menu (must be declared in the class item), if an
    * element is an empty string, then a menu item separator is added
    */
   
   void setMenu(ClassItem classItem, String[] menu);

   /**
    * Sets a item to be visible or not using the RTTI.
    *
    * <p>If a field of method item is not visible, it will not be
    * displayed by the object views. If a class item is not visible,
    * then it will not be possible to open a view on objects of this
    * class.
    *
    * <p>By default, the RTTI does not define this property and every
    * item is visble. Thus, the GUI aspect component programmer should
    * overload the default contructor of this class to set some
    * visible attributes to false.<p>
    * 
    * @param cl the class where the item is located
    * @param itemName the item name (may be a method or a field name);
    * it can be null to set the whole class visible or not
    * @param visible "true" for visible
    *
    * @see GuiAC#isVisible(MetaItem) */

   void setVisible( String cl, String itemName, String visible );

   /**
    * This configuration method allows the programmer to set class
    * members to be internally edited.
    * @param cl the class name
    * @see ValueEditor */

   void setEmbeddedEditors(String cl);

   /**
    * This configuration method allows the programmer to set a class
    * member to be internally edited.
    *
    * <p>By default, each member of a class is editable with an "edit"
    * button or link that opens a popup that allows the user to edit
    * its value (see the <code>ValueEditor</code> interface). If this
    * configuration method is called, then the editor will be embedded
    * into the object's view the member belongs to.
    *
    * @param cl the class name
    * @param itemName the name of the item to be embedded (may be a
    * field or a method)
    * @see ValueEditor */

   void setEmbeddedEditor( String cl, String itemName );

   /**
    * Sets a border to the field in an object view.
    *
    * @param className the class name
    * @param fieldName the field name
    * @param border false if no border (default is true for
    * collections, false for others) */

   void setBorder( String className, String fieldName, String border );

   /**
    * Sets the width of the field's embedded editor when exist.
    *
    * @param className the class name
    * @param fieldName the field name
    * @param width the editor width in pixel (typically used by
    * SWING)
    * @param columns the editor width in character columns (typically
    * used by the WEB)
    * @see #setEmbeddedEditor(String,String) 
    * @see #setEditorHeight(String,String,int,int)
    */

   void setEditorWidth( String className, String fieldName, int width, int columns );

   /**
    * Sets the height of the field's embedded editor when exist.
    *
    * @param className the class name
    * @param fieldName the field name
    * @param height the editor height in pixel
    * @param rows the editor height in character rows
    * @see #setEmbeddedEditor(String,String) */

   void setEditorHeight( String className, 
                         String fieldName, int height, 
                         int rows );
   
   /**
    * This SWING-specific configuration method allows the programmer
    * to define a global font for the SWING UI.
    *
    * <p>The specified font will be used by all the SWING components.
    *
    * <p>Note: this method sets a global font that will be active for
    * all the configured SWING GUIs. It is not possible to have
    * different font configurations when running several GUIs on the
    * same container.
    *
    * @param fontName the name of the font
    * @param fontType the type of the font (BOLD || ITALIC || PLAIN)
    * @param fontSize the size of the font */

   void setFont( String fontName, String fontType, int fontSize );

   /**
    * This configuration method sets some categories for a class.
    *
    * <p>The GUI will interpret this categories to split the views
    * of this class instances to be separated into several subviews.
    * For instance, the Swing GUI will show the object in
    * several parts placed into several tabs.
    *
    * <p>When this method has been called, each meta-item of the class
    * must be categorized with one of the categories by using the
    * <code>setCategory</code> method.
    *
    * @param cl the class name
    * @param categories the categories for this class
    * @see #setCategory(String,String,String) */

   void setCategories( String cl, String[] categories );

   /**
    * This configuration method sets a category to an item of a class.
    *
    * <p>The category must correspond to one of these defined on the
    * class with the <code>setCategories</code> configuration method.
    * 
    * @param cl the class name
    * @param itemName the item to categorize (a method or a field)
    * @param category the existing category name
    * @see #setCategories(String,String[]) */

   void setCategory( String cl, String itemName, String category );

   /**
    * This configuration method allows the programmer to tell the GUI
    * to insert a referenced object to be displayed as an embedded
    * view in its container object view.
    *
    * @param cl the class name
    * @param itemName the reference that must be embedded */  

   void setEmbeddedView(String cl, String itemName);

   /**
    * This configuration method allows the programmer to tell that the
    * given class should be viewed with a given customized view.
    *
    * @param type the class of the viewed objects
    * @param type the class of the view component */

   void setCustomizedView( String className, String type );

   /**
    * Sets the render of a given field (more precisely a collection)
    * to be rendered by a table.
    *
    * <p>In a table view, each item of the displayed collection fills
    * a table line. Each column represents one field of the objects
    * whithin the collection (the item within the collection whould be
    * of the same class --- or at least share a common superclass).
    *
    * @param cl the class where the collection is located
    * @param fieldName the field name that contains the collection */

   void setTableView( String cl, String fieldName );

   /**
    * This configuration method allows the user to sets sorting
    * criteria on collections.
    *
    * @param cl the class where the collection is located
    * @param fieldName the field name that contains the collection 
    * @param criteria the names of the fields on which to sort
    */

   void setCollectionSortCriteria(String cl, String fieldName, String[] criteria);

   /**
    * This configuration method tells the GUI to automatically create
    * a new instance of the field type when a setter or an adder is
    * invoke on the object that owns this field.
    *
    * <p>This features makes the GUI more usable in many cases
    * (otherwhise, a intermediate step is asked to the user for
    * creating or picking an existing instance).
    *
    * @param cl the class name
    * @param fieldName the field in this class to be auto-created */  

   void setAutoCreate( String cl, String fieldName );

   /**
    * This configuration method tells which parameters of the given
    * method should be autocreated.
    *
    * <p>It does not use the constructor view but the actual modal
    * view on an empty created object that needs to be filled by the
    * user.
    *
    * @param cl the class name
    * @param methodName the method name */

   void setAutoCreateParameters( String cl, String methodName );

   /**
    * This configuration method tells that all the class methods must
    * be in auto-creation mode.
    * @param cl the class name
    * @param methodName the method name */

   void setAutoCreateParameters(String cl);

   /**
    * This configuration method tells that all the class methods must
    * be in auto-creation mode except the excluded ones.
    * @param cl the class name
    * @param methodName the method name
    * @param excludedMethods the excluded methods names
    */

   void setAutoCreateParametersEx(String cl, String[] excludedMethods);

   /**
    * This configuration method tells which fields of the given class
    * should be auto-created.
    *
    * <p>When a wrappee of the given class is auto-created using the
    * <code>setAutoCreateParameters</code> method, all the fields
    * mentioned here are also autocreated afterwards and a modal view
    * is shown to the user so that he can fill the values. The given
    * fields should be references (on other wrappees).
    *
    * @param cl the class name
    * @param fields the fields to autocreate
    * @see #setAutoCreateParameters(String,String) */

   void setAutoCreateState( String cl, String[] fields );

   /**
    * This configuration method allows the programmer to specify that
    * the result of a given method (should be a wrappe) opens a new
    * view on this result instead of treating it as a simple result.
    *
    * @param cl the class name
    * @param methodName the method name
    * @see jac.core.Display#openView(Object) */

   void setOpenViewMethod( String cl, String methodName );

   /**
    * Sets the order in which the attributes of a class are to be
    * rendered.
    *
    * <p>Attributes not listed here will not be rendered. This order
    * is also used for tableView rendered collections if the
    * <code>setTableAttributesOrder</code> method is not used.
    *
    * @param cl the class name 
    * @param attributeNames the name of the attributes in the
    * rendering order
    * @see #setTableAttributesOrder(String,String[]) 
    * @see #setAttributesOrder(ClassItem,String,ClassItem,String[])
    */

   void setAttributesOrder( String cl, String[] attributeNames );

   void setGroupAttributes( String cl, int[] groupIndexes); 

   /**
    * Sets the order in which the methods of a class are to be
    * rendered.
    *
    * <p>Methods not listed here will not be rendered.
    *
    * @param cl the class name 
    * @param methodNames the name of the methods in the
    * rendering order */

   void setMethodsOrder( String cl, String[] methodNames );

   /**
    * Sets the order in which the attributes of a class are to be
    * rendered when an instance is created.
    *
    * @param cl the class name 
    * @param attributeNames the name of the attributes in the
    * rendering order
    * @see #setAttributesOrder(String,String[]) */

   void setCreationAttributesOrder( String cl, String[] attributeNames );

   /**
    * Sets the order in which the attributes of the elements of a
    * collection are to be rendered.
    *
    * @param cl the class of the collection
    * @param fieldName the name of the collection
    * @param targetClass the class of attributes to render
    * @param attributeNames the name of the attributes in the
    * rendering order
    * @see #setTableAttributesOrder(String,String[])
    * @see #setAttributesOrder(String,String[])
    */

   void setAttributesOrder( ClassItem cl, String fieldName, 
                            ClassItem targetClass, String[] attributeNames);

   /**
    * Sets the order in which the attributes of a class are to be
    * rendered when their instance are placed within a table.
    *
    * <p>Attributes not listed here will not be rendered.
    *
    * @param cl the class name 
    * @param attributeNames the name of the attributes in the
    * rendering order
    * @see #setAttributesOrder(String,String[])
    * @see #setAttributesOrder(ClassItem,String,ClassItem,String[])
    */

   void setTableAttributesOrder( String cl, String[] attributeNames );

   /**
    * This configuration method allows the programmer to make a set of
    * object to be proposed to the final user when an edition of this
    * field value is performed.
    *
    * <p>Most of the GUI will propose the choice within a ComboBox.
    *
    * <p>When the choices values can not be defined at programming
    * time but must be dynamically created, then the programmer can
    * use the <code>setDynamicFieldChoice</code> method.
    *
    * @param cl the class name
    * @param fieldName the field
    * @param choice the values the user will have to choose from when
    * a edition of the field is performed
    * @see #setDynamicFieldChoice(String,String,String,String,String) */

   void setFieldChoice( String cl, String fieldName, String editable, String[] choice);

   /**
    * Same as <code>setFieldChoice</code> but with dynamically defined values.
    *
    * <p>The values are dynamically defined at runtime by the
    * invocation of a target method. This target method must return an
    * array of objects that contains the possible new values for the
    * fields.</p>
    *
    * <p>If the target method is static, it will be called with the
    * object as the only parameter. If it's not static, it will called
    * <em>on</em> the object with no parameters.</p>
    *
    * @param cl the class name
    * @param fieldName the field (in the class)
    * @param targetClass the class that contains the target method
    * @param targetMethod a static method within the target class that
    * returns the values to choose from */

   void setDynamicFieldChoice( String cl, String fieldName, String editable,
                               String targetClass, String targetMethod);

   /**
    * Same as <code>setDynamicFieldChoice</code> but for method parameters
    *
    * <p>The values are dynamically defined at runtime by the
    * invocation of a static target method. This target method must
    * return an array of objects that contains the possible new values
    * for the fields.
    *
    * @param cl the class name
    * @param methodName the method name (in the class)
    * @param targetClasses the array of classes that contains the
    * target method for each parameter (elements of this array may be null)
    * @param targetMethod the static methods within the target classes that
    * returns the values to choose from (elements of this array may be null)*/

   void setMethodDynamicParameterChoice( String cl, String methodName, 
                                         String[] targetClasses, 
                                         String[] targetMethods,
                                         String[] editable); 

   /**
    * Set the type of the objects of a collection.
    *
    * <p>If this method is not used, the collection type can be
    * dynamically found out by the GUI as soon as one instance is
    * contained by the collection. However, if none instances are
    * placed in the collection or if the instances types are
    * heterogenous, this type definition can be useful.
    *
    * @param cl the class name
    * @param collectionName the collection within this class 
    * @param type the type of this collection (an exiting class name)
    */

   void setCollectionType( String cl, String collectionName, String type);

   /**
    * This configuration method allows the programmer to specify that
    * the view on an object contained in a collection will be
    * automatically opened by the GUI when the user selects it.
    *
    * @param cl the class name
    * @param collectionName collection name */ 

   void setViewOnSelection( String cl, String collectionName );

   /**
    * Set the event handler to be called when an item of the
    * collection is selected.
    *
    * @param cl the class holding the collection
    * @param collectionName the name of the collection
    * @param eventHandler the event handler method. It must be static
    * method. It will be called with the following parameters : the
    * CustomizedGUI, the CollectionItem and the selected Object. If
    * the handler returns an object, it will be used as the selected
    * object.
    */
   void setOnSelection( ClassItem cl, String collectionName, MethodItem eventHandler);

   void setContainerOperation( String className, String methodName );

   /**
    * This configuration method allows the programmer to tell which
    * height should take the table or list view of a collection, if
    * possible.
    *
    * @param className the name of the class the collection belongs to
    * @param collectionName the name of the collection
    * @param height the preferred height */

   void setPreferredHeight( String className, 
                            String collectionName, 
                            int height );

   /**
    * This configuration method allows the programmer to tell which
    * width should take the table or list view of a collection, if
    * possible.
    *
    * @param className the name of the class the collection belongs to
    * @param collectionName the name of the collection
    * @param width the preferred width */

   void setPreferredWidth( String className, 
                           String collectionName, 
                           int width );

   /**
    * Sets the view of a given setter's calling box to be a file
    * chooser.<p>
    *
    * As logically expected, the type of the set field must be a
    * string or an URL.<p>
    *
    * @param cl the class the contains the method
    * @param methodName the method item name. It can be of the form
    * "methodName" or "methodName(<types>)". The first syntax will use
    * the method with that name. The second syntax allow you to
    * specify parameter types (separated by commas, with no spaces).
    * @param paramNames the parameter names
    * @see java.net.URL
    * @see GuiAC#isFileChooserView(MethodItem) */

   void setFileChooserView( String cl, String setterName, 
                            String[] fileExtensions,
                            String fileDescription );

   /**
    * Sets the argument names of a given method item so that they can
    * be used by GUI aspect components.
    *
    * <p>The GUI aspect can automatically fill some default parameter
    * names for all the setters to lighten the GUI aspect programmer
    * work (however, these automatically generated names can be
    * overloaded if needed).
    *
    * @param cl the class that contains the method
    * @param methodName the method item name. It can be of the form
    * "methodName" or "methodName(<types>)". The first syntax will use
    * the method with that name. The second syntax allow you to
    * specify parameter types (separated by commas, with no spaces).
    * @param paramNames the parameter names
    *
    * @see #generateDefaultParameterNames(String) */

   void setParameterNames( String cl, String methodName, String[] parameterNames );

   /**
    * @see #setParameterNames(String,String,String[])
    */
   void setParameterNames( AbstractMethodItem method, String[] parameterNames );

   /**
    * Specify that a parameter of a method corresponds to a field
    * within the invoked object.
    *
    * @param cl the class name 
    * @param methodName the method 
    * @param parameterFields an array that contains the corresponding
    * field for each parameter of the method */

   void setParameterFields( String cl, String methodName, String[] parameterFields );

   /**
    * Specify that a parameter of a method is a pasword method.
    *
    * <p>A password parameter will be rendered with stars instead of
    * the actual characters.
    *
    * @param cl the class name 
    * @param methodName the method 
    * @param parameterFields an array that must contains "true" items
    * if the corresponding parameter is a password, "false" or ""
    * otherwhise. */

   void setPasswordParameters( String cl, String methodName, String[] params );

   /**
    * This configuration method allows the programmer to tell the GUI
    * to add some buttons to the collection that will invoke the
    * corresponding methods on the currently selected collection
    * element.
    *
    * @param cl the class name 
    * @param collectionName the collection 
    * @param methods the names of the methods that can be invoked on
    * the selected element */

   void setDirectCollectionMethods( String cl, String collectionName, 
                                    String[] methods );

   /**
    * Generates the default parameter names for all the setter methods
    * of a given class.<p>
    *
    * @param cl the class to treat
    * @see #setParameterNames(String,String,String[]) */
    
   void generateDefaultParameterNames( String cl );

   /**
    * Sets the default value for argument names of a given method item
    * so that they can be used by GUI aspect components.
    *
    * @param cl the class the contains the method
    * @param methodName the method item name. It can be of the form
    * "methodName" or "methodName(<types>)". The first syntax will use
    * the method with that name. The second syntax allow you to
    * specify parameter types (separated by commas, with no spaces).
    * @param values the default parameter values */

   void setDefaultValues( ClassItem classItem, String methodName, 
                          Object[] values );

   /**
    * Sets the given method to be closing (ie, when the user press the
    * corresponding button, it performs the call and closes the
    * current object view).
    *
    * @param cl the class the contains the method
    * @param methodName the method item name. It can be of the form
    * "methodName" or "methodName(<types>)". The first syntax will use
    * the method with that name. The second syntax allow you to
    * specify parameter types (separated by commas, with no spaces). */

   void setClosingMethod( String cl, String methodName );

   /**
    * Sets the style sheet URL of the generated html pages.
    *
    * @param url the URL string */

   void setStyleSheetURL( String url );

   /**
    * Tells the swing gui to capture System.out so that it appears in
    * a tab.
    */

   void captureSystemOut();

   /**
    * Tells the swing gui to capture System.err so that it appears in
    * a tab.
    */

   void captureSystemErr();
   
   /**
    * Set the default currency
    *
    * @param currencyName the name of the default currency
    * @param precision number of decimals to display
    */
   void setDefaultCurrency(String currencyName, int precision);

   /** 
    * Declare a currency and it's change rate with the default currency
    *
    * @param currencyName the name of currency
    * @param precision number of decimals to display for the currency
    * @param rate the change rate for the currency
    */ 
   void declareCurrency( String currencyName, int precision, double rate);

   /**
    * Set the type editor to be used for editing values of a given
    * type. The editor must be a java.awt.Component and it must
    * implement the ValueEditor interface.
    *
    * @param type the type
    * @param editorClass the class of the editor component 
    *
    * @see ValueEditor
    */
   void setTypeEditor(String type, ClassItem editorClass);
   
   /**
    * Set the type editor to be used for editing values of a given
    * type. The editor must implement the HTMLEditor interface.
    *
    * @param type the type
    * @param editorClass the class of the editor component 
    *
    * @see HTMLEditor
    */

   void setHTMLTypeEditor(String type, ClassItem editorClass);

   /**
    * Set the type viewer to be used for displaying values of a given
    * type. The viewer must be a java.awt.Component and implement the
    * ValueViewer interface and the TableCellRenderer interface (so
    * that it ca be used to render table cells).
    *
    * @param type the type
    * @param viewerClass the class of the viewer component 
    *
    * @see ValueViewer
    */

   void setTypeViewer(String type, ClassItem viewerClass);
   
   /**
    * Set the type viewer to be used for displaying values of a given
    * type. The viewer must implement the HTMLViewer interface.
    *
    * @param type the type
    * @param viewerClass the class of the viewer component 
    *
    * @see HTMLViewer 
    */
   void setHTMLTypeViewer(String type, ClassItem viewerClass);

   /**
    * Set the default date format used by date components. It must be
    * a valid format as defined by java.text.SimpleDateFormat
    *
    * @param dateFormat the date format
    * @see java.text.SimpleDateFormat
    */
   void setDateFormat(String dateFormat);

   /**
    * Set the commit method to use when attibutes are edited. 
    *
    * @param value If true, attribute changes are committed when the
    * widget loses focus. Otherwise, a commit and a cancel button are
    * displayed in each object view.  */

   void setCommitOnFocusLost(boolean value);

}

