/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.gui;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.List;
import jac.core.rtti.*;

/**
 * This dialog is used to ask the parameters values when a method is
 * called on a viewed JAC object.<p>
 *
 * @see View
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a>
 */

public class InputBox extends JDialog implements ActionListener /*, TextListener, ListSelectionListener*/ {

   private JButton okButton;
   private JButton cancelButton;

   private Object substance = null;
   private AbstractMethodItem method = null;
   
   private Object[] result = null;

   /**
    * Returns an array of objects that corresponds to the value
    * entered by the user in the input box.
    *
    * <p>If the user canceled the input operation, the returned value
    * is null.
    *
    * @return the values filled by the user */

   public Object[] getResult() {
      return result;
   }

   /**
    * Construct a new calling box to get the parameters of the called
    * method.<p>
    *
    * @param display the display
    * @param owner the owner top-level window (here a dialog)
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

   public InputBox( Window display, Dialog owner, Object substance, 
                    AbstractMethodItem method ) {
      super( owner, "Parameters for " + method.getName(), true );
      init(display,substance,method);
   }

   /**
    * Construct a new calling box to get the parameters of the called
    * method.<p>
    *
    * @param display the display
    * @param owner the owner top-level window (here a frame)
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

   public InputBox( Window parent, Frame owner, Object substance, 
                    AbstractMethodItem method ) {
      super( owner, "Parameters for " + method.getName(), true );
      init(parent,substance,method);
   }

   /**
    * Initializes the dialog.
    *
    * @param display the owning display
    * @param substance the object viewed by the parent window (the
    * object the method is called on)
    * @param method the called method */

  void init(Window display, Object substance, 
            AbstractMethodItem method ) {

      this.substance = substance;
      this.method = method;

      Container contentPane = getContentPane();

      addWindowListener ( new WindowAdapter () {
            public void windowClosing( WindowEvent e ) {
               e.getWindow().dispose();
            }
         }
      );

      JPanel panel = RenderFactory.createParameters(display,method,substance);

      contentPane.add("Center", panel);
      JPanel p2 = new JPanel();
      p2.setBorder( BorderFactory.createEtchedBorder() );
      okButton = addButton( p2, "Ok" );
      cancelButton = addButton( p2, "Cancel" );
      contentPane.add( "South", p2 );
      pack();

      // open the box centerd in the screen...
      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      Rectangle rect = getBounds();
      double left = (screenDim.getWidth()-rect.getWidth())/2;
      double top = (screenDim.getHeight()-rect.getHeight())/2;
      Rectangle newRect = new Rectangle(
         (int)left,(int)top,
         (int)rect.getWidth(),(int)rect.getHeight());
      setBounds(newRect);

      ((JPanel)panel.getComponent(0)).getComponent(1).requestFocus();
      setVisible ( true );
   }

   /**
    * For internal use.<p>
    */
   JButton addButton(Container c, String name)  {
      JButton button = new JButton(name);
      button.addActionListener(this);
      c.add(button);
      return button;
   }

   /**
    * Implements what is done when a button is pressed (may be either
    * OK or CANCEL).<p>
    *
    * @param evt tell what button was pressed */

   public void actionPerformed(ActionEvent evt) {

      try {
         
         Object source = evt.getSource();

         if(source == okButton) {
            result = RenderFactory.buildParameterValues(method,getContentPane());
         } else if (source == cancelButton) {
            result = null;
         }
         dispose();
         
      } catch ( Exception e ) {
         e.printStackTrace ();
      }
   }

}
