/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.gui;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

import jac.core.rtti.*;
import jac.util.*;

import jac.core.*;
import jac.wrappers.*;
import jac.aspects.naming.*;

/**
 * This class defines a Swing component view for maps (Hashtable, etc).
 *
 * @see GuiAC */

public class MapView extends JPanel
   implements FieldItemView, ActionListener, ListSelectionListener, MouseListener {

   JTable table;
   Map.Entry[] entries;
   Window parent;
   CollectionItem collection;

   Object substance = null;
   private View parentView = null;

   /**
    * Constructs a new Map view.
    *
    * @param parent the Swing dialog parent
    * @param parentView a view that aggregates this view if any
    * @param substance the object the viewed collection belongs to
    * @param collection the collection that is showed by this view */

   public MapView(Window parent, 
                  View parentView, Object substance,
                  CollectionItem collection ) 
   {
      this.parent = parent;
      this.substance = substance;
      this.parentView = parentView;
      this.collection = collection;

      Log.trace("gui.mapView","MapView "+collection);

      setLayout( new BoxLayout( this, BoxLayout.Y_AXIS ) );
      setBorder( 
         BorderFactory.createTitledBorder( 
            BorderFactory.createLineBorder( Color.black ),
            NamingConventions.textForName(" "+collection.getName()+" ")));

      JPanel upperCont = new JPanel();
      upperCont.setLayout( new BoxLayout( upperCont, BoxLayout.Y_AXIS ) );
      JPanel downCont = new JPanel();
      //downCont.setLayout( new BoxLayout( downCont, BoxLayout.X_AXIS ) );

      // Build JTable
      Vector headers = new Vector();
      Vector data = new Vector();
      buildTableHeaderAndData(headers,data);
      table = new JTable(data,headers);

      JScrollPane scrollPane = new JScrollPane(table);
      Integer height = (Integer)collection.getAttribute( "Gui.Table.viewHeight" );
      if ( height != null ) {
         table.setPreferredScrollableViewportSize(new Dimension(500, height.intValue()));
      } else {
         table.setPreferredScrollableViewportSize(new Dimension(500, 70));
      }
      //      table.setAutoResizeMode( JTable.AUTO_RESIZE_ALL_COLUMNS );
      upperCont.add( scrollPane, BorderLayout.CENTER );
      table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      BetterListSelectionModel selModel = new BetterListSelectionModel( table );
      table.setSelectionModel( selModel );         
      table.setName( collection.getName() );
      selModel.addListSelectionListener( this );

      // add, remove and view buttons
      JButton b = null;
      if ( !GuiAC.getViewOnSelection(collection) ) {
          b = new JButton ( ResourceManager.getViewIcon() );
          b.setToolTipText( "View" );
          b.setActionCommand( "open" );
          b.addActionListener( this );
          downCont.add( b );
      }

      b = new JButton ( ResourceManager.getNewIcon() );
      b.setToolTipText( "Add" );
      b.setActionCommand( "add" );
      b.addActionListener( this );
      downCont.add( b );
      
      b = new JButton ( ResourceManager.getRemoveIcon() );
      b.setToolTipText( "Remove" );
      b.setActionCommand( "remove" );
      b.addActionListener( this );
      downCont.add( b );
      
      b = new JButton ( ResourceManager.getEditIcon() );
      b.setToolTipText( "Edit" );
      b.setActionCommand( "Edit" );
      b.addActionListener( this );
      downCont.add( b );
      
      String[] directMeths = (String[])
         collection.getAttribute("Gui.directCollectionMethods");
      if( directMeths != null ) {
         for( int i=0; i < directMeths.length; i++ ) {
            b = new JButton ( NamingConventions.textForName( directMeths[i] ) );
            b.setActionCommand( "DIRECT:"+ directMeths[i] );
            b.addActionListener( this );
            downCont.add( b );
         }
      }

      add( "North", upperCont );
      add( "South", downCont );

   }

   protected void buildTableHeaderAndData(Vector headers, Vector data) {
      Map map = (Map)collection.getThroughAccessor(substance);
      headers.add("key");
      headers.add("value");
      if (map != null) {
         Set entrySet = map.entrySet();
         Iterator i = entrySet.iterator();
         entries = new Map.Entry[entrySet.size()];
         int j = 0;
         while (i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            Vector rowData = new Vector();
            rowData.add(entry.getKey());
            rowData.add(entry.getValue());
            data.add(rowData);
            entries[j++]=entry;
         }
      } else {
         entries = null;
      }
   }

   /**
    * Returns an array of the selected entries. The array is empty if
    * no object is selected, but not null.
    **/

   public Map.Entry[] getSelected() {
      Map.Entry[] selected = null;
      int[] indices = table.getSelectedRows();
      if (indices != null) {
         selected = new Map.Entry[indices.length];
         for (int i=0; i<indices.length; i++)
         {
            selected[i] = entries[indices[i]];
         }
      } else {
         selected = new Map.Entry[0];
      }
      return selected;
   }

   /**
    * Handles the actions on this view.
    *
    * <p>On a collection view, the three default possible actions are
    * to open a new view on the selected item, to add a new item to
    * the collection, and to remove the selected item from the
    * collection.
    *
    * @param evt the user event */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("open")) {
         Map.Entry[] to_open = getSelected();
         JPanel target = (JPanel) collection.getAttribute( "Gui.Swing.targetContainer" );
         if ( target == null ) {
            for (int i=0; i<to_open.length; i++) {
               ObjectView.addViewFor( to_open[i], parent, parent );
            }
         } else {
            target.removeAll();
            for (int i=0; i<to_open.length; i++) {
               target.add( new ObjectView(parent, null, to_open[i]) );
            }
            target.validate();
         }
      } else if (evt.getActionCommand().equals("add")) {
         MethodItem[] methods = collection.getAddingMethods();
         if (methods!=null && methods.length>=1) {
            MethodItem addMethod = methods[0];
            Log.trace("gui","Invoking add method "+addMethod.getName()+" on collection owner");
            if ( collection.getAttribute("Gui.autoCreate") != null ) {

               ClassItem type = ClassRepository.get().getClass(
                  addMethod.getParameterTypes()[0]);

               try {
                  Log.trace("gui", "constructors = " + type.getConstructors() );
                  Vector constructors = type.getConstructors();
                  ConstructorItem c;
                  if (constructors.size()==1) {
                     c = (ConstructorItem)type.getConstructors().get(0);
                  } else {
                     c = (ConstructorItem)type.getConstructors().get(1);
                  }
                  Object newObject = null;
                  if ( c.getParameterTypes().length > 0 ) {
                     newObject = new ConstructorBox( parent, c ).getResult();
                  } else {
                     newObject = c.newInstance();
                  }
                  
                  if ( newObject != null ) {
                     addMethod.invoke( substance, new Object[] { newObject } );
                  }

               } catch (Exception e) {
                  Log.error("Instanciation of "+type.getName()+" failed : "+e);
                  e.printStackTrace();
               }
               
            } else {
               GuiAC.invoke((Display)parent,addMethod,substance,null);
            }
         } else {
            methods = collection.getAccessingMethods();
            if (methods!=null && methods.length==1) {
               MethodItem getMethod = methods[0];
               MethodItem method = null;
               Object c;
               try {
                  c = getMethod.invoke(substance,new Object[0]);
               } catch (Exception e) {
                  Log.error("Getting collection with "+getMethod.getName()+
                            "failed : "+e);
                  return;
               }
               ClassItem cl = ClassRepository.get().getClass(c.getClass());
               if (collection.isMap()) 
                  method = cl.getMethod("put(Object,Object)");
               else
                  method = cl.getMethod("add(Object)");
               Log.trace("gui","Invoking "+method.getName()+" on collection itself");
               GuiAC.invoke((Display)parent,method,c,collection);
            }
         }
      } else if (evt.getActionCommand().equals("remove")) {
         Object[] selected = getSelected();
         MethodItem[] methods = collection.getRemovingMethods();
         if (methods!=null && methods.length>=1) {
            // try with the removing method
            MethodItem removeMethod = methods[0];            
            for (int i=0; i<selected.length; i++) {
               try {
                  Log.trace("gui","Invoking remove method "+removeMethod.getName()+" on collection owner");
                  removeMethod.invoke(substance,new Object[] {selected[i]});
               } catch (Exception e) {
                  e.printStackTrace();
               }
            }
         } else {
            // call the collection's method directly
            methods = collection.getAccessingMethods();
            if (methods!=null && methods.length==1) {
               MethodItem getMethod = methods[0];
               MethodItem method = null;
               Object c;
               try {
                  c = getMethod.invoke(substance,new Object[0]);
               } catch (Exception e) {
                  Log.error("Getting collection with "+getMethod.getName()+
                            " failed : "+e);
                  return;
               }
               ClassItem cl = ClassRepository.get().getClass(c.getClass());
               if (!collection.isMap()) {
                  method = cl.getMethod("remove(Object)");
                  Log.trace("gui","Invoking "+method.getName()+" on collection itself");
                  for (int i=0; i<selected.length; i++) {
                     try {
                        method.invoke(c,new Object[] {selected[i]});
                     } catch (Exception e) {
                        Log.error("Removing from collection with "+method.getName()+
                                  " failed : "+e);
                     }
                  }
               }
            }
            
         }
      } else if ( evt.getActionCommand().startsWith("DIRECT:") ) {

         String directMethodName = evt.getActionCommand().substring(7);
         
         int selected = table.getSelectedRow();
         if (selected!=-1) {
            Object to_invoke = entries[selected].getValue();
            MethodItem directMethod =
               ClassRepository.get().getClass( to_invoke.getClass() )
               . getMethod( directMethodName );
            if( directMethod != null )
               GuiAC.invoke((Display)parent,directMethod,to_invoke,null);
         }
         
      }
   }

   /**
    * Defines what happens when the selection changes. */

   public void valueChanged(ListSelectionEvent e) {
      if (e.getValueIsAdjusting())
         return;

      if ( e.getSource() instanceof JList ) {
         JList theList = (JList)e.getSource();
         if (! theList.isSelectionEmpty()) {
            if ( GuiAC.getViewOnSelection(collection) ) {
               actionPerformed( new ActionEvent(this, 0, "open") );
            }
         }
      } else {

         ListSelectionModel lsm = (ListSelectionModel)e.getSource();
         
         int firstIndex = e.getFirstIndex();
         int lastIndex = e.getLastIndex();
         boolean isAdjusting = e.getValueIsAdjusting();
         
         if (lsm.isSelectionEmpty() || isAdjusting) {
            return;
         } else {
            if ( GuiAC.getViewOnSelection(collection) ) {
               // Find out which indexes are selected.
               int minIndex = lsm.getMinSelectionIndex();
               int maxIndex = lsm.getMaxSelectionIndex();
               for (int i = minIndex; i <= maxIndex; i++) {
                  if (lsm.isSelectedIndex(i)) {
                     Log.trace("gui","Opening " + i);
                     actionPerformed( new ActionEvent(this, 0, "open") );
                     return;
                  }
               }
            }
         }
      }
   }

   /**
    * Shows a popup if needed.
    *
    * @param e mouse event descriptor */

   public void mousePressed(MouseEvent e) {
      //      maybeShowPopup(e);
   }

   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseClicked(MouseEvent e) {
      if (e.getClickCount() == 2) {

         System.out.println("Double click happened");

         if ( e.getSource() instanceof JList ) {
            JList theList = (JList)e.getSource();
            if (! theList.isSelectionEmpty()) {
               if ( collection.getAttribute( "Gui.viewOnDoubleClick" ) != null ) {
                  actionPerformed( new ActionEvent(this, 0, "open") );
               }
            }
         }
      }
   }
   
   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseEntered(MouseEvent e) {
   }

   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseExited(MouseEvent e) {
   }

   /**
    * Closes a popup if needed.
    *
    * @param e mouse event descriptor */

   public void mouseReleased(MouseEvent e) {
      //      maybeShowPopup(e);
   }
 
   // interface View

   public Object getSubstance() {
      return substance;
   }

   public View getParentView() {
      return parentView;
   }

   public void refreshView() {

      Vector headers = new Vector();
      Vector data = new Vector();
      buildTableHeaderAndData(headers,data);
      Log.trace("gui","refreshing collection view with headers = " + 
                headers + " and data = " + data);
      table.setModel(new DefaultTableModel( data, headers ) );
   }

   public void close() {}

   public void refreshViewItem( FieldItem item, MethodItem method, Object[] args ) {
      refreshView();
   }

   class BetterListSelectionModel extends DefaultListSelectionModel {
      JComponent list = null;
      public BetterListSelectionModel( JComponent list ) {
         super();
         this.list = list;
      }
      public JComponent getList() {
         return list;
      }
   }
   public void setFocus(FieldItem field, Object extraOption) {}

   public FieldItem getFieldItem() {
      return collection;
   }
}
