/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import java.util.*;
import java.awt.event.*;
import java.awt.*;
import javax.swing.*;

import jac.aspects.naming.*;
import jac.util.*;
import jac.core.*;
import jac.core.rtti.*;

/**
 * This is a special value editor that allows the user to choose a
 * value within a set of object of a given type. */

public class ObjectChooser extends JPanel 
   implements ActionListener,ValueEditor {

   JComboBox choice;
   Hashtable objects = new Hashtable();
   ClassItem type;
   Window parent;
   JButton viewButton;
   FieldItem reference;
   Object substance;

   /**
    * Constructs a new object chooser.
    *
    * @param type the type of the objects that will automatically fill
    * the choices presented to the user
    * @param reference the subtance reference that is affected by this
    * chooser (can be null) 
    * @param predicate if not null, further specify which objects to chose from
    */

   public ObjectChooser(Window parent, Object substance, 
                        ClassItem type, FieldItem reference,
                        Predicate predicate) 
   {
      this.reference = reference;
      this.substance = substance;
      choice =  new JComboBox();
      //choice.setEditable(true);
      this.type = type;
      this.parent = parent;
      // use JacObjects.getObjects(type) instead
      Collection all_objects = null;

      if(reference != null && reference.getAttribute("Gui.fieldChoice")!=null) {
         try {
            all_objects = (Collection)((MethodItem)reference.getAttribute("Gui.fieldChoice"))
               .invoke(null,new Object[0]);
         } catch( Exception e ) {
            e.printStackTrace();
            all_objects = ObjectRepository.getObjects(type);
         }
      } else {
         all_objects = ObjectRepository.getObjects(type);
      }
      Log.trace("gui.chooser","ObjectChooser.type = "+type);
      Iterator i = all_objects.iterator();
      while (i.hasNext()) {
         Object object = i.next();
         if (predicate==null || predicate.apply(object)) {
            addObject(object);
            Log.trace("gui.chooser",
                      "adding instance of "+object.getClass()+" : "+object);
         }
      }
      addNull();
      add(choice);

      if( JacObject.getCollaboration()
          .getAttribute("Gui.view.small") == null ) {
         viewButton = new JButton ( ResourceManager.getViewIcon() );
         viewButton.setToolTipText("View");
         viewButton.setActionCommand("open");
         viewButton.addActionListener(this);
         viewButton.setMargin(new Insets(1,1,1,1));
         add( viewButton );
         JButton button = new JButton ( "New" );
         button.setToolTipText( "Create a new instance" );
         button.setActionCommand( "new" );
         button.addActionListener( this );
         add(button);
      }
   }

   // ValueEditor interface

   public void setValue(Object value) {
      if( value == null ) value = "";
      choice.setSelectedItem(value.toString());
   }

   public Object getValue() {
      Log.trace("gui.chooser","selectedItem = "+choice.getSelectedItem());
      if ( choice.getSelectedItem() == null ) return null;
      if ( choice.getSelectedItem().equals("") ) return null;
      return objects.get(choice.getSelectedItem());
   }

   public void setWidth(int width) {
      /*      Dimension dim = choice.getSize();
      Dimension newDim = new Dimension(width,(int)dim.getHeight());
      choice.setSize(newDim);*/
      choice.setMaximumRowCount(width);
   }

   public void setHeight(int height) {
      Dimension dim = choice.getSize();
      Dimension newDim = new Dimension((int)dim.getWidth(),height);
      choice.setSize(newDim);
   }
   
   /**
    * Add an object to the choices and returns its string identifier.
    *
    * @param object the object to add
    * @return a string that will identify it within the choices */

   protected String addObject(Object object) {
      String defaultKey = object.toString();
      if (defaultKey==null) {
         defaultKey = NameRepository.get().getName(object);
      }
      if (defaultKey==null) {
         defaultKey = object.getClass().getName();
      }
      String key = defaultKey;
      int i=2;
      while (objects.containsKey(key)) {
         key = defaultKey+"<"+(i++)+">";
      }
      choice.addItem(key);
      objects.put(key,object);
      return key;
   }

   /**
    * Adds the null object.
    */

   protected void addNull() {
      if (objects.containsKey("")) return;
      choice.addItem("");
      //objects.put("",null);
   }

   /**
    * Handles the actions performed by the users on this view.
    *
    * <p>On an object chooser, a "new" action can be performed to
    * allow the user to add a new object to the choices it not present
    * yet.
    *
    * @param evt the performed action */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("new")) {
         Object newObject = null;
         ConstructorItem c = null;
         try {
            CollectionItem repositoryCollection = 
               (CollectionItem)type.getAttribute("Rtti.repositoryCollection");
            String repositoryName = 
               (String)type.getAttribute("Rtti.repositoryName");
            if (repositoryName!=null) {
               Object repository = NameRepository.get().getObject(repositoryName);
               if( repositoryCollection != null && repository!=null) {
                  Log.trace("gui", "Adding new object to repository "+
                            repositoryName+"."+repositoryCollection.getName());
                  MethodItem[] meths = repositoryCollection.getAddingMethods();
                  Object[] params = new Object[] { null };
                  if(meths!=null && meths.length>0) {
                     JacObject.getCollaboration().addTransientAttribute("Gui.display",parent);
                     JacObject.getCollaboration().addTransientAttribute("Gui.askForParameters",meths[0]);
                     meths[0].invoke(repository,params);
                     newObject = params[0];
                  }
               }
            }
            if( newObject == null ) {

               MethodItem[] meths = null;
               if(reference!=null) meths=reference.getWritingMethods();
               MethodItem setter = null;
               if(meths!=null && meths.length>0) {
                  setter=meths[0];
               }
               /*if(setter!=null && setter.getAttribute("Gui.autoCreate")!=null) {

                  JacObject.getCollaboration().addTransientAttribute("Gui.display",parent);
                  JacObject.getCollaboration().addTransientAttribute("Gui.askForParameters",setter);
                  Object[] params = new Object[1];                  
                  setter.invoke(substance,params);
                  newObject=params[0];
                  System.out.println("returning : "+Arrays.asList(params));
                  } else {*/

                  Log.trace("gui", "constructors = " + type.getConstructors() );
                  Vector constructors = type.getConstructors();
                  /*
                    if (constructors.size()==1) {
                    c = (ConstructorItem)type.getConstructors().get(0);
                    } else {
                    c = (ConstructorItem)type.getConstructors().get(1);
                    }
                  */
                  c = (ConstructorItem)type.getConstructors().get(0);
                  
                  if ( c.getParameterTypes().length > 0 ) {
                     newObject = new ConstructorBox( parent, c ).getResult();
                  } else {
                     newObject = c.newInstance();
                  }
                  //}
               
               /*
               Display display = (Display)parent;
                  //JacObject.getCollaboration().getAttribute("Gui.display");
               if (display!=null) {
                  boolean ok = display.showModal(
                     newObject,
                     "Creation of a new "+
                     NamingConventions.getShortClassName(newObject.getClass()),
                     "Fill the needed information and validate.",
                     true, true, false);
               } else {
                  Log.warning("No display available");
               }
               */
            }
         } catch (Exception e) {
            Log.error("Instanciation of "+type.getName()+" failed : "+e);
            e.printStackTrace();
         }
         if( newObject != null ) {
            choice.setSelectedItem(addObject( newObject ));
         }
         validate();
      } else if (evt.getActionCommand().equals("open")) {
         if( choice.getSelectedItem() == null ) return;
         Object object = objects.get(choice.getSelectedItem());
         if (object!=null) {
            ObjectView.addViewFor( object, parent, parent );
         }
      }
   }

   public Component getValueComponent() {
      // This is a very dirty hack to get the real data component since
      // JComboBox does not fire the focus events (certainly a bug in 
      // JDK 1.3
      if(choice.isEditable()) {
         return choice.getComponent(2);
      } else {
         return choice.getComponent(0);
      }
   }

   /**
    * Set the focus on the JComboBox
    */
   public void requestFocus() {
      choice.requestFocus();
      Log.trace("gui.focus","focusing "+choice.getClass().getName());
   }

   public void onSetFocus(Object extraOption) {}

} 
