/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.text.*;
import java.io.*;
import java.lang.reflect.*;
import jac.core.*;
import jac.core.rtti.*;
import jac.util.*;
import jac.wrappers.*;
import jac.aspects.naming.*;

/**
 * This class defines a generic view on JAC objects.
 *
 * <p>It contains the fields and an action button for each method. For
 * each field and depending on its type, it embeds the needed view.
 *
 * @see FieldView
 * @see CollectionView
 * @see ReferenceView
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> */

public class ObjectView extends JComponent 
   implements View, ActionListener
{
   private Object substance = null;
   private View parentView = null;
   private Display display = null;
   
   Vector state_components = new Vector();
   HashMap savedState = new HashMap(); // FieldItem -> Value
   Object modalResult;

   String[] categories = null;

   JTabbedPane tabbedPane = new JTabbedPane();

   Window parent;

   boolean isCustomizedView = false;

   ClassItem classItem;

   public static boolean addViewFor ( final Object substance, Window owner, 
                                      Window parent ) {
      return addViewFor( substance, owner, parent, null, null, 
                         false, false, true );
   }

   static class ObjectViewDialog extends JDialog implements ActionListener {
      public boolean ok=false;
      JButton okButton;
      JButton cancelButton;
      JButton closeButton;
      View objectView;
      Window display;

      public ObjectViewDialog(Frame owner,
                              Window display,
                              ObjectView view,
                              String title,
                              String header, 
                              boolean okay, 
                              boolean cancel, boolean close) {
         super(owner);
         init(display,view,title,header,okay,cancel,close);
      }

      public ObjectViewDialog(Dialog owner,
                              Window display,
                              ObjectView view,
                              String title,
                              String header, 
                              boolean okay, 
                              boolean cancel, boolean close) {
         super(owner);
         init(display,view,title,header,okay,cancel,close);
      }

      void init(Window display,
                ObjectView view,
                String title,
                String header, 
                boolean okay, 
                boolean cancel, boolean close) {
         setTitle(title);
         getContentPane().setLayout(new BoxLayout(getContentPane(),BoxLayout.Y_AXIS));
         if( okay || cancel ) {
            ((AbstractSwingDisplay)display).addDialog(this);
            setModal(true);
         }

         if( header != null ) {
            JPanel p=new JPanel();
            p.setBorder( BorderFactory.createEtchedBorder() );
            p.add(new JLabel(header,SwingConstants.LEFT));
            getContentPane().add(p);
         }
         if( view != null ) {
            getContentPane().add(view);
            objectView = (View)view;
         }

         this.display = display;

         if( okay || cancel || close ) {
            JPanel p2 = new JPanel();
            p2.setBorder( BorderFactory.createEtchedBorder() );
            if( okay ) {
               okButton = addButton( p2, "Ok" );
            }
            if( cancel ) {
               cancelButton = addButton( p2, "Cancel" );
            }
            if( close ) {
               closeButton = addButton( p2, "Close" );
            }
            getContentPane().add( p2 );

         }
         pack();

         // open the box centered in the screen...
         Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
         Rectangle rect = getBounds();
         double left = (screenDim.getWidth()-rect.getWidth())/2;
         double top = (screenDim.getHeight()-rect.getHeight())/2;
         Rectangle newRect = new Rectangle(
            (int)left,(int)top,
            (int)rect.getWidth(),(int)rect.getHeight());
         setBounds(newRect);

         show();
      }
      
      public void actionPerformed(ActionEvent evt) {
         try {
            Object source = evt.getSource();
            if(source == okButton || source == closeButton) {
               if( objectView != null ) {
                  objectView.close();
               }
               ok = true;
               ((AbstractSwingDisplay)display).removeDialog(this);
               dispose();
            } else if (source == cancelButton) {
               ok = false;
               ((AbstractSwingDisplay)display).removeDialog(this);
               dispose();
            }
         } catch ( Exception e ) {
            e.printStackTrace ();
         }
      }
      public JButton addButton(Container c, String name)  {
         JButton button = new JButton(name);
         button.addActionListener(this);
         c.add(button);
         return button;
      }
   }

   /**
    * Adds a view for a given Jac object.<p>
    *
    * @param substance the object to add a view for 
    * @return true if the OK button or the close button were clicked.
    */

   public static boolean addViewFor ( final Object substance, Window parent, 
                                      Window owner,
                                      String title, String header, 
                                      boolean okButton, boolean cancelButton,
                                      boolean closeButton ) {

      if( title == null ) {
         if (substance!=null) {
            Class substance_type = substance.getClass();
            String tn = substance_type.getName();
            
            title = tn.substring( tn.lastIndexOf('.') + 1) + " " +
               NameRepository.get().getName( substance ) + " -" +
               jac.core.dist.Distd.getLocalContainerName() + "-";
         } else {
            if( title == null ) {
               title= "<null> -" +
                  jac.core.dist.Distd.getLocalContainerName() + "-";
            }
         }
      }

      ObjectViewDialog view;
      if( owner instanceof Frame ) {
         view = new ObjectViewDialog(
            (Frame)owner,parent,
            substance==null?null:new ObjectView(parent,null,substance),
            title,header,okButton,cancelButton,closeButton);
      } else {
         view = new ObjectViewDialog(
            (Dialog)owner,parent,
            substance==null?null:new ObjectView(parent,null,substance),
            title,header,okButton,cancelButton,closeButton);         
      }
         //      final ObjectView v;

      return view.ok;

   }

   public ObjectView( Window parent, 
                      View parentView, final Object substance, 
                      boolean okButton, boolean cancelButton ) {
      this(parent, parentView, substance);
   }

   Object[] wrappers;

   /**
    * Constructs a view on a wrappee.<p>
    *
    * This constructor builds a panel view of the wrappee and a panel
    * view for each of its wrappers (using
    * <code>construct</code>). Then it adds them to a tabbed pane.<p>
    *
    * @param substance the object that is displayed by the view
    * @see #construct(Object,String) */

   public ObjectView( Window parent, 
                      View parentView, final Object substance ) {
      super();

      this.parent = parent;
      this.parentView = parentView;
      this.substance = substance;
      //      this.display = display;

      if (substance == null) {
         add ( construct( substance, null ) );
         return;
      }

      classItem = ClassRepository.get().getClass( substance.getClass() );

      if( JacObject.getCollaboration().getAttribute("Gui.autoCreation")
          == null ) {

         String customizedView = (String)classItem.getAttribute("Gui.customizedView");
         
         if( customizedView != null ) {
            try {
               add( (Component) Class.forName(customizedView)
                    . getConstructor(new Class[]{Display.class,Container.class,Object.class})
                    . newInstance(new Object[]{parent,this,substance}) );
               isCustomizedView = true;
               return;
            } catch(Exception e) {
               e.printStackTrace();
            }
         }
      }
      
      String tn = classItem.getName();
      
      if ( substance instanceof Wrappee ) {
         ((Wrappee)substance).forceTreatment();
         Object[] viewWrappers = ((Wrappee)substance).getWrappers(ViewControlWrapper.class);
         if( viewWrappers.length == 0 ) {
            Log.error("Wrappee "+substance+" not wrapped by ViewControlWrapper");
         } else {
            ((ViewControlWrapper)viewWrappers[0]).addView( this );
         }
         wrappers = ((Wrappee)substance).getWrappers();
      }

      JPanel pane;

      setLayout( new BoxLayout( this, BoxLayout.Y_AXIS ) );      

      categories = (String[])classItem.getAttribute( "Gui.categories" );

      ClassItem cli = ClassRepository.get().getClass(substance.getClass());

      if ( categories == null ) {
         pane = construct( substance, null ); 
         add ( pane );
         // This is not needed anymore... handeled on focus changes
         if (!GuiAC.getCommitOnFocusLost() &&
             GuiAC.hasEmbeddedEditors(cli))
            addCommitButton(pane);
         addBigGlue( this );
      } else {
         for ( int i = 0; i < categories.length; i ++ ) {
            pane = construct( substance, categories[i] );
            addBigGlue( pane );
            JScrollPane sp = new JScrollPane( pane );
            tabbedPane.addTab( categories[i], null, sp , null );         
         }
         // This is not needed anymore... handeled on focus changes
         if (!GuiAC.getCommitOnFocusLost() && GuiAC.hasEmbeddedEditors(cli)) {
            JPanel p = new JPanel();
            p.setLayout( new BoxLayout( p, BoxLayout.Y_AXIS ) );
            p.add(tabbedPane);
            addCommitButton(p);
            add(p);
         } else {
            add( tabbedPane );
         }
      }

      validate();
      
   }
   
   protected void addCommitButton(JComponent c) {
      JPanel buttonsPanel = new JPanel();
      JButton button;

      buttonsPanel.setBorder( BorderFactory.createEtchedBorder() );
      
      button = new JButton("Commit");
      button.setActionCommand("commit");
      button.addActionListener(this);
      buttonsPanel.add(button);
      
      button = new JButton("Cancel");
      button.setActionCommand("cancel");
      button.addActionListener(this);
      buttonsPanel.add(button);

      c.add(buttonsPanel);
   }

   private void addBigGlue( JComponent component ) {
      for ( int i = 0; i < 25; i++ ) {
         component.add( Box.createVerticalGlue() );
      }
   } 

   private void addBigHGlue( JComponent component ) {
      for ( int i = 0; i < 25; i++ ) {
         component.add( Box.createHorizontalGlue() );
      }
   } 

   /**
    * Builds a sub-view of a substance object within a panel.<p>
    *
    * @param substance the viewed object
    * @param category the category to display
    * @return the panel that corresponds to the view */

   public JPanel construct( Object substance, String category ) {

      JPanel resultPane = new JPanel();
      resultPane.setLayout( new BoxLayout( resultPane, BoxLayout.Y_AXIS ) );

      if (substance == null) {
         resultPane.add(new JLabel("null"));
         resultPane.setLayout(new BoxLayout( resultPane, BoxLayout.Y_AXIS ));
         return resultPane;
      }

      Class substance_type = substance.getClass();
      ClassItem cli = ClassRepository.get().getClass(substance_type);

      JPanel fieldpane = new JPanel();
      fieldpane.setLayout( new BoxLayout( fieldpane, BoxLayout.Y_AXIS ) );

      /*
        // Distribution hack : for demo only

      if( (substance instanceof Wrappee) && ((((Wrappee)substance).isExtendedBy( 
         jac.aspects.distribution.consistency.ClientServerConsistencyWrapper.class ) && 
          ! ((jac.core.dist.RemoteRef[])((Wrappee)substance).invokeRoleMethod( 
             "getKnownReplicas", new Object[0] ))[0]
              . getRemCont().isLocal()) || ((Wrappee)substance).isExtendedBy( 
             jac.core.dist.StubWrapper.class))  ) 
      {
         fieldpane.add( new JLabel( "<<CLIENT>>" ) );
      } else if( (substance instanceof Wrappee) && ((Wrappee)substance).isExtendedBy( 
         jac.aspects.distribution.consistency.StrongPushConsistencyWrapper.class ) ) 
      {
         fieldpane.add( new JLabel( "<<REPLICA-STRONG>>" ) );
      } else if( (substance instanceof Wrappee) && ((Wrappee)substance).isExtendedBy( 
         jac.aspects.distribution.LoadBalancingAC$LoadBalancingWrapper.class ) ) 
      {
         fieldpane.add( new JLabel( "<<LOAD-BALANCER>>" ) );
      }
      */



      // primitive fields and references

      //System.out.println("creating the view, gui.autocreation= "+JacObject.getCollaboration().getAttribute("Gui.autoCreation"));
      
      FieldItem[] fields=null;
      if( JacObject.getCollaboration().getAttribute("Gui.autoCreation") != null ) {
         fields = (FieldItem[])cli.getAttribute("Gui.creationAttributesOrder");
      }
      if( fields == null ) {
         fields = (FieldItem[])cli.getAttribute("Gui.attributesOrder");
      }
      if (fields == null) {
         fields = cli.getFields();
      } 
      int[] groups=(int[])cli.getAttribute("Gui.groups");
      int curgroup=0;
      JPanel subFieldPane = null;
      for (int i = 0; i < fields.length; i++) {
         if(groups!=null && groups[curgroup]==i) {
            if(subFieldPane!=null) {
               fieldpane.add(subFieldPane);
               addBigHGlue(subFieldPane);
            }
            subFieldPane = new JPanel();
            addBigHGlue(subFieldPane);
            subFieldPane.setLayout( new BoxLayout( subFieldPane, BoxLayout.X_AXIS ) );
            curgroup++;
         }
         FieldItem field = fields[i];
         //if(field==null) continue;
         FieldItem oppositeRole = 
            (FieldItem)JacObject.getCollaboration().getAttribute("Gui.oppositeRole");
         Log.trace("associations","opposite role = "+oppositeRole);
         if ( GuiAC.isVisible(field)
              && ( category == null || category != null &&
                   field.getAttribute( "Gui.category" ) != null 
                   && field.getAttribute( "Gui.category" ).equals( category ) )
              && ( oppositeRole!=field ) )
         {
            Component panel;
            if (field.isPrimitive()) {
               if(subFieldPane == null) {
                  fieldpane.add(getPrimitiveFieldPane(substance,field));
               } else {
                  subFieldPane.add(getPrimitiveFieldPane(substance,field));
               }
            } else if (field.isReference()) {
               if(subFieldPane == null) {
                  fieldpane.add(getReferenceFieldPane(substance,field));
               } else {
                  subFieldPane.add(getReferenceFieldPane(substance,field));
               }
            } else if (field instanceof CollectionItem) {
               if(subFieldPane == null) {
                  fieldpane.add(getCollectionPane(substance,(CollectionItem)field));
               } else {
                  subFieldPane.add(getCollectionPane(substance,(CollectionItem)field));
               }
            }
         }
      }
      if(subFieldPane!=null) {
         fieldpane.add(subFieldPane);
         addBigHGlue(subFieldPane);
      }
      // collections
      /*CollectionItem[] collections = cli.getCollections();
      for (int i = 0; i < collections.length; i++) {
         CollectionItem collection = collections[i];
         if ( GuiAC.isVisible(collection)
              && ( category == null || category != null && 
                   collection.getAttribute( "Gui.category" ) != null 
                   && collection.getAttribute( "Gui.category" ).equals( category ) ) ) 
         {
            fieldpane.add(getCollectionPane(substance,collection));
         }
         }*/
      
      resultPane.add( fieldpane );
      
      Log.trace("gui.methods","getMethodsPane for "+cli);

      Collection meths = new Vector();
      if( JacObject.getCollaboration().getAttribute("Gui.autoCreation") == null ) {
         MethodItem[] ms = (MethodItem[])cli.getAttribute("Gui.methodsOrder");
         if( ms != null ) {
            meths = Arrays.asList(ms);
         }
      }

      Component methodsPane = getMethodsPane(substance, meths, category );
      if (methodsPane!=null) {
         resultPane.add( methodsPane );
      }
      
      return resultPane;
   }

   /**
    * Returns the Swing coponent that will display the methods.
    *
    * @return null is no method is to be displayed */

   protected Component getMethodsPane(Object substance, 
                                      Collection methods, String category) {
      JPanel methodpane = new JPanel();
      
      //methodpane.setLayout( new BoxLayout( methodpane, BoxLayout.Y_AXIS ) );
          
      JPanel subMP = new JPanel();
         
      Iterator it = methods.iterator();
      int i = 0;
      int nbMethods = 0;

      while( it.hasNext() ) {
            
         MethodItem curmeth = (MethodItem) it.next();
            
         if ( i == 3 ) {
            methodpane.add( subMP );
            subMP = new JPanel();
            i = 0;
         }
             
         if ( curmeth == null ) continue;
            
         if ( ! GuiAC.isVisible(curmeth) ) {
            Log.trace("gui.methods","invisible, skipping "+curmeth);
            continue;
         }

         if (curmeth.getName().equals("toString")) {
            Log.trace("gui.methods","skipping toString()");
            continue;
         }

         if ( ! ( category == null || category != null && 
                  curmeth.getAttribute( "Gui.category" ) != null 
                  && curmeth.getAttribute( "Gui.category" ).equals( category ) ) ) {
            Log.trace("gui.methods","category, skipping "+curmeth);
            continue;
         }

         if ( ClassRepository.isJacMethod( curmeth.getName() ) || 
              Modifier.isStatic( curmeth.getActualMethod().getModifiers() ) ) {
            Log.trace("gui.methods","JAC method, skipping "+curmeth);
            continue;
         }
            
         if ( (substance instanceof Wrapper) && 
              Wrapper.defines( curmeth.getName() ) ) {
            Log.trace("gui.methods","Wrapper, skipping "+curmeth);
            continue;
         }
            
         if ( (substance instanceof AspectComponent) && 
              Wrapper.defines( curmeth.getName() ) ) {
            Log.trace("gui.methods","AC, skipping "+curmeth);
            continue;            
         }

         if ( ( NamingConventions.isSetter( curmeth.getName() ) && 
                curmeth.getWrittenFields() != null ) ||
              ( NamingConventions.isGetter( curmeth.getName() ) && 
                curmeth.getAccessedFields() != null ) ||
              ( NamingConventions.isAdder( curmeth.getName() ) && 
              curmeth.getAddedCollections() != null ) ||
              ( NamingConventions.isRemover( curmeth.getName() ) && 
                curmeth.getRemovedCollections() != null ) ) {
            Log.trace("gui.methods","setter..., skipping "+curmeth);
            continue;
         }
            
         if ( (substance instanceof Wrappee) && 
              ( ! ( ((Wrappee)substance).isExtendedBy( 
                 jac.aspects.distribution.consistency.ClientServerConsistencyWrapper.class ) && 
                    ! ((jac.core.dist.RemoteRef[])((Wrappee)substance)
                       . invokeRoleMethod( "getKnownReplicas", new Object[0] ))[0]
                    . getRemCont().isLocal() ) ) ) {
            if ( NamingConventions.fieldForMethod( 
               substance.getClass(), curmeth.getName() ) != null ) {
               Log.trace("gui.methods","replica, skipping "+curmeth);
               continue;
            }
         }

         String text = NamingConventions.textForName( curmeth.getName() );
         if ( curmeth.getParameterTypes().length > 0 ) {
            text += "...";
         }
         JButton button = new JButton ( text );
         button.setActionCommand( curmeth.getFullName() );
         button.addActionListener( this );
         subMP.add( button );
            
         i++;
         nbMethods++;
      }

      if (nbMethods>0) {
         methodpane.add( subMP );
         return methodpane;
      } else {
         return null;
      }
   }

   /**
    * Returns the Swing coponent that will display a given collection
    * in the viewed object.
    *
    * @param substance the object that contains the collection
    * @param collection the collection to show
    * @return the associated view
    * @see CollectionView */

   protected Component getCollectionPane(Object substance, 
                                         CollectionItem collection) {
      Log.trace("gui","collection : " + collection.getName());
      JPanel newPane;
      if (collection.isMap() && collection.getAttribute("Gui.tableView")==null) {
         newPane = new MapView(parent,this,substance,collection);
      } else {
         newPane = new CollectionView(parent,this,substance,collection);
      }
      state_components.add( newPane );
      newPane.setName( collection.getName() );
      if ( collection.getActualCollection( substance ) instanceof Wrappee ) {
         ((Wrappee)collection.getActualCollection( substance ))
            . wrapModifiers( new ViewControlWrapper( (View)newPane ), "updateView" );
      }
      return newPane;
   }

   /**
    * Returns the Swing container that will contain a field or a
    * reference view.
    *
    * @param field the field to contain
    * @param border true if bordered
    * @return the container
    * @see #getPrimitiveFieldPane(Object,FieldItem)
    * @see #getReferenceFieldPane(Object,FieldItem) */

   protected Container getFieldContainer(FieldItem field, boolean border) {
      JPanel container = new JPanel();
      if (!border) {
         if(field.getAttribute("rtti.fieldType")==null || 
            field.getAttribute("rtti.fieldType")!=ClassRepository.get().getObject("text")) {
            container.add(new JLabel(
               NamingConventions.textForName(field.getName())+": "));
         }
      } else {
         container.setBorder( 
            BorderFactory.createTitledBorder( 
               BorderFactory.createLineBorder( Color.black ),
               NamingConventions.textForName(" "+field.getName()+" ")));
      }
      container.setLayout(new FlowLayout(FlowLayout.CENTER,1,1));
      return container;
   }

   /**
    * Returns the Swing coponent that will display a given primitive
    * field in the viewed object.
    *
    * <p>A field view is contained in a field container generated by
    * <code>getFieldContainer</code>.
    *
    * @param substance the object that contains the field
    * @param field the field to show
    * @return the associated view component
    * @see FieldView
    * @see #getFieldContainer(FieldItem,boolean) */

   protected Component getPrimitiveFieldPane(Object substance, FieldItem field) {
      Log.trace("gui","primitive field : " + substance + ","+field.getName());
      boolean border = false;
      if( field.getAttribute("Gui.border" ) != null && 
          field.getAttribute("Gui.border" ).equals("true") ) {
         border = true;
      }
      Container container = getFieldContainer(field,border);
      Component pane = null;
      if( GuiAC.isEmbeddedEditor(field) ) {
         pane = new FieldEditor(parent,substance,field);
      } else {
         pane = new FieldView(parent,substance,field);
      }
      pane.setName(field.getName());
      container.add(pane);
      state_components.add(pane);
      return container;
   }

   /**
    * Returns the Swing coponent that will display a given reference
    * field in the viewed object.
    *
    * <p>A reference view is contained in a field container generated by
    * <code>getFieldContainer</code>.
    *
    * @param substance the object that contains the reference
    * @param field the reference to show
    * @return the associated view component
    * @see ReferenceView
    * @see #getFieldContainer(FieldItem,boolean) */

   protected Component getReferenceFieldPane(Object substance, FieldItem field) {
      Log.trace("gui.ref","reference field : " + field.getName());
      boolean border = false;
      if( (field.getAttribute("Gui.border" ) != null && 
          field.getAttribute("Gui.border" ).equals("true")) ||
          field.getAttribute("Gui.embeddedView")!=null ) {
         border = true;
      }
      Container container = getFieldContainer(field,border);
      Component pane = null;
      if( field.getAttribute("Gui.embeddedView")!=null ) {
         pane = new ObjectView( parent, parentView, field.getThroughAccessor(substance) );
      } else {
         if( GuiAC.isEmbeddedEditor(field) ) {
            pane = new FieldEditor(parent,substance,field);
         } else {         
            pane = new ReferenceView(parent,substance,field);
         }
      }
      pane.setName(field.getName());
      container.add(pane);
      state_components.add(pane);
      return container;
   }

   /**
    * Defines what happens when the user pushes a button corresponding
    * to a method, or when the Commit or Cancel button is pressed.<p>
    *
    * @param evt the event that tells what happened */

   public void actionPerformed(ActionEvent evt) {

      String arg = evt.getActionCommand();
      Class substance_type = substance.getClass();
      ClassItem cli = ClassRepository.get().getClass( substance.getClass() );
      FieldItem fi;
      CollectionItem ci;

      Log.trace("gui","Action : "+arg);

      if (arg.equals("commit")) {
         updateState();
      } else if (arg.equals("cancel")) {
         revertState();
      } else {
         Object tmp_substance = substance;
         /*if ( tabbedPane.getSelectedIndex() != 0 ) {
           tmp_substance = wrappers[tabbedPane.getSelectedIndex()-1];
           }*/
         cli = ClassRepository.get().getClass( tmp_substance.getClass() );
         
         if( arg.indexOf( ' ' ) != -1 ) {
            String subarg = null;
            subarg = arg.substring( arg.indexOf( ' ' ) + 1 );
            arg = arg.substring( 0, arg.indexOf( ' ' ) );
         }
         
         if ( cli.hasMethod(arg) ) {
            MethodItem mi = cli.getMethod( arg );
            invoke(mi,substance);
         } 
      }
   }

   /**
    * A GUI level method to invoke a method on an object.
    *
    * @param method the invoked method
    * @param object the invoked object
    * @param collection the corresponding collection item if the
    * invoked object is a collection */

   public void invoke(AbstractMethodItem method, Object object) 
   {
      /* TODO: factorize with ProgramView.invoke() */
      if ( method.getParameterTypes().length == 0 ) {
         Log.trace("gui","Invoking " + method + " on " + object );
         InvokeThread.run( 
            object, method, new Object[0],
            null, null,
            new String[] { "Gui.display", "Session.sid" }, 
            new Object[] { (Display)parent, 
                           "Swing"+jac.core.dist.Distd.getLocalContainerName() } );
      } else {
         Log.trace("gui","Invoking "+method+" on "+object+" (ask for parameters is on)");
         InvokeThread.run( 
            object, method, 
            new Object[method.getParameterTypes().length], 
            null, null, 
            new String[] { "Gui.display", "Session.sid", "Gui.askForParameters" }, 
            new Object[] { (Display)parent, 
                           "Swing"+jac.core.dist.Distd.getLocalContainerName(), 
                           method } );
      }
   }

   /**
    * Updates the state of the substance using the current values of
    * the fields editors if any. */

   protected void updateState() {
      
      JacObject.getCollaboration().addTransientLocalAttribute(
         "Gui.display", parent);
      Iterator it = state_components.iterator();
      while (it.hasNext()) {
         Component cur = (Component)it.next();
         if( cur instanceof FieldEditor ) {
            FieldEditor editor = (FieldEditor)cur;
            if (editor.valueHasChanged()) {
               FieldItem editedField = editor.getField();
               editedField.setThroughWriter(
                  substance, editor.getValue() );
               savedState.put(editedField,editor.getValue());
            }
         }
      }
   }

   protected void revertState() {
      Iterator it = state_components.iterator();
      while (it.hasNext()) {
         Component cur = (Component)it.next();
         if( cur instanceof FieldEditor ) {
            FieldEditor editor = (FieldEditor)cur;
            if (editor.valueHasChanged()) {
               FieldItem editedField = editor.getField();
               editor.setValue(savedState.get(editedField));
            }
         }
      }
   }

   // interface View

   public Object getSubstance() {
      return substance;
   }
   
   public View getParentView() {
      return parentView;
   }

   public void refreshView() {
      Log.trace("gui.viewControl","refreshView");
      for ( int i = 0; i < state_components.size(); i++ ) {
         ((View)state_components.get(i)).refreshView();
      }
   }

   public void refreshViewItem( FieldItem field, MethodItem method, Object[] args ) {
      Log.trace("gui.viewControl","refreshViewItem("+field+") on "+classItem);
      for ( int i = 0; i < state_components.size(); i++ ) {
         if ( ((Component) state_components.get(i)).getName().equals( field.getName() ) ) {
            ((View)state_components.get(i)).refreshView();
         }
      }
   }  

   FieldItemView getViewForField(FieldItem field) {
      Iterator i = state_components.iterator();
      while(i.hasNext()) {
         FieldItemView view = (FieldItemView)i.next();
         if (view.getFieldItem()==field)
            return view;
      }
      return null;
   }

   public void setFocus(FieldItem field, Object extraOption) {
      String category = GuiAC.getCategory(field);
      if (category!=null) {
         tabbedPane.setSelectedIndex(tabbedPane.indexOfTab(category));
         FieldItemView view = getViewForField(field);
         if (view!=null) {
            view.setFocus(field,extraOption);
         }
      }
   }

   public void close() {
      if ( substance instanceof Wrappee ) {
         ((Wrappee)substance).invokeRoleMethod("removeView",new Object[] {this} );
      }
      if( isCustomizedView ) {
         ((View)getComponent(0)).close();
      } else {
         updateState();
      }
   }

}
