/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.reflect.*;
import jac.core.*;
import jac.core.rtti.*;
import jac.util.*;
import jac.wrappers.*;
import jac.aspects.naming.*;
import java.net.*;
import java.io.*;

/**
 * This class implements the default view that allows the user to
 * introspect and manipulate JAC programs (an that objects that run
 * whithin the system).
 *
 * <p>This display is used for the JAC administration interface.
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> */

public class ProgramView extends AbstractSwingDisplay 
   implements ActionListener, TextListener,
              MouseListener, ItemListener, TreeSelectionListener {
   
   static String[] BROWSER_HEADER = { "Browser", "Navigateur" };
   static String[] OBJECTS_HEADER = { "Components", "Composants" };
   static String[] ASPECTS_HEADER = { "Aspects", "Aspects" };
   static String[] M_SYSTEM = { "System", "Systeme" };
   static String[] DM_SYSTEM = { "This menu contains operations on the running JAC system", "Ce menu contient les operations possibles a effectuer sur ce systeme JAC" };
   static String[] M_SYSTEM_CONTAINERS = { "Containers", "Conteneurs" };
   static String[] M_SYSTEM_TRACE = { "Trace", "Trace" };
   static String[] M_SYSTEM_EXIT = { "Exit", "Quitter" };
   static String[] M_PROGRAMS = { "Applications", "Applications" };
   static String[] DM_PROGRAMS = { "This menu allows the user to manage JAC applications", "Ce menu permet a l'utilisateur de gerer des applications JAC" };
   static String[] M_PROGRAMS_LAUNCH = { "Launch", "Lancer" };

   static String[] MSG_EXIT = { "Do you really want to quit JAC?", "Voulez-vous vraiment quitter JAC ?" };
   static String[] PROGRAM = { "Application", "Application" };
   static String[] PROGRAM_TITLE = { "Select an application to launch", "Selectionnez une application a lancer" };

   static String PROGRAMS_PATH = Jac.getJacRoot() + "src/jac/samples";   

   String programName;

   //Display display;

   JTree objects = null;
   JTree wrappersTree;
   DefaultMutableTreeNode wrappersRoot = new DefaultMutableTreeNode("root");
   
   Hashtable programRoots = new Hashtable();

   JLabel text = new JLabel();

   JPanel aspects = new JPanel();
   Hashtable aspectChecks = new Hashtable();

   Hashtable programClassNodes = new Hashtable();

   JPopupMenu classPopup;
   JPopupMenu wrapperPopup;
   JPopupMenu methodPopup;
   JPopupMenu defaultPopup;

   Object actionedObject = null;
   DefaultMutableTreeNode actionedNode = null;

   Vector programs = new Vector();
   ButtonGroup programMenuGroup;
   JMenu programMenu = null;
   JTabbedPane tabbedPane = new JTabbedPane();

   JTextArea consoleView = new JTextArea();
   //Console console = null;
   JPanel objectView = new JPanel();

   JTabbedPane tabbedAspects = new JTabbedPane();

   javax.swing.Timer timer = null;

   JProgressBar memoryBar = new JProgressBar();
   JProgressBar outNetworkBar = new JProgressBar();
   JProgressBar inNetworkBar = new JProgressBar();

   long lastOutput = 0;
   long lastInput = 0;

   static ProgramView instance;

   /**
    * Gets the sole instance of the program view.
    * 
    * @return the instance
    */

   public static ProgramView get() {
      return instance;
   }

   /**
    * Creates a new program view with no program viewed (by default,
    * the JAC system is shown). */
 
   //public ProgramView( ) {
   //   this("jac.system");
   //}

   // display interface

   public void close() {}

   public String getDisplayID() {
      return "admin";
   }

   public void setDisplayID(String displayID) {}

   public void show( Object object ) {
      if (object != null) 
         Log.trace("gui.show","show("+object.getClass().getName()+")");
      else 
         Log.trace("gui.show","show(null)");
      if (object == null) {
         ProgramView.addViewFor( null, this );
      } else if (object instanceof Throwable) {
         ((Throwable)object).printStackTrace();
      } else if (object instanceof Reader) {
         saveStreamToFile((Reader)object);
      } else if (object.getClass().isArray()) {
         ObjectView.addViewFor( Arrays.asList((Object[])object), this, this );
      } else {
         ObjectView.addViewFor( object, this, this );
      }
   }

   public void newObject( Object newObject ) {
      addObject( newObject );
   }   

   /**
    * Constructs a view on a given JAC program.
    *
    * <p>To be compliant with this view, the JAC program must be in a
    * package that corresponds to its name. The package must contain a
    * <code>Main</code> or a <code>Run</code> class to launch the
    * program. Each aspects of the program must be in the package an
    * the naming convention imposes a name of the following form
    * "<programName><aspectName>AC".
    *
    * @param programName the program that is displayed by the viewer
    */

   public ProgramView() {
      super();

      UIManager.put( "Label.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "ComboBox.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "Menu.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "MenuBar.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "Panel.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "Border.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "MenuItem.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "Button.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "RadioButton.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "RadioButtonMenuItem.font", new Font( "SansSerif", Font.PLAIN, 12 ) );
      UIManager.put( "CheckBox.font", new Font( "SansSerif", Font.PLAIN, 12 ) );      
      UIManager.put( "CheckBoxMenuItem.font", new Font( "SansSerif", Font.PLAIN, 12 ) );     
      UIManager.put( "TabbedPane.font", new Font( "SansSerif", Font.PLAIN, 12 ) );

      instance = this;
      
      //      display = new SwingDisplay();

      addWindowListener ( new WindowAdapter () {
            public void windowClosing( WindowEvent e ) {
               ((ACManager)ACManager.get()).whenCloseDisplay((Display)e.getWindow());
               e.getWindow().dispose();
            }
         }
      );

      construct();
      createMenus();
      createPopups();      

      // fill the program menu with the application repository
      addProgram("jac.system",true);
      Collection applications = ApplicationRepository.get().
         getApplications().keySet();
      Iterator it = applications.iterator();
      while( it.hasNext() ) {
         String name = (String)it.next();
         addProgram(name,false);
      }

      // add all the objects in the tree view
      addObjects();

      setProgram("jac.system");
      pack();

      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      setBounds( new Rectangle( screenDim ) );

      show();

   }

   /** adds a new program */
   void addProgram(String name,boolean select) {
      if ( ! programs.contains( name ) ) {
         programs.add( name );
         JRadioButtonMenuItem rbMenuItem = new JRadioButtonMenuItem( name );         
         programMenuGroup.add( rbMenuItem );
         programMenu.add( rbMenuItem );
         rbMenuItem.setSelected(select);
         rbMenuItem.addItemListener( this );
      }
      if( ! programRoots.containsKey( name ) ) {
         programRoots.put( name, 
                           new DefaultMutableTreeNode( 
                              OBJECTS_HEADER[GuiAC.LANGUAGE]+" for "+name ) );
      }
   }
   
   /** adds all the system's objects in the tree view */
   void addObjects() {
      java.util.List objects = JacObject.getObjects();
      Iterator it = objects.iterator();
      while( it.hasNext() ) {
         Object cur = it.next();
         addObject( cur );
      }
   }

   /**
    * Capture the System.out stream so that it appears in a tab of the
    * gui.  
    */

   public void captureSystemOut() {
      System.setOut( new IObservedPrintStream( System.out ) );
   }

   /**
    * Capture the System.err stream so that it appears in a tab of the
    * gui.  
    */

   public void captureSystemErr() {
      System.setErr( new IObservedPrintStream( System.err ) );
   }

   /**
    * Sets the program to display in this window.
    *
    * @param name the program name */
   
   public void setProgram( String name ) {

      Log.trace("gui.setProgram", "setting program " + name );
      
      addProgram(name,true);
      programName = name;
      if ( jac.core.dist.Distd.getLocalContainerName().equals( "" ) ) {
         setTitle( "JAC: " + programName );   
      } else {
         setTitle( "JAC: " + programName + " -" +
                   jac.core.dist.Distd.getLocalContainerName() + "-" );   
      }
      DefaultMutableTreeNode objectsRoot = (DefaultMutableTreeNode) 
         programRoots.get( programName );

      Log.trace("gui.setProgram", "new objects root is " + objectsRoot +
                ", programRoots = "+programRoots);

      createAspects();

      DefaultTreeModel newTreeModel = new DefaultTreeModel( objectsRoot );
      objects.setModel( newTreeModel );
      objects.validate();

   }

   /** Returns the current program name. */
   public String getProgram() {
      return programName;
   }
   
   /**
    * Initializes the view. 
    */

   public void construct () {

      JPanel left = new JPanel();
      JPanel right = new JPanel();
      left.setLayout( new BoxLayout( left, BoxLayout.Y_AXIS ) );
      right.setLayout( new BoxLayout( right, BoxLayout.Y_AXIS ) );
      
      JLabel label = null;
      label = new JLabel ( OBJECTS_HEADER[GuiAC.LANGUAGE] );
      left.add( label );
   
      JScrollPane treeView;

      createObjects();
      treeView = new JScrollPane(objects);
      //objects.setPreferredSize( new Dimension( 300, 500 ) );

      JScrollPane aspectView = new JScrollPane( aspects );
      aspects.setLayout( new BoxLayout( aspects, BoxLayout.Y_AXIS ) );
      //aspects.setPreferredSize( new Dimension( 300, 200 ) );

      tabbedAspects.addTab(ASPECTS_HEADER[GuiAC.LANGUAGE], null, aspectView, 
                           "Select/unselect to weave/unweave aspects");
      tabbedAspects.setSelectedIndex(0);
      wrappersTree = new JTree( wrappersRoot );
      //wrappersTree.putClientProperty("JTree.lineStyle", "Horizontal");
      wrappersTree.setCellRenderer( new ProgramNodeRenderer() );
      //wrappersTree.addTreeSelectionListener( this );
      wrappersTree.setRootVisible( false );

      tabbedAspects.addTab("Wrappers", null, new JScrollPane( wrappersTree ), 
                           "The wrappers and wrapping methods for a given component or component method");


      JSplitPane leftSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
                                                treeView, tabbedAspects );
      leftSplitPane.setOneTouchExpandable(true);
      leftSplitPane.setDividerSize( 8 );

      left.add( leftSplitPane );

      JScrollPane textView = new JScrollPane( text );
      //      textView.setPreferredSize(new Dimension(500, 600));

      textView.setBackground( Color.white );

      tabbedPane.addTab("Code", ResourceManager.getCodeIcon(), textView, 
                        "Click right on the component class to see the Java code");
      tabbedPane.setSelectedIndex(0);
      
      consoleView.setColumns( 80 );
      tabbedPane.addTab("Console", ResourceManager.getInformationIcon(), new JScrollPane( consoleView ), 
                        "Display program traces and errors");

      tabbedPane.addTab("View", ResourceManager.getViewIcon(), new JScrollPane( objectView ), 
                        "Display a component view");
      objectView.setLayout( new BoxLayout( objectView, BoxLayout.Y_AXIS ) );


      right.add( tabbedPane );

      JSplitPane browserPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                                              left, right);
      browserPane.setOneTouchExpandable(true);
      browserPane.setDividerSize( 8 );

      browserPane.setBorder( 
         BorderFactory.createTitledBorder( 
            BorderFactory.createEtchedBorder(),
            " " + BROWSER_HEADER[GuiAC.LANGUAGE] + " " ) );

      //
      // State panel
      //
      JPanel statePane = new JPanel();
      
      JLabel caption = null;
      
      caption = new JLabel( "Memory: " );
      statePane.add( caption );
      statePane.add( memoryBar );      
      statePane.add( Box.createHorizontalStrut( 20 ) );

      caption = new JLabel( "Network input: " );
      statePane.add( caption );
      statePane.add( inNetworkBar );
      statePane.add( Box.createHorizontalStrut( 20 ) );

      caption = new JLabel( "Network output: " );
      statePane.add( caption );
      statePane.add( outNetworkBar );       

      memoryBar.setMaximum( (int) ( Runtime.getRuntime().totalMemory() / 1000 ) );

      outNetworkBar.setMaximum( 3000 );
      inNetworkBar.setMaximum( 3000 );

      statePane.setBorder( BorderFactory.createLoweredBevelBorder() );
      
      getContentPane().add( browserPane );
      getContentPane().add( statePane, BorderLayout.SOUTH );

      timer = new javax.swing.Timer( 1000, this );
      timer.start();

      leftSplitPane.setDividerLocation(400);
      browserPane.setDividerLocation(300);
      
      //addClasses();
   }

   private void createPopups() {
      JMenuItem menuItem;

      //Create the popup menus.

      // default popup
      defaultPopup = new JPopupMenu();
      menuItem = new JMenuItem( "View..." );
      menuItem.addActionListener( this );
      defaultPopup.add( menuItem );

      // class popup
      classPopup = new JPopupMenu();
      menuItem = new JMenuItem( "Display code" );
      menuItem.addActionListener( this );
      classPopup.add( menuItem );
      menuItem = new JMenuItem( "Display Javadoc" );
      menuItem.addActionListener( this );
      classPopup.add( menuItem );

      // wrapper popup
      wrapperPopup = new JPopupMenu();
      menuItem = new JMenuItem( "Open view (new window)..." );
      menuItem.addActionListener( this );
      wrapperPopup.add( menuItem );
      wrapperPopup.addSeparator();
      menuItem = new JMenuItem( "Display code" );
      menuItem.addActionListener( this );
      wrapperPopup.add( menuItem );
      menuItem = new JMenuItem( "Display Javadoc" );
      menuItem.addActionListener( this );
      wrapperPopup.add( menuItem );

      // method popup
      methodPopup = new JPopupMenu();
      menuItem = new JMenuItem( "Invoke..." );
      menuItem.addActionListener( this );
      methodPopup.add( menuItem );

      //Add listener to components that can bring up popup menus.
      //MouseListener popupListener = new PopupListener();
      objects.addMouseListener(this);   
      wrappersTree.addMouseListener(this);
   }

   private void createMenus() {
      JMenuBar menuBar;
      JMenu menu, submenu;
      JMenuItem menuItem;
      JCheckBoxMenuItem cbMenuItem;
      JRadioButtonMenuItem rbMenuItem;
      String rcDir = ResourceManager.RESOURCES_PATH;

      //Create the menu bar.
      menuBar = new JMenuBar();
      setJMenuBar(menuBar);

      //Build SYSTEM menu.
      menu = new JMenu( M_SYSTEM[GuiAC.LANGUAGE] );
      menu.setMnemonic( KeyEvent.VK_S );
      menu.getAccessibleContext().setAccessibleDescription( DM_SYSTEM[GuiAC.LANGUAGE] );
      menuBar.add(menu);

      //a group of JMenuItems
      menuItem = new JMenuItem( M_SYSTEM_CONTAINERS[GuiAC.LANGUAGE],
                                KeyEvent.VK_C);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_1, ActionEvent.ALT_MASK));
      menuItem.addActionListener(this);
      if ( jac.core.dist.Distd.getLocalContainerName().equals( "" ) ) {
         menuItem.setEnabled( false );
      } else {
         menuItem.setEnabled( true );
      }
      menu.add(menuItem);

      menuItem = new JMenuItem( "Start WEB server",
                                KeyEvent.VK_C);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_W, ActionEvent.ALT_MASK));
      menuItem.addActionListener(this);
      menu.add(menuItem);

      menuItem = new JMenuItem( M_SYSTEM_TRACE[GuiAC.LANGUAGE],
                                KeyEvent.VK_C);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_T, ActionEvent.ALT_MASK));
      menuItem.addActionListener(this);
      menu.add(menuItem);

      menuItem = new JMenuItem( "Force GC",
                                KeyEvent.VK_C);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_G, ActionEvent.ALT_MASK));
      menuItem.addActionListener(this);
      menu.add(menuItem);

      menu.addSeparator();

      menuItem = new JMenuItem( M_SYSTEM_EXIT[GuiAC.LANGUAGE] );
      menuItem.setIcon( ResourceManager.createIcon( "jac/aspects/gui/resources/wordexit.gif") );
      menuItem.setMnemonic(KeyEvent.VK_X);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_X, ActionEvent.CTRL_MASK));
      menuItem.addActionListener(this);
      menu.add(menuItem);

      //Build PROGRAMS menu
      programMenu = new JMenu( M_PROGRAMS[GuiAC.LANGUAGE] );
      programMenu.setMnemonic(KeyEvent.VK_P);
      programMenu.getAccessibleContext().setAccessibleDescription( DM_PROGRAMS[GuiAC.LANGUAGE] );
      menuBar.add(programMenu);

      menuItem = new JMenuItem( M_PROGRAMS_LAUNCH[GuiAC.LANGUAGE],
                                KeyEvent.VK_L);
      menuItem.setAccelerator(KeyStroke.getKeyStroke(
         KeyEvent.VK_2, ActionEvent.ALT_MASK));
      menuItem.addActionListener(this);
      programMenu.add(menuItem);

      menuItem = new JMenuItem( "Applications repository",
                                KeyEvent.VK_R);
      menuItem.addActionListener(this);
      programMenu.add(menuItem);

      programMenu.addSeparator();

      programMenuGroup = new ButtonGroup();

   }

   /** creates the JTree */
   private void createObjects() {
      objects = new JTree();      
      objects.putClientProperty("JTree.lineStyle", "Angled");
      ToolTipManager.sharedInstance().registerComponent( objects );
      objects.setCellRenderer( new ProgramNodeRenderer() );
      objects.addTreeSelectionListener( this );
   }

   
   /**
    * Adds all wrapped classes in the tree view excepts system classes.
    */
   /*   public void addClasses() {
      Object[] classes = ClassRepository.get().getClasses();
      Log.trace("gui.Treeview","Found "+classes.length+" classes");
      for (int i=0;i<classes.length; i++) {
         ClassItem c = (ClassItem)classes[i]; 
         if (!c.isSystemClass()) {
            Log.trace("gui.Treeview","Adding "+c);
            addClass(c.getActualClass());
         }
      }
      }*/

   /**
    * Adds a class in the tree view.
    *
    * @param cl the class to add
    * @return the tree node the class was embedded into
    */

   public DefaultMutableTreeNode addClass( Class cl, 
                                           DefaultMutableTreeNode objectsRoot, 
                                           String program  ) {
      DefaultMutableTreeNode node = null;

      ClassItem cli = ClassRepository.get().getClass(cl);
      Hashtable classesNodes = null;
      if( programClassNodes.containsKey( program ) ) {
         classesNodes = (Hashtable)programClassNodes.get(program);
      } else {
         classesNodes = new Hashtable();
         programClassNodes.put(program, classesNodes);
      }
      if( classesNodes.containsKey( cl.getName() ) ) {
         node = (DefaultMutableTreeNode)classesNodes.get( cl.getName() );
      } else {
         node = new DefaultMutableTreeNode( cli );
         classesNodes.put( cl.getName(), node );
         objectsRoot.add( node );
      }
      return node;
   }
   
   /**
    * Adds an object to the objects tree view.
    *
    * <p>This method should be called each time a new object is
    * created whithin the system (e.g. by a
    * <code>whenNewInstance</code> method).
    *
    * <p>If the object's class that is added is not in the free view,
    * then the class is added.
    *
    * @param o the object to add
    * @see #addClass(Class,DefaultMutableTreeNode,String)
    * @see jac.core.AspectComponent#whenNewInstance(Wrappee) */
 
   public void addObject( Object o ) {
      if ( o == null ) return;
      String applicationName = ((Wrappee)o).getApplication();
      if( applicationName == null ) applicationName = "jac.system";
      Log.trace("gui.addObject"," Adding " + o + " : " + applicationName );
     
      DefaultMutableTreeNode objectsRoot = null;
      if(!programs.contains(applicationName)) {
         addProgram(applicationName,false);
      }
      objectsRoot = (DefaultMutableTreeNode) programRoots.get( applicationName );

      DefaultMutableTreeNode cl = null;
      DefaultMutableTreeNode object = null;
      DefaultMutableTreeNode method = null;
      
      NameRepository nr = (NameRepository) NameRepository.get();

      object = new DefaultMutableTreeNode( o );

      cl = addClass( o.getClass(), objectsRoot, applicationName );
         
      ClassItem cli = ClassRepository.get().getClass( o.getClass().getName() );
      Collection methods = cli.getMethods();
      Iterator it = methods.iterator();
      while( it.hasNext() ) {
         MethodItem[] meth = (MethodItem[]) it.next();
         for (int i=0; i<meth.length;i++) {
            if ( ClassRepository.isJacMethod( meth[i].getName() ) ) continue;
            method = new DefaultMutableTreeNode( meth[i] );
            object.add( method );
         }
      }

      for (Enumeration e = cl.children(); e.hasMoreElements() ;) {
         Object cur = ((DefaultMutableTreeNode)e.nextElement()).getUserObject();
         if ( cur == o ) return;
     }

      cl.add( object );
      //      pack();

      DefaultTreeModel newTreeModel = new DefaultTreeModel( objectsRoot );
      objects.setModel( newTreeModel );

      Log.trace("gui", "End adding " + o );

      //objects.invalidate();
      
   }

   private void createWrappers( Wrappee wrappee, MethodItem method ) {
      wrappersRoot.removeAllChildren();
      if ( wrappee != null && method == null ) {
         wrappee.forceTreatment();
         Object[] wrappers = wrappee.getWrappers();
         for( int i = 0; i < wrappers.length; i++ ) {
            wrappersRoot.add( new DefaultMutableTreeNode( wrappers[i] ) );
         }
      } else if ( wrappee != null && method != null ) {
         Vector wrappingChain = wrappee.getWrappingChain( method.getName() );
         if ( wrappingChain != null ) {
            Iterator it = wrappingChain.iterator();
            while( it.hasNext() ) {
               Object[] cur = (Object[])it.next();
               DefaultMutableTreeNode w = new DefaultMutableTreeNode( cur[0] );
               DefaultMutableTreeNode m = new DefaultMutableTreeNode( cur[1] );
               w.add( m );
               wrappersRoot.add( w );
            }  
         }
      }
      DefaultTreeModel newTreeModel = new DefaultTreeModel( wrappersRoot );
      wrappersTree.setModel( newTreeModel );
      int rc = 0;
      int i = 0;
      do {
         rc = wrappersTree.getRowCount();
         wrappersTree.expandRow( i );
         i += wrappersTree.getRowCount() - rc;
         i++;
      } while ( i < wrappersTree.getRowCount() );
      
   }

   boolean creatingAspects=false;
   
   private void createAspects() {
      Log.trace("gui", "Creating aspects view" + (Vector) aspectChecks.get( programName ) + aspectChecks );
      creatingAspects = true;
      try {
         Vector checks = null;
         if ( ( checks = (Vector) aspectChecks.get( programName ) ) == null ) {
            
            checks = new Vector();
            
            Log.trace("gui", "============" + programName );
            
            Application app = ApplicationRepository.get().getApplication( programName );
            Log.trace("gui", "============" + app );
            if ( app != null ) {
               Collection acConfs = app.getAcConfigurations();
               Iterator it = acConfs.iterator();
               while( it.hasNext() ) {
                  ACConfiguration acConf = (ACConfiguration) it.next();
                  Log.trace("gui", "creating checkbox for " + acConf );
                  JCheckBox cb = new JCheckBox( acConf.getName() );
                  cb.setActionCommand( programName + "." + 
                                       acConf.getName() );
                  checks.add( cb );
                  cb.addItemListener( this );
                  cb.addMouseListener(this);   
               }
               
            } 
            aspectChecks.put( programName, checks );
         }
         
         aspects.removeAll();
         Iterator it = checks.iterator();
         while( it.hasNext() ) {
            JCheckBox cb = (JCheckBox) it.next();
            aspects.add( cb );
            if ( getWovenAC( programName + "." + cb.getText() ) != null ) {
               cb.setSelected( true );
            } else {
               cb.setSelected( false );
            }
         }
         
         aspects.validate();
         aspects.repaint();
      } catch( Exception e ) {
         e.printStackTrace();
      } finally {
         creatingAspects = false;
      }
   
   }

   private AspectComponent getWovenAC( String name ) {
      AspectComponent ac = (AspectComponent) ACManager.get().getObject( name );
      return ac;
   }           

   /**
    * Adds an object view for the given substance in the given program
    * view.
    *
    * @param substance the new viewed object
    * @param parent the parent program view
    * @see ObjectView */  

   public static void addViewFor( Object substance, ProgramView parent ) {

      if ( substance instanceof Wrappee ) {
         Wrappee wrappee = (Wrappee)substance;
         wrappee.forceTreatment();
         if ( wrappee.getWrappers( ViewControlWrapper.class ).length != 0 ) {
            ViewControlWrapper vw = (ViewControlWrapper) wrappee
               . getWrappers( ViewControlWrapper.class )[0];
            if ( parent.objectView.getComponentCount() != 0 ) {
               vw.removeView( (ObjectView) parent.objectView.getComponent( 0 ) );
            }
            if ( vw.getViews().size() == 0 ) {
               wrappee.unwrap( vw );
            }
         }
      }
     
      parent.tabbedPane.setSelectedIndex( 2 );
      if ( parent.objectView.getComponentCount() != 0 ) {
         parent.objectView.remove( 0 );
      }
      parent.objectView.add( new ObjectView( parent, null, substance ) );
      parent.objectView.validate();
   }

   boolean isProgram( String name ) {
      return name.endsWith( ".jac" );
   }
   
   /**
    * Defines what happens when the user pushes a button corresponding
    * to a method.<p>
    *
    * @param evt the event that tells what happened */

   public void actionPerformed(ActionEvent evt) {
      String arg = evt.getActionCommand();

      if ( arg == null ) {

         long curOutput = jac.core.dist.Distd.outputCount;
         long curInput = jac.core.dist.Distd.inputCount;

         int tot = (int) ( Runtime.getRuntime().totalMemory() / 1000 );
         if ( memoryBar.getMaximum() < tot ) {
            memoryBar.setMaximum( tot );
         }
         memoryBar.setValue( tot - (int) ( Runtime.getRuntime().freeMemory() / 1000 ) );

         outNetworkBar.setValue( (int) (curOutput - lastOutput) );
         inNetworkBar.setValue( (int) (curInput - lastInput) );

         lastOutput = curOutput;
         lastInput = curInput;
         
         return;

      }

      Log.trace("gui","--- Action performed: " + arg );
      
      if ( arg.startsWith("New") ) {

         ClassItem classItem = (ClassItem)actionedObject;
         Vector constructors = classItem.getConstructors();
         for (int i=0; i < constructors.size(); i++) {
            ConstructorItem constructor = (ConstructorItem)constructors.get(i);
            if ( arg.equals("New "+NamingConventions.getShortClassName(classItem)+
                            NamingConventions.getPrintableParameterTypes( constructor )) ) 
            {
               if (constructor.getParameterTypes().length==0) {
                  try {
                     constructor.newInstance(new Object[0]);
                  } catch (Exception e) {
                     Log.error("*** Instanciation failed");
                     e.printStackTrace();
                  }
               } else {
                  new ConstructorBox(this,constructor);
               }
            }
         }
         
      } else if ( arg.equals( "Open view (new window)..." ) ) {

         ObjectView.addViewFor( actionedObject, this, this );
         return;

      } else  if ( arg.equals( "View" ) ) {

         addViewFor( actionedObject, this );
         return;
      
      } else if ( arg.equals( "Invoke..." ) ) {
         invoke((MethodItem)actionedObject,
                ((DefaultMutableTreeNode)actionedNode.getParent()).getUserObject(),
                null);
         //View.addViewFor( actionnedObject );
         return;
      } else if ( arg.equals( "Display code" ) ) {
         String object = null;
         if ( actionedObject instanceof ClassItem ) {
            object = ((ClassItem)actionedObject).getName();
         }

         if ( actionedObject instanceof Wrapper ) {
            object = actionedObject.getClass().getName();
         }

         if ( actionedObject instanceof JCheckBox ) {
            object = ((JCheckBox)actionedObject).getActionCommand();
         }

         if ( object != null ) {
            try {
               /*text.setPage( new URL( 
                 "file://" + JacObject.jacRootPath + "src/" +  
                 ((Class)object).getName().replace( '.', '/' ) +
                 ".java" ) );*/
               text.setText( JavaHighlight.getHighlightedText ( 
                  Jac.getJacRoot() + "src/" +
                  object.replace( '.', '/' ) +
                  ".java" ).toString() );
               tabbedPane.setSelectedIndex( 0 );
            } catch ( Exception e ) {
               e.printStackTrace();
            }
         }
         return;
      } else if ( arg.equals( "Display Javadoc" ) ) {
         String object = null;
         if ( actionedObject instanceof ClassItem ) {
            object = ((ClassItem)actionedObject).getName();
         }

         if ( actionedObject instanceof Wrapper ) {
            object = actionedObject.getClass().getName();
         }

         if ( actionedObject instanceof JCheckBox ) {
            object = ((JCheckBox)actionedObject).getActionCommand();
         }
         if ( object != null ) {
            try {
               Runtime.getRuntime().exec( 
                  "netscape " + Jac.getJacRoot() + "doc/javadoc/" +
                  object.replace( '.', '/' ) +
                  ".html" );
            } catch ( Exception e ) {
               e.printStackTrace();
            }

         }
         return;

      } else if ( arg.equals( M_SYSTEM_CONTAINERS[GuiAC.LANGUAGE] ) ) {

         ObjectView.addViewFor( jac.core.dist.Topology.get(), this, this );
         return;         

      } else if ( arg.equals( M_SYSTEM_TRACE[GuiAC.LANGUAGE] ) ) {

         ObjectView.addViewFor( new WrappableMap(jac.util.Log.getLevels()), this, this);
         return;         

      } else if ( arg.equals( "Force GC" ) ) {

         GarbageCollector.forceGC();
         return;         

      } else if ( arg.equals( "Start WEB server" ) ) {

         ((GuiAC)ACManager.get().getObject("gui"))
            .createDisplay("web",jac.aspects.gui.WebGui.class);
         return;         

      } else if ( arg.equals( M_SYSTEM_EXIT[GuiAC.LANGUAGE] ) ) {

         exit();         
         return;
      }
      else if ( arg.equals( "Applications repository" ) ) {
         
         ((GuiAC)ACManager.get().getObject("gui")).createDisplay( 
            "application-repository", ConsoleAdmin.class );         
         return;
         
      } else if ( arg.equals( M_PROGRAMS_LAUNCH[GuiAC.LANGUAGE] ) ) {
         final class ProgramFilter extends javax.swing.filechooser.FileFilter {
            public boolean accept( java.io.File f ) { 
               return ((!f.isFile()) || (isProgram(f.getName()))); 
            }
            public String getDescription() { 
               return "JAC programs"; 
            }
         }
         final class ProgramFileView extends javax.swing.filechooser.FileView {
            public Icon getIcon( java.io.File f ) {
               if ( isProgram( f.getName() ) ) {
                  return ResourceManager.getGroupIcon();
               }
               return super.getIcon( f );
            }
         }
         JFileChooser chooser = new JFileChooser();
         chooser.setFileFilter( new ProgramFilter() );
         chooser.setFileView( new ProgramFileView() );
         chooser.setCurrentDirectory( new java.io.File( PROGRAMS_PATH ) );    
         chooser.setDialogTitle( PROGRAM_TITLE[GuiAC.LANGUAGE] );
         int returnVal = chooser.showOpenDialog( this );
         if( returnVal == JFileChooser.APPROVE_OPTION ) {
            String path = chooser.getSelectedFile().getPath();
            Log.trace("gui", "path: " + path );
            ApplicationRepository.launchProgram(new String[] {path});
            ((ACManager)ACManager.get()).afterApplicationStarted();

         }
         return;
      } else {
      
         Object o = actionedObject;
         ClassItem cli = ClassRepository.get().getClass( o.getClass() );
         MethodItem method = cli.getMethod( arg );
         invoke(method,o,null);
         //View.addViewFor( actionnedObject );
         return;
      }

   }

   /**
    * Exits the application (and the current JAC container).
    */

   public void exit() {
      int option = JOptionPane.showConfirmDialog( 
         this, MSG_EXIT[GuiAC.LANGUAGE], M_SYSTEM_EXIT[GuiAC.LANGUAGE],
         JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE );
      if ( option == JOptionPane.YES_OPTION ) {
         System.exit( 0 );
      }
   }

   /**
    * Handdles checkboxes and radio buttons changes.
    *
    * @param e the change event
    */

   public void itemStateChanged(ItemEvent e) {

      if ( e.getItem() instanceof JRadioButtonMenuItem ) {
         if ( e.getStateChange() == ItemEvent.SELECTED ) {
            setProgram( ((JRadioButtonMenuItem)e.getItem()).getText() );
         }
      }
      if ( e.getItem() instanceof JCheckBox ) {
         String action = ((JCheckBox)e.getItem()).getActionCommand();
         String text = ((JCheckBox)e.getItem()).getText();

         try {

            Application app = ApplicationRepository.get().getApplication( programName );
            Log.trace("gui", "item state changed, app = " + app );
            if ( app == null ) return;
            ACConfiguration acConf = app.getAcConfiguration( text );
            Log.trace("gui", "item state changed, acConf = " + acConf );
            if ( acConf == null ) return;
            
            ACManager acm = (ACManager)ACManager.get();

            if ( e.getStateChange() == ItemEvent.DESELECTED ) {
               if( acm.getObject(programName+"."+text) != null ) { 
                  Log.trace("gui", "Unweaving " + text );
                  ApplicationRepository.get().unextend(programName,text);
                  //acConf.unweave();
               }
            } else if ( e.getStateChange() == ItemEvent.SELECTED ) {
               if( acm.getObject(programName+"."+text) == null ) { 
                  Log.trace("gui", "Weaving " + text );
                  ApplicationRepository.get().extend(programName,text);
                  //acConf.weave();
               }
            }

         } catch ( Exception ex ) {
            ex.printStackTrace();
         }
      }
   }

   /**
    * Do nothing.
    *
    * @param evt changing descriptor */

   public void textValueChanged(TextEvent evt) {
   }

   /**
    * Handles the tree view selection changes. */

   public void valueChanged(TreeSelectionEvent event) {
      DefaultMutableTreeNode node = (DefaultMutableTreeNode)
         objects.getLastSelectedPathComponent();
      
      if (node == null) return;      
      Object o = node.getUserObject();
      if ( o instanceof Wrappee ) {
         createWrappers( (Wrappee) o, null );
      } else if ( o instanceof MethodItem ) {
         Object wrappee = ((DefaultMutableTreeNode)node.getParent()).getUserObject();
         if ( wrappee instanceof Wrappee ) {
            createWrappers( (Wrappee)wrappee, (MethodItem)o );
         }
      } else {
         createWrappers ( null, null );
      }

   }

   /**
    * Shows a popup if needed.
    *
    * @param e mouse event descriptor */

   public void mousePressed(MouseEvent e) {
      maybeShowPopup(e);
   }

   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseClicked(MouseEvent e) {
   }
   
   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseEntered(MouseEvent e) {
   }

   /**
    * Do nothing.
    *
    * @param e mouse event descriptor */

   public void mouseExited(MouseEvent e) {
   }

   /**
    * Closes a popup if needed.
    *
    * @param e mouse event descriptor */

   public void mouseReleased(MouseEvent e) {
      maybeShowPopup(e);
   }
     
   private void maybeShowPopup(MouseEvent e) {
      if ( e.getComponent() == objects || e.getComponent() == wrappersTree ) {
         TreePath tp = ((JTree)e.getComponent()).getPathForLocation( e.getX(), e.getY() );
         if ( tp != null ) {
            Object o = ((DefaultMutableTreeNode) tp.getLastPathComponent()).getUserObject();
            if (e.isPopupTrigger()) {
               actionedObject = o;
               actionedNode = (DefaultMutableTreeNode) tp.getLastPathComponent();
               if ( o instanceof MethodItem ) {

                  methodPopup.show(e.getComponent(), e.getX(), e.getY());

               } else if ( o instanceof ClassItem ) {

                  JPopupMenu classPopup = new JPopupMenu();
                  JMenuItem menuItem = null;
                  menuItem = new JMenuItem( "Display code" );
                  menuItem.addActionListener( this );
                  classPopup.add( menuItem );
                  menuItem = new JMenuItem( "Display Javadoc" );
                  menuItem.addActionListener( this );
                  classPopup.add( menuItem );

                  classPopup.addSeparator();

                  // constructors
                  ClassItem type = (ClassItem)actionedObject;
                  Vector constructors = type.getConstructors();
                  for (int i=0; i < constructors.size(); i++) {
                     ConstructorItem constructor= (ConstructorItem)constructors.get(i);
                     String ret = NamingConventions.getShortClassName( type );
                     ret += NamingConventions.getPrintableParameterTypes( constructor );
                     menuItem = new JMenuItem( "New " + ret );
                     menuItem.addActionListener( this );
                     classPopup.add( menuItem );
                  }

                  classPopup.show(e.getComponent(), e.getX(), e.getY());

               } else if ( o instanceof Wrapper ) {

                  wrapperPopup.show(e.getComponent(), e.getX(), e.getY());

               } else if ( (o instanceof Wrappee) || (o instanceof AspectComponent) ) {

                  JPopupMenu dynPopup = new JPopupMenu();
                  JMenuItem menuItem = null;
                     
                  menuItem = new JMenuItem( "View" );
                  menuItem.addActionListener( this );
                  dynPopup.add( menuItem );
                  menuItem = new JMenuItem( "Open view (new window)..." );
                  menuItem.addActionListener( this );
                  dynPopup.add( menuItem );
                  dynPopup.addSeparator();
                     
                  ClassItem cli = (ClassItem)((DefaultMutableTreeNode)actionedNode.getParent()).getUserObject();
                  Collection meths = cli.getMethods();
                  Iterator it = meths.iterator();
                  while( it.hasNext() ) {                     
                     MethodItem mi = ((MethodItem[])it.next())[0];
                     if ( ClassRepository.isJacMethod( mi.getName() ) ) continue;
                     menuItem = new JMenuItem( mi.getName() );
                     menuItem.addActionListener( this );
                     dynPopup.add( menuItem );
                  }
                  dynPopup.show(e.getComponent(), e.getX(), e.getY());
               }
            }
         }
      } else {
         if (e.isPopupTrigger()) {
            actionedObject = e.getComponent();
            classPopup.show(e.getComponent(), e.getX(), e.getY());
         }
      }
   }
   


   class IObservedPrintStream extends PrintStream {
      public IObservedPrintStream( OutputStream out ) {
         super( out, true );
      }
      public void write(byte[] buf, int off, int len) {
         byte[] newBytes = new byte[len];
         System.arraycopy( buf, off, newBytes, 0, len );
         tabbedPane.setSelectedIndex( 1 );
         consoleView.append( new String( newBytes ) );
         consoleView.setCaretPosition( consoleView.getText().length() );
      }
   }

}

class ProgramNodeRenderer extends DefaultTreeCellRenderer {
   ImageIcon objectIcon;
   ImageIcon classIcon;
   ImageIcon aspectIcon;
   ImageIcon methodIcon;
   ImageIcon fieldIcon;
   ImageIcon privFieldIcon;
   ImageIcon protFieldIcon;
   ImageIcon wrapperIcon;

   public ProgramNodeRenderer() {
      String rootDir = ResourceManager.RESOURCES_PATH;
      objectIcon = ResourceManager.createIcon( rootDir + "object.gif" );
      classIcon = ResourceManager.createIcon( rootDir + "class.gif" );
      aspectIcon = ResourceManager.createIcon( rootDir + "group.gif" );
      methodIcon = ResourceManager.createIcon( rootDir + "pubfunc.gif" );
      fieldIcon = ResourceManager.createIcon( rootDir + "pubslot.gif" );
      privFieldIcon = ResourceManager.createIcon( rootDir + "privslot.gif" );
      wrapperIcon = ResourceManager.createIcon( rootDir + "wrapper.gif" );
   }
   
   public Component getTreeCellRendererComponent( JTree tree,
                                                  Object value,
                                                  boolean sel,
                                                  boolean expanded,
                                                  boolean leaf,
                                                  int row,
                                                  boolean hasFocus) {
      
      super.getTreeCellRendererComponent(
         tree, value, sel, expanded, leaf, row, hasFocus);

      setDefaultAppearence( value );
      
      return this;
   }

   protected void setDefaultAppearence( Object value ) {
      DefaultMutableTreeNode node =
         (DefaultMutableTreeNode)value;
      Object o = node.getUserObject();
      ImageIcon icon = null;
      String text = null;
      String toolTipText = null;
      
      if ( o instanceof Wrappee ) {
         icon = objectIcon;
         text = NameRepository.get().getName( o );
         toolTipText = null;
      }         

      if ( o instanceof Wrapper ) {
         icon = wrapperIcon;
         text = NamingConventions.getShortClassName( o.getClass() );
         if (((Wrapper)o).getAspectComponent()!=null) {
            toolTipText = "from " + 
               NamingConventions.getShortClassName( 
                  ((Wrapper)o).getAspectComponent().getClass() );
         }
      }

      if ( o instanceof Class ) {
         icon = classIcon;
         text = NamingConventions.getShortClassName( (Class)o );
         toolTipText = null;
      }

      if ( o instanceof ClassItem ) {
         icon = classIcon;
         text = NamingConventions.getShortClassName( (ClassItem)o );
         toolTipText = null;
      }

      if ( o instanceof MethodItem ) {
         icon = methodIcon;
         text = o.toString();
         toolTipText = null;
      }

      if ( o instanceof Method ) {
         icon = methodIcon;
         text = ((Method)o).getName();
         toolTipText = null;
      }

      if ( icon == null ) {
         icon = objectIcon;
         text = o.toString();
         toolTipText = null;
      }

      //      Log.trace("gui","set apperance for " + o + ": " + icon + ":" + text );

      if ( icon != null ) setIcon( icon );
      if ( text != null ) setText( text );
      if ( toolTipText != null ) setToolTipText( toolTipText );
      
      
   }

   public void actionPerformed(ActionEvent evt) {
      String arg = evt.getActionCommand();
   }

   //   class PopupListener extends MouseAdapter {
   //   }
   
}

/**
The JavaHighlight class implements a java source code highlighter.
@author Marius Scurtescu
@version CMPT 883, Assignemnt 2, October 1997
*/

//package mas.util;

//import java.util.Properties;
//import java.util.Hashtable;
//import java.util.Date;
//import java.io.*;
//import mas.util.CommandParser;


/**
A very simple Java source code tokenizer, it is used for highlighting.
@see mas.util.JavaHighlight
*/

class JavaTokenizer
{
   /**
      Creates a tokenizer based on an input stream
      @param in The input stream which will be tokenized
   */
   JavaTokenizer (InputStream in)
   {
      strmIn = new PushbackInputStream (in);
      initJavaKeywords ();
   }
   
   /**
      Parses and returns the next token. The token string
      representation can be retrieved from the value variable.
      @return The parsed token id.
      @see value */
   int nextToken ()
   {
      int c;
      
      try
      {
         c = strmIn.read ();
	 
         if (c == -1)
            return TT_EOF;
         else if (c == '/')
         {
            int c2 = strmIn.read ();
            if (c2 == '*')
            {
               value = parseSlashStar ();
               return TT_COMMENT;
            }
            else if (c2 == '/')
            {
               value = parseSlashSlash ();
               return TT_COMMENT;
            }
	    
            strmIn.unread (c2);
         }
         else if (c == '(' || c == ')' || c == '[' 
                  || c == ']' || c == '{' || c == '}')
         {
            value = new String () + (char)c;
            return TT_BRACKET;
         }
         else if (Character.isJavaLetterOrDigit ((char)c))
         {
            value = parseIdentifier (c);
            return identifierType (value);
         }
         else if (Character.isDigit ((char)c))
         {
            value = parseNumber (c);
            return TT_NUMBER;
         }
         else if (c == '\"' || c == '\'')
         {
            value = parseString (c);
            return TT_STRING;
         }
	 
         value = new String () + (char)c;
         return TT_OTHER;
      }
      catch (IOException e)
      {
         Log.error ("Error reading!");
         Log.error (e.getMessage ());
         Log.error (e.toString ());
         return TT_EOF;
      }
   }
        
   /**
      Stores the string representation of the last parsed token.
   */
   String value;
        
   final static int TT_EOF = -1;
   final static int TT_OTHER = 0;
   final static int TT_COMMENT = 1;
   final static int TT_BRACKET = 2;
   final static int TT_CLASS = 3;
   final static int TT_CONSTANT = 4;
   final static int TT_MODIFIER = 5;
   final static int TT_TYPE = 6;
   final static int TT_CONTROL = 7;
   final static int TT_PROCLAMATION = 8;
   final static int TT_LITERAL = 9;
   final static int TT_RESERVED = 10;
   final static int TT_NUMBER = 11;
   final static int TT_STRING = 12;
        
   /**
        Parses a slash star comment. It assumes the a slash and a star were already consumed from
        the inpit stream and it considers nested commens.
        @return The parsed comment, including start and stop sequences.
        */
   private String parseSlashStar () throws IOException
   {
      int crnt, prev;
      int level;
      StringBuffer buf = new StringBuffer ("/*");
                
      crnt = '*';
      level = 1;
      while (level > 0)
      {
         prev = crnt;
         crnt = strmIn.read ();
         if (crnt == -1) // can't be EOF for a correct program, but...
         {
            Log.error ("Unexpected end of file during slash-star comment parse!");
            break;
         }
         buf.append ((char)crnt);
         if (crnt == '*' && prev == '/')
            level++;
         else if (crnt == '/' && prev == '*')
            level--;
      }
                
      return buf.toString ();
   }
        
   /**
        Parses a slash slash comment. It assumes the two slashes where consumed.
        @return The parsed comment, including start and stop sequences.
        */
   private String parseSlashSlash () throws IOException
   {
      int c;
      StringBuffer buf = new StringBuffer ("//");
                
      while (true)
      {
         c = strmIn.read ();
         if (c == -1)
            break;
         if (c == '\r' || c == '\n')
         {
            strmIn.unread (c);
            break;
         }
         buf.append ((char)c);
      }
                
      return buf.toString ();
   }
        
   /**
        Parses a Java identifier.
        @param c The first character of the identifier.
        @return The parsed identifier.
        */
   private String parseIdentifier (int c) throws IOException
   {
      StringBuffer buf = new StringBuffer ();
                
      do
      {
         buf.append ((char)c);
         c = strmIn.read ();
      } while (c != -1 && Character.isJavaLetter ((char)c));
                
      if (c != -1) // can't be EOF for a correct program, but...
         strmIn.unread (c);
                
      return buf.toString ();
   }
        
   /**
        Decides what type of identifier is the passed argumemt. If it is a keyword then
        it is looked up in the keyword table to get its type. If not, then if it is all 
        uppercase the it is a constant, if the first charcter is uppercase then it is
        a class name otherwise it is just an ordinary identifier.
        @param Java identifier.
        @return The identifier code.
        */
   private static int identifierType (String word)
   {
      if (javaKeywords.containsKey (word))
         return ((Integer)javaKeywords.get (word)).intValue ();
                
      if (Character.isUpperCase (word.charAt (0)))
      {
         boolean bAllUpper = true;
                        
         for (int i = 1; i < word.length () && bAllUpper; i++)
            bAllUpper = !Character.isLetter (word.charAt (i)) || 
               Character.isUpperCase (word.charAt (i));
                        
         return bAllUpper ? TT_CONSTANT : TT_CLASS;
      }
                
      return TT_OTHER;
   }
        
   /**
        Parses a number. Only unsigned integers are parsed.
        @param c The first digit.
        @return The parsed number as a string.
        */
   private String parseNumber (int c) throws IOException
   {
      StringBuffer buf = new StringBuffer ();
                
      do
      {
         buf.append ((char)c);
         c = strmIn.read ();
      } while (c != -1 && Character.isDigit ((char)c));
                
      if (c != -1) // can't be EOF for a correct program, but...
         strmIn.unread (c);
                
      return buf.toString ();
   }
        
   /**
        Parses a qutoed string. It is aware of back slash escape sequences.
        @param c The starting quote (it can be single or double quote).
        @return The parsed string, including start and stop quotes.
        */
   private String parseString (int c) throws IOException
   {
      StringBuffer buf = new StringBuffer ();
      char limit = (char)c;
                
      buf.append (limit);
      do
      {
         c = strmIn.read ();
         if (c == -1) // unexpected EOF
         {
            Log.error ("Unexpected end of file during string parse!");
            break;
         }
         if (c == '\n') // syntax error, but we don't care
         {
            Log.error ("Unescaped \n in string! String parse aborted.");
            strmIn.unread (c);
            break;
         }
         buf.append ((char)c);
         if (c == '\\')
            buf.append ((char)(strmIn.read ()));
      } while (c != limit);
                
      return buf.toString ();
   }
        
   /**
        Initializes the java keyword lookup table.
        */
   private static void initJavaKeywords ()
   {
      if (javaKeywords == null)
      {
         javaKeywords = new Hashtable ();
                        
         javaKeywords.put ("abstract", new Integer (TT_MODIFIER));
         javaKeywords.put ("boolean", new Integer (TT_TYPE));
         javaKeywords.put ("break", new Integer (TT_CONTROL));
         javaKeywords.put ("byte", new Integer (TT_TYPE));
         javaKeywords.put ("case", new Integer (TT_CONTROL));
         javaKeywords.put ("catch", new Integer (TT_CONTROL));
         javaKeywords.put ("char", new Integer (TT_TYPE));
         javaKeywords.put ("class", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("const", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("continue", new Integer (TT_CONTROL));
         javaKeywords.put ("default", new Integer (TT_CONTROL));
         javaKeywords.put ("do", new Integer (TT_CONTROL));
         javaKeywords.put ("double", new Integer (TT_TYPE));
         javaKeywords.put ("else", new Integer (TT_CONTROL));
         javaKeywords.put ("extends", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("final", new Integer (TT_MODIFIER));
         javaKeywords.put ("finally", new Integer (TT_CONTROL));
         javaKeywords.put ("float", new Integer (TT_TYPE));
         javaKeywords.put ("for", new Integer (TT_CONTROL));
         javaKeywords.put ("goto", new Integer (TT_CONTROL));
         javaKeywords.put ("if", new Integer (TT_CONTROL));
         javaKeywords.put ("implements", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("import", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("instanceof", new Integer (TT_RESERVED));
         javaKeywords.put ("int", new Integer (TT_TYPE));
         javaKeywords.put ("interface", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("long", new Integer (TT_TYPE));
         javaKeywords.put ("native", new Integer (TT_MODIFIER));
         javaKeywords.put ("new", new Integer (TT_RESERVED));
         javaKeywords.put ("package", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("private", new Integer (TT_MODIFIER));
         javaKeywords.put ("protected", new Integer (TT_MODIFIER));
         javaKeywords.put ("public", new Integer (TT_MODIFIER));
         javaKeywords.put ("return", new Integer (TT_CONTROL));
         javaKeywords.put ("short", new Integer (TT_TYPE));
         javaKeywords.put ("static", new Integer (TT_MODIFIER));
         javaKeywords.put ("super", new Integer (TT_RESERVED));
         javaKeywords.put ("switch", new Integer (TT_CONTROL));
         javaKeywords.put ("synchronized", new Integer (TT_MODIFIER));
         javaKeywords.put ("this", new Integer (TT_RESERVED));
         javaKeywords.put ("throws", new Integer (TT_PROCLAMATION));
         javaKeywords.put ("transient", new Integer (TT_MODIFIER));
         javaKeywords.put ("try", new Integer (TT_CONTROL));
         javaKeywords.put ("void", new Integer (TT_TYPE));
         javaKeywords.put ("volatile", new Integer (TT_MODIFIER));
         javaKeywords.put ("while", new Integer (TT_CONTROL));
         javaKeywords.put ("null", new Integer (TT_LITERAL));
         javaKeywords.put ("true", new Integer (TT_LITERAL));
         javaKeywords.put ("false", new Integer (TT_LITERAL));

         javaKeywords.put ("proceed", new Integer (TT_CONTROL));
      }
   }
        
   /**
      Push back stream based on the input stream specified by the
      constructor.  */
   private PushbackInputStream strmIn;

   /**
      Java keyword lookup table.
   */
   private static Hashtable javaKeywords;

   /**
      Prevent this constructor from being used.
   */
   private JavaTokenizer ()
   {
   }
}


/**
Java source code highlighter. Generates an HTML file representing the
highlighted souce code of a Java file.
@see JavaTokenizer */

class JavaHighlight
{
   /**
      This function does the job of highlighting of java source code.
      @param strJavaFile The name of the Java file to be highlighted.
      @param strHtmlFile The name of the output file.
      @param strPropFile The name of the property file.
   */

   public static ByteArrayOutputStream getHighlightedText (String strJavaFile ) throws IOException
   { 
      //Properties propHtml = new Properties ();
      //FileInputStream strmCfg =  new FileInputStream (strPropFile);
      FileInputStream strmJava;
      ByteArrayOutputStream strmHtml = new ByteArrayOutputStream();
      
      //   propHtml.load (strmCfg);
      
      strmJava = new FileInputStream (strJavaFile);
      //strmHtml = new FileOutputStream (strHtmlFile);
      
      PrintStream outHtml = new PrintStream (strmHtml);
      JavaTokenizer tok = new JavaTokenizer (strmJava);
      
      String strDocTxColor = "#000000";
      String strDocBgColor = "#FFFFFF";
                
      outHtml.println ("<HTML><HEAD><TITLE>"+strJavaFile+"</TITLE></HEAD>");
      outHtml.println ("<BODY TEXT=\""+strDocTxColor+"\" BGCOLOR=\""+strDocBgColor+"\" LINK=\""+strDocTxColor+"\" VLINK=\""+strDocTxColor+"\" ALINK=\""+strDocTxColor+"\">");
      outHtml.println ("<P ALIGN=\"LEFT\"><FONT SIZE=-1>"+strJavaFile+"</FONT></P><HR><PRE><FONT SIZE=-1>");
      
      
   loop:   while (true)
   {
      String strColor = "";
      String strStyle = "";
      int type = tok.nextToken ();
      
      switch (type)
      {
         case JavaTokenizer.TT_EOF:
            break loop;
         case JavaTokenizer.TT_OTHER:
            break;
         case JavaTokenizer.TT_COMMENT:
            strColor = "#AA0000";
            strStyle = "";
            break;
         case JavaTokenizer.TT_BRACKET:
            strColor = "BLACK";
            strStyle = "";
            break;
         case JavaTokenizer.TT_CLASS:
            strColor = "#009000";
            strStyle = "";
            break;
         case JavaTokenizer.TT_CONSTANT:
            strColor = "#A0A0A0";
            strStyle = "";
            break;
         case JavaTokenizer.TT_MODIFIER:
            strColor = "#A000A0";
            strStyle = "";
            break;
         case JavaTokenizer.TT_TYPE:
            strColor = "#009000";
            strStyle = "";
            break;
         case JavaTokenizer.TT_CONTROL:
            strColor = "#0000A0";
            strStyle = "";
            break;
         case JavaTokenizer.TT_PROCLAMATION:
            strColor = "#A000A0";
            strStyle = "";
            break;
         case JavaTokenizer.TT_LITERAL:
            strColor = "GRAY";
            strStyle = "";
            break;
         case JavaTokenizer.TT_RESERVED:
            strColor = "#A000A0";
            strStyle = "";
            break;
         case JavaTokenizer.TT_NUMBER:
            strColor = "#AA00AA";
            strStyle = "";
            break;
         case JavaTokenizer.TT_STRING:
            strColor = "#902000";
            strStyle = "";
            break;
         default:
            Log.error ("Unknown token!");
            Log.error ("code:  " + type);
            Log.error ("value: " + tok.value);
            break loop;
      }
                        
      if (strColor.length () > 0)
         outHtml.print ("<FONT COLOR=\""+strColor+"\">");
      if (strStyle.length () > 0)
         outHtml.print ("<"+strStyle+">");
      outHtml.print (HtmlEscape (tok.value));
      if (strStyle.length () > 0)
         outHtml.print ("</"+strStyle+">");
      if (strColor.length () > 0)
         outHtml.print ("</FONT>");
   }
                
      outHtml.println ("</FONT></PRE><HR>");
      outHtml.println ("<FONT SIZE=-2>Generated for "+ System.getProperty ("user.name",""));
      outHtml.println (" on " + (new Date ()).toLocaleString ());
      outHtml.print ("<BR>Java Highlighter, (c) 1997 - ");
      outHtml.println ("<A HREF=\"mailto:mas@cs.sfu.ca\">Marius Scurtescu</A>");
      outHtml.print ("<BR>Modified by Renaud Pawlak, (c) 2001 - ");
      outHtml.println ("<A HREF=\"mailto:pawlak@cnam.fr\">Renaud Pawlak</A><BR></FONT>");
      outHtml.println ("</BODY></HTML>");
      outHtml.flush ();
      outHtml.close ();
                
      try
      {
         strmHtml.close ();
         strmJava.close ();
      }
      catch (IOException e)
      {
         Log.error ("Cannot close file!");
         Log.error (e.getMessage ());
         Log.error (e.toString ());
      }

      return strmHtml;
   } 

   
   /**
      Replaces all occurences of &, < and > with the HTML escape sequence in a string.
      @param text String to be escaped
      @return HTML escaped string.
   */
   protected static String HtmlEscape (String text)
   {
      StringBuffer strBuf = new StringBuffer ();
      for (int i = 0; i < text.length (); i++)
      {
         char c = text.charAt (i);
         switch (c)
         {
            case '<':
               strBuf.append ("&lt;");
               break;
            case '>':
               strBuf.append ("&gt;");
               break;
            case '&':
               strBuf.append ("&amp;");
               break;
            default:
               strBuf.append (c);
               break;
         }
      }
                
      return strBuf.toString ();
   }

}

