/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import jac.core.*;
import jac.core.rtti.*;
import jac.util.*;

/**
 * This class defines a Swing component view for references in
 * objects.
 *
 * <p>By default this view constructs an embedded <code>JLabel</code>
 * containing the string representation of the referenced object. However,
 * the field can be attributed to be displayed with a customized
 * rendering by the GUI aspect component.
 *
 * @see GuiAC
 * @see CollectionView
 * @see FieldView */

public class ReferenceView extends JPanel implements View, ActionListener {

   Window parent;
   Object substance;
   FieldItem reference;
   Object object = null;
   MethodItem setter = null;
   JButton viewButton;
   JLabel label;

   /**
    * Constructs a new reference view.
    *
    * @param parent the Swing dialog parent
    * @param substance the object the viewed field belongs to
    * @param reference the reference that is showed by this view */   

   public ReferenceView(Window parent, 
                        Object substance, FieldItem reference) {
      this.parent = parent;
      this.substance = substance;
      this.reference = reference;

      label = new JLabel();
      add(label);

      add( Box.createHorizontalStrut( 20 ) );

      JButton b;
      viewButton = new JButton ( ResourceManager.getViewIcon() );
      viewButton.setToolTipText("View");
      viewButton.setActionCommand("open");
      viewButton.addActionListener(this);
      viewButton.setMargin(new Insets(1,1,1,1));
      add( viewButton );

      if ( reference.getWritingMethods() != null ) {
         // we should be more specific
         setter = reference.getWritingMethods()[0];
         if ( setter != null ) {
            b = new JButton ( ResourceManager.getEditIcon() );
            b.setToolTipText("Edit");
            b.setActionCommand("edit");
            b.addActionListener(this);
            b.setMargin(new Insets(1,1,1,1));
            add( b );
         }
      }
      setLayout(new FlowLayout(FlowLayout.CENTER,1,0));
      refreshView();
   }

   /**
    * Handles the actions on this view.
    *
    * <p>On a reference view, the two default possible actions are to
    * open a new view on the referenced object, or to edit the
    * reference value.
    *
    * @param evt the user event */

   public void actionPerformed(ActionEvent evt) {
      if (evt.getActionCommand().equals("open")) {
         if (object!=null) {
            ObjectView.addViewFor( object, parent, parent );
         }
      } else if (evt.getActionCommand().equals("edit")) {
         GuiAC.invoke((Display)parent,setter,substance,null);
      }
   }

   // View interface

   public Object getSubstance() {
      return substance;
   }

   public View getParentView() {
      return null;
   }

   public void refreshViewItem( FieldItem item, MethodItem method, Object[] args ) {
      Log.trace("gui.viewControl","refreshViewItem("+item+") on "+getClass().getName());
      refreshView();
   }

   public void refreshView( ) {
      Log.trace("gui.viewControl","refreshViewItem on "+getClass().getName());
      if (object!=null && object instanceof Wrappee) {
         ((Wrappee)object).forceTreatment();
         ViewControlWrapper vw = (ViewControlWrapper) ((Wrappee)object)
            . getWrappers( ViewControlWrapper.class )[0];
         vw.removeView(this);
      }
      object = reference.getThroughAccessor(substance);
      if (object!=null && object instanceof Wrappee) {
         ((Wrappee)object).forceTreatment();
         ViewControlWrapper vw = (ViewControlWrapper) ((Wrappee)object)
            . getWrappers( ViewControlWrapper.class )[0];
         vw.addView(this);
      }

      Log.trace("gui.viewControl","new value of reference "+reference+" is "+object);
      String name;
      if (object!=null) {
         viewButton.enable();
         name = object.toString();
      } else {
         viewButton.disable();
         name = "null";
      }
      label.setText(name);
   }

   public void close() {}

   public void setFocus(FieldItem field, Object extraOption) {
   }
}
