/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Collection;
import java.io.*;
import java.lang.reflect.*;

import java.awt.*;
import javax.swing.*;

import jac.core.*;
import jac.util.*;
import jac.core.rtti.*;

/**
 * A class with static methods to build components.<p>
 *
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a>
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 */

public class RenderFactory {

   /**
    * This method creates the parameters panel that is used when when
    * a method or a constructor is invoked within the GUI.<p>
    *
    * @param method the method item that is invoked
    * @param substance the object the method is invoked on. Use null for constructors
    * @return a <code>Component</code> that displays the parameter
    * panel */

   static public JPanel createParameters(Window parent, 
                                         final AbstractMethodItem method,
                                         Object substance) {

      Log.trace("gui","createParameters("+method+")");
      
      JPanel panel = new JPanel();

      panel.setBorder( 
         BorderFactory.createTitledBorder( 
            BorderFactory.createEtchedBorder(),
            " Parameters " ) );
      panel.setLayout( new BoxLayout( panel, BoxLayout.Y_AXIS ) );
      JComponent extra_component = null;
      Class[] paramTypes = method.getParameterTypes();
      String[] paramNames = (String[]) method.getAttribute( "Gui.parameterNames" );

      for ( int i = 0; i < paramTypes.length; i++ ) {

         ValueEditor ve = getEditorComponent(parent, substance, method, i);

         String paramName = null;
         if( paramNames != null ) {
            paramName = paramNames[i];
         } else {
            paramName = "arg" + i + " (" + paramTypes[i] + ")";
         }

         JPanel container = new JPanel();
         container.add(new JLabel(paramName+" : "));
         ((Component)ve).setName( "arg" + i );
         container.add( (Component)ve );
         panel.add( container );
      }
      return panel;
   }

   /**
    * Returns a ValueEditor suitable for editing the parameter i of a
    * method.
    *
    * @param parent the display
    * @param subtance the substance object
    * @param method the method that contains the parameter
    * @param i the parameter index
    */

   public static ValueEditor getEditorComponent(Window parent, 
                                                Object substance,
                                                AbstractMethodItem method,
                                                int i)
      //                                                FieldEditor fieldEditor)
   {
      Class[] paramTypes = method.getParameterTypes();
      String[] paramNames = (String[]) method.getAttribute( "Gui.parameterNames" );
      String[] passwordParams = (String[]) method.getAttribute( "Gui.passwordParameters" );
      Object[] defaultValues = (Object[]) method.getAttribute( "Gui.defaultValues" );
      String[] parameterFields = (String[]) method.getAttribute ( "Gui.parameterFields" );
      Object[] parameterChoices = (Object[]) method.getAttribute ( "Gui.parameterChoices" );
      boolean[] editableChoices = (boolean[]) method.getAttribute ( "Gui.editableChoices" );
      String[] parametersType = (String[]) method.getAttribute ( "rtti.parametersType" );
      FieldItem editedField = null;
         
      Object defaultValue = null;
      if( defaultValues != null ) {
         defaultValue = defaultValues[i];
      };
      Object choiceObject = null;
      Object[] choice = null;
      boolean editableChoice = false;
      MetaItem type = null;
      if (parametersType!=null)
         type = ClassRepository.get().getClass(parametersType[i]);
      // if it's a set-method, get the default value from the field
      // as well as the editableChoice attribute
      if (substance != null) {
         FieldItem[] fs = ((MethodItem)method).getWrittenFields();
         if ( fs!=null && fs.length==1) {
            if ( method.getParameterTypes()[0].isAssignableFrom( fs[0].getType() ) ) {
               editedField = fs[0];
               defaultValue = fs[0].get( substance );
               choiceObject = fs[0].getAttribute( "Gui.fieldChoice" );
               editableChoice = GuiAC.isEditableChoice(fs[0]);
               Log.trace("gui.choice","choice = "+choice);
               type = RttiAC.getFieldType(fs[0]);
            }
         }
      }

      if ( parameterFields != null ) {
         FieldItem cfi = ((ClassItem)method.getParent()).getField( parameterFields[i] );
         if ( cfi != null ) {
            choiceObject = cfi.getAttribute( "Gui.fieldChoice" );
         }
      }

      if (parameterChoices != null) {
         choiceObject = parameterChoices[i];
      }

      if ( choiceObject != null ) {
         if ( choiceObject instanceof MethodItem ) {
            MethodItem choiceMethod = (MethodItem)choiceObject;
            try {
               if (Modifier.isStatic(choiceMethod.getModifiers())) {
                  choiceObject = choiceMethod.invoke(null,new Object[] {substance});
               } else {
                  choiceObject = choiceMethod.invoke(substance,new Object[0]);
               }
            } catch (Exception e) {
               Log.error("Invocation of choice "+choiceMethod+" method failed : "+e);
               e.printStackTrace();
            }
         }
         if ( choiceObject instanceof Object[] ) {
            choice = (Object[])choiceObject;
         } else if( choiceObject instanceof Collection ) {
            choice = ((Collection)choiceObject).toArray();
         }
      }

      ValueEditor ve = null;
      if (choice!=null) {
         if (editableChoices != null)
            editableChoice = editableChoices[i];
         ve = new ChoiceFieldEditor(choice, editableChoice);
      } else if ( Wrappee.class.isAssignableFrom( paramTypes[i] ) ||
                  paramTypes[i] == Object.class) {
         // jac objects
         CollectionItem collection = (CollectionItem)JacObject.
            getCollaboration().getAttribute("Gui.collectionItem");
         ClassItem collectionType = null;
         if (collection!=null) {
             collectionType = (ClassItem)collection.getAttribute("Gui.collectionType");
         }
         if( collectionType != null ) {
            ve = new ObjectChooser(parent, substance, 
                                   ClassRepository.get().getClass(collectionType),
                                   editedField,null);
         } else {
            Predicate predicate = null;
            if (method instanceof MethodItem) {
               CollectionItem[] collections = ((MethodItem)method).getAddedCollections();
               if (collections!=null && collections.length>0) {
                  if (Set.class.isAssignableFrom(collections[0].getType())) {
                     predicate = new NotInCollectionPredicate((Collection)collections[0].get(substance));
                  }
               }
            }
            ve = new ObjectChooser(parent, substance,
                                   ClassRepository.get().getClass(paramTypes[i]),
                                   editedField,predicate);
         }
      } else if ( paramTypes[i].isArray() || Collection.class.isAssignableFrom ( paramTypes[i] ) ) {
         // array and collections
         ve = new ArrayEditor(paramTypes[i]);
      } else {
         ClassItem editorClass = null;
         if (type != null) {
            editorClass = GuiAC.getTypeEditor(type);
         }
         if (editorClass==null) {
            editorClass = GuiAC.getTypeEditor(
               ClassRepository.get().getClass(paramTypes[i]));
         }
         if (editorClass!=null && 
             !(passwordParams != null && passwordParams[i].equals("true"))) {
            try {
               ConstructorItem constructor = editorClass.getConstructor(new Class[] {Class.class});
               if (constructor == null) {
                  constructor = editorClass.getConstructor(new Class[0]);
                  if (constructor != null) {
                     ve = (ValueEditor)constructor.newInstance();
                  } else {
                     Log.error("Could not find a suitable constructor for editor component "+
                               editorClass.getName());
                  }
               } else {
                  ve = (ValueEditor)constructor.newInstance(new Object[] { paramTypes[i] });
               }
            } catch (Exception e) {
               Log.error("Instanciation of editor component "+editorClass.getName()+" failed");
               e.printStackTrace();
            }
         } else if ( ( paramTypes[i].isPrimitive() || 
                       paramTypes[i].getName().startsWith( "java.lang" ) ) &&
                     ( ! paramTypes[i].isArray() ) ) {
            // primitive types (int, float, ...
            ve = new PrimitiveFieldEditor(
               paramTypes[i],
               passwordParams != null && passwordParams[i].equals("true"));
            if( defaultValues != null ) {
               ve.setValue(defaultValues[i]);
            }
         } else {
            Log.error("Could not find an editor component for type "+paramTypes[i]);
         }  // we should have a fallback editor
      }

      /*if( fieldEditor != null && ve!=null ) {
         System.out.println("added focus listener for "+ve);
         System.out.println("--- Listener is: "+fieldEditor);
         ve.getValueComponent().addFocusListener(fieldEditor);
         //((Component)ve).addFocusListener(fieldEditor);
         }*/

      if (ve!=null && defaultValue!=null) {
         ve.setValue(defaultValue);
      }
      return ve;
   }

   /**
    * Builds the parameters values of a method invocation from a
    * container (an input box for instance).
    *
    * @param method the invoked method
    * @param container the swing container of the user inputted values
    * @return the parameters values */

   public static Object[] buildParameterValues(AbstractMethodItem method, Container container) 
   {
      try {      
         JPanel panel = (JPanel)(container.getComponents()[0]);
         Class[] pts = method.getParameterTypes();
         Object[] v = new Object[pts.length];
         for ( int i = 0; i < panel.getComponentCount(); i++ ) {
            Component component = ((JPanel)panel.getComponent(i)).getComponent(1);
            if ( component instanceof ValueEditor) {
               ValueEditor ve = (ValueEditor)component;
               v[i] = ve.getValue();
            } else {
               JTextField tf = (JTextField) component;
               try {
                  v[i] = pts[i].getConstructor(
                     new Class [] { String.class }).newInstance(
                        new Object[] {tf.getText()} );
               } catch (Exception e) {
                  e.printStackTrace();
                  v[i] = pts[i].newInstance();
               }
            }
         }
         return v;
      } catch ( Exception e ) {
         e.printStackTrace ();
         return null;
      }
   }
}



/**
 * This private class has been taken from the SUN jdk
 * (demo/jfc/FileChooserExample). Thanks. */

class CustomFileFilter extends javax.swing.filechooser.FileFilter {

   private static String TYPE_UNKNOWN = "Type Unknown";
   private static String HIDDEN_FILE = "Hidden File";
   
   private Hashtable filters = null;
   private String description = null;
   private String fullDescription = null;
   private boolean useExtensionsInDescription = true;
   
   /**
    * Creates a file filter. If no filters are added, then all
    * files are accepted.
    *
    * @see #addExtension */

   public CustomFileFilter() {
      this.filters = new Hashtable();
   }
   
   /**
    * Creates a file filter that accepts files with the given extension.
    * Example: new ExampleFileFilter("jpg");
    *
    * @see #addExtension
    */

   public CustomFileFilter(String extension) {
      this(extension,null);
   }
   
    /**
     * Creates a file filter that accepts the given file type.
     * Example: new ExampleFileFilter("jpg", "JPEG Image Images");
     *
     * Note that the "." before the extension is not needed. If
     * provided, it will be ignored.
     *
     * @see #addExtension
     */

   public CustomFileFilter(String extension, String description) {
      this();
      if(extension!=null) addExtension(extension);
      if(description!=null) setDescription(description);
   }

   /**
    * Creates a file filter from the given string array.
    * Example: new ExampleFileFilter(String {"gif", "jpg"});
    *
    * Note that the "." before the extension is not needed adn
    * will be ignored.
    *
    * @see #addExtension
    */
   public CustomFileFilter(String[] filters) {
      this(filters, null);
   }
   
   /**
    * Creates a file filter from the given string array and description.
    * Example: new ExampleFileFilter(String {"gif", "jpg"}, "Gif and JPG Images");
    *
    * Note that the "." before the extension is not needed and will be ignored.
    *
    * @see #addExtension
    */

   public CustomFileFilter(String[] filters, String description) {
      this();
      for (int i = 0; i < filters.length; i++) {
         // add filters one by one
         addExtension(filters[i]);
      }
      if(description!=null) setDescription(description);
   }
   
   /**
    * Return true if this file should be shown in the directory pane,
    * false if it shouldn't.
    *
    * Files that begin with "." are ignored.
    *
    * @see #getExtension
    * @see FileFilter#accepts
    */

   public boolean accept(File f) {
      if(f != null) {
         if(f.isDirectory()) {
            return true;
         }
         String extension = getExtension(f);
         if(extension != null && filters.get(getExtension(f)) != null) {
            return true;
         };
      }
      return false;
   }
   
    /**
     * Return the extension portion of the file's name .
     *
     * @see #getExtension
     * @see FileFilter#accept
     */

   public String getExtension(File f) {
      if(f != null) {
         String filename = f.getName();
         int i = filename.lastIndexOf('.');
         if(i>0 && i<filename.length()-1) {
            return filename.substring(i+1).toLowerCase();
         };
      }
      return null;
   }
   
   /**
    * Adds a filetype "dot" extension to filter against.
    *
    * For example: the following code will create a filter that filters
    * out all files except those that end in ".jpg" and ".tif":
    *
    *   ExampleFileFilter filter = new ExampleFileFilter();
    *   filter.addExtension("jpg");
    *   filter.addExtension("tif");
    *
    * Note that the "." before the extension is not needed and will be ignored.
    */

   public void addExtension(String extension) {
      if(filters == null) {
         filters = new Hashtable(5);
      }
      filters.put(extension.toLowerCase(), this);
      fullDescription = null;
   }
   
   
   /**
    * Returns the human readable description of this filter. For
    * example: "JPEG and GIF Image Files (*.jpg, *.gif)"
    *
    * @see setDescription
    * @see setExtensionListInDescription
    * @see isExtensionListInDescription
    * @see FileFilter#getDescription
    */

   public String getDescription() {
      if(fullDescription == null) {
         if(description == null || isExtensionListInDescription()) {
            fullDescription = description==null ? "(" : description + " (";
            // build the description from the extension list
            Enumeration extensions = filters.keys();
            if(extensions != null) {
               fullDescription += "." + (String) extensions.nextElement();
               while (extensions.hasMoreElements()) {
                  fullDescription += ", " + (String) extensions.nextElement();
               }
            }
            fullDescription += ")";
         } else {
            fullDescription = description;
         }
      }
      return fullDescription;
   }

   /**
    * Sets the human readable description of this filter. For
    * example: filter.setDescription("Gif and JPG Images");
    *
    * @see setDescription
    * @see setExtensionListInDescription
    * @see isExtensionListInDescription
    */

   public void setDescription(String description) {
      this.description = description;
      fullDescription = null;
   }
   
   /**
    * Determines whether the extension list (.jpg, .gif, etc) should
    * show up in the human readable description.
    *
    * Only relevent if a description was provided in the constructor
    * or using setDescription();
    *
    * @see getDescription
    * @see setDescription
    * @see isExtensionListInDescription
    */

   public void setExtensionListInDescription(boolean b) {
      useExtensionsInDescription = b;
      fullDescription = null;
   }
   
   /**
    * Returns whether the extension list (.jpg, .gif, etc) should
    * show up in the human readable description.
    *
    * Only relevent if a description was provided in the constructor
    * or using setDescription();
    *
    * @see getDescription
    * @see setDescription
    * @see setExtensionListInDescription
    */

   public boolean isExtensionListInDescription() {
      return useExtensionsInDescription;
   }

}
