/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.gui;

import jac.core.*;
import jac.util.Log;
import jac.core.rtti.*;
import java.lang.reflect.*;
import java.util.*;
import java.awt.*;
import javax.swing.JTextArea;

/**
 * This wrapper updates the views of a given object when its state
 * changes, that is to say when a write method is called on the
 * wrappee.<p>
 *
 * A view controller can control several views of the same wrappee at
 * the same time.<p>
 *
 * This mecanism is similar to the famous MVC design pattern. The
 * model is the wrappee and the controller is the wrapper.<p>
 *
 * @see View
 * @see #addView(View)
 * @see #removeView(View) */

public class ViewControlWrapper extends Wrapper {
  
   /** Stores the wrappee's views that are controlled by this
       controller. */
   protected Vector controlledViews = new Vector();

   /**
    * Creates a wrapper that links the given view to the wrappee.<p>
    *
    * @param view the GUI component that will be used to display the
    * wrappee */

   public ViewControlWrapper ( View view ) {
      if ( view != null ) {
         controlledViews.add( view );
      }
   }

   /**
    * Creates an empty  view control wrapper.<p>
    */

   public ViewControlWrapper () {}

   /**
    * Adds a view to the wrappee.<p>
    *
    * @param view the GUI component that will be used to display the
    * wrappee
    * @see #removeView(View) */

   public void addView( View view ) {
      if ( view != null && (!controlledViews.contains(view))) {
         Log.trace("gui.viewControl","addView("+view.getClass()+") on "+wrappee());
         controlledViews.add( view );
      }
   }      

   /**
    * Removes a view that has been previously added with
    * <code>addView</code>.
    *
    * @param view the view GUI component to remove
    * @see #addView(View) */

   public void removeView( View view ) {
      if ( view != null ) {
         Log.trace("gui.viewControl","removeView("+view.getClass()+") on "+wrappee());
         controlledViews.remove( view );
      }
   }

   /**
    * Get the views controlled by this wrapper.<p>
    *
    * @return the set of controlled views
    */

   public Vector getViews() {
      return controlledViews;
   }

   /**
    * A wrapping method that updates the views of the objects.<p>
    *
    * It uses the RTTI aspect to know the fields and the collections
    * that are written, added, or removed by the currently wrapped
    * method. Then it upcall the <code>refreshStateComponent</code> of
    * all the controlled views to refresh the implied GUI
    * components.<p>
    *
    * @see jac.core.rtti
    * @see View#refreshViewItem(FieldItem,MethodItem,Object[])
    */

   public Object updateView() {

      Object ret = proceed();

      Log.trace("gui","checking view updating for method " + wrappee().getClass()+"."+method());

      ClassItem cli = ClassRepository.get().getClass( wrappee().getClass() );
      MethodItem method = null;// = cli.getMethod( method() );
      // DIRTY HACK
      MethodItem[] methods = cli.getMethods( method() );
      for (int i=0; i<methods.length; i++) {
         if (methods[i].getParameterTypes().length == args().length) {
            method = methods[i];
            break;
         }
      }

      if( method == null ) {
         Log.warning("did not found method in update view");
         return ret;
      }
      JTextArea console = (JTextArea) method.getAttribute("Gui.loggingMethod"); 
      if( console != null ) {
         console.append( (String)arg(0) );
         console.setCaretPosition( console.getText().length() );
      }
      
      FieldItem[] writtenFields = method.getWrittenFields();
      CollectionItem[] addedCollections = method.getAddedCollections();
      CollectionItem[] removedCollections = method.getRemovedCollections();

      boolean stateRefreshed = false;
      
      Vector toRefresh = new Vector();
      if ( writtenFields != null ) 
         toRefresh.addAll( Arrays.asList( writtenFields ) );
      if ( addedCollections != null ) 
         toRefresh.addAll( Arrays.asList( addedCollections ) );
      if ( removedCollections != null ) 
         toRefresh.addAll( Arrays.asList( removedCollections ) );

      Iterator itFields = toRefresh.iterator();      
      Iterator itViews = null;

      while ( itFields.hasNext() ) {
         FieldItem field = (FieldItem) itFields.next();
         for( int i=0; i<controlledViews.size(); i++) {
            View view = (View) controlledViews.get(i);
            Log.trace("gui.viewControl","refreshing " + view.getClass().getName());
            stateRefreshed = true;
            view.refreshViewItem( field, method, args() );
         }

      }

      if ( (!stateRefreshed) && method.isStateModifier() ) {
         for( int i=0; i<controlledViews.size(); i++) {
            View view = (View) controlledViews.get(i);
            Log.trace("gui.viewControl","refreshing " + view);
            stateRefreshed = true;
            view.refreshView();
         }
      }

      if (!stateRefreshed) Log.trace("gui","no view refreshed");

      return ret;
   }

}




