/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.aspects.gui;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import jac.core.*;
import jac.aspects.naming.*;
import jac.core.rtti.*;
import jac.aspects.distribution.thinclient.*;
import jac.util.*;
import jac.wrappers.*;

/**
 * This class provides a server for web clients using a thin client
 * protocol and implements the <code>Display</code> interface to
 * assume data inputs and outputs between JAC objects and web clients.
 *
 * <p>This inferface is typically used by Java web clients and more
 * specifically by servlets. The implementation provided by JAC is
 * <code>JacServlet</code>.
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> 
 */

public class WebGui extends AbstractServer {

   /** Stores the web GUI instance. */
   protected static WebGui webGui;

   /** Stores the style sheet url. */
   protected String styleSheetURL = "http://jac.aopsys.com/style.css";

   /** Stores the charset */
   protected String charset = "iso-8859-1";

   /**
    * A getter of the the WebGui instance (should be only one per JAC
    * container).
    *
    * @return the web GUI instance */

   public static WebGui get() {
      if ( webGui == null ) {
         webGui = new WebGui();
      }
      return webGui;
   }
   
   /**
    * Creates a new web gui display and automatically registers it in
    * the name repository under "JAC.guiWebServer".
    *
    * <p>As it is a full JAC object and as it is registered, the web
    * GUI can be remotely accessed and can serve web clients that
    * implement the JLP. */

   public WebGui() {
      Log.trace("webgui",">>> New WebGui");
      webGui = this;
      NameRepository.get().unregisterObject( this );
      NameRepository.get().register( "JAC.guiWebServer", this );
   }

   /**
    * Sets the style sheet URL of the generated html pages.
    *
    * @param url the URL string */

   public void setStyleSheetURL( String url ) {
      styleSheetURL = url;
   }

   /**
    * Sets the charset of the generated html pages
    *
    * @param charset the charset */

   public void setCharset( String charset ) {
      this.charset = charset;
   }

   /*************************************************************
    * The following methods set the response for the request 
    * that is currently waiting
    *************************************************************/

   /**
    * Creates an HTML page embedding the given object view as a
    * response for the currently waiting request thread of the current
    * session.
    *
    * @param substance the object to build the view of
    * @see #viewRequest(String,String)
    * @see #refreshRequest(String) */

   public void viewResponse( Object substance ) {
      Log.trace("webgui","viewResponse("+substance+") "+getSession());
      Log.trace("webgui","Session.context = "+getSession().getCurrentRequest().getContext());
      String name = null;
      try {
         name = NameRepository.get().getName(substance);
      } catch( Exception e ) { 
         name = substance.toString();
      }

      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      PrintStream s = new PrintStream( dataStream );

      String title = getSession().getCurrentRequest().title;
      addHtmlHeader(s,title==null?"JAC component view: "+name:title,
                    getSession().getCurrentRequest().header);

      Class substance_type = substance.getClass();
      ClassItem cli = ClassRepository.get().getClass(substance_type);
      FieldItem[] fields = cli.getFields();
      CollectionItem[] collections = cli.getCollections();

      //      s.println("<H1>"+cli.getName()+"</H1>");

      Collection methods = new Vector();
      if( JacObject.getCollaboration().getAttribute("Gui.autoCreation") == null ) {
         MethodItem[] ms = (MethodItem[])cli.getAttribute("Gui.methodsOrder");
         if( ms != null ) {
            methods = Arrays.asList(ms);
         }
      }

      s.println("<FORM class=update ACTION=\"JacServlet\">");

      Method tmpMeth = null;

      Iterator it = methods.iterator();
      int i = 0;
      int nbMethods = 0;

      boolean hadEmbeddedEditor=false;

      // we draw methods only if we are not in a modal view so 
      // that the user cannot make any manipulation mistakes
      if( ! getSession().getCurrentRequest().isModalView() ) {

         // Methods
         s.println( "<div class=methods>" );
         while( it.hasNext() ) {
            
            MethodItem curmeth = (MethodItem) it.next();
            
            if ( curmeth == null ) continue;
            
            if ( ClassRepository.isJacMethod( curmeth.getName() ) || 
                 Modifier.isStatic( curmeth.getModifiers() ) ) {
               continue;
            }
            
            if ( ! GuiAC.isVisible(curmeth) || 
                 curmeth.getName().equals("toString") ) continue;            
            
            if ( ( NamingConventions.isSetter( curmeth.getName() ) && 
                   curmeth.getWrittenFields() != null ) ||
                 ( NamingConventions.isGetter( curmeth.getName() ) && 
                   curmeth.getAccessedFields() != null ) ||
                 ( NamingConventions.isAdder( curmeth.getName() ) && 
                   curmeth.getAddedCollections() != null ) ||
                 ( NamingConventions.isRemover( curmeth.getName() ) && 
                   curmeth.getRemovedCollections() != null ) ) 
               continue;

            /*
              if ( (substance instanceof Wrappee) && 
              ( ! ( ((Wrappee)substance).isExtendedBy( 
              jac.aspects.distribution.consistency.ClientServerConsistencyWrapper.class ) && 
              ! ((jac.core.dist.RemoteRef[])((Wrappee)substance)
              . invokeRoleMethod( "getKnownReplicas", new Object[0] ))[0]
              . getRemCont().isLocal() ) ) ) {
              if ( NamingConventions.fieldForMethod( 
              substance_type, curmeth.getName() ) != null ) {
              continue;
              }
              }*/

            addMethodCode(s,curmeth,name);
            i++;
            nbMethods++;
         }
         s.println( "</div>" );
      }

      //      Log.trace("webgui","fields...");
      // Fields
      if( (substance instanceof Wrappee) && ((Wrappee)substance).isExtendedBy( 
         jac.aspects.distribution.consistency.ClientServerConsistencyWrapper.class ) && 
          ! ((jac.core.dist.RemoteRef[])((Wrappee)substance).invokeRoleMethod( 
             "getKnownReplicas", new Object[0] ))[0]
          . getRemCont().isLocal() ) {
         
         s.println( "<B>NO STATE (pure client)</B>" );

      } else {

         boolean headDisplayed = false;

         // primitive fields
         // use specified order if av available
         fields = (FieldItem[])cli.getAttribute("Gui.attributesOrder");
         if (fields==null) {
            fields = cli.getFields();
         }

         Session session = getSession();
         
         for (i = 0; i < fields.length; i++) {
            if ( GuiAC.isVisible(fields[i]) ) {
               if ( fields[i].getClass() != CollectionItem.class ) {

                  if ( ! headDisplayed ) {
                     s.println( "<TABLE class=attributes>" );
                     //  s.println( "<THEAD><TR><TH>Name</TH><TH>Value</TH></TR></THEAD>" );
                     s.println( "<TBODY>" );
                     headDisplayed = true;
                  }

                  String attributeName = NamingConventions.textForName(fields[i].getName());
                  s.println( "<TR><TD class=attrName>"+attributeName+"</TD>" );
                  s.print( "<TD class=attrValue>" );

                  if ( fields[i].getWritingMethods() != null &&
                       fields[i].getWritingMethods()[0] != null ) {
                     MethodItem setter = fields[i].getWritingMethods()[0];
                     if ( GuiAC.isEmbeddedEditor(fields[i]) ) {
                        hadEmbeddedEditor = true;
                        addEditorCode( s, substance, fields[i] );
                        s.println( "</TD>" );
                     } else {
                        addFieldCode( s, substance, fields[i], false );
                        s.println( "</TD>" );
                        s.print( "<TD class=method>"+
                                 "<input type=submit class=method "+
                                 "name=\"_EDIT_"+setter.getName()+"\" "+
                                 "value=\"edit\"></TD>" );
                     }
                  } else {
                     addFieldCode( s, substance, fields[i], false );
                     s.println( "</TD>" );
                  }
               }
            }
         }

         if ( headDisplayed )
            s.println( "</TBODY></TABLE>" );


         // we draw collections only if we are not in a modal view so 
         // that the user cannot make any manipulation mistakes
         if( ! getSession().getCurrentRequest().isModalView() ) {
         
            // collections
            //         Log.trace("webgui","collections...");
            
            for (i = 0; i < fields.length; i++) {
               if ( GuiAC.isVisible(fields[i]) && 
                    fields[i].getClass() == CollectionItem.class ) {
                  addCollectionCode(s,substance,(CollectionItem)fields[i]);
               }
            }
            
         }
      }

      s.println("<INPUT TYPE=HIDDEN NAME=jid VALUE=\""+
                NameRepository.get().getName(substance)+"\">");
      if( hadEmbeddedEditor ) {
         s.println("<BR><DIV class=actions>");
         s.println("<INPUT TYPE=HIDDEN NAME=action VALUE=update>");
         s.println( "<INPUT class=button TYPE=submit NAME=choice VALUE=OK>" );
         s.println( "<INPUT class=button TYPE=submit NAME=choice VALUE=CANCEL>" );
         s.println("</DIV>");
      } else {
         // show the close button if several views on the stack
         if( getSession().getRequestCount() > 1 ) {
            s.println("<BR><DIV class=actions>");
            s.println("<INPUT TYPE=HIDDEN NAME=action VALUE=close>");
            s.println( "<INPUT class=button TYPE=submit NAME=choice VALUE=CLOSE>" );
            s.println("</DIV>");
         }
      }
      s.println("</FORM>");

      addHtmlFooter(s);
      s.flush();

      setResponse( dataStream.toByteArray() );

   }

   /*
   protected void frameModeResponse( String jid1, String jid2, String jid3 ) {
      
      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      PrintStream s = new PrintStream( dataStream );

      s.println("<html><head>");
      s.println("<frameset cols=\"22%,*\"><frameset rows=\"7%,*\">");
      s.println("<frame scrolling=yes name=frame1 src=\"JacServlet?action=view&amp;jid="+jid1+"\">");
      s.println("<frame scrolling=yes name=frame2 src=\"JacServlet?action=view&amp;jid="+jid2+"\">");
      s.println("</frameset>");
      s.println("<frameset cols=\"100%,*\">");
      s.println("<frame scrolling=yes name=frame3 src=\"JacServlet?action=view&amp;jid="+jid3+"\">");
      s.println("</frameset>");
      s.println("</frameset>");
      s.println("</head></html>");

      s.flush();

      setResponse( dataStream.toByteArray() );

      }*/

   /**
    * Response with a stream of a given mime type
    * @param reader the stream
    * @param mimeType the mime type of the stream */

   protected void fileResponse(Reader reader, String mimeType) {
      StringWriter writer = new StringWriter();
      try {
         int b = reader.read();
         while (b!=-1) {
            writer.write(b);
            b = reader.read();
         }
         reader.close();
      } catch(IOException e) {
         Log.error("Exception in fileResponse: "+e);
         e.printStackTrace();
      }
      setResponse( writer.toString().getBytes() );
   }

   /**
    * Creates an HTML page embedding a textual message as a response
    * for the currently waiting request thread of the current session.
    *
    * @param text the text of the response */
   
   protected void messageResponse( String text ) {
      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      PrintStream s = new PrintStream( dataStream );

      String title = getSession().getCurrentRequest().title;
      addHtmlHeader( s, title==null?"JAC":title, getSession().getCurrentRequest().header );
      s.println( text );
      s.println( "<P><A HREF=\"JacServlet?action=refresh\">Back to main menu</A>" );
      addHtmlFooter(s);

      s.flush();

      setResponse( dataStream.toByteArray() );
   }

   /**
    * Creates an HTML page embedding a string as an error response for
    * the currently waiting request thread of the current session.
    *
    * @param description the error description */
   
   protected void errorResponse( String description ) {
      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      PrintStream s = new PrintStream( dataStream );

      addHtmlHeader( s, "Error", getSession().getCurrentRequest().header );
      s.println( "An error was reported by the JAC thin-client server: " + description );
      s.println( "<P><A HREF=\"JacServlet?action=refresh\">Back to main menu</A>" );
      addHtmlFooter(s);

      s.flush();

      setResponse( dataStream.toByteArray() );
   }

   /**
    * Creates an HTML page embedding a table as a response for the
    * currently waiting request thread of the current session.
    *
    * <p>Each item within the collection corresponds to one row of the
    * table. Headers and collums are formed with the fields of the
    * objects within the collection. These objects must be of the same
    * type.
    *
    * @param c the collection to build the table from */

   protected void collectionResponse( Collection c ) {
      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      PrintStream s = new PrintStream( dataStream );

      addHtmlHeader( s, "JAC", getSession().getCurrentRequest().header );
      addTableCode( s, c, null );
      s.println( "<P><A HREF=\"JacServlet?action=refresh\">Back to main menu</A>" );
      addHtmlFooter(s);

      s.flush();

      setResponse( dataStream.toByteArray() );
   }

   /**
    * Creates an HTML page embedding an input form as a response for
    * the currently waiting "invoke" request thread of the current
    * session.
    *
    * @param object the object being invoked
    * @param method the method being invoked
    * @see #invokeRequest(String,String,String) */

   protected void inputResponse( Object object, AbstractMethodItem method ) {
      ByteArrayOutputStream dataStream = new ByteArrayOutputStream();
      try {
         String jid = NameRepository.get().getName(object);
         Log.trace("webgui","inputResponse("+object+"["+jid+"],"+method+") "+getSession());
         PrintStream s = new PrintStream( dataStream );
         addHtmlHeader( s, "JAC method call: " + method, getSession().getCurrentRequest().header );
         
         s.println( "<FORM ACTION=\"JacServlet\" METHOD=GET>\n"+
                    "<INPUT TYPE=hidden NAME=action VALUE=input>\n"+
                    "<INPUT TYPE=hidden NAME=jid VALUE="+jid+">\n"+
                    "<INPUT TYPE=hidden NAME=method VALUE="+method.getName()+">\n");
         
         addParametersCode(s,object,method);

         s.println( "<BR><INPUT class=button TYPE=submit NAME=choice VALUE=OK>" );
         s.println( "<INPUT class=button TYPE=submit NAME=choice VALUE=CANCEL>" );
         s.println( "</FORM>" );
         
         addHtmlFooter(s);
         s.flush();
         
      } catch ( Exception e ) {
         e.printStackTrace();
      }

      setResponse( dataStream.toByteArray() );
   }

   /***********************************************************
    * Some useful methods to add HTML code into a stream
    ***********************************************************/

   /**
    * Adds an HTML header into the given stream.
    *
    * @param s the stream to add into
    * @param title the title of the page that will use this header */

   protected void addHtmlHeader(PrintStream s, String title, String header) {
      s.println( "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">" );
      s.println( "<HTML>" );
      s.println( "<HEAD>" );
      s.println( "<META HTTP-EQUIV=\"Content-Type\" "+
                 "CONTENT=\"text/html; charset="+charset+"\">" );
      s.println( "<META NAME=\"Author\" CONTENT=\"JAC web-gui server\">" );
      s.println( "<META NAME=\"GENERATOR\" CONTENT=\"JAC web-gui server by Renaud Pawlak\">" );
      s.println( "<TITLE>"+title+"</TITLE>" );
      if( ! styleSheetURL.equals( "" ) ) {
         s.println( "<link rel=\"stylesheet\" type=\"text/css\" "+
                    "href=\""+styleSheetURL+"\" title=JAC>");
      }
      s.println( "</HEAD>" );
      s.println( "<BODY>" );
      s.println("<div style=\"text-align: center\">");
      if(header!=null) {
         s.println(header+"<BR>");
      }
   }

   /**
    * Adds an HTML footer into the given stream.
    *
    * @param s the stream to add into */

   protected void addHtmlFooter(PrintStream s) {
      s.println("</div>");
      s.println("<DIV class=footer>Generated by JAC server version 0.8.1 - "+new Date()+"</DIV>");
      s.println( "</BODY>" );
      s.println( "</HTML>" );
   }

   /**
    * Adds the HTML code that corresponds to the parameters of a given
    * method into the given stream.
    *
    * @param s the stream to add into */

   protected void addParametersCode( PrintStream s, Object object, 
                                     AbstractMethodItem method ) {

      addInputCode( s, object, method, true, false, null, false, null );
   }

   /*
   public void getViewPageByName( String name ) {
      Log.trace("webgui","getViewPageByName("+name+")");
      Object substance = NameRepository.get().getObject( name ) ;
      if ( substance == null ) {
         setResult( "JAC gui-web server error: no such object (" + name + ")" );
      } else {
         getViewPage( substance );
      }
      }*/

   /**
    * Adds the HTML code that corresponds to the printable
    * representation of a field into the given stream.
    *
    * @param s the stream to add into
    * @param substance the object that contains the field
    * @param field the field to generate the HTML code from
    * @param small tells if the field value can be expanded or not 
    */

   protected void addFieldCode( PrintStream s, Object substance, FieldItem field,
                                boolean small ) {
      Object value = field.getThroughAccessor( substance );

      MetaItem type = (MetaItem)field.getAttribute("rtti.fieldType");
      if (type==null) {
         if (value!=null)
            type = ClassRepository.get().getClass(value);
         else
            type = ClassRepository.get().getClass(field.getType());
      }
      ClassItem viewerClass = GuiAC.getHTMLTypeViewer(type);
      HTMLViewer viewer = null;
      if (viewerClass != null) {
         try {
            viewer = (HTMLViewer)viewerClass.newInstance();
         } catch (Exception e) {
            Log.error("Instanciation of HTML viewer component "+viewerClass.getName()+" failed");
            e.printStackTrace();
         }
      } else {
         viewer = new HTMLPrimitiveFieldEditor();
      }

      if (value==null) {
         s.print( "null" );
      } else if (field.isReference() || value instanceof Wrappee) {
         String name = NameRepository.get().getName(value);
         s.print( "<A HREF=\"JacServlet?action=view&amp;jid="+name+"\">"+value+"</A>" );
      } else if ( field.getType().isArray() ) {
         s.print( Arrays.asList((Object[])value).toString() );
      } else {
         s.print( viewer.genHTML(value,small) );
      }
   }

   private void addInputCode( PrintStream s, Object object,
                              AbstractMethodItem method, boolean showNames, 
                              boolean argsNaming, String[] names, 
                              boolean forceValue, Object forcedValue ) 
   {

      Class[] pts = method.getParameterTypes();
      String[] paramNames = (String[]) method.getAttribute( "Gui.parameterNames" );
      String[] passwordParams = (String[])method.getAttribute( "Gui.passwordParameters" ); 
      Object[] defaultValues = (Object[]) method.getAttribute( "Gui.defaultValues" );
      String[] parameterFields = (String[]) method.getAttribute ( "Gui.parameterFields" );
      Object[] parameterChoices = (Object[]) method.getAttribute ( "Gui.parameterChoices" );

      Object origValue = null;

      for ( int i = 0; i < pts.length; i++ ) {
         String paramName = null;
         if( paramNames != null ) {
            paramName = paramNames[i];
         } else {
            paramName = "arg" + i + " (" + pts[i] + ")";
         }

         String inputName = null;
         if (argsNaming) {
            if ( names != null ) {
               inputName = "_FIELD_"+names[i];
            } else if ( paramNames != null ) {
               inputName = "_FIELD_"+paramNames[i];
            } else {
               inputName = "_FIELD_arg"+i;
            }
         } else {
            inputName = "arg"+i;
         }

         Log.trace("webgui", "constructing parameter " + i + 
                   " for inputing " + method + 
                   " object = "+ ( object == null ? "null" : object.getClass().getName() ) +
                   "["+object+"]" ); 
         
         if ( showNames )
            s.println( "<span class=paramName>"+paramName+" : </span>" );

         Map context = getSession().getCurrentRequest().getContext();
         HTMLEditor editor = 
            (HTMLEditor)context.get(inputName);

         if (editor==null) {
            editor = getEditorComponent(object,method,i);
            context.put(inputName,editor);
            Log.trace("webgui","New editor in context : "+inputName+" -> "+editor);
            if (editor!=null && forceValue) {
               editor.setValue(forcedValue);
            }
         } else {
            Log.trace("webgui","Found editor in context : "+inputName+" -> "+editor);
         }

         if (editor!=null)
            s.println(editor.genHTML(inputName));
         else
            s.println("null");
         s.println("<br>");
      }

   }


   /**
    * Add some objects in the session's current request's context.
    * 
    */
   public void addContext( String sid, Map map ) {
      Log.trace("webgui","addContext("+sid+","+map+")");
      Map context = initSession( sid ).getCurrentRequest().getContext();
      Set entries = map.entrySet();
      Iterator i = entries.iterator();
      while (i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         HTMLEditor editor = (HTMLEditor)context.get(entry.getKey());
         if (editor!=null) {
            editor.setValue(editor.getValue((String)entry.getValue()));
         }
      }
   }

   /**
    * Returns an editor code suitable for editing the i-th parameter
    * of a method.
    *
    * @param subtance the substance object
    * @param method the method that contains the parameter
    * @param i the parameter index
    */

   public HTMLEditor getEditorComponent(Object substance,
                                        AbstractMethodItem method,
                                        int i)
   {
      // IT SHOULD BE FACTORIZED WITH RenderFactory.getEditorComponent
      Class[] paramTypes = method.getParameterTypes();
      String[] paramNames = (String[]) method.getAttribute( "Gui.parameterNames" );
      String[] passwordParams = (String[]) method.getAttribute( "Gui.passwordParameters" );
      Object[] defaultValues = (Object[]) method.getAttribute( "Gui.defaultValues" );
      String[] parameterFields = (String[]) method.getAttribute ( "Gui.parameterFields" );
      Object[] parameterChoices = (Object[]) method.getAttribute ( "Gui.parameterChoices" );
      boolean[] editableChoices = (boolean[]) method.getAttribute ( "Gui.editableChoices" );
      String[] parametersType = (String[]) method.getAttribute ( "rtti.parametersType" );
      FieldItem editedField = null;
         
      Object defaultValue = null;
      if( defaultValues != null ) {
         defaultValue = defaultValues[i];
      };
      Object choiceObject = null;
      Object[] choice = null;
      boolean editableChoice = false;
      MetaItem type = null;
      if (parametersType!=null)
         type = ClassRepository.get().getClass(parametersType[i]);
      // if it's a set-method, get the default value from the field
      // as well as the editableChoice attribute
      if (substance != null) {
         FieldItem[] fs = ((MethodItem)method).getWrittenFields();
         if ( fs!=null && fs.length==1) {
            if ( method.getParameterTypes()[0].isAssignableFrom( fs[0].getType() ) ) {
               editedField = fs[0];
               defaultValue = fs[0].get( substance );
               choiceObject = fs[0].getAttribute( "Gui.fieldChoice" );
               editableChoice = GuiAC.isEditableChoice(fs[0]);
               Log.trace("gui.choice","choice = "+choice);
               type = RttiAC.getFieldType(fs[0]);
            }
         }
      }

      if ( parameterFields != null ) {
         FieldItem cfi = ((ClassItem)method.getParent()).getField( parameterFields[i] );
         if ( cfi != null ) {
            choiceObject = cfi.getAttribute( "Gui.fieldChoice" );
         }
      }

      if (parameterChoices != null) {
         choiceObject = parameterChoices[i];
      }

      if ( choiceObject != null && substance!=null) {
         if ( choiceObject instanceof MethodItem ) {
            MethodItem choiceMethod = (MethodItem)choiceObject;
            try {
               if (Modifier.isStatic(choiceMethod.getModifiers())) {
                  choiceObject = choiceMethod.invoke(null,new Object[] {substance});
               } else {
                  choiceObject = choiceMethod.invoke(substance,new Object[0]);
               }
            } catch (Exception e) {
               Log.error("Invocation of choice method failed : "+e);
               e.printStackTrace();
            }
         }
         if ( choiceObject instanceof Object[] ) {
            choice = (Object[])choiceObject;
         }
      }

      HTMLEditor editor = null;
      if (choice!=null) {
         if (editableChoices != null)
            editableChoice = editableChoices[i];
         editor = new HTMLChoiceFieldEditor(choice, editableChoice);
      } else if ( Wrappee.class.isAssignableFrom( paramTypes[i] ) ||
                  paramTypes[i] == Object.class) {
         // jac objects
         CollectionItem collection = (CollectionItem)JacObject.
            getCollaboration().getAttribute("Gui.collectionItem");
         ClassItem collectionType = null;
         if (collection!=null) {
             collectionType = (ClassItem)collection.getAttribute("Gui.collectionType");
         }
         if( collectionType != null ) {
            editor = new HTMLObjectChooser(ClassRepository.get().getClass(collectionType),
                                           editedField,null, true, true);
         } else {
            Predicate predicate = null;
            if (method instanceof MethodItem && substance!=null) {
               CollectionItem[] collections = ((MethodItem)method).getAddedCollections();
               if (collections!=null && collections.length>0) {
                  if (Set.class.isAssignableFrom(collections[0].getType())) {
                     predicate = new NotInCollectionPredicate(
                        (Collection)collections[0].get(substance));
                  }
               }
            }
            editor = new HTMLObjectChooser( ClassRepository.get().getClass(paramTypes[i]),
                                            editedField,predicate, true, true);
         }
      } else if ( paramTypes[i].isArray() || 
                  Collection.class.isAssignableFrom ( paramTypes[i] ) ) {
         // array and collections
         //editor = new HTMLArrayEditor(paramTypes[i]);
         Log.warning("*** HTMLArrayEditor is NOT IMPLEMENTED YET");
      } else {
         ClassItem editorClass = null;
         if (type != null) {
            editorClass = GuiAC.getHTMLTypeEditor(type);
         }
         if (editorClass==null) {
            editorClass = GuiAC.getHTMLTypeEditor(
               ClassRepository.get().getClass(paramTypes[i]));
         }
         if (editorClass!=null && 
             !(passwordParams != null && passwordParams[i].equals("true"))) {
            try {
               ConstructorItem constructor = editorClass.getConstructor(new Class[] {Class.class});
               if (constructor == null) {
                  constructor = editorClass.getConstructor(new Class[0]);
                  if (constructor != null) {
                     editor = (HTMLEditor)constructor.newInstance();
                  } else {
                     Log.error("Could not find a suitable constructor for editor component "+
                               editorClass.getName());
                  }
               } else {
                  editor = (HTMLEditor)constructor.newInstance(new Object[] { paramTypes[i] });
               }
            } catch (Exception e) {
               Log.error("Instanciation of editor component "+editorClass.getName()+" failed");
               e.printStackTrace();
            }
         } else if ( ( paramTypes[i].isPrimitive() || 
                       paramTypes[i].getName().startsWith( "java.lang" ) ) &&
                     ( ! paramTypes[i].isArray() ) ) {
            // primitive types (int, float, ...
            editor = new HTMLPrimitiveFieldEditor(
               passwordParams != null && passwordParams[i].equals("true"));
         } else {
            Log.error("Could not find an editor component for type "+paramTypes[i]);
         }  // we should have a fallback editor
      }
      if (editor!=null) {
         editor.setValue(defaultValue);
         editor.setType(paramTypes[i]);
      }
      return editor;
   }

   /**
    * Adds the HTML code for a field editor.
    *
    * @param s the stream the code is added to
    * @param substance the object that contains the field
    * @param field the field */

   protected void addEditorCode( PrintStream s, Object substance, FieldItem field ) {
      MethodItem method = field.getWritingMethods()[0];
      Object value = null;
      value = field.getThroughAccessor( substance );
      Log.trace("webgui","Editor code : "+substance.getClass().getName()+"["+substance+"] "+
                field.getName()+": "+value);
      addInputCode( s, substance, method, false, true, 
                    new String[] { field.getName() }, true, value);

   }

   /**
    * Adds the HTML code for a method.
    *
    * @param s the stream the code is added to
    * @param method the concerned method
    * @param field the name */
   
   public void addMethodCode(PrintStream s, 
                             MethodItem method, 
                             String name) {
      String dots = method.getParameterTypes().length == 0 ? "" : "(...)";
      String methodName = NamingConventions.textForName( method.getName() );
      
      /*
      s.println( "<div class=method><A HREF=\"JacServlet?"+
                 "action=invoke"+
                 "&amp;method="+method.getFullName()+
                 "&amp;jid="+name+
                 "\">"+methodName+"</A></div>" );
      */
      s.println(
         "<input type=submit class=method name=\"_INVOKE_"+method.getFullName()+"\" "+
         "value=\""+methodName+"\">");
   }

   /**
    * Adds the HTML code that corresponds to a given collection item
    * of a given object method into the given stream.
    *
    * @param s the stream to add into
    * @param substance the object that contains the collection
    * @param collection the collection to generate the HTML code from */

   protected void addCollectionCode( PrintStream s, 
                                     Object substance, 
                                     CollectionItem collection ) {

      Log.trace("webgui","addCollectionCode("+substance+","+collection+")");
      s.println("<div class=collection>");

      String name = null;
      try {
         name = NameRepository.get().getName(substance);
      } catch( Exception e ) { 
         name = substance.toString();
      }

      // add and remove methods
      MethodItem[] addingMeths = collection.getAddingMethods();
      if ( addingMeths != null ) {
         addMethodCode(s,addingMeths[0],name);
      }
      MethodItem[] removingMeths = collection.getRemovingMethods();
      if ( removingMeths != null ) {
         addMethodCode(s,removingMeths[0],name);
      }

      Collection c = null;
      if (Map.class.isAssignableFrom(collection.getType()))
         c = ((Map)collection.getThroughAccessor(substance)).entrySet();
      else
         c = collection.getActualCollectionThroughAccessor(substance);
      if( c!=null && c.size() > 0 ) {
         s.println("<div class=collName>"+
                   NamingConventions.textForName(collection.getName())+
                   ":</div>");
      }
      addTableCode(s, c, collection);
      s.println("</div>");

   }

   /**
    * Adds the HTML code of a table where values are picked from a
    * collection.
    *
    * @param s the stream to add into
    * @param c the collection to create the HTML table from */

   protected void addTableCode( PrintStream s, Collection c, CollectionItem collection ) {
      Log.trace("webgui","addTableCode("+c+")");
      if(c==null) return;
      s.println( "<TABLE class=collection>" );
      Iterator it = c.iterator();
      FieldItem[] fields = null;
      
      try {
         
         int rownum = 0;
         while( it.hasNext() ) {
            Object curObj = it.next();
            if( curObj == null ) continue;
            ClassItem clit = ClassRepository.get().getClass(curObj.getClass());
            //         Log.trace("webgui","line -> "+clit.getName());
            if (fields==null) {
               if (collection!=null)
                  fields = (FieldItem[])collection.getAttribute("Gui.attributesOrder");
               if (fields==null)
                  fields = (FieldItem[])clit.getAttribute("Gui.tableAttributesOrder");
               if (fields==null)
                  fields = (FieldItem[])clit.getAttribute("Gui.attributesOrder");
               if (fields==null) {
                  fields = clit.getFields();
               }
               s.print( "<THEAD><TR>" );
               for ( int i = 0; i < fields.length; i++ ) {
                  if ( fields[i] != null ) {
                     s.print( "<TH>"+NamingConventions.textForName(fields[i].getName())+"</TH>");
                  } else {
                     s.print( "<TH>null</TH>");
                  }
               }
               //if( curObj instanceof Wrappee && (
               //   clit.getAttribute("Gui.tableOpenColumn")==null ||
               //   (clit.getAttribute("Gui.tableOpenColumn")!=null && 
               //   clit.getAttribute("Gui.tableOpenColumn").equals("true")))) {
               //   s.print( "<TH></TH>" );
               //}
               s.println( "</TR></THEAD>" );
               s.println( "<TBODY>" );
            }
            // odd colums have different class
            String rowtype = ( rownum%2 == 0 ) ? " class=even" : " class=odd";
            s.println( "<TR"+rowtype+">" );
            for ( int i = 0; i < fields.length; i++ ) {
               s.print( "<TD>" );
               addFieldCode( s, curObj, fields[i], true );
               s.println( "</TD>" );
            }
            if( curObj instanceof Wrappee && (
               clit.getAttribute("Gui.tableOpenColumn")==null ||
               (clit.getAttribute("Gui.tableOpenColumn")!=null && 
                clit.getAttribute("Gui.tableOpenColumn").equals("true"))) ) {
               String name = NameRepository.get().getName(curObj);
               s.println( "<TD class=innerButton>"+
                          "<input type=submit class=method "+
                          "name=\"_VIEW_"+name+"\" "+
                          "value=\"view\"></TD>" );
            }
            s.println( "</TR>" );
            rownum++;
         }

      } catch ( Exception e ) {
         e.printStackTrace();
      }

      s.println( "</TBODY></TABLE>" );         //      Log.trace("webgui","createTableCode done.");
   }

   public Object[] buildParameterValues(AbstractMethodItem method, String[] params) {
      Object[] result = new Object[params.length];
      Class[] pts = method.getParameterTypes();
      for(int i=0; i<params.length; i++)
      {
         HTMLEditor editor = getEditorComponent(null,method,i);
         result[i] = editor.getValue(params[i]);
      }
      return result;
   }

   /*
   public void setResult( String result ) {
      Log.trace("webgui","setResult");
      this.result = result;
      semaphore.release();
      }*/

   // display interface

   public String getDisplayID() {
      return "web";
   }

   public void setDisplayID(String displayID) {}

   public void close() {}

}

class InvalidActionException extends RuntimeException {}
