/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.naming;

import jac.core.*;
import jac.util.*;
import jac.core.dist.*;

import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import java.util.Hashtable;

/**
 * This aspect component implements the default binding policy for
 * the JAC system.
 *
 * <p>The binding aspect uses the naming aspect. Do not try to use it
 * alone.
 *
 * @see NamingAC */

public class BindingAC extends AspectComponent {
   
   /**
    * Bind a deserialized JAC object.
    *
    * <p>This method takes the name filled by the naming aspect within
    * the serialized infos and wraps the final object by a
    * <code>BindingWrapper</code> so that it can be resolved later
    * on. If the forwarding is disabled for this object, the new
    * object is registered within the repository and is not wrapped by
    * a binding wrapper.
    *
    * @param orgObject the JAC object that is being deserialized.
    *
    * @see jac.util.Repository#register(String,Object) 
    * @see jac.aspects.naming.NamingAC#whenSerialized(SerializedJacObject)
    * @see #whenDeserialized(SerializedJacObject)
    * @see jac.core.SerializedJacObject#getACInfos(String)
    * @see BindingWrapper */
   
   public void whenDeserialized ( SerializedJacObject orgObject ) {
      
      String name = (String) orgObject.getACInfos( "naming" );
      //System.out.println ( "<<< when deserialied binding 1 " );
      if ( name != null ) {
         Wrappee finalObject = (Wrappee) attr("finalObject");
         //System.out.println ( "<<< when deserialied binding 2 " + finalObject );
         //if ( ! orgObject.isForwarder() ) {
         //   NameRepository.get().dump();
         //}
         Object newFinal = NameRepository.get().getObject( name );
         //if ( ! orgObject.isForwarder() ) {
         //   NameRepository.get().dump();
         //   System.out.println ( finalObject + " was replaced by " + newFinal );
         //}
         if ( newFinal != null ) {
            //System.out.println( "BINDING: deserializing existing object " + name );
            lattrdef( "finalObject", newFinal );
            finalObject = (Wrappee) newFinal;
         } else {
            if ( orgObject.isForwarder() ) {
               //System.out.println( "BINDING: deserializing non exisiting object (forwarding) " + name );
               ((Wrappee)finalObject).wrapAll ( new BindingWrapper ( name ), "bind" );
            } else {
               //System.out.println( "BINDING: deserializing non existing object (not forwarding) " + name );
               //NameRepository.get().dump();
               //NameRepository.get().register( name, finalObject );
               
               // Upcall the ACManager to notify it from a kind of 
               // remote instantiation
               ((ACManager) ACManager.get()).whenRemoteInstantiation( 
                  (Wrappee) finalObject, name );

               //NameRepository.get().dump();
            }
         }
      }
      
   }

   /**
    * Add the name to the SerializedJacObject when an object that is
    * wrapped by a <code>BindingWrapper</code> serialized.
    *
    * @param finalObject the corresponding serialized structure.
    *
    * @see #whenDeserialized(SerializedJacObject)
    * @see jac.core.SerializedJacObject#setACInfos(String,Object)
    * @see BindingWrapper */

   public void whenSerialized ( SerializedJacObject finalObject ) {

      NameRepository nr = (NameRepository) NameRepository.get();
      Wrappee orgObject = (Wrappee)attr("orgObject");
      if ( orgObject.isExtendedBy( BindingWrapper.class ) ) {
         //System.out.println( "Serializing binder for " + orgObject );
         String name = (String) orgObject.invokeRoleMethod(
            "getLogicalName", new Object[0] );
         if ( name != null ) {
            //System.out.println( "BINDING: serializing object " + name );
            finalObject.setACInfos( "naming", name );
         }
      }
   }

   public void whenObjectMiss(String name) {
      Log.trace("binding", "object miss in name repository: "+name);
      if( name.equals("parserimpl0") ) {
         Parser acParser = new ParserImpl();
         String crName = DistdClassLoader.classRepositoryName;
         // crName == null means that we are on the master site
         if( crName != null ) {
            RemoteRef rr = RemoteContainer.resolve( crName ).bindTo("parserimpl0");
            ((Wrappee)acParser).wrapAll( new StubWrapper( rr ), "invoke" );
         }
         attrdef( "foundObject", acParser );
      }
   }
}



