/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.naming;

import jac.core.*;
import java.util.Hashtable;
import jac.util.Log;

/**
 *
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 */

/**
 * Handles the naming aspect within the JAC system.
 *
 * <p>This aspect component automatically registers all the created
 * JAC object into the sole instance of <code>NameRepository</code>.
 *
 * @see NameRepository */

public class NamingAC extends AspectComponent {

   /**
    * Creates a naming aspect component.
    *
    * <p>It creates the name repository if it does not exist yet.
    *
    * @see NameRepository */

   public NamingAC() {

      NameRepository nr = (NameRepository) NameRepository.get();
      if ( nr == null ) {
         System.out.println ("Error: cannot create name repository.");
         System.exit ( -1 );
      }
   }
   
   private static Hashtable counters = new Hashtable();

   /**
    * Generate default JAC object names.
    *  
    * <p>The programmer should overload this method to generically
    * generate a name for the given JAC object. By default, the
    * generated name is the name of the class concatenated with the
    * index of the JAC object in the array of all the current
    * instances of its class.
    *
    * <p>This method is used by this aspect component to name the JAC
    * objects.
    * 
    * @param object the object to give a name for
    * @return the proposed name
    *
    * @see #whenNewInstance(Wrappee) */

   public String generateName ( Object object ) {
      String className = object.getClass().getName();
      String res = className.substring( className.lastIndexOf('.') + 1 )
         . toLowerCase() + counters.get(className);
      Log.trace( "naming","generated name " + res );
      //jac.core.dist.Distd.getLocalContainerName()+"/"+
      return res;
   }

   public static void forceName( String name ) {
      JacObject.getCollaboration().addTransientAttribute("forceName", name);
   }

   /**
    * Name a new instance using the <code>nameObject()</code> and
    * <code>generateName()</code> methods.
    *
    * @param newInstance the new object to name
    *
    * @see #nameObject(Object,String)
    * @see #generateName(Object) */

   public void whenNewInstance ( Wrappee newInstance ) {
      //System.out.println( "Naming " + newInstance );
      
      if ( ! ( newInstance instanceof AspectComponent ) ) {
         String className = newInstance.getClass().getName();
         Long n = (Long)counters.get(className);
         if (n==null) {
            n = new Long(0);
         }
         counters.put(className,new Long(n.longValue()));
         nameObject ( newInstance, (String)attr("forceName") );
         counters.put(className,new Long(n.longValue()+1));
         attrdef("forceName", null);
      }
      //System.out.println( "EO Naming " + newInstance );
   }

   /**
    * This method is upcalled when a new object is instanciated from a
    * remote site.
    * 
    * <p>The name that is passed is the name of the remote reference
    * that has been used to create the object. Thus, the default
    * behavior of the naming aspect is to register the new object into
    * the name repository.
    * 
    * @param newInstance the remotly created new instance
    * @param name the name that was forwarded from the creator host
    * 
    * @see NameRepository
    * @see jac.util.Repository#register(String,Object) */

   public void whenRemoteInstantiation( Wrappee newInstance, String name ) {
      NameRepository.get().register( name, newInstance );
   }
   
   /**
    * Name a single object by registering it in the name repository
    * (see the <code>NameRepository</code> class).
    * 
    * <p>If the repository does not exist, it is created.
    *
    * @param object the object to name
    * @param name the given name
    *
    * @see NameRepository
    * @see jac.util.Repository#register(String,Object) */

   public void nameObject ( Object object, String name) {
      NameRepository nr = (NameRepository) NameRepository.get();
      if ( nr == null ) {
         System.out.println ("Error: cannot create name repository.");
         System.exit ( -1 );
      }
      if ( name == null ) {
         name = generateName ( object );
      }
      attrdef( "finalName", name );
      ((ACManager)ACManager.get()).whenNameObject( object );
      name = (String) attr( "finalName" );
      nr.register ( name, object );
   }

   /**
    * Add the name to the <code>SerializedJacObject</code> structure
    * when serialized.
    *
    * <p>This allows the name to be serialized with a JAC object so
    * that the object can be named with the same name when deserialized
    * (see <code>BindingAC</code>).
    *
    * @param finalObject the structure corresponding to the object
    * that is being serialized
    *
    * @see BindingAC#whenDeserialized(SerializedJacObject)
    * @see jac.core.SerializedJacObject#setACInfos(String,Object) */

   public void whenSerialized ( SerializedJacObject finalObject ) {

      NameRepository nr = (NameRepository) NameRepository.get();
      Wrappee orgObject = (Wrappee)attr("orgObject");
      String name = nr.getName( orgObject );
      if ( name != null ) {
         //         System.out.println( "NAMING: serializing object " + name );
         finalObject.setACInfos( ACManager.get().getName( this ), name );
      } else {
         if ( orgObject.isExtendedBy( jac.wrappers.ForwardingWrapper.class ) ) {
            //System.out.println( "NAMING: Serializing forwardee for " + orgObject );
            Object newOrg = orgObject.invokeRoleMethod( "getForwardee", new Object[0] );
            if ( newOrg != null ) {
               lattrdef( "orgObject", newOrg );
               orgObject = (Wrappee) newOrg;
            } else {
               System.out.println( "NAMING: Oooops!! Forwardee is " + newOrg );
            }
         }
         name = nr.getName( orgObject );
         if ( name != null ) {
            //System.out.println( "NAMING: serializing object " + name );
            finalObject.setACInfos( ACManager.get().getName( this ), name );
         } else {
            System.out.println( "NAMING: Oooops!! Object is still unamed: " + orgObject );
            //System.out.println( "NAMING: still a forwarder? => " + orgObject.isExtendedBy( jac.wrappers.ForwarderWrapper.class) );
         }
      }
   }
}













