/*

  Copyright (C) 2001 Laurent Martelli.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.persistence;

import java.util.*;
import java.lang.reflect.*;
import java.io.*;

import jac.core.rtti.*;
import jac.util.Log;
import jac.util.Strings;
import jac.util.Semaphore;

/**
 * A FileSystem storage
 */

public class FSStorage implements Storage {

   // names -> oid 
   private Hashtable names = new Hashtable();
   // oid -> name
   private Hashtable oids = new Hashtable();
   // oid -> classid
   private Hashtable classes = new Hashtable();
   
   private File basedir;
   private File oidsFile;
   private File classesFile;

   private long lastOID = 0;

   static Semaphore semaphore = new Semaphore(1);

   public FSStorage(String basedirName) {
      basedir = new File(basedirName);
      if (!basedir.isAbsolute()) {
         basedir = new File(jac.core.Jac.getJacRoot(),basedir.toString());
      }
      if (! basedir.isDirectory()) {
         basedir.mkdirs();
      }
      oidsFile = new File(basedir,"oids");
      try {
         oidsFile.createNewFile();
         readOids();
      } catch(Exception e) {
         e.printStackTrace();
      }

      classesFile = new File(basedir,"classes");
      try {
         classesFile.createNewFile();
         readClasses();
      } catch(Exception e) {
         e.printStackTrace();
      }

      Runtime.getRuntime().addShutdownHook(
         new Thread() {
               public void run() {
                  // wait for completion of deleteObject
                  semaphore.acquire();
                  System.out.println("FSStorage shutdown hook completed");
                  System.out.println("Bye bye...");
               }
            }
      );
   }

   void readOids() throws Exception 
   {
      StreamTokenizer tokens = getStreamTokenizer(oidsFile);

      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         Log.trace("storage",2,"token : "+tokens.sval+"/"+tokens.nval+" "+tokens.ttype);
         LongOID oid = new LongOID(Long.parseLong(tokens.sval));
         long oidval = oid.getOID();
         if (oidval>lastOID)
            lastOID = oidval;
         tokens.nextToken();
         Log.trace("storage",2,"token : "+tokens.sval+"/"+tokens.nval+" "+tokens.ttype);
         String name = tokens.sval;
         oids.put(oid,name);
         names.put(name,oid);
      }
   }

   void writeOids() throws Exception 
   {
      Log.trace("storage","writeOids");
      PrintWriter writer = 
         new PrintWriter(new FileWriter(oidsFile.toString(),false),false);

      Set entries = oids.entrySet();
      Iterator i = entries.iterator();
      while(i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         writer.println(entry.getKey().toString()+" "+entry.getValue());
      }
      writer.flush();
   }

   void readClasses() throws Exception 
   {
      Log.trace("storage","readClasses");
      StreamTokenizer tokens = getStreamTokenizer(classesFile);
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         LongOID oid = new LongOID(Long.parseLong(tokens.sval));
         long oidval = oid.getOID();
         if (oidval>lastOID)
            lastOID = oidval;
         tokens.nextToken();
         String classid = tokens.sval;
         Log.trace("storage",2,oid.toString()+" -> "+classid);
         classes.put(oid,classid);
      }
   }

   void writeClasses() throws Exception 
   {
      Log.trace("storage","writeClasses");
      PrintWriter writer = 
         new PrintWriter(new FileWriter(classesFile.toString(),false),false);

      Set entries = classes.entrySet();
      Iterator i = entries.iterator();
      while(i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         writer.println(entry.getKey().toString()+" "+entry.getValue());
      }
      writer.flush();
   }

   protected StreamTokenizer getStreamTokenizer(File file) 
      throws FileNotFoundException, IOException
   {
      file.createNewFile();
      StreamTokenizer tokens = new StreamTokenizer(new FileReader(file));
      tokens.resetSyntax();
      tokens.wordChars('\000','\377');
      tokens.whitespaceChars(' ',' ');
      tokens.whitespaceChars('\t','\t');
      tokens.whitespaceChars('\n','\n');
      tokens.whitespaceChars('\r','\r');
      return tokens;
   }

   public OID createObject(String className) throws Exception
   {
      semaphore.acquire();
      try {
         lastOID++;
         LongOID oid = new LongOID(lastOID);
         classes.put(oid,className);
         PrintWriter writer = 
            new PrintWriter(new FileWriter(classesFile.toString(),true),true);
         writer.println(oid.toString()+" "+className);
         return oid;
      } finally {
         semaphore.release();
      }
   }

   public void deleteObject(OID oid) throws Exception
   {
      semaphore.acquire();
      try {
         Log.trace("storage","deleteObject("+oid+")");
         String name = (String)oids.get(oid);
         if (name!=null) {
            names.remove(name);
            oids.remove(oid);
            writeOids();
         } else {
            Log.warning("FSStorage.deleteObject: oid "+oid+" does not have name");
         }
         if ( classes.contains(oid)) {
            classes.remove(oid);
            writeClasses();
         } else {
            Log.warning("FSStorage.deleteObject: oid "+oid+" does not have a class");
         }
             //         Thread.currentThread().sleep(1000);
      } catch(Exception e) {
         e.printStackTrace();
      } finally {
         Log.trace("storage","deleteObject("+oid+") DONE");
         semaphore.release();
      }
   }

   public void setField(OID oid, FieldItem field, Object value) 
      throws Exception
   {
      Log.trace("storage","setField("+oid+","+field+","+value+")");
      Properties props = new Properties();
      File objectFile = new File(basedir,oid.toString());
      objectFile.createNewFile();
      props.load(new FileInputStream(objectFile));
      props.setProperty(field.getName(),ValueConverter.objectToString(value));
      props.store(new FileOutputStream(objectFile),"Object "+oid);
   }

   public void updateField(OID oid, FieldItem field, Object value) 
      throws Exception
   {
      Log.trace("storage","updateField("+oid+","+field+","+value+")");
      setField(oid,field,value);
   }

   public Object getField(OID oid, FieldItem field)
      throws Exception
   {
      Log.trace("storage","getField("+oid+","+field+")");
      Properties props = new Properties();
      File objectFile = new File(basedir,oid.toString());
      objectFile.createNewFile();
      props.load(new FileInputStream(objectFile));
      String value = (String)props.get(field.getName());
      Object ret;
      if (value != null) {
         ret = ValueConverter.stringToObject(value);
      } else {
         if (field.isPrimitive()) {
            Log.warning("no such field in storage "+oid+","+field.getName());
         }
	 ret = null;
      }
      Log.trace("storage",2," -> "+ret);
      return ret;
   }

   protected Properties getFields(OID oid) throws IOException
   {
      Properties props = new Properties();
      props.load(new FileInputStream(new File(basedir,oid.toString())));
      return props;
   }

   public StorageField[] getFields(OID oid,ClassItem cl,FieldItem[] fields) 
      throws Exception
   {
      Properties props = getFields(oid);
      StorageField ret[] = new StorageField[fields.length];
      int count = 0;
      for (int i=0; i<fields.length; i++) {
         String value = (String)props.get(fields[i].getName());
         if (value!=null)
            ret[i] = new StorageField(cl,fields[i],ValueConverter.stringToObject(value));
         else
            ret[i] = new StorageField(cl,fields[i],null);
      }
      return ret;
   }

   // Collection functions

   public void clearCollection(OID cid) throws Exception
   {
      saveList(cid,new Vector());
   }

   public long getCollectionSize(OID cid) throws Exception
   {
      return getList(cid).size();
   }

   public OID getCollectionID(OID oid, CollectionItem collection) 
      throws Exception
   {
      return (OID)getField(oid,collection);
   }

   // List functions

   public List getList(OID oid, CollectionItem collection)
      throws Exception
   {
      return getList(getCollectionID(oid,collection));
   }

   public List getList(OID cid)
      throws Exception
   {
      File file = new File(basedir,cid.toString());
      StreamTokenizer tokens = getStreamTokenizer(file);
      Vector ret = new Vector();
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         ret.add(ValueConverter.stringToObject(Strings.unslashify(tokens.sval)));
      }
      return ret;
   }

   public Object getListItem(OID cid, long index)
      throws Exception
   {
      return getList(cid).get((int)index);
   }

   public boolean listContains(OID cid, Object value) 
      throws Exception
   {
      return getList(cid).contains(value);
   }

   protected void saveList(OID cid, List list) 
      throws IOException
   {
      Log.trace("storage","saveList("+cid+","+list+")");
      File file = new File(basedir,cid.toString());
      Log.trace("storage","file = "+file);
      PrintWriter writer = new PrintWriter(new FileWriter(file.toString(),false),true);
      for (int i=0; i<list.size(); i++) {
         String value = ValueConverter.objectToString(list.get(i));
         Log.trace("storage","  -> "+value);
         writer.println(Strings.slashify(value));
      }
   }

   public void addToList(OID cid, long position, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+position+","+value+")");
      List list = getList(cid);
      list.add((int)position,value);
      saveList(cid,list);
   }

   public void addToList(OID cid, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+value+")");
      List list = getList(cid);
      list.add(value);
      saveList(cid,list);
   }

   public void setListItem(OID cid, long index, Object value)
      throws Exception
   {
      List list = getList(cid);
      list.set((int)index,value);
      saveList(cid,list);
   }

   public void removeFromList(OID cid, long position)
      throws Exception
   {
      List list = getList(cid);
      list.remove((int)position);
      saveList(cid,list);
   }

   public void removeFromList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      list.remove(value);
      saveList(cid,list);
   }

   public long getIndexInList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      return list.indexOf(value);
   }

   public long getLastIndexInList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      return list.lastIndexOf(value);
   }


   // Set functions

   public List getSet(OID oid, CollectionItem collection) 
      throws Exception 
   {
      return getSet(getCollectionID(oid,collection));
   }

   public List getSet(OID cid) 
      throws Exception 
   {
      return getList(cid);
   }

   public boolean addToSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","addToSet("+cid+","+value+")");
      List list = getList(cid);
      boolean ret = list.add(value);
      saveList(cid,list);
      return ret;
   }

   public boolean removeFromSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","addToSet("+cid+","+value+")");
      List list = getList(cid);
      boolean ret = list.remove(value);
      saveList(cid,list);
      return ret;
   }

   public boolean setContains(OID cid, Object value) 
      throws Exception
   {
      return getList(cid).contains(value);
   }

   // Map functions

   public Map getMap(OID oid, CollectionItem collection) 
      throws Exception
   {
      return getMap(getCollectionID(oid,collection));
   }

   public Map getMap(OID cid) 
      throws Exception
   {
      Log.trace("storage","getMap("+cid+")");
      File file = new File(basedir,cid.toString());
      StreamTokenizer tokens = getStreamTokenizer(file);
      Hashtable ret = new Hashtable();
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         Object key = ValueConverter.stringToObject(Strings.unslashify(tokens.sval));
         tokens.nextToken();
         Object value = ValueConverter.stringToObject(Strings.unslashify(tokens.sval));
         Log.trace("storage","  -> "+key+","+value);
         ret.put(key,value);
      }
      return ret;
   }

   protected void saveMap(OID cid, Map map) 
      throws IOException
   {
      Log.trace("storage","saveMap("+cid+","+map+")");
      File file = new File(basedir,cid.toString());
      Log.trace("storage","file = "+file);
      PrintWriter writer = new PrintWriter(new FileWriter(file.toString(),false),true);
      Set entrySet = map.entrySet();
      Iterator i = entrySet.iterator();
      while (i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         String value = ValueConverter.objectToString(entry.getValue());
         String key = ValueConverter.objectToString(entry.getKey());
         Log.trace("storage","  -> "+key+","+value);
         writer.println(Strings.slashify(key)+" "+Strings.slashify(value));
      }
   }

   public Object putInMap(OID cid, Object key, Object value) 
      throws Exception
   {
      Map map = getMap(cid);
      Object ret = map.put(key,value);
      saveMap(cid,map);
      return ret;
   }

   public Object getFromMap(OID cid, Object key) 
      throws Exception
   {
      return getMap(cid).get(key);
   }

   public boolean mapContainsKey(OID cid, Object key)
      throws Exception
   {
      return getMap(cid).containsKey(key);
   }

   public boolean mapContainsValue(OID cid, Object value)
      throws Exception
   {
      return getMap(cid).containsValue(value);
   }

   public Object removeFromMap(OID cid, Object key)
      throws Exception
   {
      Map map = getMap(cid);
      Object ret = map.remove(key);
      saveMap(cid,map);
      return ret;
   }

   // others...

   public void removeField(OID oid, FieldItem field, Object value) 
      throws Exception
   {}

   public String newName(String className)
   {
      lastOID++;
      return className+lastOID;
   }

   public OID getOIDFromName(String name) throws Exception
   {
      return (OID)names.get(name);
   }

   public String getNameFromOID(OID oid) throws Exception
   {
      return (String)oids.get(oid);
   }

   public void bindOIDToName(OID oid,String name) throws Exception
   {
      semaphore.acquire();
      try {
         names.put(name,oid);
         oids.put(oid,name);
         PrintWriter writer = new PrintWriter(new FileWriter(oidsFile.toString(),true),true);
         writer.println(oid.toString()+" "+name);
      } finally {
         semaphore.release();
      }
   }

   public void deleteName(String name) throws Exception
   {}

   public String getClassID(OID oid) throws Exception
   {
      Log.trace("storage","getClassID("+oid+") -> "+(String)classes.get(oid));
      return (String)classes.get(oid);
   }

   public Collection getRootObjects() throws Exception
   {
      return oids.keySet();
   }

   public Collection getObjects(ClassItem cl) throws Exception
   {
      Log.trace("storage","getObjects("+cl+")");
      if (cl == null) {
         return classes.keySet();
      } else {
         Vector ret = new Vector();
         Set entries = classes.entrySet();
         Iterator i = entries.iterator();
         while(i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            Log.trace("storage",2,"testing "+entry.getValue());
            if (cl==null || entry.getValue().equals(cl.getName())) {
               ret.add(entry.getKey());
            }
         }
         return ret;
      }
   }
}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***

