/*
  Copyright (C) 2001 Laurent Martelli.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.persistence;

import jac.core.*;
import jac.util.*;
import jac.aspects.naming.*;
import jac.core.rtti.*;
import java.util.*;

/**
 * This AC defines a generic and configurable persistence aspect.<p>
 */

public class PersistenceAC extends AspectComponent implements PersistenceConf {   

   /** The global identifier for a persistence root class in the
       RTTI. */
   public static String ROOT = "root";

   /** The global identifier for a persistent class in the RTTI. */
   public static String PERSISTENT = "persistent";

   private boolean connected = false;
   /**
    * The used storage. */
   protected Storage storage = null;
   /**
    * The loaded objects from the storage. */
   protected Hashtable objects = new Hashtable();

   ClassItem storageClass;
   String[] storageParameters;

   /*
    * PersistenceConf interface
    */

   public void configureStorage( ClassItem storageClass, String[] storageParameters ) {
      try {
         this.storageClass = storageClass;
      } catch ( Exception e ) {
         e.printStackTrace();
      }
      this.storageParameters = storageParameters;
   }
   
   public void configureClass( ClassItem type,
                               String persistenceType ) {
      Log.trace ( "persistence", 1,"configureClass " +
                  type.getName()+","+persistenceType);
      type.setAttribute( "Persistence.type", persistenceType );
   }

   public static void wrapCollection( ClassItem cli, String fieldName ) {
      CollectionItem collection = (CollectionItem)cli.getField( fieldName );
      if ( collection != null ) {
         collection.setAttribute( "Persistence.wrapped", new Boolean(true) );
      }
   }

   public void registerStatic(String name) {
      statics.add(name);
   }



   /** 
    * Tells whether a collection is wrapped.
    */

   public static boolean isWrapped( FieldItem field ) {
      Object value = field.getAttribute( "Persistence.wrapped" );
      if (java.util.Collection.class.isAssignableFrom(field.getType()) ||
          java.util.Set.class.isAssignableFrom(field.getType()) ||
          java.util.Map.class.isAssignableFrom(field.getType()))
         return true;
      if ( value == null )
         return false;
      else
         return ((Boolean)value).booleanValue();
   }

   private HashSet statics = new HashSet();

   /**
    * Returns the storage.
    *
    * @return the storage
    */
   public Storage getStorage() {
      if ( storage == null ) {
         Log.trace("persistence",getClass().getName()+".connectStorage("+storageClass+","+Arrays.asList(storageParameters)+")");
         try {
            Vector constructors = storageClass.getConstructors();
            ConstructorItem constructor = null;
            for (int i=0; i<constructors.size(); i++) {
               if (((ConstructorItem)constructors.get(i)).getParameterTypes().length==storageParameters.length) {
                  constructor = ((ConstructorItem)constructors.get(i));
                  break;
               }
            }
            if (constructor != null) {
               storage = (Storage) constructor.newInstance( storageParameters );
            } else {
               throw new Exception("Could not find a constructor for class "+storageClass+
                                   " with parameters "+Arrays.asList(storageParameters));
            }
         } catch (Exception e) {
            Log.error("failed to connect to storage : "+e);
            e.printStackTrace();
         }
      }
      return storage;
   }

   /*
   public void whenNewInstance ( Wrappee newInstance ) {
      Repository nr = NameRepository.get();
      String name = nr.getName(newInstance);
      if ( !statics.contains(name) ) {
         // names of non static root objects will be handle when they are wrapped.
         ClassItem cli = ClassRepository.get().getClass( newInstance.getClass() );
         String type = (String) cli.getAttribute( "Persistence.type" );
         if ( type!=null && (type.equals(ROOT) || type.equals(PERSISTENT)) ) {
            Log.trace("naming","unregistering persistent instance "+newInstance);
            nr.unregisterObject(newInstance);
         }
         }
         }*/

   /**
    * The persistence aspect checks if the given name is okay
    * regarding the objects that are already within the storage. */

   public void whenNameObject( Object object ) {      
      if( storage == null ) return;
      if( attr("forceName") != null ) return;
      String name = (String)attr("finalName");
      if ( !statics.contains(name) ) {
         // names of non static root objects will be handled when they are wrapped.
         ClassItem cli = ClassRepository.get().getClass( object.getClass() );
         String type = (String) cli.getAttribute( "Persistence.type" );
         try {
            if ( type!=null && (ROOT.compareToIgnoreCase(type)==0 || 
                                PERSISTENT.compareToIgnoreCase(type)==0 ) ) {
               Log.trace("naming","persistence checks instance name "+
                         object.getClass().getName()+", name="+attr("finalName"));
               OID oid = getStorage().getOIDFromName( name );
               if( oid != null ) {
                  attrdef("finalName", name = getStorage().newName(cli.getName()));
                  Log.trace("naming","name collision with persistence "+
                            object.getClass().getName()+" : "+name);
               }
            }
         } catch( Exception e ) {
            e.printStackTrace();
         }
      }
   }
   
   /**
    * The persistence aspect checks whether an object was in the
    * storage when a <code>NameRepository.getObject</code> call
    * failed.
    */

   public void whenObjectMiss( String name ) {
      Log.trace("naming","persistence recieves an object miss for name "+name);
   }

   /**
    * Dynamically wrap the instances of the base program to be
    * persistent.<p>
    */

   public void whenUsingNewInstance ()
   {      
      Wrappee wrappee = wrappee();
      Log.trace ( "persistence.wrap",
                  "<<< "+getClass().getName()+".whenUsingNewInstance " +
		  wrappee.getClass().getName() + " >>>");

      ClassItem cli = ClassRepository.get().getClass( wrappee.getClass() );
      String type = (String) cli.getAttribute( "Persistence.type" );

      // only wrap ROOT objects by default
      if ( type != null && ( ROOT.compareToIgnoreCase(type)==0 || 
                             PERSISTENT.compareToIgnoreCase(type)==0 ) &&
           ! wrappee.isExtendedBy( PersistenceWrapper.class ) ) 
      {
         wrap(wrappee,null);
      }

      // Performs the load of the collections...
      if ( wrappee.isExtendedBy( ListWrapper.class ) ) {
         wrappee.invokeRoleMethod("load",new Object[0]);
      } else if ( wrappee.isExtendedBy( MapWrapper.class ) ) {
         wrappee.invokeRoleMethod("load",new Object[0]);
      } else if ( wrappee.isExtendedBy( SetWrapper.class ) ) {
         wrappee.invokeRoleMethod("load",new Object[0]);
      }

      // performs the load of the references (for instance, the tree
      // view does not use the references but can print them with a
      // toString so that they must be loaded)

      FieldItem[] refs = cli.getReferences();
      if( refs != null ) {
         for( int i=0; i<refs.length; i++ ) {
            Log.trace("persistence","trying to load reference "+refs[i]);
            wrappee.invokeRoleMethod("getReference",new Object[]{refs[i]});
            //refs[i].getThroughAccessor(wrappee);
         }
      }
   }

   /**
    * Wrap an object with a presistence wrapper.
    *
    * <p>Same as <code>wrap(wrappee,oid,true)</code>.
    *
    * @param wrappee the object to wrap
    * @param oid the oid for this object, may be null
    * @return the persistence wrapper that wraps the wrappee */

   public PersistenceWrapper wrap(Wrappee wrappee, OID oid) {
      return wrap(wrappee,oid,true);
   }

   /**
    * Wraps an object with a presistence wrapper.
    *
    * @param wrappee the object to wrap
    * @param oid the oid for this object, may be null
    * @param doRegister performs the loading and the registering of
    * the object
    * @return the persistence wrapper that wraps the wrappee */

   public PersistenceWrapper wrap(Wrappee wrappee, OID oid, 
                                  boolean doRegister) {
      Log.trace("persistence",
                "wrapping instance of "+wrappee.getClass().getName()+"("+oid+")");
      try {
         boolean isNew = false;
         PersistenceWrapper persistence_wrapper = null;
         ClassItem cli = ClassRepository.get().getClass( wrappee.getClass() );
         String type = (String) cli.getAttribute( "Persistence.type" );

         String name = NameRepository.get().getName(wrappee);

         if (oid == null) {
            // new or static instance
            if (name==null) {
               // generate a new name for non static instance
               name = getStorage().newName(cli.getName());
               Log.trace("naming","new name for persistent instance "+
                         wrappee.getClass().getName()+" : "+name);
               NameRepository.get().register(name,wrappee);
            }
            oid = getStorage().getOIDFromName( name );
            if ( oid == null ) {
               oid = storage.createObject(cli.getName());
               storage.bindOIDToName( oid, name );
               Log.trace("persistence.wrap",wrappee.getClass().getName()+" isNew");
               isNew = true;
            } else {
               isNew = false;
            }
         } else {
            // restored instance : make the object has the same
            // name as the one stored in the storage.
            if (name == null ) {
               name = getStorage().getNameFromOID( oid );
               Log.trace("naming","restored name for persistent instance "+
                         wrappee.getClass().getName()+" : "+name);
            }
            Log.trace("naming","registered instance "+name+" -> "+wrappee.getClass().getName());
            NameRepository.get().register( name, wrappee );
            isNew = false;
         }

         persistence_wrapper = new PersistenceWrapper(oid);
         wrappee.wrap( persistence_wrapper );
         // cur_AC may not be set
         /*
         String ac = ACManager.get().getName(this);
         persistence_wrapper.setAspectComponent(ac);
         */
         Log.trace("persistence.wrap","wrapper = "+persistence_wrapper);

         // wrap the methods
         Collection methods = cli.getMethods();
         Iterator it = methods.iterator();
         while ( it.hasNext() ) {
            MethodItem[] method_items = (MethodItem[]) it.next();
            if (method_items!=null) {
               for (int i=0; i<method_items.length; i++) {
                  MethodItem method = method_items[i];
                  // wrap this method if it accesses or writes some fields or collections
                  if (method.hasAccessedReferences()
                      || method.getWrittenFields() != null
                      || method.getAddedCollections() != null 
                      || method.getRemovedCollections() != null) 
                  {  
                     Log.trace("persistence.wrap",2,"wrapping method "+method);
                     wrappee.wrap( persistence_wrapper, 
                                   "applyPersistence", method.getName() );
                  }
               }
            }
         }

         // Wrap collections
         Vector collectionsToInit = new Vector();
         CollectionItem[] collections = cli.getCollections();
         for ( int i=0; i<collections.length; i++) {
            Object coll_value = collections[i].get(wrappee);
            Class collType = collections[i].getType();
            if ( coll_value == null ) {
               //               Log.warning("persistence.wrap",
               //                           "uninitialized collection "+collections[i].getName());
               // uninitialized field
               if (collType.isArray()) {
                  coll_value = java.lang.reflect.Array.newInstance(
                     collType.getComponentType(),0);
               } else {
                  coll_value = collType.newInstance();
               }
               collections[i].set(wrappee,coll_value);
            }

            if ( isWrapped(collections[i]) || 
                 collType == jac.lib.java.util.Vector.class ||
                 collType == jac.lib.java.util.HashMap.class ||
                 collType == jac.lib.java.util.Hashtable.class ||
                 collType == jac.lib.java.util.HashSet.class ) 
            {
               OID coll_oid;
               if (isNew) {
                  coll_oid = getStorage().createObject(collections[i].getType().getName());
               } else {
                  coll_oid = (OID)getStorage().getField(oid,collections[i]);
                  // classes may change after objects were created and made persistent
                  if (coll_oid==null) {
                     coll_oid = getStorage().createObject(collections[i].getType().getName());
                     collectionsToInit.add(collections[i]);
                  }
               }

               if ( collections[i].isList() ) {
                  // lists
                  Log.trace("persistence.wrap",2,
                            "wrapping List "+collections[i].getName());
                  ListWrapper wrapper = new ListWrapper(coll_oid);
                  //                  wrapper.setAspectComponent(ac);
                  Wrappee list = (Wrappee)coll_value;
                  list.wrap(wrapper);
                  list.wrapSameName(wrapper,
                                    new String[] {"add","clear","contains",
                                                  "get","remove","set","size",
                                                  "indexOf","lastIndexOf","isEmpty"});
                  list.wrap(wrapper,"preload",
                            new String[] {"equals","toArray","iterator"});
               } else if ( collections[i].isSet() ) {
                  // sets
                  Log.trace("persistence.wrap",2,
                            "wrapping Set "+collections[i].getName());
                  SetWrapper wrapper = new SetWrapper(coll_oid);
                  //                  wrapper.setAspectComponent(ac);
                  Wrappee set = (Wrappee)coll_value;
                  if ( set != null ) {
                     set.wrap(wrapper);
                     set.wrapSameName(wrapper,
                                      new String[] {"add","clear","contains",
                                                    "remove","size","isEmpty"});
                     set.wrap(wrapper,"preload",
                              new String[] {"equals","toArray","iterator"});
                  } else {
                     Log.warning("persistence",
                                 "Set "+cli.getName()+"."+collections[i].getName()+
                                 " is null, not wrapped");
                  }
               } else if ( collections[i].isMap() ) {
                  // maps
                  Log.trace("persistence.wrap",2,
                            "wrapping Map "+collections[i].getName());
                  MapWrapper wrapper = new MapWrapper(coll_oid);
                  Wrappee map = (Wrappee)coll_value;
                  if ( map != null ) {
                     map.wrap(wrapper);
                     map.wrapSameName(wrapper,
                                      new String[] {"clear","isEmpty","size",
                                                    "containsKey","containsValue","get","put"});
                     map.wrap(wrapper,"preload",
                              new String[] {"equals","keySet","entrySet","values"});
                  } else {
                     Log.warning("persistence",
                                 "Map "+cli.getName()+"."+collections[i].getName()+
                                 " is null, not wrapped");
                  }
               }
            }
         }

         if( doRegister ) {
            if ( isNew ) {
               persistence_wrapper.initAllFields( wrappee );
            } else {
               Log.trace("persistence",2,"collectionsToInit.size = "+collectionsToInit.size());
               Log.trace("persistence",2,"collectionsToInit = "+collectionsToInit);
               CollectionItem[] array = new CollectionItem[0];
               persistence_wrapper.initCollections( 
                  wrappee, 
                  (CollectionItem[])collectionsToInit.toArray(array));
               loadAllFields(wrappee,oid);
            }

            registerObject(oid,wrappee);

         }

         Log.trace("persistence.wrap",">>> instance wrapped "+persistence_wrapper);
         return persistence_wrapper;
      } catch (Exception e) {
         Log.error("Persistence: wrapping failed : "+wrappee.getClass().getName());
         e.printStackTrace();
         return null;
      }
   }

   /**
    * Ensure that an object is persistent.
    *
    * @param wrappee the object to make persistent
    * @return the persistence wrapper that wraps the persistent object */

   public PersistenceWrapper makePersistent( Wrappee wrappee ) throws Exception {

      if( wrappee == null ) return null;
      PersistenceWrapper wrapper = null;
      wrappee.forceTreatment();
      Object[] wrappers = wrappee.getWrappers( PersistenceWrapper.class );

      if ( wrappers.length == 0 ) {
         Log.trace("persistence",
                   "makePersistent : no persistence wrapper for "+wrappee.getClass().getName());
         wrapper = wrap( wrappee, null );
      } else {
         wrapper = (PersistenceWrapper)wrappers[0];
      }
      Log.trace("persistence",
                "makePersistent : already persistent "+wrapper.getOid());
      return wrapper;
   }

   /**
    * Returns the PersistenceWrapper of an object.
    *
    * @param wrappe a <code>Wrappe</code>
    * @return the <code>PersistenceWrapper</code> of the wrappee
    */
   public static PersistenceWrapper getPersistenceWrapper(Wrappee wrappee) 
      throws NoSuchWrapperException
   {
      wrappee.forceTreatment();
      Object[] wrappers = wrappee.getWrappers( PersistenceWrapper.class );

      if ( wrappers.length == 0 )
         throw new NoSuchWrapperException();
      if (wrappers.length>1)
         Log.warning("persistence",
                     wrappers.length+" PersistenceWrappers found "+
                     "for object "+wrappee.getClass().getName());
      return (PersistenceWrapper)wrappers[0];
   }   

   /**
    * Add an object in the list of persistent objects.
    *
    * @param oid the OID of the object
    * @param object the object
    */
   protected void registerObject(OID oid, Object object) {
      Log.trace("persistence","registerObject("+oid+","+object.getClass()+")");
      //      Log.trace("persistence",2,"objects = "+objects);
      //      new Exception().printStackTrace();
      if (objects.containsKey(oid))
         throw new Error("Persistence: an object is already registered with this oid: "+oid);
      objects.put(oid,object);
      Log.trace("persistence",2,"object "+oid+" added");
   }

   /**
    * Gets a persistent object in the storage from its OID.
    *
    * <p>By default, this method loads all the fields that are
    * primitive to initialize the object state. Collections and
    * references are initialized later on when they are accessed by
    * some methods.<p>
    *
    * @param oid the OID of the seeked object
    * @return the object
    *
    * @see PersistenceWrapper#initAllFields(Wrappee) */
 
   public synchronized Object getObject(OID oid) throws Exception {

      Object result = objects.get( oid );

      if ( result != null ) {
         Log.trace("persistence","Object "+oid+" found in cache");
         return result;
      } else {
         Log.trace("persistence",
                   this+".Object "+oid+" NOT found in cache; Loading from storage\n");
         String lClassID = getStorage().getClassID( oid );
         if (lClassID==null)
            Log.error("persistence: getClassID("+oid+") -> NULL");
         ClassItem lClass = ClassRepository.get().getClass(lClassID);
         Log.trace("persistence","Class = "+lClass.getName());
         attrdef("forceName", getStorage().getNameFromOID(oid));
         Object newObject = lClass.newInstance();
         
         Wrappee wrappee = (Wrappee)newObject;
         PersistenceWrapper wrapper = wrap(wrappee,oid);
         Log.trace("persistence",
                   "New object " + oid + " : " +newObject.getClass().getName() );
         Log.trace("persistence","Object loaded");
         return newObject;
      }
   }

   /**
    * Load all the fields of an object from the storage and initialize
    * the object.
    *
    * @param wrappee the object
    * @param oid the object's oid
    */
   public void loadAllFields(Wrappee wrappee, OID oid) throws Exception
   {
      //      checkOid();
      String lClassID = wrappee.getClass().getName();
      Log.trace("persistence","loadAllFields("+oid+")");

      // Load all fields from the storage
      ClassItem cli = ClassRepository.get().getClass(lClassID);

      // this should be faster because only one call to the storage is done
      FieldItem[] fields = cli.getFields();
      if ( fields.length>0 ) {
         StorageField[] values = getStorage().getFields(oid,cli,fields);
         for ( int i=0; i<values.length; i++) {
            if ( values[i]!=null ) {
               FieldItem field = values[i].fieldID;
               if (field.isPrimitive()) {
                  field.set(wrappee,values[i].value);
               } else if (field.isReference() && 
                          ( field.getAccessingMethods() == null || 
                            field.getAccessingMethods().length == 0 ) ) {
                  wrappee.invokeRoleMethod( "getReference", new Object[] { field } );
               } else if ( (field instanceof CollectionItem) ) {
                  if ( ! ((CollectionItem)field).isWrappable() ) {
                     Log.trace("persistence","loadAllFields -> invoking getcollection" );
                     wrappee.invokeRoleMethod( "getCollection", new Object[] { field } );
                  }
               }
            }
         }
      }
   }

   /**
    * This method allows the deserialization of the OID of a
    * persistent object.<p>
    *
    * @param orgObject the JAC object structure that is beeing
    * deserialized
    * @see #whenSerialized(SerializedJacObject) */

   public void whenDeserialized ( SerializedJacObject orgObject ) {
      
      Log.trace("persistence","PersistenceAC.whenDeserialize");
      OID oid = (OID) orgObject.getACInfos( "persistence" );
      if ( oid != null ) {
         Wrappee finalObject = (Wrappee)attr("finalObject");
         PersistenceWrapper pw = wrap( finalObject, oid, false );
         //pw.setOid( oid );
      }
      
   }

   /**
    * This method add the OID info to the serialized JAC object when a
    * serialization is requested by another aspect.<p>
    *
    * This adding allows, for instance, the OID to be transmitted to
    * remote containers.<p>
    *
    * @param finalObject the object that is being serialized */

   public void whenSerialized ( SerializedJacObject finalObject ) {
      
      Log.trace("persistence","PersistenceAC.whenSerialize");
      Wrappee orgObject = (Wrappee)attr("orgObject");
      if ( orgObject.isExtendedBy( PersistenceWrapper.class ) ) {
         finalObject.setACInfos(
            "persistence", orgObject.invokeRoleMethod( "getOid", new Object[0] ) );
      }
   }

   /**
    * Load objects from the storage when required.
    */
   public void whenGetObjects(List objects, ClassItem cl)
   {
      if( cl == null ) return;
      try {
         Collection oids = getStorage().getObjects(cl);
         Iterator i = oids.iterator();
         while (i.hasNext()) {
            OID oid = (OID)i.next();
            Object object = getObject(oid);
            if (!objects.contains(object))
               objects.add(object);

         }
      } catch (Exception e) {
         e.printStackTrace();
      }
   }

   public boolean beforeRunningWrapper( Wrapper wrapper, 
                                        String wrappingMethod ) {
      if( attr("Persistence.disabled") != null ) {
         return false;
      } else {
         return true;
      }
   }

   public void whenDeleted(Wrappee object) {
      if(storage!=null) {
         try {
            storage.deleteObject(
               ((OID)object.invokeRoleMethod("getOid",new Object[0])));
         } catch(Exception e) {
            //e.printStackTrace();
         }
      }
   }

   public static class NoSuchWrapperException extends RuntimeException {}

}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***
