/*

  Copyright (C) 2001 Laurent Martelli.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.aspects.persistence;

import java.util.*;
import java.lang.reflect.*;
import java.io.*;
import jac.core.*;
import jac.aspects.naming.*;
import jac.core.rtti.*;
import jac.util.*;

/**
 * This wrapper defines persistence extensions for objects that a
 * defined persitent by a persistent aspect component.
 *
 * <p>The unique and generic wrapping method that handles persistency is
 * <code>applyPersistence</code>.<p>
 *
 * @see #applyPersistence() */

public class PersistenceWrapper extends Wrapper {

   /**
    * The OID of the persitent object in the used database. */
   protected OID oid;

   private HashSet loadedVectors = new HashSet();
   private HashSet loadedReferences = new HashSet();

   /**
    * Create a persistence wrapper for a given class that is
    * configured by a <code>Config</code> instance.<p>
    *
    * @param config the class configuration
    * @param oid the oid of the persistent object */

   public PersistenceWrapper(OID oid) {
      super();
      if (oid==null) 
         throw new Error("oid cannot be null in PersistenceWrapper constructor");
      this.oid = oid;
   }

   /**
    * The setter method for the oid.
    *
    * @param oid the new oid
    */

   public void setOid(OID oid ) {
      Log.trace("persistence","setOid("+oid+")");
      if (this.oid!=null && !this.oid.equals(oid))
         throw new Error("Changing OID "+this.oid+" -> "+oid);
      this.oid = oid;
   }

   /**
    * The getter method for the oid.
    *
    * @return the oid of the wrappee
    */

   public OID getOid() {
      return oid;
   }

   protected Object normalizeOutput(Object value) throws Exception {
      if (value instanceof OID) {
         value =  getAC().getObject((OID)value);
      }
      return value;
   }

   protected Object normalizeInput(Object value) throws Exception {
      if (value instanceof Wrappee) {
         Wrappee wrappee = (Wrappee)value; // added object
         PersistenceWrapper wrapper = getAC().makePersistent(wrappee);
         
         if ( wrapper.getOid()==null ) {
            // this should never happen
            Log.error("oid is NULL in addToCollection, method = "+
                      method()+Arrays.asList(args()) );
            throw new InvalidOidException("oid is NULL in addToCollection, method = "+
                                          method()+Arrays.asList(args()));
         }
         value = wrapper.getOid();
      }
      return value;
   }

   /**
    * Add a vector to the loaded vectors list.
    * 
    * @param name the name of the vector to load */

   public void setVectorLoaded( String name ) {
      Log.trace("persistence","Set vector loaded: " + name );
      loadedVectors.add(name);
   }

   /**
    * Tells if the current object is persistent (role method).
    *
    * @return true if persistent */

   public boolean isPersistent() {
      return (oid != null);
   }

   /**
    * This wrapping method apply the persistence aspect for a given
    * call on a persistent object.
    *
    * <p>It uses the RTTI aspect to determine the accessed and written
    * field types.<p>
    *
    * Default behavior is:<p>
    *
    * <ul>
    * <li>get the fields that are accessed for reading by this wrapped
    * method
    * <li>if the field is a collection, get the collection from the
    * storage
    * <li>if the field is a reference, get the referenced object from
    * the storage
    * <li>proceed the wrappee (and other wrappers if needed)
    * <li>get the fields that were written by this wrapped method
    * <li>if the field is a collection, update the storage depending
    * on the wrapped method type (adder or remover)
    * <li>if the field is a reference, update the storage to change
    * the referenced object
    * <li> if the field is a simple field, just update the storage
    * with its value.
    * </ul>
    *
    * @return the value returned by the wrapped method
    *
    * @see jac.core.rtti.MethodItem#getAccessedFields()
    * @see #getCollection(CollectionItem)
    * @see #getReference(FieldItem)
    * @see #addToCollection(CollectionItem,Object)
    * @see #removeFromCollection(CollectionItem,Object)
    * @see #setReference(FieldItem,Wrappee)
    * @see #setField(FieldItem)
    * @see jac.core.rtti.MethodItem#getWrittenFields()
    * @see jac.core.rtti */

   public Object applyPersistence() throws Exception 
   {
      Log.trace("persistence",2,"applyPersistence on "+method());

      // the object may be wrapped but not persistent yet
      if (oid==null ) {
	 return proceed();
      }

      ClassItem cli = ClassRepository.get().getClass(wrappee().getClass());
      MethodItem method = cli.getMethod(method());
      
      FieldItem[] fields;
      // preload accessed fields and collections
      fields = method.getAccessedFields();
      for( int i = 0; fields != null && i < fields.length; i++ ) {
         FieldItem field = fields[i];
	 
         if (field instanceof CollectionItem && ! PersistenceAC.isWrapped( field ) ) {
            getCollection( (CollectionItem) field );
         } else if (field.isReference()) {
            getReference(field);
         } 
         
      }
      
      Object result = proceed();
      
      // save written fields
      fields = method.getWrittenFields();
      for( int i = 0; fields != null && i < fields.length; i++ ) {
         FieldItem field = fields[i];
         Log.trace("persistence","Save written field in "+method()+" : "+fields[i].getName());
         if (JacObject.class.isAssignableFrom( field.getType() )) {
            setReference(field,(Wrappee)arg(0));
         } else {
            setField(field);
         }
      }
      
      CollectionItem[] collections = method.getAddedCollections();
      for( int i = 0; collections != null && i < collections.length; i++ ) {
         if ( ! PersistenceAC.isWrapped( collections[i] ) ) {
            addToCollection(collections[i],arg(0));
         }
      }
      collections = method.getRemovedCollections();
      for( int i = 0; collections != null && i < collections.length; i++ ) {
         if ( ! PersistenceAC.isWrapped( collections[i] ) ) {
            removeFromCollection(collections[i],arg(0));
         }
      }
      
      return result;
   }
   
   /**
    * This method loads alls the persistent objects that are contained
    * in this collection from the storage.
    *
    * <p>If the collection has already been loaded from the storage,
    * do nothing.<p>
    *
    * @param collection the collection to load */

   public void getCollection(CollectionItem collection) throws Exception
   {
      checkOid();
      Log.trace("persistence",oid+".getCollection("+collection.getName()+")");
      if ( ! loadedVectors.contains( collection.getName() ) ) {
         String classID = collection.getParent().getName();
         List vector;
         if (collection.isList() || collection.isArray()) {
            vector = getStorage().getList( oid, collection );
         } else if (collection.isSet()) {
            vector = getStorage().getSet( oid, collection );
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            Log.error("please see the JAC programming guidelines for more details");
            //new Exception().printStackTrace();
            return;
         }

         collection.clear( wrappee() );

         for( int i = 0; i < vector.size(); i++) {
            collection.add( wrappee(), normalizeOutput(vector.get(i)), null );
         }
         loadedVectors.add( collection.getName() );
      }
   }

   /**
    * This method stores the persistent object that is added to the
    * collection into the storage.<p>
    *
    * @param collection the collection to add to
    *
    * @see #removeFromCollection(CollectionItem,Object)
    * @see #getCollection(CollectionItem)
    */   

   protected void addToCollection( CollectionItem collection, Object value ) 
      throws Exception 
   {
      checkOid();
      Log.trace("persistence",
                oid+".addToCollection: "+method()+Arrays.asList(args()));

      if ( isPersistent() && (args().length == 1) )
      {
         // the method's arguments look correct
         value = normalizeInput(value);
         OID cid = getStorage().getCollectionID(oid, collection);
         if ( collection.isList() || collection.isArray() ) {
            getStorage().addToList( cid, value );
         } else if ( collection.isSet() ) {
            getStorage().addToSet( cid, value );
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            new Exception().printStackTrace();
         }
         
      } else {
         Log.error("NOT IMPLEMENTED YET !!!");
         // the method's arguments are weird, so let's diff
         //         storage.updateCollection(collection);
      }
   }

   /**
    * This method delete the persistent object that is removed from
    * the collection from the storage.<p>
    *
    * @param collection the collection to remove from
    *
    * @see #addToCollection(CollectionItem,Object)
    * @see #getCollection(CollectionItem) */

   protected void removeFromCollection( CollectionItem collection, Object value ) 
      throws Exception 
   {
      checkOid();
      if (isPersistent()) {

         Log.trace("persistence",oid+".removeFromCollection "+method());

         value = normalizeInput(value);
         OID cid = getStorage().getCollectionID(oid,collection);
         if ( collection.isList() || collection.isArray() ) {
            getStorage().removeFromList( cid, value );
         } else if ( collection.isSet() ) {
            getStorage().removeFromSet( cid, value );
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            new Exception().printStackTrace();
         }
      }
   }

   /**
    * This method loads the persistent object that is pointed by the
    * reference from the storage.<p>
    *
    * If the reference has already been loaded from the storage, do
    * nothing.<p>
    *
    * @param reference the reference to load
    *
    * @see #setReference(FieldItem,Wrappee) */

   public void getReference( FieldItem reference ) throws Exception 
   {
      try {
         checkOid();
         Log.trace("persistence.ref",oid+".getReference("+reference.getName()+")");
         if ( isPersistent() ) {
            if ( ! loadedReferences.contains( reference.getName() ) ) {
               OID lOid = (OID)getStorage().getField(oid, reference);
               Log.trace("persistence.ref",oid+"."+reference+" -> "+lOid+" ("+
                         wrappee().getClass().getName()+")");
               Object obj = null;
               if ( lOid != null ) {
                  obj = getAC().getObject( lOid );
               }
               reference.set( wrappee(),obj );
               loadedReferences.add( reference.getName() );
            } else {
               Log.trace("persistence",2,oid+"."+reference+" already loaded");
            }
         } else {
            Log.trace("persistence",oid+" is not persistent, "+reference+" not loaded");
         }
      } catch (Exception e) {
         Log.error("getReference");
         e.printStackTrace();
      }
   }

   /**
    * This method stores the persistent object that is pointed by the
    * reference into the storage.<p>
    *
    * @param reference the reference to store
    *
    * @see #getReference(FieldItem) */

   protected void setReference( FieldItem reference, Wrappee value ) 
      throws Exception { 
      
      if( value != null ) {
         checkOid();
         if ( isPersistent() ) {
            PersistenceWrapper wrapper = getAC().makePersistent( value );
            if ( wrapper.oid == null ) {
               // this should never happen
               Log.error("oid is NULL in setReference, method = "+method());
               throw new Exception("Null OID Exception in "+method());
            }
            getStorage().updateField( oid, reference, wrapper.oid );
         }
      } else {
         getStorage().updateField( oid, reference, null );
      }
   }

   /**
    * This method stores the field into the storage.<p>
    *
    * @param field the field to store */

   protected void setField( FieldItem field ) throws Exception { 
      checkOid();
      if ( isPersistent() ) {
         getStorage().updateField( oid, field, field.get( wrappee() ) ) ;
      }
   }

   public void initCollections( Wrappee wrappee, CollectionItem[] collections ) 
      throws Exception 
   {
      Log.trace("persistence", "wrappee = "+wrappee.getClass().getName() );
      Log.trace("persistence", "collections = "+collections );
      Log.trace("persistence", "collections.length = "+collections.length );
      for ( int i = 0; i < collections.length; i++ ) {
         Log.trace("persistence", "collections["+i+"] = "+collections[i] );
         Log.trace("persistence", "Init collection "+collections[i].getName() );
         if ( collections[i].isList() || collections[i].isSet() || collections[i].isArray() ) {
            OID cid;
            if ( collections[i].isWrappable() ) {
               cid = getAC().getPersistenceWrapper((Wrappee)collections[i].get(wrappee)).getOid();
            } else {
               cid = getStorage().createObject( collections[i].getType().getName() );
            }
            getStorage().setField( oid, collections[i], cid);
            Collection coll = collections[i].getActualCollection( wrappee );
            Log.trace("persistence","Coll value = " + coll );
            Iterator it = coll.iterator();
            while( it.hasNext() ) {
               Object value = normalizeInput(it.next());
               Log.trace("persistence","Coll elt value = " + value );
               if (value!=null) {
                  if ( collections[i].isList() || collections[i].isArray() ) {
                     getStorage().addToList( cid, value );
                  } else {
                     getStorage().addToSet( cid, value );
                  }
               }
            }
         } else if ( collections[i].isMap() ) {
            OID cid;
            if ( collections[i].isWrappable() ) {
               cid = getAC().getPersistenceWrapper((Wrappee)collections[i].get(wrappee)).getOid();
            } else {
               cid = getStorage().createObject( collections[i].getType().getName() );
            }
            getStorage().setField( oid, collections[i], cid);
            Map map = (Map)collections[i].get(wrappee);
            Set entries = map.entrySet();
            Iterator it = entries.iterator();
            while ( it.hasNext() ) {
               Map.Entry entry = (Map.Entry)it.next();
               getStorage().putInMap(cid,normalizeInput(entry.getKey()),
                                     normalizeInput(entry.getValue()));
            }
         } else {
            Log.error("unhandled collection type : "+collections[i].getType());
            new Exception().printStackTrace();
         }
      }
   }

   /**
    * Initialize all fields of an object in the storage.
    *
    * <p>This method is called when a new object is made persistent. The
    * references and collections are also recursively saved into the
    * storage.<p>
    *
    * @param oid the object ID in the storage
    * @param wrappee the initializer object
    *
    * @see PersistenceAC#getObject(OID) */

   public void initAllFields( Wrappee wrappee ) throws Exception {
      checkOid();
      String classID = wrappee.getClass().getName();
      Log.trace("persistence","initAllfields("+oid+") : "+classID);
      ClassItem cli = ClassRepository.get().getClass(classID);
      FieldItem[] fields = cli.getPrimitiveFields();

      // fields
      for ( int i = 0; i < fields.length; i++ ) {
         Object fieldValue = fields[i].get( wrappee );
         Log.trace("persistence",
                   "Init field "+fields[i].getName()+"->"+fieldValue);
         if ( fieldValue!=null ) {
            getStorage().setField( oid, fields[i], fieldValue );
         }
      } 

      FieldItem[] references = cli.getReferences();

      // references
      for ( int i = 0; i < references.length; i++ ) {
         Log.trace("persistence","Init reference " + references[i].getName() );
         Wrappee obj = (Wrappee) references[i].getActualField().get( wrappee );
         Log.trace("persistence","Ref value = " + 
                    ( obj!=null ? obj.getClass().getName() : "null" ) );
         
         if ( obj!=null ) {
            PersistenceWrapper wrapper = getAC().makePersistent( obj );

            if ( wrapper.oid == null ) {
               // this should never happen
               Log.error("wrapper.oid is NULL in initAllFields, reference = "+
			 references[i].getName());
            } else {
               Log.trace("persistence","Ref value = " + wrapper.oid );
               getStorage().setField( oid, references[i], wrapper.oid);
               
               Log.trace("persistence","Ref saved" );
            }
         }
         
      }

      initCollections(wrappee,cli.getCollections());

   }

   public Storage getStorage() throws Exception {
      Log.trace("bug",this+"AC -> "+getAspectComponent());
      return ((PersistenceAC)getAspectComponent()).getStorage();
   }

   public PersistenceAC getAC() {
      return (PersistenceAC)getAspectComponent();
   }

   public void checkOid() throws InvalidOidException
   {
      if (oid == null) 
         throw new InvalidOidException("oid is NULL");
   }

   public class InvalidOidException extends Exception {
      public InvalidOidException(String msg) { super(msg); }
   }
}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***
