/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.tracing;

import jac.core.*;
import jac.wrappers.*;
import jac.aspects.naming.*;

import java.util.*;

/**
 * This sample aspect component traces the calls on all the objects of
 * the program.
 *
 * <p>To be active this aspect component must be configured with the
 * <code>addTrace</code> method.
 *
 * @see jac.wrappers.VerboseWrapper */

public class TracingAC extends AspectComponent implements TracingConf  {

   public void addTrace( String wrappeeExpr, 
                         String wrappeeClassExpr, 
                         String wrappeeMethodExpr ) {

      pointcut( wrappeeExpr, wrappeeClassExpr, wrappeeMethodExpr,
                VerboseWrapper.class.getName(), "printCallingInfos", false );
   }

   public void addRecording( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr ) {

      pointcut( wrappeeExpr+" && !recorder0",
                wrappeeClassExpr,
                wrappeeMethodExpr,
                RecordingWrapper.class.getName(), "record", false );
   }

   Hashtable counters = new Hashtable();

   public void addCounter( String name,
                           String wrappeeExpr,
                           String wrappeeClassExpr, 
                           String wrappeeMethodExpr ) {

      Counter c = (Counter) counters.get( name );
      if( c == null ) {
         c = new Counter();
         counters.put( name, c );
      }
      pointcut( wrappeeExpr, wrappeeClassExpr, wrappeeMethodExpr,
                new SimpleCountingWrapper(c), "incr" );
   }

   public void addOptimizedCounter( String name,
                                    String wrappeeExpr, 
                                    String wrappeeClassExpr, 
                                    String wrappeeMethodExpr,
                                    String fieldName,
                                    String argNumber) {
      
      Counter c = (Counter) counters.get( name );
      if( c == null ) {
         c = new Counter();
         counters.put( name, c );
      }
      if( ! fieldName.equals("") ) {
         pointcut( wrappeeExpr, wrappeeClassExpr, 
                   wrappeeMethodExpr, 
                   new OptimizedCountingWrapper(c,fieldName), 
                   "incrWithField" );
      }
      if( ! argNumber.equals("") ) {
         pointcut( wrappeeExpr, wrappeeClassExpr, 
                   wrappeeMethodExpr, 
                   new OptimizedCountingWrapper(
                      c, (new Integer(argNumber)).intValue()), 
                   "incrWithArg" );
      }
   }

   /**
    * Skips the counting wrapper if this call is part of a global
    * conting optimization performed by an
    * <code>OptimizedCountingWrapper</code>
    *
    * @param wrapper the wrapper 
    * @param wrappingMethod the wrapping method that is about to be
    * run
    * @see OptimizedCountingWrapper */

   public boolean beforeRunningWrapper ( Wrapper wrapper, 
                                         String wrappingMethod ) {
      if( attr("tracing.globalIncr") != null ) return false;
      return true;
   }


}

/**
 * The recording wrapper (upcalls the recorder).
 */

public class RecordingWrapper extends Wrapper {
   public Object record() {
      Object ret = proceed();
      if(Recorder.get().isRecording()) {
         Recorder.get().recordMethodCall(wrappee(), method(), args());
      }
      return ret;
   }
}






