/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package jac.aspects.transaction;

import jac.core.AspectComponent;
import jac.core.Wrapper;
import jac.core.Wrappee;

import jac.util.Log;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Iterator;

/**
 * This aspect component handle the transaction aspect within JAC
 * applications.
 *
 * @author Renaud Pawlak */

public class TransactionAC extends AspectComponent implements TransactionConf {

   /**
    * The default constructor. Wraps every application objects with a
    * <code>DispatchTransactionWrapper</code> instance.
    *
    * @see DispatchTransactionWrapper
    * @see DispatchTransactionWrapper#dispatch
    */

   public TransactionAC() {
      pointcut("ALL", "ALL", "ALL", 
               DispatchTransactionWrapper.class.getName(), 
               "dispatch", true);      
   }

   public void defineTransactionalMethods(String classExpr, 
                                          String methodExpr, 
                                          String objectExpr) {
      pointcut(objectExpr, classExpr, methodExpr, 
               new TransactionWrapper(), "handleTransaction");
   }

   /**
    * This wrapper defines the transaction wrapper.
    *
    * <p>It defines a wrapping method that starts a transaction when a
    * method is called and that commits it when the method returns
    * normally.
    *
    * <code>Within the transaction, the cloning and the dispatching is
    * ensured by an external
    * <code>DispatchTransactionWrapper</code>.
    *
    * @see DispatchTransactionWrapper */

   public class TransactionWrapper extends Wrapper {
      /** The current transaction count (increments on each
          transaction). */
      int transactionCount=0;
      /**
       * This wrapping method wraps a transactional method.
       */
      public Object handleTransaction() {
         beginOfTransaction();
         Object ret = proceed();
         endOfTransaction();
         return ret;
      }
      /**
       * Initializes the transaction. */
      public void beginOfTransaction() {
         Log.trace("transaction","begin of transactional method: "+method());
         Integer id = new Integer(++transactionCount);
         attrdef("Transaction.id",id);
         pattrdef("Transaction"+id+".affectedObjects",new Vector());
      }
      /**
       * Ends the transaction (when success). */
      public void endOfTransaction() {
         Log.trace("transaction","end of transactional method: "+method());
         Integer id = (Integer)attr("Transaction.id");
         Vector v = (Vector)attr("Transaction"+id+".affectedObjects");
         attrdef("Transaction.commit",id);
         Iterator it = v.iterator();
         while( it.hasNext() ) {
            Wrappee w = (Wrappee)it.next();
            w.invokeRoleMethod("commit",new Object[] {attr("Transaction.id")});
         }
         pattrdef("Transaction"+id+".affectedObjects",null);
      }
      /**
       * Handles an exception within a transaction (rollback).
       */
      public void catchException(Exception e) throws Exception {
         Log.trace("transaction","exception in transactionnal method: "+method());
         Integer id = (Integer)attr("Transaction.id");
         Vector v = (Vector)attr("Transaction"+id+".affectedObjects");
         attrdef("Transaction.rollback",id);
         if( id != null && v != null ) {
            Iterator it = v.iterator();
            while( it.hasNext() ) {
               Wrappee w = (Wrappee)it.next();
               w.invokeRoleMethod("rollback",new Object[] {attr("Transaction.id")});
            }
            pattrdef("Transaction"+id+".affectedObjects",null);
         }
         throw e;
      }
   }
}


