/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Array;
import java.util.*;
import jac.util.*;
import java.net.*;
import java.io.*;
import jac.core.rtti.*;

/**
 * This class defines aspect component configurations so that the
 * programmer or the system administrator will be able to configure
 * the available aspects for a given application.
 *
 * <p>The default available aspects in the system are the ones that
 * are declared to the AC manager in the <code>jac.prop</code> (see
 * the <code>jac.acs</code> property).
 *
 * @see ApplicationRepository
 * @see Application
 * @see ACParser
 * @see ACManager
 */

public class ACConfiguration implements Serializable {

   /** The application this configuration belongs to. */
   protected Application application;
   /** The name of the configured AC. */
   protected String name;
   /** The aspect component that corresponds to this configuration
       once instantiated (null before). This is a local aspect
       component. */
   transient AspectComponent instance = null;
   
   /** The configuration file's URL. */
   protected URL filePath;

   String acPath = null;

   /**
    * Sets the URL of the configuration file that defines the
    * configuration operations.
    *
    * @param filePath a valid file path
    * @see #getURL() */

   public void setURL( URL filePath ) {
      this.filePath = filePath;
   }

   /**
    * The getter of the configuration file's URL.
    *
    * @return the URL
    * @see #setURL(URL) */

   public URL getURL() {
      return filePath;
   }

   /**
    * This flag tells if the aspect that is configured by the current
    * configuration will be woven on demand (by the administrator or
    * by a configuration program) or if the aspect will be
    * automatically woven and restored by the system.
    *
    * <p>For instance, a persistence aspect should always have this
    * configuration flag to false whilst a debugging aspect should
    * most of the time be woven on demand (when debugging is
    * needed). */

   protected boolean weaveOnDemand = true;
   
   /**
    * Gets the <code>weaveOnDemand</code> flag value.
    * @return the flag value */

   public boolean getWeaveOnDemand() {
      return weaveOnDemand;
   }
   
   /**
    * Sets the <code>weaveOnDemand</code> flag value.
    * @param b the new flag value */

   public void setWeaveOnDemand(boolean b) {
      weaveOnDemand = b;
   }

   /**
    * Creates a new aspect component configuration.
    *
    * @param application the application this configuration belongs to 
    * @param name the name of the AC as defined in the declared ACs of
    * the AC manager
    * @param filePath the path of the configuration file; it can be
    * absolute but, if relative, it is automatically concatened to the
    * application's path
    * @param weaveNow a true value means that the aspect that
    * configured by this configuration will be automatically woven at
    * the application's start, a false value means that the user will
    * have to weave it with a program (or with the administration
    * GUI); default is true */

   public ACConfiguration( Application application, String name, 
                           String filePath, boolean weaveNow ) {
      this.name = name;
      this.application = application;
      this.weaveOnDemand = ! weaveNow;
      if (filePath != null) {
         try {
            if( filePath.startsWith( "file:" ) ) {
               filePath = filePath.substring(5);
            }
            //if( ! filePath.startsWith("/") ) {
            //   filePath = application.getPath() + "/" + filePath;
            //}
            this.filePath = new URL( "file:"+filePath );
         } catch ( Exception e ) {
            e.printStackTrace();
         }
      }
   }

   /**
    * This method returns the list of the allowed configuration
    * operations that can be used by this configuration to configure
    * the aspect component it corresponds to.
    *
    * <p>These operations are automatically deduced by introspecting
    * the aspect component (it is in fact all the modifying public
    * instance methods that takes strings or arrays of string as
    * arguments).
    *
    * @return the allowed configuration operations for this aspect */

   public Object[] getAllowedOperations() {
      String acName = ((ACManager)ACManager.get())
         .getACPathFromName(name);

      Vector names = new Vector();
      if ( acName != null ) {
         try {
            java.lang.reflect.Method[] methods = 
               Class.forName(acName).getDeclaredMethods();
            for ( int i=0; i<methods.length; i++ ) {
               if( NamingConventions.isModifier( methods[i].getName() ) ) {
                  names.add( methods[i].getName() );
               }
            }
         } catch ( Exception e ) {
            e.printStackTrace();
         }
      }
      return names.toArray();
   }

   /**
    * Gets the name of the configured AC as defined in the declared
    * ACs of the AC manager.
    *
    * @return the AC name
    * @see #setName(String)
    * @see ACManager */

   public String getName() {
      return name;
   }

   /**
    * The aspect name setter. Must be declared in the
    * <code>jac.prop</code>.
    *
    * @param name the aspect name 
    * @see ACManager */

   public void setName(String name) { 
      this.name = name;
   }
   
   /**
    * Gets the aspect component instance that corresponds to this
    * configuration.
    *
    * @return the AC instance */

   public AspectComponent getInstance() {
      return instance;
   }

   /**
    * Gets the owning application.
    *
    * @return the application that owns this configuration */

   public Application getApplication() {
      return application;
   }

   /**
    * Return the class item for the configured aspect component.
    *
    * @return the AC type */

   public ClassItem getAspectClass() {
      String acPath = ((ACManager)ACManager.get()).getACPathFromName( name );
      return ClassRepository.get().getClass( acPath );
   }

   /**
    * Instantiates a new aspect component.
    *
    * @return the new aspect component, null if something went wrong
    */

   protected AspectComponent instantiate() {
      Log.trace( "aspects", "instantiating "+name );
      ACManager acm = (ACManager)ACManager.get(); 
      Object oldAC = JacObject.getCollaboration()
         .getAttribute("cur_AC");
      try {
         instance = (AspectComponent) acm.getObject(
            application.getName()+"."+name);
         if (instance == null ) {
            if (acPath == null ) 
               acPath = acm.getACPathFromName( name );
            Log.trace( "aspects", name + " : " + acPath );
            if( acPath == null ) return null;
            Class cl = Class.forName( acPath );
            Log.trace( "aspects", "instantiating " + cl );
            instance = (AspectComponent) cl.newInstance();
         }
         JacObject.getCollaboration()
            .addTransientAttribute("cur_AC",
                                   getApplication().getName() + 
                                   "." + name );
         return instance;
      } catch( Exception e ) {
         e.printStackTrace();
         return null;
      } finally {
         JacObject.getCollaboration()
            .addTransientAttribute("cur_AC", oldAC );
      }
   }

   /**
    * Configures the aspect component.
    *
    * <p>This method takes the aspect component instance that corresponds
    * to this configuration, parse the configuration file, and calls
    * all the configuration operations that are defined in this file.
    *
    * @see ACParser
    * @see #getInstance() */


   protected void configure() {
      //      configure(Jac.getJacRoot()+name+".acc");
      configure(filePath.getFile());
   }

   protected void configure(String filePath) {
      
      if (filePath == null) {
         return;
      }
      Log.trace("aspects.config", "configuring aspect " + name + " with file "+filePath); 
      Vector configMethods = null;
      try {
         // acparser0 is the only instance of ACParser, if we are on the 
         // master site, it is automatically instantiated by the binding 
         // aspect, if not, the binding aspect creates a stub for   
         // acparser0@mastersite
         Parser acp = (Parser)NameRepository.get()
            .getObject( "parserimpl0" );
         configMethods = acp.parse(filePath,instance.getClass().getName());
      } catch (FileNotFoundException e) {
         Log.warning("cannot find config file "+filePath+" for "+name+
                     "("+e+"), using serialized infos.");
      } catch ( Exception e ) {
         e.printStackTrace();
         return;
      }
      
      for( int i = 0; i < configMethods.size(); i++ ) {
         MethodItem method = null;
         ConfigMethod cm = (ConfigMethod)configMethods.get(i);
         try {
            Log.trace("aspects.config", 2, "invoking configuration method '" + 
                      cm.getMethod() + "' on " + instance.getClass() );
            if (cm.getMethod().equals("")) continue;
            MethodItem[] methodItems = ClassRepository.get().getClass( instance.getClass() ).
               getMethods(cm.getMethod());
            Log.trace("aspects.config",2,"matching methods : "+Arrays.asList(methodItems));
            // determine which method to call based on number of parameters
            Object[] paramValues = cm.getArgs();
            for (int j=0; j<methodItems.length; j++) {
               if (methodItems[j].getParameterTypes().length==paramValues.length) {
                  method = methodItems[j];
                  Log.trace("aspects.config",2,"selecting method "+method);
                  break;
               }
            }
            if (method==null) {
               throw new Exception("Wrong number of parameters for method "+cm.getMethod());
            }

            Class[] paramTypes = method.getParameterTypes();
            for (int j=0; j<paramTypes.length; j++) {
               paramValues[j] = convertValue(paramValues[j],paramTypes[j]);
            }
            Log.trace("aspects.config",filePath+": line "+ cm.getLineNumber() +
                        ", invoke configuration method '" + 
                        (method!=null?method.toString():cm.getMethod()) + "' on " + 
                        instance.getClass() + " with " + Arrays.asList(cm.getArgs()));
            method.invoke(instance,paramValues);
         } catch ( Throwable e ) {
            if (e instanceof InvocationTargetException)
               e = ((InvocationTargetException)e).getTargetException();
            Log.warning(cm.getLineNumber() +
                        ", failed to invoke configuration method '" + 
                        (method!=null?method.toString():cm.getMethod()) + "' on " + 
                        instance.getClass() + " with " + Arrays.asList(cm.getArgs())+" : "+e);
         }
      }
   }

   Object convertArray(Object[] array, Class componentType) {
      Object result  = Array.newInstance(componentType, array.length);
      for (int i=0; i<array.length; i++) {
         Array.set(result,i,convertValue(array[i],componentType));
      }
      return result;
   }

   Object convertValue(Object object, Class type) {
      if (object != null && object.getClass() != type) {
         try {
            if (type.isArray()) {
               object = convertArray((Object[])object,type.getComponentType());
            } else if (type==double.class) {
               object = new Double((String)object);
            } else if (type==int.class) {
               object = new Integer((String)object);
            } else if (type==long.class) {
               object = new Long((String)object);
            } else if (type==float.class) {
               object = new Float((String)object);
            } else if (type==boolean.class) {
               object = new Boolean((String)object);
            } else if (type==short.class) {
               object = new Short((String)object);
            } else if (type==Class.class) {
               object = Class.forName((String)object);
            } else if (type==ClassItem.class) {
               object = ClassRepository.get().getClass((String)object);
            } else if  (AbstractMethodItem.class.isAssignableFrom(type)) {
               String str = (String)object;
               int index = str.indexOf("(");
               if (index!=-1)
                  index = str.lastIndexOf(".",index);
               else
                  index = str.lastIndexOf(".");
               if (index!=-1) {
                  ClassItem classItem = ClassRepository.get().getClass(
                     str.substring(0,index));
                  object = classItem.getAbstractMethod( str.substring(index+1) );
               } else {
                  new Exception("Failed to convert "+str+" into a "+type.getName());
               }

            } else if  (FieldItem.class.isAssignableFrom(type)) {
               String str = (String)object;
               int index = str.lastIndexOf(".");
               if (index!=-1) {
                  ClassItem classItem = ClassRepository.get().getClass(
                     str.substring(0,index));
                  object = classItem.getField( str.substring(index+1) );
               } else {
                  new Exception("Failed to convert "+str+" into a "+type.getName());
               }
            }
         } catch (Exception e) {
            Log.error("Failed to convert "+object+" into "+type+" : "+e);
            e.printStackTrace();
         }
      }
      return object;
   }
   
   /**
    * Instantiates, configures, and weaves the aspect component that
    * corresponds to this configuration.
    *
    * @param name the name of the aspect configuration
    * @see #instantiate()
    * @see #configure() */

   public void weave() {
      //acs.put( acName, ac );
      instantiate();
      if( instance == null ) {
         Log.warning("could not instantiate aspect "+name);
         return;
      }
      if ( ((ACManager)ACManager.get()).getObject( 
         getApplication().getName() + "." + name ) == null ) {
         
         Log.trace( "aspects", "registering " + instance + 
                    " on application " + getApplication() );
         ((ACManager)ACManager.get()).register( getApplication().getName() + 
                                                "." + name, instance );
      } else {
         Log.trace( "aspects", "aspect " + instance + 
                    " is already registered on application " + 
                    getApplication() );
      }

      try {
         configure();
      } catch (Exception e) {
         unweave();
      }
   }

   /**
    * Unweaves the aspect component that corresponds this
    * configuration.
    *
    * @param name the name of the aspect configuration */

   public void unweave() {
      ((ACManager)ACManager.get()).unregister( application.getName() + "." + name );
      instance = null;
   }   

   /**
    * Returns a string representation of this configuration.
    * @return a string */

   public String toString() {
      return "AC " + name + " configuration";   
   }

}
