/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import jac.util.*;
import java.util.*;
import java.io.*;

import jac.core.rtti.*;

/**
 *
 * @version 0.8.1
 * @author <a href="mailto:pawlak@cnam.fr">Renaud Pawlak</a>
 */

/**
 * This class manages all the aspect components that are present in
 * the JAC system.
 *
 * <p>When a new aspect component is registered, it is
 * woven regarding its weaving properties. Once woven, the aspect
 * component manager dispatches all the events of the
 * <code>BaseProgramListener</code> to it.
 *
 * @see AspectComponent
 * @see BaseProgramListener */

public class ACManager extends OrderedRepository 
   implements CollaborationParticipant, BaseProgramListener {

   /**
    * The jac.prop property string to declare initially woven aspect
    * components. */

   protected static String acsProp = "jac.acs";

   /**
    * The jac.prop property string to declare when the initially woven
    * aspect components are woven. */

   protected static String startWeavingPlacesProp = "jac.startWeavingPlaces";

   /**
    * The jac.prop property string to set the composition aspect class
    * name. */

   protected static String compositionAspectProp = "jac.comp.compositionAspect";

   /**
    * A internally-used flag to bootstrap the AC manager. */

   public boolean registering = false;

   /**
    * Get the sole instance of Aspect Component Manager for this
    * JAC container.
    * 
    * <p>If this instance does not exist yet, then it is created.
    *
    * <p>This method returns a <code>Repository</code> so that the
    * result must be casted to used specific method of
    * <code>ACManager</code>.
    *
    * @return the aspect component manager of the local container
    * @see jac.util.Repository */
   
   public static Repository get () {
      if ( acManager == null ) return new ACManager();
      return acManager;
   }
   
   /** Stores the sole instance of Aspect Component Manager. */
   protected static ACManager acManager = null;

   /** Stores all the declared aspect components. */
   protected Hashtable declaredACs = new Hashtable();

   /**
    * Declares a new aspect component.
    *
    * @param name the name of the aspect component
    * @param path its path (must be accessible from the current
    * classpath) */

   public void declareAC( String name, String path ) { 
      Log.trace( "aspects", "declaring " + path + " as " + name );
      declaredACs.put( name, path );
   }

   public boolean isACDeclared( String name ) {
      return (declaredACs.get(name) != null);
   }

   /**
    * Gets a declared aspect component path from its name.
    *
    * @param acName the name of the AC
    * @return the corresponding path, null if not declared */

   public String getACPathFromName( String acName ) {
      return (String) declaredACs.get( acName );
   }

   /**
    * Gets the declared aspect components.
    *
    * @return the set of the declared aspect components names */

   public Set getDeclaredACs() {
      return declaredACs.keySet();
   }

   /**
    * Registers a declared aspect as is (not linked to an application
    * and with no configuration).
    *
    * @param name the aspect name */

   public AspectComponent registerDeclaredAC( String name ) {
      AspectComponent ac = null;
      try {
         if ( ! acManager.isRegistered( name ) ) {
            String className = getACPathFromName( name );
            Log.trace( "aspects", "registering the " + name + " ac" );
            acManager.register(
               name, ac = (AspectComponent) Class.forName( className ).newInstance());
         } else {
            Log.trace( "aspects", name + " is already registered" );
         }            
      } catch( Exception e ) {
         e.printStackTrace();
      }
      return ac;
   }

   /**
    * Called by the JAC launcher to create and declare the aspect components
    * defined in jac.prop. */

   public static void main( String[] args ) throws Throwable {

      if ( acManager == null ) new ACManager();

      Log.trace( "aspects", "initializing the AC manager" );

      if ( JacObject.props != null ) {
         
         try {

            /** Create the Composition Aspect */

            String prop1 = JacObject.props.getProperty(compositionAspectProp);

            AspectComponent ac = null;
            try {
               ac = (AspectComponent) Class.forName( prop1 ).newInstance();
            } catch ( Exception e ) {
               e.printStackTrace();
               System.exit( 1 );
            }

            acManager.register( "JAC_composition_aspect", ac );

            /** Create the regular Aspect Components. */

            prop1 = JacObject.props.getProperty(acsProp);

            int acs_count = 0;
         
            if ( prop1 != null ) {
                        
               StringTokenizer st1 = new StringTokenizer( prop1 );
            
               while ( st1.hasMoreElements() ) {
                  acs_count++;
                  String ac_path = (String) st1.nextElement();
                  String ac_name = (String) st1.nextElement();

                  acManager.declareAC( ac_name, ac_path );
               }
            }
            
            /** Creating the needed aspects. */
            
            acManager.registerDeclaredAC( "naming" );
            acManager.registerDeclaredAC( "binding" );
            acManager.registerDeclaredAC( "rtti" );
            acManager.registerDeclaredAC( "deployment" );

            /*
            AspectComponent pac = acManager.registerDeclaredAC( "persistence" );
            ClassItem pacClass = ClassRepository.get().getClass( pac.getClass() );
            pacClass.invoke( pac, "configureStorage", new Object[] {
               Class.forName( JacObject.props.getProperty("jac.appStorage.type") ),
               new Object[] { 
                  JacObject.props.getProperty("jac.appStorage.url"),
                  JacObject.props.getProperty("jac.appStorage.user"),
                  JacObject.props.getProperty("jac.appStorage.password") } } );
            pacClass.invoke( pac, "configureClass", new Object[] {
               ApplicationRepository.class, "root" } );
            pacClass.invoke( pac, "registerStatic", new Object[] {
               "applicationrepository0" } );
            pacClass.invoke( pac, "configureClass", new Object[] {
               Application.class, "persistent" } );
            pacClass.invoke( pac, "configureClass", new Object[] {
               ACConfiguration.class, "persistent" } );
            pacClass.invoke( pac, "wrapCollection", new Object[] {
               ApplicationRepository.class, "applications" } );
            pacClass.invoke( pac, "wrapCollection", new Object[] {
               Application.class, "acConfigurations" } );
            pacClass.invoke( pac, "wrapCollection", new Object[] {
               ACConfiguration.class, "configOperations" } );
            */

            /** Create the needed GUIs. */

            AspectComponent guiAC = null;
            guiAC = acManager.registerDeclaredAC( "gui" );

            //if ( Jac.startGUI() || Jac.startWebGUI() ) {
            //   Log.trace( "aspects", "creating the gui (" + guiAC + ")" );
            //}          

            if ( Jac.startGUI() ) {
               guiAC.getClass().getMethod( "createDisplay", 
                                           new Class[] { String.class, Class.class } )
                  .invoke( guiAC, new Object[] { "admin", Class.forName("jac.aspects.gui.ProgramView") } );
            }

            if ( Jac.startWebGUI() ) {
               guiAC.getClass().getMethod( "createDisplay", 
                                           new Class[] { String.class, Class.class } )
                  .invoke( guiAC, new Object[] { "web", Class.forName("jac.aspects.gui.WebGui") } );
            }

               /*
               String guiACName = null;
               ac = null;
               try {

                  String programDir = Jac.getJacRoot() + "classes/" + 
                     Jac.getLaunchedProgram().replace( '.', '/' );

                  File rdir = new File( programDir );

                  Log.trace("core","Program dir "+programDir);

                  String[] acPaths = rdir.list( new FilenameFilter() {
                        public boolean accept( File dir, String name ) {
                           Log.trace("core","trying "+name);
                           if ( name.endsWith( "GuiAC.class" ) ) {
                              Log.trace("core","Found aspect component "+name);
                              return true;
                           } else {
                              Log.trace("core","Discard "+name);
                              return false;
                           }
                        }
                     } );
                  
                  if ( acPaths != null && acPaths.length > 0 ) {
                     guiACName = acPaths[0].substring( 0, acPaths[0].indexOf('.') )
                        . replace( '/', '.' );
                  }
                  Class acClass = Class.forName( Jac.getLaunchedProgram() + "." + guiACName );
                  Object guiAC;
                  
                  if ( ! acManager.isRegistered ( Jac.getLaunchedProgram() + ".gui" ) ) {
                     acManager.register ( Jac.getLaunchedProgram() + ".gui", guiAC = acClass.newInstance() );
                  } else {
                     guiAC = acManager.getObject( Jac.getLaunchedProgram() + ".gui" );
                  }
                  acClass.getMethod( "createSwingDisplay", new Class[0] ).invoke( guiAC, new Object[0] );
                  
               } catch ( Exception e ) {
                  e.printStackTrace();
                  System.out.println( "Warning: no GUI found for '" + Jac.getLaunchedProgram() + "'." );
                  try {
                     acManager.register ( 
                        "gui", (AspectComponent) Class.forName( "jac.aspects.gui.GuiAC" )
                        . newInstance() );
                  } catch ( Exception e1 ) {
                     System.out.println( "Warning: cannot launch the GUI." );
                  }
               }
            }

            if ( Jac.startWebGUI() ) {
               String guiACName = null;
               ac = null;
               try {

                  String programDir = Jac.getJacRoot() + "classes/" + 
                     Jac.getLaunchedProgram().replace( '.', '/' );

                  File rdir = new File( programDir );

                  String[] acPaths = rdir.list( new FilenameFilter() {
                        public boolean accept( File dir, String name ) {
                           if ( name.endsWith( "GuiAC.class" ) ) {
			       //    Log.trace("core","Found aspect component "+name);
                              return true;
                           } else {
                              return false;
                           }
                        }
                     } );
                  
                  if ( acPaths != null && acPaths.length > 0 ) {
                     guiACName = acPaths[0].substring( 0, acPaths[0].indexOf('.') )
                        . replace( '/', '.' );
                  }

                  Class acClass = Class.forName( Jac.getLaunchedProgram() + "." + guiACName );
                  Object guiAC;
                  
                  if ( ! acManager.isRegistered ( Jac.getLaunchedProgram() + ".gui" ) ) {
                     acManager.register ( Jac.getLaunchedProgram() + ".gui", guiAC = acClass.newInstance() );
                  } else {
                     guiAC = acManager.getObject( Jac.getLaunchedProgram() + ".gui" );
                  }
                  acClass.getMethod( "createWebDisplay", new Class[0] ).invoke( guiAC, new Object[0] );

               } catch ( Exception e ) {
                  e.printStackTrace();
                  System.out.println( "Warning: no Web GUI found for '" + 
                                      Jac.getLaunchedProgram() + "'." );
                  try {
                     acManager.register ( 
                        "gui", (AspectComponent) Class.forName( "jac.aspects.gui.WebGuiAC" )
                        . newInstance() );
                  } catch ( Exception e1 ) {
                     System.out.println( "Warning: cannot launch the Web GUI." );
                  }
               }
            } 
               */

         } catch ( Exception e ) {
            e.printStackTrace();
         }
              
      }
   }

   /**
    * The default constructor will set the acManager field to the
    * right value and bind the JAC insternal MOP to it so that it will
    * be notified of the base program events.
    *
    * @see BaseProgramListener */

   public ACManager () {
      acManager = this;
      JacObject.acm = this;
   }

   /**
    * A flag that is set to true when all the aspect components are
    * woven (for optimization purpose). */

   public boolean allWoven = false;

   /**
    * Store the non-woven yet Aspect Components. */

   protected Vector nonWovenACs = new Vector();

   /**
    * Dispatch this event to all the registered Aspect Components
    * (even if not woven).
    *
    * @param newInstance the new instance
    * @see AspectComponent#whenNewInstance(Wrappee)
    * @see BaseProgramListener#whenNewInstance(Wrappee) */

   public final void whenNewInstance ( Wrappee newInstance ) {
      Log.trace("aspects","------ NEWINST -> "+newInstance.getClass().getName());
      String old_ac = (String)attr("cur_AC");
      try {
         Vector toNotify = (Vector)orderedObjects.clone();
         for ( int i = 0; i < toNotify.size(); i ++ ) { 
            AspectComponent curac = (AspectComponent)toNotify.get(i);
            attrdef( "cur_AC", names.get( curac ) );
            curac.whenNewInstance ( newInstance );
         }
         ((JacObject)newInstance).toNotifyACs.addAll( toNotify );
      } catch ( Exception e ) {
         e.printStackTrace();
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the registered Aspect Components.
    *
    * @param newInstance the remotly created instance
    * @param name the original remote name 
    * @see AspectComponent#whenRemoteInstantiation(Wrappee,String)
    * @see BaseProgramListener#whenRemoteInstantiation(Wrappee,String) */

   public final void whenRemoteInstantiation ( Wrappee newInstance, String name ) {
      System.out.println("------ REMINST "+name+"-> "+newInstance);
      Iterator it = orderedObjects.iterator();
      while ( it.hasNext() ) {
         AspectComponent curac = (AspectComponent)it.next();
         attrdef( "cur_AC", names.get( curac ) );
         curac.whenRemoteInstantiation ( newInstance, name );
      }
      ((JacObject)newInstance).toNotifyACs.addAll( orderedObjects );
   }

   /**
    * Dispatch this event to all the registered Aspect Components that
    * are woven.
    * 
    * @see AspectComponent#whenUsingNewInstance()
    * @see BaseProgramListener#whenUsingNewInstance() */

   public final void whenUsingNewInstance () {

      if ( registering ) return;
      
      //registering = true;
      
      if ( ! allWoven ) {
         checkWeaving();
      }

      Collection org = ((JacObject)wrappee()).toNotifyACs;
      Object[] toNotify = org.toArray();

      /** First remove all the acs that will be notified */
      for ( int i = 0; i < toNotify.length; i++ ) {
         AspectComponent curac = (AspectComponent)toNotify[i];
         if ( curac.woven ) {
            org.remove( curac );
         }
      }

      /** Then notify them... */
      String old_ac = (String)attr("cur_AC");
      try {
         for ( int i = 0; i < toNotify.length; i++ ) {
            AspectComponent curac = (AspectComponent)toNotify[i];
            if ( curac.woven ) {
               attrdef( "cur_AC", names.get( curac ) );
               if ( (! wrappee().getClass().getName().startsWith("jac.core")) || 
                    curac.isSystemListener() ) {
                  curac.whenUsingNewInstance();
               }
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }

      //registering = false;
      
   }

   /**
    * Dispatch this event to all the woven Aspect Components.
    *
    * <p>Note: the current collaboration contains an attribute called
    * "orgObject" and that is a reference to the object that is
    * currently serialized.
    *
    * @param finalObject the corresponding serialized structure.
    * @see AspectComponent#whenSerialized(SerializedJacObject)
    * @see BaseProgramListener#whenSerialized(SerializedJacObject) */

   public final void whenSerialized ( SerializedJacObject finalObject ) {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            attrdef( "cur_AC", names.get( curac ) );
            curac.whenSerialized ( finalObject );
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the woven Aspect Components.
    *
    * <p>Note: the current collaboration contains an attribute called
    * "finalObject" and that is a reference to the object that is
    * will be finally deserialized.
    *
    * @param orgObject the corresponding serialized structure.
    * @see AspectComponent#whenDeserialized(SerializedJacObject)
    * @see BaseProgramListener#whenDeserialized(SerializedJacObject) */

   public final void whenDeserialized ( SerializedJacObject orgObject ) {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            attrdef( "cur_AC", names.get( curac ) );
            curac.whenDeserialized ( orgObject );
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /** 
    * Apply all the woven aspect components to the given wrappee.
    *
    * <p>This is done by upcalling
    * <code>AspectComponent.forceTreatment()</code> on the woven
    * aspect components.
    *
    * @param wrappee the wrappee that have to be treated
    * @see AspectComponent#forceTreatment(Wrappee) */

   public final void forceTreatment ( Wrappee wrappee ) {
      Log.trace("forceTreatment",this.getClass().getName());
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = ((Vector)((JacObject)wrappee).toNotifyACs.clone()).iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            if ( curac.woven ) {
               attrdef( "cur_AC", ACManager.get().getName( curac ) );
               curac.forceTreatment( wrappee );
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /** 
    * Apply all the woven aspect components to all the currently
    * untreated wrapees.
    *
    * <p>This is done by upcalling
    * <code>AspectComponent.forceTreatment()</code> on the woven
    * aspect components.
    *
    * @see AspectComponent#forceTreatment() */

   public final void forceTreatment () {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            if ( curac.woven ) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.forceTreatment ();
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the registered Aspect Components that
    * are woven.
    * 
    * @param cloned the wrappee that is being cloned
    * @param clone the new clone of cloned
    * @see AspectComponent#whenClone(Wrappee,Wrappee)
    * @see BaseProgramListener#whenClone(Wrappee,Wrappee) */

   public final void whenClone ( Wrappee cloned, Wrappee clone ) {
      String old_ac = (String)attr("cur_AC");
      try {
         Object[] acs = orderedObjects.toArray();
         for ( int i = 0; i < acs.length; i++ ) {
            AspectComponent curac = (AspectComponent)acs[i];
            if (curac.woven) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenClone ( cloned, clone );
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   public final void whenDeleted(Wrappee object) {
      String old_ac = (String)attr("cur_AC");
      try {
         Object[] acs = orderedObjects.toArray();
         for ( int i = 0; i < acs.length; i++ ) {
            AspectComponent curac = (AspectComponent)acs[i];
            if (curac.woven) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenDeleted ( object );
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the registered Aspect Components that
    * are woven.
    * 
    * @see AspectComponent#afterApplicationStarted() */

   public final void afterApplicationStarted () {
      String old_ac = (String)attr("cur_AC");
      try {
         Object[] acs = orderedObjects.toArray();
         for ( int i = 0; i < acs.length; i++ ) {
            AspectComponent curac = (AspectComponent)acs[i];
            if (curac.woven) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.afterApplicationStarted();
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the registered Aspect Components that
    * are woven.
    * 
    * @see AspectComponent#whenTopologyChanged() */

   public final void whenTopologyChanged() {
      String old_ac = (String)attr("cur_AC");
      try {
         Object[] acs = orderedObjects.toArray();
         for ( int i = 0; i < acs.length; i++ ) {
            AspectComponent curac = (AspectComponent)acs[i];
            if (curac.woven) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenTopologyChanged();
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Dispatch this event to all the registered Aspect Components that
    * are woven.
    * 
    * @see AspectComponent#whenCloseDisplay(Display) */

   public final void whenCloseDisplay(Display display) {
      String old_ac = (String)attr("cur_AC");
      try {
         Object[] acs = orderedObjects.toArray();
         for ( int i = 0; i < acs.length; i++ ) {
            AspectComponent curac = (AspectComponent)acs[i];
            if (curac.woven) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenCloseDisplay(display);
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Forward this event to the aspect component that owns the
    * wrapper.
    *
    * @param wrapper the wrapper that is going to be runned
    * @param wrappingMethod the name of the may-be runned wrapping
    * @see AspectComponent#beforeRunningWrapper(Wrapper,String)
    * @see BaseProgramListener#beforeRunningWrapper(Wrapper,String) */

   public final boolean beforeRunningWrapper ( Wrapper wrapper,
                                               String wrappingMethod ) {
      boolean ret = true;
      AspectComponent ac = wrapper.getAspectComponent();      
      
      if ( ac != null ) {
         if ( ac.woven ) {
            attrdef( "cur_AC", names.get( ac ) );
            if ( ! ac.beforeRunningWrapper( wrapper, wrappingMethod ) ) {
               ret = false;
            }
         }
      }
      return ret;
   }

   /**
    * Upcall the beforeRunningWrapper method of all the aspect
    * component that owns the wrapper.
    *
    * @param wrapper the wrapper that has just been runned
    * @param wrappingMethod the name of the runned wrapping method
    * @see AspectComponent#afterRunningWrapper(Wrapper,String)
    * @see BaseProgramListener#afterRunningWrapper(Wrapper,String) */

   public final void afterRunningWrapper ( Wrapper wrapper,
                                           String wrappingMethod ) {
      AspectComponent ac = wrapper.getAspectComponent();
      if ( ac != null ) {
         if ( ac.woven ) {
            attrdef( "cur_AC", names.get( ac ) );
            ac.afterRunningWrapper( wrapper, wrappingMethod );
         }
      }
   }

   /**
    * Upcall the beforeRunningWrapper method of all the aspect
    * component that owns the wrapper.
    *
    * @see AspectComponent#afterWrap(Wrappee,Wrapper,String[],String[][]) */

   public final void afterWrap ( Wrappee wrappee, Wrapper wrapper, 
                                 String[] wrapping_methods, 
                                 String[][] wrapped_methods ) {
      Object[] acs = orderedObjects.toArray();
      for ( int i = 0; i < acs.length; i++ ) {
         AspectComponent curac = (AspectComponent)acs[i];
         if (curac.woven) {
            attrdef( "cur_AC", names.get( curac ) );
            curac.afterWrap( wrappee, wrapper, wrapping_methods, wrapped_methods );
         }
      }
   }

   public void whenGetObjects(List objects, ClassItem cl)
   {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            if ( curac.woven ) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenGetObjects (objects,cl);
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   public void whenNameObject( Object object )
   {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            if ( curac.woven ) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenNameObject (object);
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   public void whenObjectMiss( String name )
   {
      String old_ac = (String)attr("cur_AC");
      try {
         Iterator it = orderedObjects.iterator();
         while ( it.hasNext() ) {
            AspectComponent curac = (AspectComponent)it.next();
            if ( curac.woven ) {
               attrdef( "cur_AC", names.get( curac ) );
               curac.whenObjectMiss (name);
            }
         }
      } finally {
         attrdef( "cur_AC", old_ac );
      }
   }

   /**
    * Check if a registered Aspect Component has to be woven for the
    * currently called method.
    *
    * @see AspectComponent#weave() */

   protected final void checkWeaving () {
      boolean temp_allWoven = true;

      Wrappee callee = wrappee();
      String called_method_name = method();
      
      Iterator it = orderedObjects.iterator();
      
      while ( it.hasNext() ) {
         AspectComponent curac = (AspectComponent) it.next();

         
         if ( ! curac.woven ) {

            if ( curac.firstCall < curac.startWeavingCCount ) {
               
               if ( (JacObject.getObjects().size() >= curac.startWeavingCount) && 
                   (curac.startWeavingType == null || 
                    callee.getClass() == curac.startWeavingType) &&
                   (curac.startWeavingMethod == null || 
                    curac.startWeavingMethod.equals( called_method_name )))
               {
                  curac.firstCall++;
                  if (!(curac.firstCall < curac.startWeavingCCount)) {
                     weave ( curac );
                  } else {
                     temp_allWoven = false;
                  }
               } else {
                  temp_allWoven = false;
               }
            } else {
               temp_allWoven = false;
            }
         }

         allWoven = temp_allWoven;
         

      }
   }

   /**
    * Register a new aspect component.
    *
    * <p>When a new aspect component is registered, the
    * <code>allWoven</code> flag is set to false. And the
    * <code>checkWeaving</code> method will be called for each base
    * method call until all the aspect components are woven.
    *
    * @param name the aspect component key for the AC manager
    * @param aspectComponent the registered aspect component 
    * @see AspectComponent#weave()
    * @see ACManager#checkWeaving() */

   public final boolean register ( String name, Object aspectComponent ) {
      Object o = names.get( name );
      if ( o != null ) {
         /** reject if already registered */
         return false;
      }
      registering = true;
      AspectComponent ac = (AspectComponent) aspectComponent;
      ac.init();
      //AspectComponent compAC = (AspectComponent) names.get( 
      //   "JAC_composition_aspect" );
      
      Iterator it = JacObject.getObjects().iterator();
      while( it.hasNext() ) {
         JacObject cur = (JacObject) it.next();
         String className = cur.getClass().getName();
         if ( ! ( className.equals("jac.core.NameRepository") ||
                  className.equals("jac.core.ACManager") || 
                  className.equals("jac.core.dist.Topology") ) ) {
            if ( (! className.startsWith("jac.core")) || ac.isSystemListener() ) {
               cur.toNotifyACs.add( ac );
            }
         }
      }
      boolean ret = super.register( name, ac );
      nonWovenACs.add ( ac );
      allWoven = false;
      registering = false;
      return ret;
   }

   /**
    * A resister method that can be used by a UI.
    *
    * <p>This method creates a new aspect component with the name of
    * the given class and registers it. It will be woven immediatly.
    *
    * @param name the aspect component key for the AC manager
    * @param className the aspect component class name
    * @see ACManager#register(String,Object)
    * @see AspectComponent#weave() */

   public final boolean registerWithName ( String name, String className ) {
      boolean ret = false;
      try {
         ret = register ( name, Class.forName( className ).newInstance() );
      } catch ( Exception e ) {
         e.printStackTrace();
      }
      return ret;
   }

   /**
    * Register (if not allready registered) and weaves a new Aspect
    * Component.
    *
    * @param ac the aspect component to weave
    * @see ACManager#register(String,Object)
    * @see AspectComponent#weave() */

   public final void weave ( AspectComponent ac ) {
      if ( names.get( ac ) == null ) {
         register( ac.toString(), ac );
      }
      ac.woven = true;
      nonWovenACs.remove ( ac );
      if ( nonWovenACs.size() == 0 ) {
         allWoven = true;
      }
   }   

   /**
    * Unregister a given aspect component (it is then unwoven). 
    *
    * @param name the aspect component key for the AC manager
    * @see #register(String,Object)
    * @see AspectComponent#unweave() */
   
   public final void unregister( String name ) {
      AspectComponent ac = (AspectComponent) objects.get( name );
      if ( ac == null ) {
         System.out.println( "'" + name + "' aspect component not found. ("  + names + ")" );
         return;
      }
      Iterator it = JacObject.getObjects().iterator();
      while( it.hasNext() ) {
         JacObject cur = (JacObject) it.next();
         cur.toNotifyACs.remove( ac );
      }
      ac.unweave();
      super.unregister( name );
   }

   /**
    * Dump the registered aspect components. */

   public void dump () {
      
      Log.trace("aspects", "Dumping the registered aspect components " + (allWoven?"(all woven)":"(not all woven)") + ":");
      Object[] acs = orderedObjects.toArray();
      for ( int i = 0; i < acs.length; i++ ) {
         AspectComponent curac = (AspectComponent) acs[i];
         Log.trace("aspects", " - " + names.get( curac ) + (curac.woven?" (woven) [":" [") + curac.firstCall + ":" + curac.startWeavingType + ":" + curac.startWeavingMethod + ":" + curac.startWeavingCount + ":" + curac.startWeavingCCount + "]" );
         Log.trace ("aspects", "   wrappers:            " + curac.getWrappers() );
      }
   }

   // following methods implement the CollaborationParticipant interface

    public final void pattrdef( String name, Object value ) {
       JacObject.getCollaboration().addPersistentAttribute( name, value );
    }

    public final void attrdef( String name, Object value ) {
       JacObject.getCollaboration().addTransientAttribute( name, value );
    }

   public void lattrdef( String name, Object value ) {
      JacObject.getCollaboration().addTransientLocalAttribute( name, value );
   }
    
    public final Object attr( String name ) {
       return JacObject.getCollaboration().getAttribute( name );
    }

   public final Wrappee wrappee() {
      return (Wrappee)JacObject.getCollaboration()
         . getCurrentInteraction()[0];
   }

   public final String method() {
      return (String)JacObject.getCollaboration()
         . getCurrentInteraction()[1];
   }

   public final Object[] args() {
      return (Object[])JacObject.getCollaboration()
         . getCurrentInteraction()[2];
   }

   public final Object arg( int nth ) {
      return ((Object[])JacObject.getCollaboration()
              .getCurrentInteraction()[2])[nth];
   }

   public final void setarg( int nth, Object value ) {
      ((Object[])JacObject.getCollaboration()
       . getCurrentInteraction()[2])[nth] = value;
   }

   public final void setargs( Object[] values ) {
      JacObject.getCollaboration().getCurrentInteraction()[2] = values;
   }

}
