/*
  Copyright (C) 2001 Hicham Alhasan, Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.*;
import java.util.*;

import jac.util.*;

/**
 * This class parses the <code>*.acc</code> files.
 *
 * <p>The <code>*.acc</code> files are the Aspect Component
 * Configuration files. The are used to configure the aspect
 * components when they are woven to an application. Each application
 * must define one configuration file for each aspect they want to
 * use.
 *
 * <p>For the moment, the syntax of a configuration file is very
 * simple and allow the programmer that uses an existing aspect to
 * invoke a set of public methods taking strings or string arrays as
 * parameters on the configured aspect. For instance, if the
 * authentication aspect has the following configuration interface:
 *
 * <pre>
 * public class AuthenticationAC extends AspectComponent {   
 *    public void addTrustedUser(String user, String password);
 *    public void addRestrictedMethods(String clName, String[] methodNames);
 * }
 * </pre>
 *
 * <p>Then, a possible configuration file could be:
 *
 * <pre>
 * addTrustedUser "renaud" "password_renaud"
 * addTrustedUser "laurent" "password_laurent"
 * addRestrictedMethods "myClass" { "m1" "m2" "m3" }
 * </pre>
 *
 * <p>Note that:
 * <ul>
 * <li>all parameters or parameter items must be double-quoted</li>
 * <li>you must not use comas to separate parameters</li>
 * <li>an array value is defined beetween curly brackets</li>
 * <li>arrays of arrays are allowed (but untested)</li>
 * </ul> */

public class ACParser implements Parser 
{
   static final String COMMENT     = "//";
   static final String OPEN_ARRAY  = "{";
   static final String CLOSE_ARRAY = "}";
   static final String SEPARATOR   = ",";

   static final String DELIM   = OPEN_ARRAY + CLOSE_ARRAY + SEPARATOR + "\"" + " ";

   private Vector data = new Vector();
   //private String source    = new String();

   private BufferedReader  in           = null; 
   private String          line         = new String();
   private StringTokenizer stk; 	
   private int             nline        = 1;
   private boolean parsing = false;

   private String token;	
   private Stack  stack = new Stack();

   /** The default constructor. Call the <code>parse</code> method to
       parse an acc file. */
   public ACParser() {}

   /**
    * Parses a configuration file defined as the source.
    *
    * @param source the name of the configuration file (must be a
    * valid file path)
    * @param return an array containing the result: res[0] = a vector
    * containing the methods names (in the right order), res[1] = a
    * vector containing the parameters (in the right order), res[2] =
    * a vector containing line numbers (in the right order)
    */
    
   public Vector parse(String streamName) throws IOException 
   {
      Vector methods   = new Vector();
      Vector lineNumbers = new Vector();
      boolean parsingString   = false;
      boolean neutralizeQuote = false;

      Vector attribut; 

      nline = 1;

      Log.trace("parser", "beginning parsing file "+streamName);
      
      methods.clear();
      data.clear();

      in = new BufferedReader( new InputStreamReader( is )); 

      line = in.readLine();			

      while(line != null) {
		
         stk   = new StringTokenizer(line, DELIM, true);			

         if(stk.hasMoreTokens()) {
			
            token = stk.nextToken();
			
            if(token.startsWith(COMMENT)) {
               line = in.readLine(); 
               nline++;
               continue;
            }
			
            else if(!token.equals(" ") && isDelimiter(token)) 
               throw(new Exception(streamName+":"+nline+" illegal method name at line"));
				
            else {
		               int n = token.indexOf(COMMENT);
					
               if( n != -1) {
                  Log.trace("parser", 2, "new method found: "
                            +token.substring(0, n));
                  methods.addElement(token.substring(0, n));	
                  data.addElement(new Vector());
                  lineNumbers.addElement(new Integer(nline));
                  line = in.readLine(); 
                  nline++;
                  continue;						
               }

               else 
               {
                  int end = token.indexOf(' ');
                  if (end != -1) {
                     methods.addElement(token.substring(0, end));
                  } else {
                     methods.addElement(token);
                  }
                  lineNumbers.addElement(new Integer(nline));
               }
            }


            stack.push(new Vector());
            
            while(stk.hasMoreTokens()) {
				
               token = stk.nextToken();
               //   System.out.println(token);
					
               if(token.equals(SEPARATOR)) {
                  
                  if(parsingString) {
                     String s = ((String)((Vector)stack.peek()).lastElement());
                     ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                     ((Vector)stack.peek()).addElement(s + (neutralizeQuote ? "\\" : "") + SEPARATOR);
                     neutralizeQuote = false;
                  }
                  
                  else if(stack.size() == 1)
                     throw(new Exception(streamName+":"+nline+" illegal attribute '" + SEPARATOR + "'"));
                  
                  else
                     continue;
               }

               else if(token.equals("\"")) {
                  
                  if(neutralizeQuote) {
                     String s = ((String)((Vector)stack.peek()).lastElement());
                     ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                     ((Vector)stack.peek()).addElement(s + "\"");
                     neutralizeQuote = false;
                  }

                  else {

                     if(parsingString) { 
                        String s = ((String)((Vector)stack.peek()).lastElement());
                        ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                        //System.out.println(s);
                        ((Vector)stack.peek()).addElement(new Data(s));
                     }
                  
                     else
                        ((Vector)stack.peek()).addElement(new String());
                  
                     parsingString = !parsingString;
                     continue;
                  }

               }
					
               					
               else if(token.equals(OPEN_ARRAY)) {
                  
                  if(parsingString) {
                     String s = ((String)((Vector)stack.peek()).lastElement());
                     ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                     ((Vector)stack.peek()).addElement(s + (neutralizeQuote ? "\\" : "") + OPEN_ARRAY);
                     neutralizeQuote = false;
                  }
                  else 
                     stack.push(new Vector());
               }

               else if(token.equals(CLOSE_ARRAY)) {
                 
                  if(parsingString) {
                     String s = ((String)((Vector)stack.peek()).lastElement());
                     ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                     ((Vector)stack.peek()).addElement(s + (neutralizeQuote ? "\\" : "") + CLOSE_ARRAY);
                     neutralizeQuote = false;
                  }

                  else if(stack.size() == 1)
                     throw(new Exception(streamName+":"+nline+" no matching '" + OPEN_ARRAY + "' for " + token));
                  
                  else {
                     Vector v = (Vector)stack.pop();
                     ((Vector)stack.peek()).addElement(new Data(v));
                  }
               }
					
               else {
			
                  int n;

                  if((!parsingString)) {
                     
                     if(token.startsWith(COMMENT)) 
                        break;
                     else if(((n = token.indexOf(COMMENT)) != -1)) {
                        if(!isEmptyString(token.substring(0, n)))
                           throw(new Exception(streamName+":"+nline+" "+token.substring(0, n) + " should be double quoted"));
                     }

                     else if(isEmptyString(token))
                        continue;
                     else if(token.trim().equals("null"))
                        ((Vector)stack.peek()).addElement(new Data((String)null));
                     else
                        ((Vector)stack.peek()).addElement(new Data((String)token.trim()));
                        
                        //                        throw(new Exception(streamName+":"+nline+" "+token + " should be double quoted"));
                  }
                  
                  else {
                     
                      if(token.endsWith("\\")) {
      
                        neutralizeQuote = true;
                        String s = ((String)((Vector)stack.peek()).lastElement());
                        ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                        ((Vector)stack.peek()).addElement(s + token.substring(0, token.length() - 1));

                     }

                     else {
                        
                        String s = ((String)((Vector)stack.peek()).lastElement());
                        ((Vector)stack.peek()).remove(((Vector)stack.peek()).size() - 1);
                        ((Vector)stack.peek()).addElement(s + token);
                        
                     }

                  }

               }

            }

            if(stack.size() != 1)
               throw(new Exception(streamName+":"+nline+" unclosed " + OPEN_ARRAY));
            else if(parsingString)
               throw(new Exception(streamName+":"+nline+" missing \""));
            data.addElement(stack.pop());

         }
         
         line = in.readLine(); 
         nline++;
		
      }

      Vector result = new Vector();
      for(int i = 0; i < data.size(); i++) {
         Vector params = (Vector)data.get(i);
         for( int j = 0; j < params.size(); j++ ) {
            Log.trace("parser",2,"parameter conversion: " + params.get(j) );
            params.set(j, ((Data)params.get(j)).toArray() );
         }
         result.add(new ConfigMethod((String)methods.get(i), ((Vector)data.get(i)).toArray(), ((Integer)lineNumbers.get(i)).intValue()));
      }

      return result; //new Object[] { methods, data, lineNumbers };
   }
				

   private boolean isDelimiter(String str) {
      for(int i = 0; i < DELIM.length(); i++)
         if(str.equals(DELIM.substring(i, i+1)))
            return true;
      return false;
   }  	

   private boolean isEmptyString(String str) {
      for(int i = 0; i <str.length(); i++)
         if(str.charAt(i) != ' ') return false;
      return true;
   }

   /*
   public String toString() {
		
      String str = new String();
      for(int i = 0; i < methods.size(); i++) {
         str = str + (String)methods.elementAt(i) + " : ";
         for(int j = 0; j < ((Vector)data.elementAt(i)).size(); j++) {
            Log.trace("parser", ((Vector)data.elementAt(i)).elementAt(j).toString() );
            str = str + ((Data)((Vector)data.elementAt(i)).elementAt(j)).toString() + " ";
         }
         str = str + "\n";
      }
      return str;

   }


   public static void main(String[] args) {

      try {
         ACParser mp = new ACParser("fichier.txt");
         Log.trace("parser",mp.toString());
      }

      catch(Exception e) { e.printStackTrace(); }

      }*/

}
			
			
class Data {

   private String type;
   private Vector vect = new Vector();
   private String str  = new String();

   public Data(Vector vect) {
      type     = "Vector";
      this.vect = vect;
   }

   public Data(String str) {
      type = "String";
      this.str = str;	
   }

   public String getType() { return type; }
	
   public String getStringValue() { return str; }
	
   public Vector getVectorValue() { return vect; }

   public String toString() {
      if(type.equals("String"))
         return str;
      else {
         String s = new String();
         for(int i = 0; i < vect.size(); i++)
            s = s + ((Data)vect.elementAt(i)).toString() + " ";
         return ("[" + s + "]");
      }
   }

   public Object toArray() {
      if( type.equals("Vector") ) {
         //Log.trace("parser", "TO ARRAY: " + vect + "====> " );
         Object[] array = new String[vect.size()];
         for( int i = 0; i < vect.size(); i ++ ) {
            array[i] = ((Data)vect.get(i)).toArray();
         }
         //Log.trace("parser", "TO ARRAY: " + vect + "====> " + Arrays.asList( array ) );
         return array;
      } else {
         return str;
      }
   }

}
