/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import jac.core.rtti.*;
import jac.util.*;
import java.io.*;
import jac.lib.java.util.*;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * This class is a repository for all the applications defined in the
 * JAC system.
 *
 * @see Application */

public class ApplicationRepository {

   HashMap applications = new HashMap();

   static ApplicationRepository applicationRepository;

   /**
    * Returns the application of the current thread.
    *
    * @return an application, null if none */

   public static Application getCurrentApplication() {
      String appName = (String)JacObject.getCollaboration().getAttribute("cur_App");
      if( appName == null ) return null;
      return (Application)get().applications.get(appName);
   }

   /**
    * Launches a JAC program.
    *
    * @param args the program arguments (first is the .jac file) */

   public static void launchProgram( String [] args ) {
      FileInputStream fis = null;
      Properties applicationDescriptor = new Properties();
      
      try {
         fis = new FileInputStream( args[0] );
         applicationDescriptor.load( fis );
      } catch(Exception e) {
         Log.warning( "Cannot read application descriptor "+args[0] );
         return;
      }
      String applicationName = applicationDescriptor.getProperty("applicationName");
      String launchingClass = applicationDescriptor.getProperty("launchingClass");
      String aspects = applicationDescriptor.getProperty("aspects");
      String topology = applicationDescriptor.getProperty("topology");

      if( applicationName == null ) {
         Log.warning( "Cannot read application descriptor "+args[0] );
         return;
      }
      
      Log.trace("application","launching application "+applicationName);

      // bind to the given topology if any
      if( topology != null ) {
         try {
            Class remoteContainerClass = Class.forName("jac.core.dist.RemoteContainer");
            StringTokenizer st1 = new StringTokenizer( topology );
            while ( st1.hasMoreElements() ) {
               String container = (String) st1.nextElement();
               remoteContainerClass.getMethod("bindNewContainer",new Class[] {String.class})
                  .invoke(null,new Object[] {container});
            }
            Log.trace("application","application topology successfully bound");
         } catch( Exception e ) {
            e.printStackTrace();
         }
      }

      // create the application
      Application app = new Application(applicationName,null,
                                        launchingClass,null);
      
      // create the aspects configurations
      if( aspects != null ) {
         StringTokenizer st1 = new StringTokenizer( aspects );
         while ( st1.hasMoreElements() ) {
            String ac_name = (String) st1.nextElement();
            String ac_path = (String) st1.nextElement();
            String ac_weave = (String) st1.nextElement();
            Log.trace("application","adding aspect "+ac_name+" to "+app);
            app.addAcConfiguration( 
               new ACConfiguration( app, ac_name, 
                                    ac_path, 
                                    (ac_weave.equals("true")?true:false) ) );
         }
      }

      Log.trace("application","adding application "+app);
     ApplicationRepository.get().addApplication( app );

     try {
         String[] appargs = new String[args.length - 1];
         if (appargs.length > 0) {
            System.arraycopy(args, 1, appargs, 0, args.length - 1);
         }
         Class lc = Class.forName(launchingClass);
         lc.getMethod("main",new Class[] {String[].class})
            .invoke(null,new Object[] {appargs});
      } catch(Exception e) {
         Log.error(applicationName+" launch failed");
         e.printStackTrace();
      }
   }

   /**
    * Launches and initializes the application repository (this method
    * is called by the system).
    *
    * @param args the launching arguments of JAC */

   public static void main( String[] args ) throws Throwable {
      if ( applicationRepository == null ) 
         applicationRepository = new ApplicationRepository();
      launchProgram(args);
   }

   public ApplicationRepository() {
   }

   /**
    * Gets the sole instance of the application repository within the
    * local JAC system.
    *
    * @return the application repository */

   public static ApplicationRepository get() {
      if ( applicationRepository == null ) {
         applicationRepository = new ApplicationRepository();
      }
      return applicationRepository;
   }

   /**
    * Adds an application within the repository.
    *
    * <p>When added, the application is initialized so that the needed
    * aspects are woven.
    *
    * @param app the application to add */

   public void addApplication( Application app ) {
      System.out.println("--- Launching " +app+ " ---"); 
      applications.put( app.getName(), app );
      app.init();
   }

   /**
    * Returns the applications that have been added to the repository.
    *
    * @return a hash map (application's name -> application)
    * @see #addApplication */
 
   public HashMap getApplications() {
      return applications;
   }

   /**
    * Gets an application from its name.
    *
    * @param name the application's name
    * @return the application named <code>name</code>
    * @see #getApplications() */
   
   public Application getApplication( String name ) {
      return (Application) applications.get( name );
   }

   /**
    * Extends an application with a given aspect (only if this
    * application has a configuration for this aspect and if this
    * aspect is not yet woven).
    *
    * @param applicationName the application's name
    * @param aspectName the aspect's name */

   public void extend( String applicationName, String aspectName ) {
      Log.trace("aspects", "extending application "+applicationName+
                " with "+aspectName);
      Application application = getApplication(applicationName);
      if( application == null ) {
         Log.error("No such application to extend: "+applicationName);
      }
      ACConfiguration acConf = application.getAcConfiguration( aspectName );
      if( acConf == null ) {
         Log.error("No such AC configuration: "+aspectName+
                   " found in "+applicationName);
      }
      acConf.weave();
   }

   /**
    * Un-extends an application with a given aspect (only if this
    * application has a configuration for this aspect and if this
    * aspect is woven).
    *
    * @param applicationName the application's name
    * @param aspectName the aspect's name */

   public void unextend( String applicationName, String aspectName ) {
      Log.trace("aspects", "un-extending application "+applicationName+
                " with "+aspectName);
      Application application = getApplication(applicationName);
      if( application == null ) {
         Log.error("No such application to extend: "+applicationName);
      }
      ACConfiguration acConf = application.getAcConfiguration( aspectName );
      if( acConf == null ) {
         Log.error("No such AC configuration: "+aspectName+
                   " found in "+applicationName);
      }
      acConf.unweave();
   }

}




