/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.Serializable;
import java.util.*;
import jac.util.*;
import jac.core.rtti.*;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.lang.reflect.InvocationTargetException;

/**
 * This class is the root class for the aspect components defined by a
 * Jac application programmers.
 * 
 * <p>When programming a new aspect, the programmer must extends this
 * class and define new pointcuts to link a set of base method
 * invocation to a wrapping method. If necessary, the programmer can
 * also redefine the methods of the interface
 * <code>BaseProgramListener</code> to be able to react on several
 * events happenning in the base programm such as object
 * (de)serialization, object instantiation, object naming, object
 * cloning, or wrappers execution.
 *
 * @version 0.8.1
 * @author <a href="mailto:pawlak@cnam.fr">Renaud Pawlak</a>
 *
 * @see Pointcut */

public class AspectComponent 
   implements Serializable, CollaborationParticipant, BaseProgramListener {

   /**
    * Returns true if the method is defined in the aspect component
    * class.
    *
    * @param methodName the method to test
    * @return true if aspect component method */

   public static boolean defines( String methodName ) {
      if ( ClassRepository.getDirectMethodAccess( AspectComponent.class, methodName )[0] != null )
         return true;
      return false;
   }

   /**
    * Tell if the aspect component is already woven to the base
    * program.<br>
    *
    * @see #weave() */

   public boolean woven = false;

   /** 
    * Memorize how many calls have been performed on
    * <code>start_weaving_type.start_weaving_method</code>. The
    * weaving will start when this number reaches
    * <code>start_weaving_count</code>.<br>
    *
    * @see #weave() */

   public int firstCall = -1;

   /** 
    * The type where the weaving has to start.
    *
    * @see #weave() */
   
   public Class startWeavingType;

   /** 
    * The method where the weaving has to start.<br>
    *
    * @see #weave() */

   public String startWeavingMethod;

   /** 
    * The number of calls of
    * <code>startWeavingType.startWeavingMethod</code> where the
    * weaving has to start.<br>
    *
    * @see #weave() */
    
   public int startWeavingCount;

   /**
    * The number of instance of <code>startWeavingType</code> where
    * the weaving has to start.<br>
    *
    * @see #weave() */

   public int startWeavingCCount;

   /** 
    * The wrappers handled by this aspect component.<br>
    *
    * @see #getWrappers() 
    * @see #addWrapper(Wrapper) */
   
   public transient Vector wrappers = null;

   /**
    * Tells if this aspect component is a system listener one so that
    * it can receive events from the jac.core objects. */

   protected boolean systemListener = false;

   /**
    * Sets this aspect component to be a system or non-system
    * listener.
    *
    * @param systemListener if true, the component can receive events
    * from the JAC core objects (e.g. whenUsingNewInstance) ; if
    * false, these events are filtered */

   public void setSystemListener( boolean systemListener ) {
      this.systemListener = systemListener;
   }

   /**
    * Tells if this aspect component is a system listener.
    *
    * @return true if system listener
    * @see #setSystemListener(boolean) */

   public boolean isSystemListener() {
      return systemListener;
   }

   /**
    * The default constructor of the aspect component. By default,
    * there is no start weaving type or method so that the
    * <code>weave()</code> is called as the aspect component is
    * registering to the aspect component manager<br>. 
    *
    * @see #weave()
    * @see ACManager#register(String,Object) */
   
   public AspectComponent () {   
      Log.trace("aspects","New AC: "+this);
      wrappers = new Vector();
      firstCall = -1;
      startWeavingType = null;
      startWeavingMethod = null;
      startWeavingCount = 0;
      startWeavingCCount = 0;
      init();
      attrdef( "cur_AC", ACManager.get().getName( this ) );
   }

   /**
    * Initializes the aspect component so that its state is resetted
    * to the default state.<br> */
   
   public void init() {      
      if ( wrappers == null ) wrappers = new Vector();
      woven = false;
      wrappers.clear();
   }

   /**
    * Returns the wrappers handled by this aspect component. When you
    * have a reference on a wrapper, you can also know which aspect
    * component is currently handling it by using
    * <code>Wrapper.getAspectComponent()</code>.<br>
    *
    * @return a vector containing the handdled wrappers
    * @see #addWrapper(Wrapper)
    * @see Wrapper#getAspectComponent() */
   
   public Vector getWrappers() {
      if ( wrappers == null ) wrappers = new Vector();
      return wrappers;
   }
   
   /**
    * Add a wrapper to the wrappers handled by this aspect
    * component.
    * 
    * <p>The programmer should not call this method explicitly unless
    * s/he really hnows what s/he is doing.  In fact, this method is
    * automatically upcalled when the aspect component wraps a base
    * program object (see <code>Wrappee.wrap()</code>).<br>
    *
    * @param wrapper the wrapper to add to the handdled wrappers
    * @see #getWrappers()
    * @see Wrappee#wrap(Wrapper,String,String) */

   public void addWrapper( Wrapper wrapper ) {
      if ( wrappers == null ) wrappers = new Vector();
      if ( ! wrappers.contains( wrapper ) ) {
         wrappers.add( wrapper );
      }
   }

   /**
    * Force the treatment of the untreated objects by calling the
    * <code>simulateUsingNewInstance(Wrappee)</code> method.
    * 
    * <p>PRE: the aspect component has to be woven. 
    *
    * <p>To force the treatment on a give wrappee by this aspect
    * component, use <code>forceTreatment(Wrappee)</code><br>.
    * 
    * @see #forceTreatment(Wrappee)
    * @see #simulateUsingNewInstance(Wrappee)
    * @see #whenUsingNewInstance() 
    * @see #weave() */
   
   public final void forceTreatment () {
      Iterator it = JacObject.getObjects().iterator();
      while ( it.hasNext() ) {
         forceTreatment( (Wrappee) it.next() );
      }
   }


   /**
    * Force the treatment of a given wrappee by calling the
    * <code>simulateUsingNewInstance(Wrappee)</code> method.
    * 
    * <p>PRE: the aspect component has to be woven. 
    *
    * <p>To force the treatment of all the untreated wrappees by this
    * aspect component, use <code>forceTreatement(Wrappee)</code><br>.
    *
    * <p>IMPORTANT NOTE: to force the treatment of a given wrappe by
    * ALL the woven aspect components (and not by only one), the
    * programmer should use <code>JacObject.forceTreatment()</code>.<br>
    * 
    * @see #forceTreatment(Wrappee)
    * @see #simulateUsingNewInstance(Wrappee)
    * @see #whenUsingNewInstance() 
    * @see #weave()
    * @see JacObject#forceTreatment() */
   
   public final void forceTreatment ( Wrappee wrappee ) {
      Log.trace("forceTreatment",
                ACManager.get().getName( this )+".simulateUsingNewInstance");
      simulateUsingNewInstance( wrappee );
   }

   /**
    *  The programmer should define this method to wrap all the
    *  currently instantiated Jac objects so that it plugs a given
    *  aspect. Since the objects might allready be wrapped by other
    *  weavers, the programmer should be careful that it does not wrap
    *  the same object with the same wrapper several times. This can
    *  be done in the weave function by using the
    *  <code>Wrappee.isExtendedBy()</code> method. When a new AC is
    *  registered the weave method is automatically called. In the
    *  jac.prop file, the time when it is called can be parametrized
    *  with the jac.startWeavingPlaces property.
    *
    *  <p>This method is only called once. For the objects that are
    *  created after the weaving, the programmer must define the
    *  <code>whenUsingNewInstance</code> method to wrap them.
    *
    *  <p>The default implementation of the weave method is the
    *  following. In most of the case, it will work if the
    *  <code>whenUsingNewInstance()</code> is correctly defined (thus,
    *  the programer do not have to redefine the weave method).
    *
    *  <ul><pre>
    *  for (int i=0; i < JacObject.objectCount(); i++) {
    *     simulateUsingNewInstance ( (Wrappee)JacObject.getObject(i) );
    *  }
    *  </pre></ul>
    *
    *  <b>IMPORTANT NOTE</b>: this method is not deprecated but the
    *  programmer should not overload it since it is much cleaner to
    *  define the <code>whenUsingNewInstance()</code> to implement the
    *  weaving.<br>
    *
    * @see #whenUsingNewInstance()
    * @see #simulateUsingNewInstance(Wrappee)
    * @see Wrappee#isExtendedBy(Class) */

   public void weave () {
      Iterator it = JacObject.getObjects().iterator();
      while ( it.hasNext() ) {
         simulateUsingNewInstance ( (Wrappee)it.next() );
      }
   }

   /**
    * This method is called when a new instance (that has not been
    * wrapped by the aspect component yet) is used by a peer
    * object.
    * 
    * <p>By default, this method check out all the pointcuts that have
    * been defined within this aspect component and accordingly wraps
    * the wrappee.
    * 
    * <p>However, for performance or flexibility reasons, the
    * programmer can also define it from scratch so that the aspects
    * can dynamically apply while the base program creates new base
    * objects. The aspect component is notified only once for each new
    * instance.
    * 
    * <p>Here is a typical implementation of this method for a
    * many-to-one wrappees-wrapper relationship:
    *
    * <ul><pre>
    * public class MyAC extends AspectComponent {
    *   public MyWrapper myWrapper = null;
    *   public void whenUsingNewInstance() {
    *     // creates the sole instance of MyWrapper
    *     if ( myWrapper == null ) myWrapper = new MyWrapper();
    *     // make sure we do not wrap an object several times 
    *     // (this is usually not useful!!)
    *     if ( wrappee().isExtendedBy( MyWrapper.class ) ) return;
    *     // Do not wrap system or aspect objects
    *     if ( wrappee().getClass().getName().startsWith( "jac.core." ) ||
    *          wrappee().getClass().getName().startsWith( "jac.aspects." ) ) 
    *       return;
    *     // wrap it...
    *     wrappee().wrapAll( myWrapper, "myWrappingMethod" );
    *   }
    * }
    * </pre></ul>
    *
    * <p>The same but with one-to-one wrappees-wrappers
    * relationship:
    *
    * <ul><pre>
    * public class MyAC extends AspectComponent {
    *   public void whenUsingNewInstance() {
    *     // make sure we do not wrap an object several times 
    *     // (this is usually not useful!!)
    *     if ( wrappee().isExtendedBy( MyWrapper.class ) ) return;
    *     // Do not wrap system or aspect objects
    *     if ( wrappee().getClass().getName().startsWith( "jac.core." ) ||
    *          wrappee().getClass().getName().startsWith( "jac.aspects." ) ) 
    *       return;
    *     // one wrapper for each new instance
    *     MyWrapper myWrapper = new MyWrapper();
    *     // wrap it...
    *     wrappee().wrapAll( myWrapper, "myWrappingMethod" );
    *   }
    * }
    * </pre></ul>
    *
    * <b>NOTE</b>: this method is not upcalled until the aspect component
    * is woven.<br>
    * 
    * @see #weave() */

   public void whenUsingNewInstance() {
      Iterator it = pointcuts.iterator();
      while ( it.hasNext() ) {
         Pointcut pointcut = (Pointcut) it.next();
         pointcut.applyTo( wrappee() );
      }

   }

   /**
    * This method simulates an instance first use so that the system
    * upcalls <code>whenUsingNewInstance()</code> with a fake
    * collaboration point (<code>method()</code> is null and
    * <code>args</code> is null too).
    *
    * @param wrappee the wrappe on which the call is simulated.
    * @see #whenUsingNewInstance() */

   protected void simulateUsingNewInstance( Wrappee wrappee ) {
      /** simulates a new interaction */
       //      Log.trace("core","simulateUsingNewInstance "+this);
      JacObject.getCollaboration().newInteraction( wrappee, null, null );
      ((JacObject)wrappee).toNotifyACs.remove(this);
      whenUsingNewInstance();
      /** end of simulated interaction */
      JacObject.getCollaboration().endOfInteraction();
   }

   /**
    * This method is called when a new instance is created.
    * 
    * <p>Since the instance is not initialized yet, the aspect
    * component must not wrap it at this stage. To wrap a new
    * instance, the aspect component must use the
    * <code>whenUsingNewInstance()</code> method. This method is not
    * upcalled when the instantiation is remotly performed (from a
    * deployer site), see the <code>whenRemoteInstantiation()</code>
    * method.<br>
    *
    * @param newInstance the newly created instance
    * @see #whenUsingNewInstance()
    * @see #whenRemoteInstantiation(Wrappee,String) */

   public void whenNewInstance( Wrappee newInstance ) {}

   /**
    * This method is upcalled when a new object is instanciated from a
    * remote site.
    *
    * <p>The name that is passed is the name of the remote
    * reference that has been used to create the object. Typically, it
    * is the name that will be used by the naming aspect to name the
    * new object.
    * 
    * @param newInstance the instance that have been created by a
    * remote host
    * @param the name of the new instance
    * @see jac.aspects.naming.NamingAC#whenRemoteInstantiation(Wrappee,String) */

   public void whenRemoteInstantiation( Wrappee newInstance,
                                        String name ) {}

   /**
    * This method unwraps all the instances that have been wrapped by
    * the aspect component.
    * 
    * <p>This is done in a generic way that is not very efficient. To
    * optimize this process, the user may overload the
    * <code>unweave()</code> method to change its default
    * behavior.<br>
    *
    * @see #unweave() */

   protected final void unwrapAll () {
      Iterator it1 = JacObject.getObjects().iterator();
      while ( it1.hasNext() ) {
         Wrappee wrappee = (Wrappee)it1.next();
         Iterator it2 = wrappers.iterator();
         while ( it2.hasNext() ) {
            Wrapper wrapper = (Wrapper)it2.next();
            Log.trace( "wrapper","Unwrapping " + wrappee + " with " + wrapper );
            if ( wrappee.isExtendedBy( wrapper.getClass() ) ) {
               wrappee.unwrap( wrapper );
            }
         }
      }
      wrappers.clear();
   }
   
   /**
    * The programmer should overload this method to unwrap all the
    * currently wrapped Jac objects so that it unplugs a given aspect.
    * 
    * <p>By default, calls the <code>unwrapAll</code> method but the
    * programmer can implement the unveaving process in a more
    * efficient way.<br>
    *
    * @see #unwrapAll() */

   public void unweave () {
      unwrapAll();
   }

   /**
    * This method is automatically called when a Jac Object is
    * cloned. */

   public void whenClone ( Wrappee cloned, Wrappee clone ) {}

   /**
    * This method is called when a JAC object is serialized and can
    * parametrize the serialization by filling the <code>finalObject</code>
    * parameter.
    *
    * <p>Note: the current collaboration contains an attribute called
    * "orgObject" and that is a reference to the object that is
    * currently serialized.
    *
    * <p><b>IMPORTANT</b>: this method is upcalled only if the
    * serialization is done with a
    * <code>JacObjectOutputStream</code>. To ensure the correct use of
    * this class, only use <code>JacObject.serialize()</code> to
    * serialize an object.
    *
    * @param finalObject the corresponding serialized structure
    * @see SerializedJacObject
    * @see JacObjectOutputStream
    * @see JacObject#serialize(Object) */

   public void whenSerialized ( SerializedJacObject finalObject ) {
   }

   /**
    * This method is called when a base object is deserialized and can
    * parametrize the deserialization by reading the
    * SerializedJacObject instance to get some extra infos on the
    * aspects.
    *
    * <p>Note: the current collaboration contains an attribute called
    * "finalObject" and that is a reference to the object that is
    * will be finally deserialized.<br>
    *
    * <p><b>IMPORTANT</b>: this method is upcalled only if the
    * deserialization is done with a
    * <code>JacObjectInputStream</code>. To ensure the correct use of
    * this class, only use <code>JacObject.deserialize()</code> to
    * deserialize an object.
    *
    * @param orgObject the corresponding serialized structure
    * @see SerializedJacObject
    * @see JacObjectInputStream
    * @see JacObject#deserialize(byte[]) */

   public void whenDeserialized ( SerializedJacObject orgObject ) {
   }

   /**
    * This method is called when a wrapper is going to be applied to a
    * wrappee.
    *
    * <p>By overloading this method, the aspect component can skip the
    * wrapper, for instance if the current collaboration shows that
    * the wrapper has allready been applied.<br>
    *
    * <p>If the method returns true, then the wrapper is run, else it
    * is skipped. Default: return true.<br>
    *
    * @param wrapper the wrapper that is going to be runned
    * @param wrappingMethod the name of the may-be runned wrapping
    * method
    * @return a boolean that tells if the wrapper has to be runned
    * (true) or not (false)
    * @see Wrappee
    * @see Wrappee#wrap(Wrapper,String,String)
    * @see Wrapper
    * @see Wrapper#proceed() */

   public boolean beforeRunningWrapper ( Wrapper wrapper, 
                                         String wrappingMethod ) {
      return true;
   }
   
   /**
    * This method is called after the application of the
    * wrapper.
    *
    * <p>Typically, the aspect component can set an attribute to
    * the collaboration that indicates that the wrapper has been
    * applied (and that can be used by the beforeRunningWrapper
    * method).
    * 
    * @param wrapper the wrapper that has just been runned
    * @param wrappingMethod the name of the runned wrapping method
    * @see Wrappee
    * @see Wrappee#wrap(Wrapper,String,String)
    * @see Wrapper
    * @see Wrapper#proceed() */

   public void afterRunningWrapper ( Wrapper wrapper, 
                                     String wrappingMethod ) {}

   /**
    * This method is called after a new wrapper wraps a wrappee.
    */

   public void afterWrap ( Wrappee wrappee, Wrapper wrapper,
                           String[] wrapping_methods,
                           String[][] wrapped_methods  ) {}
   
   /**
    * This method is called when a program gets a set of object from
    * the object repository.
    *
    * <p>At this step, the aspect can change the returned list so that
    * the program will see the right set of objects (for instance a
    * security aspect may filter this set so that the user will not
    * see the objects s/he is not allowed to see or a persistance
    * aspect may force the load of a set of objects from the storage
    * and add them to the list).
    *
    * @param objects the returned list (can be modified)
    * @param cl the class that was used by the program to filter its
    * instances (can be null for all the objects)
    * @see ObjectRepository */

   public void whenGetObjects(List objects, ClassItem cl) {}

   /**
    * This method is upcalled by JAC when an object is named by the
    * naming aspect.
    *
    * <p>This method allows the current aspect to override the naming
    * policy of the naming aspect in a secure way. For instance, the
    * persistence AC could change the naming policy if the
    * automatically given name is already owned by a stored object.
    *
    * <p>By default, this method does nothing.
    *
    * <p>The final choosen name is a contextual attribute called
    * "finalName".
    *
    * @param object the named object */

   public void whenNameObject( Object object ) {}

   /**
    * This method is upcalled by JAC when an object was seeked into
    * the name repository and was not found.
    *
    * <p>The reason of this miss can be multiple. For instance, the
    * persistence aspect may have not already load the object from the
    * storage, or the distribution aspect may need to bind to a remote
    * object. Thus, this method allows the aspects to resolve the
    * object.
    *
    * <p>By default, this method does nothing.
    *
    * <p>The final choosen name is a contextual attribute called
    * "foundObject".
    *
    * @param object the named object */
   
   public void whenObjectMiss( String name ) {}

   public void whenDeleted(Wrappee object) {}

   public void afterApplicationStarted() {}

   public void whenCloseDisplay(Display display) {}

   public void whenTopologyChanged() {}

   /** Store the pointcuts. */
   Vector pointcuts = new Vector();

   /**
    * Defines and adds a new method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrappingClassName the name of the wrapper class
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param one2One true if each new wrapped instance corresponds to
    * one different wrapper
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             String wrappingClassName,
                             String wrappingMethodName,
                             boolean one2One ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              null,
                                              wrappingClassName,
                                              wrappingMethodName,
                                              null,
                                              null,
                                              ".*",
                                              one2One );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrappingClassName the name of the wrapper class
    * @param initParameters the initialization parameters of the
    * wrapper (passed to the wrappers constructor that matches the
    * parameters types)
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param one2One true if each new wrapped instance corresponds to
    * one different wrapper
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             String wrappingClassName,
                             Object[] initParameters,
                             String wrappingMethodName,
                             boolean one2One ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              null,
                                              wrappingClassName,
                                              wrappingMethodName,
                                              null,
                                              initParameters,
                                              ".*",
                                              one2One );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new localized method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrappingClassName the name of the wrapper class
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param hostExpr a regular expression that macthes the hosts
    * where the pointcut has to be applied (if null or empty string,
    * default is ".*" which means that the pointcut will be applied on
    * all the hosts)
    * @param one2One true if each new wrapped instance corresponds to
    * one different wrapper
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             String wrappingClassName,
                             String wrappingMethodName,
                             String hostExpr,
                             boolean one2One ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              null,
                                              wrappingClassName,
                                              wrappingMethodName,
                                              null,
                                              null,
                                              hostExpr,
                                              one2One );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new localized method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrappingClassName the name of the wrapper class
    * @param initParameters the initialization parameters of the
    * wrapper (passed to the wrappers constructor that matches the
    * parameters types)
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param hostExpr a regular expression that macthes the hosts
    * where the pointcut has to be applied (if null or empty string,
    * default is ".*" which means that the pointcut will be applied on
    * all the hosts)
    * @param one2One true if each new wrapped instance corresponds to
    * one different wrapper
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             String wrappingClassName,
                             Object[] initParameters,
                             String wrappingMethodName,
                             String hostExpr,
                             boolean one2One ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              null,
                                              wrappingClassName,
                                              wrappingMethodName,
                                              null,
                                              initParameters,
                                              hostExpr,
                                              one2One );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrapper the wrapper that contains the wrapping method,
    * cannot be null
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             Wrapper wrapper,
                             String wrappingMethodName ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              wrapper,
                                              wrapper.getClass().getName(),
                                              wrappingMethodName,
                                              null,
                                              null,
                                              ".*",
                                              false );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new localized method pointcut.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param wrapper the wrapper that contains the wrapping method,
    * cannot be null
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param hostExpr a regular expression that macthes the hosts
    * where the pointcut has to be applied (if null or empty string,
    * default is ".*" which means that the pointcut will be applied on
    * all the hosts)
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String wrappeeMethodExpr,
                             Wrapper wrapper,
                             String wrappingMethodName,
                             String hostExpr ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              wrappeeMethodExpr,
                                              wrapper,
                                              wrapper.getClass().getName(),
                                              wrappingMethodName,
                                              null,
                                              null,
                                              hostExpr,
                                              false );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new pointcut that upcalls an aspect method
    * when the matching object is created.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param methodName a method of the aspect component to call
    * @param methodArgs the arguments of this methods when called
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String methodName,
                             Object[] methodArgs ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              null,
                                              null,
                                              null,
                                              null,
                                              methodName,
                                              methodArgs,
                                              ".*",
                                              false );
      pointcuts.add( pc );
      return pc;
   }

   /**
    * Defines and adds a new localized pointcut that upcalls an aspect
    * method when the matching object is created.
    *
    * <p>For more details on how to use pointcuts, see the
    * <code>MethodPointcut</code> class.
    *
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param methodName a method of the aspect component to call
    * @param methodArgs the arguments of this methods when called
    * @see MethodPointcut */

   public Pointcut pointcut( String wrappeeExpr, 
                             String wrappeeClassExpr, 
                             String methodName,
                             Object[] methodArgs,
                             String hostExpr ) {

      MethodPointcut pc = new MethodPointcut( this,
                                              wrappeeExpr, 
                                              wrappeeClassExpr, 
                                              null,
                                              null,
                                              null,
                                              null,
                                              methodName,
                                              methodArgs,
                                              hostExpr,
                                              false );
      pointcuts.add( pc );
      return pc;
   }
 
   /**
    * Configures the current aspect component by parsing the given
    * configuration file.
    *
    * @param filePath the configuration file path */

   public void configure(String filePath) {

      String name = ACManager.get().getName(this);

      if (filePath == null) {
         return;
      }
      Log.trace("aspects.config", "configuring aspect " + name + " with file "+filePath); 
      Vector res;
      try {
         // acparser0 is the only instance of ACParser, if we are on the 
         // master site, it is automatically instantiated by the binding 
         // aspect, if not, the binding aspect creates a stub for   
         // acparser0@mastersite
         Parser acp = (Parser)NameRepository.get()
            .getObject( "parserimpl0" );
         
         res = acp.parse(filePath,this.getClass().getName());
         /*      } catch (FileNotFoundException e) {
         Log.warning("cannot find config file "+filePath+" for "+name+
         "("+e+"), using serialized infos.");*/
      } catch ( Exception e ) {
         e.printStackTrace();
         return;
      }
      
      for( int i = 0; i < res.size(); i++ ) {
         MethodItem method = null;
         ConfigMethod cm = (ConfigMethod)res.get(i);
         try {
            Log.trace("aspects.config", "invoking configuration method '" + 
                      cm.getMethod() + "' on " + getClass() );
            MethodItem[] methodItems = ClassRepository.get().getClass( getClass() ).
               getMethods(cm.getMethod());
            Log.trace("aspects.config",2,"matching methods : "+Arrays.asList(methodItems));
            // determine which method to call based on number of parameters
            Object[] paramValues = cm.getArgs();
            for (int j=0; j<methodItems.length; j++) {
               if (methodItems[j].getParameterTypes().length==paramValues.length) {
                  method = methodItems[j];
                  Log.trace("aspects.config",2,"selecting method "+method);
                  break;
               }
            }
            if (method==null) {
               throw new Exception("Wrong number of parameters for method "+cm.getMethod());
            }

            Class[] paramTypes = method.getParameterTypes();
            for (int j=0; j<paramTypes.length; j++) {
               if (paramTypes[j]!=String.class && paramValues[j].getClass()==String.class) {
                  String str = (String)paramValues[j];
                  Class type = paramTypes[j];
                  if (type==double.class) {
                     paramValues[j] = new Double(str);
                  } else if (type==int.class) {
                     paramValues[j] = new Integer(str);
                  } else if (type==long.class) {
                     paramValues[j] = new Long(str);
                  } else if (type==float.class) {
                     paramValues[j] = new Float(str);
                  } else if (type==boolean.class) {
                     paramValues[j] = new Boolean(str);
                  } else if (type==short.class) {
                     paramValues[j] = new Short(str);
                  } else if (type==Class.class) {
                     paramValues[j] = Class.forName(str);
                  } else if (type==ClassItem.class) {
                     paramValues[j] = ClassRepository.get().getClass(str);
                  } else if  (type==AbstractMethodItem.class) {
                     int index = str.indexOf("(");
                     if (index!=-1)
                        index = str.lastIndexOf(".",index);
                     else
                        index = str.lastIndexOf(".");
                     ClassItem classItem = ClassRepository.get().getClass(str.substring(0,index));
                     paramValues[j] = classItem.getAbstractMethod( str.substring(index+1) );
                  }
               }
            }
            method.invoke(this,paramValues);
         } catch ( Throwable e ) {
            if (e instanceof InvocationTargetException)
               e = ((InvocationTargetException)e).getTargetException();
            Log.warning(filePath+":"+ cm.getLineNumber() +
                        " - failed to invoke configuration method '" + 
                        (method!=null?method.getName():cm.getMethod()) + "' on " + 
                        getClass() + " with " + Arrays.asList(cm.getArgs())+" : "+e);
            e.printStackTrace();
         }
      }
   }

   // following methods implement the CollaborationParticipant interface

    public final void pattrdef( String name, Object value ) {
       JacObject.getCollaboration().addPersistentAttribute( name, value );
    }

    public final void attrdef( String name, Object value ) {
       JacObject.getCollaboration().addTransientAttribute( name, value );
    }

   public void lattrdef( String name, Object value ) {
      JacObject.getCollaboration().addTransientLocalAttribute( name, value );
   }

    public final Object attr( String name ) {
       return JacObject.getCollaboration().getAttribute( name );
    }

   public final Wrappee wrappee() {
      return (Wrappee)JacObject.getCollaboration()
         . getCurrentInteraction()[0];
   }
   
   public final String method() {
      return (String)JacObject.getCollaboration()
         . getCurrentInteraction()[1];
   }
   
   public final Object[] args() {
      return (Object[])JacObject.getCollaboration()
         . getCurrentInteraction()[2];
   }

   public final Object arg( int nth ) {
      return ((Object[])JacObject.getCollaboration()
              .getCurrentInteraction()[2])[nth];
   }

   public final void setarg( int nth, Object value ) {
      ((Object[])JacObject.getCollaboration()
       . getCurrentInteraction()[2])[nth] = value;
   }

   public final void setargs( Object[] values ) {
      JacObject.getCollaboration().getCurrentInteraction()[2] = values;
   }

}
