/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.Serializable;
import java.util.*;

import jac.core.rtti.*;

/**
 * This interface specify the prototypes of the methods that can be
 * notified when an event occurs within the base program.
 * 
 * <p>This can be regarded as the JAC Meta-Object Protocol (MOP).
 *
 * @version 0.8.1
 * @author Renaud Pawlak
 */

public interface BaseProgramListener {

   /**
    * This method is upcalled by JAC when a given instance is used for
    * the first time (when method call occurs on it).
    * 
    * <p>Informations about the called method can be retrieved by
    * using the <code>CollaborationParticipant</code> methods.<br>
    *
    * @see CollaborationParticipant */

   void whenUsingNewInstance();

   /**
    * This method is upcalled by JAC when a new instance is created.
    *
    * @param newInstance the newly created instance
    * @see #whenUsingNewInstance()
    * @see #whenRemoteInstantiation(Wrappee,String) */

   void whenNewInstance( Wrappee newInstance );

   /**
    * This method is upcalled by JAC when a new object is instantiated
    * from a remote site.
    *
    * <p>The name that is passed is the name of the remote
    * reference that has been used to create the object.
    * 
    * @param newInstance the instance that have been created by a
    * remote host
    * @param the name of the new instance */

   void whenRemoteInstantiation( Wrappee newInstance,
                                        String name );

   /**
    * This method is upcalled by JAC when a JAC object is cloned.
    * 
    * @param cloned the object that is being cloned
    * @param clone the new object (the clone of cloned) */

   void whenClone ( Wrappee cloned, Wrappee clone );

   /**
    * This method is upcalled when a JAC object is serialized
    * through the <code>JacObject.serialize()</code> method.
    *
    * @param finalObject the corresponding serialized structure
    * @see SerializedJacObject
    * @see JacObject#serialize(Object) */

   void whenSerialized ( SerializedJacObject finalObject );

   /**
    * This method is called when a JAC object is deserialized through
    * the <code>JacObject.deserialize()</code> method.
    *
    * @param orgObject the corresponding serialized structure
    * @see SerializedJacObject
    * @see JacObject#deserialize(byte[]) 
    */

   void whenDeserialized ( SerializedJacObject orgObject );

   /**
    * This method is upcalled by JAC when a wrapper is going to be
    * applied to a wrappee.
    *
    * @param wrapper the wrapper that is going to be runned
    * @param wrappingMethod the name of the may-be runned wrapping
    * method
    * @return a boolean that tells if the wrapper has to be runned
    * (true) or not (false)
    * @see Wrappee
    * @see Wrappee#wrap(Wrapper,String,String)
    * @see Wrapper
    * @see Wrapper#proceed() */

   boolean beforeRunningWrapper ( Wrapper wrapper, 
                                         String wrappingMethod );
   
   /**
    * This method is upcalled by JAC after the application of the
    * wrapper.
    * 
    * @param wrapper the wrapper that has just been runned
    * @param wrappingMethod the name of the runned wrapping method
    * @see Wrappee
    * @see Wrappee#wrap(Wrapper,String,String)
    * @see Wrapper
    * @see Wrapper#proceed() */

   void afterRunningWrapper ( Wrapper wrapper, 
                                     String wrappingMethod );

   /**
    * This method is upcalled after the wrappee is wrapped by the
    * wrapper.
    *
    * @see Wrapper
    * @see Wrappee
    * @see Wrappee#wrap(Wrapper,String,String) */

   void afterWrap ( Wrappee wrappee, Wrapper wrapper,
                           String[] wrapping_methods,
                           String[][] wrapped_methods  );

   /**
    * Calls whenGetObjects on all aspects.
    * @param objects list of objects already in memory
    * @param cl return only instances of this class
    */
   void whenGetObjects(List objects, ClassItem cl);

   /**
    * Upcalled when the naming aspect names an object.
    *
    * <p>The final choosen name is a contextual attribute called
    * "finalName".
    *
    * @param object the currently named object */

   void whenNameObject( Object object );

   /**
    * Upcalled when an object is been seeked within the name
    * repository and is not found.
    *
    * <p>The finally found object is a contextual attribute called
    * "foundObject".
    *
    * @param name the name that has not been found */

   void whenObjectMiss( String name );

   /**
    * Upcalled when an object is deleted. */
   
   void whenDeleted(Wrappee objet);

   /**
    * Upcalled after a new application is started on the JAC system.  */

   void afterApplicationStarted (); 

   /**
    * Upcalled when a display is closing.
    *
    * @param display the closing display */

   void whenCloseDisplay(Display display);

   /**
    * This method is upcalled when a topological change occurs whithin
    * the distributed application. */

   void whenTopologyChanged();

}
