/*
  Copyright (C) 2001 Renaud Pawlak, Eddy Truyen.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.util.*;
import jac.util.*;

/**
 * This class represents a set of contextual informations for a given
 * thread of a running JAC system.
 * 
 * <p>For each method call on a JAC object, a new interaction is
 * started on the current collaboration. At any point of the program,
 * the programmer can get the current collaboration for the current
 * thread by using <code>JacObject.getCollaboration()</code>.
 *
 * <p>The <code>CollaborationParticipant</code> interface provides a
 * more friendly interface to access the current collaboration. In a
 * clean design, only the classes that implement this interface may
 * access the collaboration.
 *
 * @see #newInteraction(Wrappee,String,Object[])
 * @see JacObject#getCollaboration()
 * @see CollaborationParticipant
 *
 * @version 0.8.1
 * @author Renaud Pawlak
 * @author Eddy Truyen
 */

public class Collaboration implements java.io.Serializable {
   
   /** Store the persistent attributes of the interaction. */
   private HashMap persistentAttrs;

   /** Store the transient attributes along the collaboration path. */
   private Stack taStack;

   /** Store the transient local attributes along the collaboration
       path. */
   transient private Stack tlaStack;

   /** Store the method and arguments along the collaboration path. */
   transient private Stack maStack;

   /**
    * Creates an new collaboration.
    * 
    * <p>The programmer should not explicitly create a new
    * collaboration since this is automatically done by the system for
    * any new thread.
    *
    * <p> At any point of the program, the programmer can get the
    * current collaboration for the current thread by using
    * <code>JacObject.getCollaboration()</code>.
    *
    * @see JacObject#getCollaboration() */

   public Collaboration() {
      persistentAttrs = new HashMap();
      taStack = new Stack();
      taStack.push ( new HashMap() );
      tlaStack = new Stack();
      tlaStack.push ( new HashMap() );
      maStack = new Stack();
      maStack.push ( new Object[] { null, "", new Object[0], new Integer(0) } );
   }

   /**
    * Reset the collaboration.
    * 
    * <p>This method clears and inititialize the method calls
    * (interactions) and the transient local attributes
    * stacks. */

   public void resetInteractions () {
      maStack = new Stack();
      maStack.push ( new Object[] { null, "", new Object[0], new Integer(0) } );
      tlaStack = new Stack();
      tlaStack.push ( new HashMap() );
   }

   /**
    * Add a transient attribute to the collaboration (can be done by
    * any JAC object).
    * 
    * <p>NOTE: If the attribute already exist, then it is overloaded for
    * all the following collaboration. However, if the method where
    * the attribute has been added returns, then the attribute value
    * is the old one for all the set of calling objects.<br>
    *
    * @param name the name of the attribute
    * @param att the value of the attribute
    * @return the just added attribute */

   public Object addTransientAttribute(String name, Object att) {
      if ( name != null ) {
         Map map = (HashMap)taStack.peek();
         /*
         if (map.containsKey(name)) {
            Log.trace("attributes",
                      "addTransientAttribute "+name+" = "+map.get(name)+" -> "+att);
         } else {
            Log.trace("attributes",
                      "addTransientAttribute "+name+" = "+att);
         }
         */
         map.put(name, att);
      }
      return att;
   }

   /**
    * Add a transient local attribute to the collaboration (can be
    * done by any JAC object).
    *
    * <p>A transient local attribute is the same as a transient
    * attribute, except that it is not serialized (not transmitted
    * when a remote call is performed).
    * 
    * <p>NOTE: If the attribute already exist, then it is overloaded for
    * all the following collaboration. However, if the method where
    * the attribute has been added returns, then the attribute value
    * is the old one for all the set of calling objects.<br>
    *
    * @param name the name of the attribute
    * @param att the value of the attribute
    * @return the just added attribute */

   public Object addTransientLocalAttribute(String name, Object att) {
      if ( name != null ) {
         ((HashMap)tlaStack.peek()).put(name, att);
      }
      return att;
   }

   /**
    * Add a persistent attribute to the collaboration (can be done by
    * any Jac object).
    * 
    * <p>NOTE: a persitent attribute is visible for all the objects of a
    * collaboration.
    *
    * @param name the name of the attribute
    * @param att the value of the attribute */

   public void addPersistentAttribute(String name, Object att) {
      if ( name != null ) {
         //         Log.trace("core","addPersistentAttribute "+name+" -> "+att);
         persistentAttrs.put(name, att);
      }
   }

   /**
    * Return the value of an attribute for the current collaboration.
    * 
    * <p>This method first seeks into the persistent attributes. If
    * none matches, it seeks into the transient attributes. If still
    * not found, finally seeks into the transient local attributes. If
    * all the lookups failed, return null.
    *
    * @param name the name of the attribute
    * @return the value of the attribute */

   public Object getAttribute(String name) {
      //      Log.trace("core","getAttribute("+name+")");
      Object ret = null;
      ret = persistentAttrs.get(name);
      if ( ret == null ) {
         //         Log.trace("core",name+" not in persistentAttrs");
         ret = ((HashMap)taStack.peek()).get(name);
      }
      if ( ret == null ) {
         //         Log.trace("core",name+" not in taStack");
         ret = ((HashMap)tlaStack.peek()).get(name);
      }
      /*
      if ( ret == null ) {
         Log.trace("core",name+" not in tlaStack");
      }
      */
      return ret;
   }

   /**
    * This method is called by the JAC system when an interaction
    * starts within the collaboration, that is to say when a method is
    * called on a wrappee.
    *
    * <p>This method pushes current method call informations on the
    * collaboration stack so that all the objects within this
    * interaction (the collaboration participants) can access them.
    *
    * <p>Note: a new interaction occurs only when a JAC object is called
    * (not when a regular object is called).
    *
    * @param wrappee the wrappee that handles the method that is
    * called
    * @param method_name the method that initiates the interaction
    * @param args the arguments of this method for the interaction
    *
    * @see #endOfInteraction()
    * @see CollaborationParticipant */

   public void newInteraction ( Wrappee wrappee, String method_name, 
                                Object[] args ) {
      //      Log.trace("interaction","newInteraction("+wrappee+","+method_name+")");
      //if ( (Hashtable)taStack.peek() == null ) {
      //   taStack.push ( new Hashtable() );
      //}
      try {
	  //rank = 0;
         //         System.out.println ( "***** BOI ******" );
         //         Log.trace("interaction","before : "+taStack);
         taStack.push ( ((HashMap)taStack.peek()).clone() );
         //         Log.trace("interaction","after : "+taStack);
         tlaStack.push ( ((HashMap)tlaStack.peek()).clone() );
         maStack.push ( new Object[] { wrappee, method_name, args, new Integer(0) } );
      } catch (Exception e) {
         System.out.println (e);
      }
   }
   
   /**
    * Return the current interaction (information on the current
    * method call).
    *
    * <p>It is a three objects array that represents the called
    * object, the method name and the arguments.
    *
    * <ul><pre>
    * wrappee = getCurrentInteraction()[0]
    * method_name = getCurrentInteraction()[1]
    * method_args = getCurrentInteraction()[2]
    * </pre></ul>
    *
    * @return an array containing the interaction infos
    *
    * @see #newInteraction(Wrappee,String,Object[])
    * @see #endOfInteraction() */

   public Object[] getCurrentInteraction () {
      return (Object[]) maStack.peek();
   }

   /**
    * Return a given interaction (information on a given point of the
    * method calling flow).
    *
    * <p>It is a three objects array that represents the called
    * object, the method name and the arguments at this interaction
    * step.
    *
    * <p>Index 0 represents the current interaction, index -1 the
    * calling interaction, and so on.
    *
    * <p>For instance, to get the informations about the calling
    * interaction:
    *
    * <ul><pre>
    * wrappee = getInteraction(-1)[0]
    * method_name = getInteraction(-1)[1]
    * method_args = getInteraction()[2]
    * </pre></ul>
    *
    * @return an array containing the interaction infos
    *
    * @see #newInteraction(Wrappee,String,Object[])
    * @see #endOfInteraction() */

   public Object[] getInteraction (int index) {
      return (Object[]) maStack.get( maStack.size() + index );
   }   

   /**
    * This method is called by the JAC system when an interaction
    * (method call) ends within the collaboration. 
    *
    * <p>Precisely, it pops the interaction informations and the
    * transient attributes so that they are not visible any more.
    *
    * @see #newInteraction(Wrappee,String,Object[]) */

   public void endOfInteraction () {
      //           System.out.println ( "***** EOI ******" );
      maStack.pop();
      taStack.pop();
      tlaStack.pop();
   }

   /**
    * Return the wrapping rank within a given interaction (method
    * call).
    *
    * <p>When the called method is wrapped by a wrapping chain, then
    * this rank correspond to the current wrapping method index within
    * the wrapping chain. If the wrapping chain is empty, then this
    * index stays to -1.
    *
    * @return the index of the currently runned wrapper
    * @see Wrapper
    * @see #setRank(int) */

    public int getRank() {
       return ((Integer) ((Object[])maStack.peek())[3]).intValue();
    }

   /**
    * Change the index of the currently executed wrapper.
    *
    * <p><b>Do not call unless you really know what you are doing.</b>
    *
    * @param rank the new rank
    * @see Wrapper
    * @see #getRank() */

   public void setRank( int rank ) {
      ((Object[])maStack.peek())[3] = new Integer( rank );
   }
   
   /**
    * Returns a textual representation of the collaboration. */

   public String toString() {
      return "Collaboration: \n" + 
         "transient attributes = " + taStack.toString() + "\n" +
         "transient local attributes = " + tlaStack.toString() + "\n" +
         "persistent attributes = " + persistentAttrs.toString();
   }
   
   
}






