/*

  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

/**
 * The classes that implement this interface are objects that can
 * participate to a collaboration.
 * 
 * <p>This includes the ability to get the current wrappee, method and
 * arguments of the call that is currently proceeded, and the ability
 * to define and retrieve attributes from the current collaboration
 * flow.
 *
 * <p>The classes that implement this interface use the
 * <code>Collaboration</code> class.
 * 
 * @see Collaboration
 *
 * @version 0.8.1
 * @author Renaud Pawlak
 */

public interface CollaborationParticipant {

   /**
    * Add a persistent attribute to the current collaboration.
    *
    * <p>A persistent attribute is an attribute that is visible from
    * all the objects of the local JAC container and that propagates
    * to remote containers when remote objects are called on this
    * container. If the pattrdef method is called twice with the same
    * name and different values, the changes are made visible to all
    * the objects of the local container.
    *
    * @param name the name of the persistent attribute.
    * @param value its value (must be serializable), null undefines
    * the attribute
    * @see Collaboration#addPersistentAttribute(String,Object) */

   void pattrdef( String name, Object value );

   /**
    * Add a transient attribute to the current collaboration.
    * 
    * <p>A transient attribute is visible from the object that defines
    * it and from all the subsquenlty called objects on local and
    * remote containers. When the method that defined the transient
    * attribute returns, the attribute is destroyed from the context
    * so that it is not defined anymore. If an object redefines an
    * attribute by calling again the attrdef method with the same
    * name, then the value is changed for all the subsequently called
    * objects but remains the same for all the previous callers.
    * 
    * @param name the name of the transient attribute
    * @param value its value (must be serializable), null undefines
    * the attribute
    * @see Collaboration#addTransientAttribute(String,Object) */

   void attrdef( String name, Object value );

   /**
    * Add a transient local attribute to the current collaboration.
    * 
    * <p>A transient local attribute is the same as a transient
    * attribute except that it is not transmitted when a remote call
    * is performed.
    *
    * <p>Typically, the programmer should only use this kind of
    * attribute when memorizing a JAC object reference into the
    * context. However, this manipulation should be avoided since it
    * does not ensure compatibility with the distribution aspect. For
    * this reason, prefer <code>attrdef</code>.
    *
    * @param name the name of the transient local attribute.
    * @param value its value
    * @see Collaboration#addTransientLocalAttribute(String,Object) */

   void lattrdef( String name, Object value );

   /**
    * Get an attribute value for the current collaboration. This
    * attribute can be persistent or transient.
    *
    * @param name the name of the collaboration attribute.
    * @return the value of the attribute
    * @see Collaboration#getAttribute(String) */
    
   Object attr( String name );

   /**
    * Returns the wrappee of the current call, that is to say the base
    * program object that have been called during the current
    * collaboration point (method call).
    * 
    * @return the currently called wrappee
    * @see Collaboration#newInteraction(Wrappee,String,Object[]) */

   Wrappee wrappee();
   
   /**
    * Returns the method name that have been called on the wrappee
    * during the current collaboration point (method call).
    *
    * @return the currently called method
    * @see Collaboration#newInteraction(Wrappee,String,Object[]) */

   String method();
   
   /**
    * Returns the args that have been passed to the method (see
    * <code>method()</code>).
    *
    * @return the array arguments of the currently called method
    * @see Collaboration#newInteraction(Wrappee,String,Object[]) */

   Object[] args();

   /**
    * Returns the nth argument of the current collaboration point
    * method (see <code>args()</code>).
    * 
    * @param nth the zero-indexed argument index
    * @return the nth argument of the currently called method
    * @see Collaboration#newInteraction(Wrappee,String,Object[]) */

   Object arg( int nth );

   /**
    * Sets the nth argument value.
    *
    * @param nth the zero-indexed argument index
    * @param value the new value
    */

   void setarg( int nth, Object value );

   /**
    * Sets the argument values.
    *
    * @param values the new values
    */

   void setargs( Object[] values );

}









