/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import jac.util.Log;
import jac.util.Semaphore;
import jac.core.rtti.RttiAC;
import jac.core.rtti.ClassRepository;
import jac.core.rtti.ClassItem;
import jac.core.rtti.FieldItem;
import jac.core.rtti.CollectionItem;

import java.util.Collection;
import java.util.Vector;
import java.util.Iterator;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;

public class GarbageCollector {

   static {
      /*
      new Thread() {
            public void run() {
               while(true) {
                  try {
                     // GC is running every 30s
                     sleep(10000);
                  } catch(Exception e) {
                     e.printStackTrace();
                  }
                  GarbageCollector.run();
               }
            }
         }.start();
      */
   }

   static Semaphore semaphore = new Semaphore(1);
   static boolean frozen = false;

   public static void freeze() {
      semaphore.acquire();
      frozen=true;
      semaphore.release();
   } 

   public static void unfreeze() {
      frozen=false;
   } 

   /**
    * This method can be called to force the GC. */

   public static void forceGC() {
      run();
   }

   /**
    * Runs the garbage collection process.
    *
    * <p>The used algorithm is a thread-safe variant of the
    * stop-and-copy algorithm. All the non-reachable object, except
    * those created during the GC are marked to be deleted. They will
    * be removed by the Java GC on its next run. */
   
   public static void run() {

      semaphore.acquire();

      try {

         if(frozen)
            return;

         Vector rootObjects = RttiAC.getRoots();
         List jacObjects = JacObject.getObjects();
         Vector treatedApplications = new Vector();
         HashSet reachableObjects = new HashSet();
         
         int startIndex = JacObject.objectCount();
         
         Log.trace("gc","jac objects: "+jacObjects);
         Log.trace("gc","root objects: "+rootObjects);
         
         for(int i=0; i<jacObjects.size(); i++) {
            JacObject cur = (JacObject)jacObjects.get(i);
            String appName = cur.getApplication();
            if( appName == null ) {
               // it is a system object... do not garbage
               Log.trace("gc","add (null app) "+cur);
               reachableObjects.add(cur);
               continue;
            }
            if( !treatedApplications.contains(appName) ) {
               // the first instantiated object of an application
               // is a root by default
               Log.trace("gc","root for "+appName+" : "+cur);
               treatedApplications.add(appName);
               reachableObjects.add(cur);
               reachableObjects.addAll(getReachableObjectsFrom(cur));
               continue;
            }
            if(rootObjects.contains(NameRepository.get().getName(cur))) {
               // it is a user-defined root
               Log.trace("gc","add (root object) "+cur);
               reachableObjects.add(cur);
               reachableObjects.addAll(getReachableObjectsFrom(cur));
               continue;
            }
            // cur is not reachable by any means for the moment (candidate 
            // for garbage at this step)...
         }
         
         Log.trace("gc","accessible objects = "+reachableObjects);
         
         Vector toGarbage = new Vector();
         // garbage all the objects that are not in the reachable objects set
         for( int i=0; i<startIndex; i++ ) {
            JacObject cur = (JacObject)jacObjects.get(i);
            if( ! reachableObjects.contains(cur) ) {
               toGarbage.add(cur);
            }
         }
         Log.trace("gc","to garbage objects = "+toGarbage);
         Iterator it = toGarbage.iterator();
         while(it.hasNext()) {
            JacObject cur = (JacObject)it.next();
            // garbage it!
            Log.trace("gc","garbaging "+cur);
            cur.delete();
         }
      } finally {
         semaphore.release();
      }
   }

   /**
    * Returns all the objects that can be reached by any relation path
    * from the given object.
    *
    * @param root the root object
    * @return all the objects reachable from the root */

   protected static HashSet getReachableObjectsFrom(JacObject root) {
      HashSet reachables = new HashSet();
      getReachableObjectsFrom(root, reachables);
      return reachables;
   }
      
   static void getReachableObjectsFrom(JacObject o, HashSet reachables) {

      Log.trace("gc","getting reachable objects from "+o+
                ", currently reached from the root="+reachables);

      ClassRepository cr = ClassRepository.get();
      ClassItem cli = cr.getClass( o.getClass() );

      // add referenced objects
      FieldItem[] refs = cli.getReferences();
      for( int i=0; i<refs.length; i++ ) {
         if( refs[i].isReference() ) {
            Object ref = refs[i].get( o );
            if( ref != null && (ref instanceof JacObject) ) {
               if( ! reachables.contains(ref) ) {
                  Log.trace("gc",o.toString()+"."+refs[i].getName()+" -> "+ref);
                  reachables.add( ref );
                  getReachableObjectsFrom((JacObject)ref,reachables);
               }
            }
         } 
      }

      // add objects in referenced collections
      CollectionItem[] colls = cli.getCollections();
      for( int i=0; i<colls.length; i++ ) {
         Object collection = colls[i].get(o);
         if (collection!=null && collection instanceof JacObject && 
             !reachables.contains(collection)) {
            Log.trace("gc","collection "+o.toString()+"."+colls[i].getName());
            reachables.add(collection);
         }

         Collection cref = colls[i].getActualCollection( o );
         if( cref != null ) {
            Iterator it = cref.iterator();
            while( it.hasNext() ) {
               Object cur = it.next();
               if( cur != null && (cur instanceof JacObject) ) {
                  if( ! reachables.contains(cur) ) {
                     Log.trace("gc",o.toString()+"."+colls[i].getName()+" -> "+cur);
                     reachables.add( cur );
                     getReachableObjectsFrom((JacObject)cur,reachables);
                  }
               }
            }
         }
      }
      Log.trace("gc","getting reachable DONE objects from "+o+
                ", currently reached from the root="+reachables);
   }



}
