/*

  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.lang.reflect.*;
import jac.core.rtti.*;
import jac.util.Log;

/**
 * This class allows the programmer to invoke a given method in a new
 * thread.
 *
 * <p>JAC programmers should use JAC to pass some attibutes of the
 * current thread to the new thread.
 *
 * <p>Typical use:
 * <ul><pre>
 * InvokeThread.run( myObject, "myMethodName",
 *                   new Object[] { arg0, ...},
 *                   new String[] { attrName0, ... },
 *                   new Object[] { attrValue0, ... },
 *                   new String[] { lattrName0, ... },
 *                   new Object[] { lattrValue0, ... } );
 * </pre></ul> */

public class InvokeThread extends Thread {
   AbstractMethodItem method;
   Object object;
   Object[] parameters;
   Object returnValue;
   String[] attrNames;
   Object[] attrValues; 
   String[] lattrNames;
   Object[] lattrValues; 
   boolean showResult = true;
   //String applicationName;
   
   /**
    * Runs a method in a new thread and sets a display for this
    * thread.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method
    * @param attrNames the attribute names to set into the new thread
    * collaboration
    * @param attrValues the values of these attributes
    * @param lattrNames the local attribute names to set into the new
    * thread collaboration
    * @param lattrValues the values of these local attributes */


   public static InvokeThread run( Object object, AbstractMethodItem method,
                                   Object[] parameters, 
                                   String[] attrNames, Object[] attrValues,
                                   String[] lattrNames, Object[] lattrValues ) {
      InvokeThread ret = new InvokeThread( object, method, parameters );
      ret.attrNames = attrNames;
      ret.attrValues = attrValues;
      ret.lattrNames = lattrNames;
      ret.lattrValues = lattrValues;
      //ret.applicationName = (String)JacObject.getCollaboration()
      //   .getAttribute("cur_App");
      ret.start();
      return ret;
   }

   /**
    * Runs a method in a new thread and sets a display for this
    * thread. Do not show any results on the display.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method
    * @param attrNames the attribute names to set into the new thread
    * collaboration
    * @param attrValues the values of these attributes
    * @param lattrNames the local attribute names to set into the new
    * thread collaboration
    * @param lattrValues the values of these local attributes */

   public static InvokeThread quietRun( Object object, AbstractMethodItem method,
                                        Object[] parameters, 
                                        String[] attrNames, Object[] attrValues,
                                        String[] lattrNames, Object[] lattrValues ) {
      InvokeThread ret = new InvokeThread( object, method, parameters );
      ret.attrNames = attrNames;
      ret.attrValues = attrValues;
      ret.lattrNames = lattrNames;
      ret.lattrValues = lattrValues;
      ret.showResult = false;
      //ret.applicationName = (String)JacObject.getCollaboration()
      //   .getAttribute("cur_App");
      ret.start();
      return ret;
   }

   /**
    * Runs a method in a new thread.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method */

   public static InvokeThread run( Object object, AbstractMethodItem method,
                                   Object[] parameters ) {
      InvokeThread ret = new InvokeThread( object, method, parameters );
      //ret.applicationName = (String)JacObject.getCollaboration()
      //   .getAttribute("cur_App");
      ret.start();
      return ret;
   }

   /**
    * Creates a new thread that will invoke a method when started.
    *
    * <p>The programmer should use the static <code>run</code>
    * methods.
    *
    * @param object the object that contains the method to invoke
    * @param method the method to invoke when started
    * @param parameters the parameters to pass to the method */

   public InvokeThread( Object object, AbstractMethodItem method,
                        Object[] parameters ) {
      this.method = method;
      this.object = object;
      this.parameters = parameters;
   }

   /**
    * Runs the thread (and invoke the method that was given to the
    * constructor with the right display in the collaboration).
    *
    * <p>Do not call this method directly. */

   public void run() {
      /*Log.trace("application","running thread"+
                " for application "+applicationName);
      // put the application in the context
      JacObject.getCollaboration().addPersistentAttribute("cur_App",applicationName);*/
      // put the asked attributes in the context
      if ( attrNames != null ) {
         for ( int i = 0; i < attrNames.length; i++ ) {
            jac.util.Log.trace("threads",
                               "setting attribute " + attrNames[i] + " to " + attrValues[i] );
            JacObject.getCollaboration().addTransientAttribute( 
               attrNames[i], attrValues[i] );
         }
      }
      if ( lattrNames != null ) {
         for ( int i = 0; i < lattrNames.length; i++ ) {
            jac.util.Log.trace("threads",
                               "setting local attribute " + 
                               lattrNames[i] + " to " + lattrValues[i] );
            JacObject.getCollaboration().addTransientLocalAttribute( 
               lattrNames[i], lattrValues[i] );
         }
      }
      Display display = (Display) JacObject.getCollaboration()
         . getAttribute( "Gui.display" );
      try {
         jac.util.Log.trace("threads","InvokeThread " + method + "," + object + "," + 
                            java.util.Arrays.asList(parameters) );
         
         Class[] paramTypes = method.getParameterTypes();
         for (int i=0; i< parameters.length; i++) {
            if (parameters[i]==null) {
               if (paramTypes[i] == float.class) {
                  parameters[i] = new Float(0.0);
               } else if (paramTypes[i] == long.class) {
                  parameters[i] = new Long(0);
               } else if (paramTypes[i] == double.class) {
                  parameters[i] = new Double(0.0);
               } else if (paramTypes[i] == byte.class) {
                  parameters[i] = new Byte((byte)0);
               } else if (paramTypes[i] == char.class) {
                  parameters[i] = new Character(' ');
               } else if (paramTypes[i] == short.class) {
                  parameters[i] = new Short((short)0);
               } else if (paramTypes[i] == int.class) {
                  parameters[i] = new Integer(0);
               } else if (paramTypes[i] == boolean.class) {
                  parameters[i] = new Boolean(false);
               }
            }
         }

         if ( method instanceof ConstructorItem ) {
            returnValue = ((ConstructorItem)method).newInstance( parameters );
         } else {
            /*if( applicationName == null ) {
               // if none application is defined, then we use the application
               // of the invoked object.
               if( object instanceof Wrappee ) {
                  JacObject.getCollaboration()
                     .addPersistentAttribute(
                        "cur_App", ((Wrappee)object).getApplication());
                  Log.trace("application","application was null -> "+
                            ((Wrappee)object).getApplication());
               }
               }*/
            returnValue = ((MethodItem)method).invoke( object, parameters );
         }

         if ( method.getType() != void.class ) {
            if ( display != null && showResult ) {
               if (JacObject.getCollaboration()
                   .getAttribute("Gui.openView")!=null) {
                  display.openView( returnValue );
               } else {
                  display.show( returnValue );
               }
            }
         } else {
            if ( display != null && showResult ) 
               display.refresh();
         }
      } catch ( Exception e ) {
         System.out.println("invoke thread is handling exception");
         if ( e instanceof InvocationTargetException ) { 
            Throwable te = ((InvocationTargetException)e).getTargetException();
            if (te instanceof WrappedThrowableException) {
               te = ((WrappedThrowableException)te).getWrappedThrowable();
            }
            if ( display != null && showResult ) display.show( te );
            return;
         } else {
            e.printStackTrace();
            if ( display != null && showResult ) display.show( e );
            return;
         }  
      }
   }

   Object getReturnValue() {
      return returnValue;
   }
}

class NotAvailableException extends Exception {}
