/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli, Lionel Seinturier.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import javassist.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import jac.core.*;
import jac.util.*;


/**
 * This class runs a JAC program by using a Javassist loader
 * customized with a <code>JacLoader</code> user classpath.
 *
 * <p>Use the following command to run <code>MyApp.main()</code> with
 * <code>arg1, ...</code>:
 *
 * <pre> 
 * <b>% java jac.core.Jac [options] app.jac arg1, ...</b>
 *
 * where 'app.jac' is a JAC application descriptor
 *
 * where options are:
 *    -r (release): JAC displays the release number
 *    -v (verbose): the JAC class loader is set to verbose
 *    -d (debug)  : the JAC class loader is set to debugging mode
 *    -c (clean)  : JAC cleans the tempory classes directory (previously created 
 *                  with 'write' option)
 *    -w (write)  : the JAC class loader writes on disk all the translated 
 *                  classes (and use them for the next run)
 *    -V logname  : set the log to verbose mode
 *    -L file     : redirect all the logs to a file
 *    -D [name] (dist): JAC runs in distributed mode, i.e. it constructs a distributed
 *                  JAC system. If no name is given, the default is s0.
 *                  NOTE: this option is incompatible with the -S and -C options.
 *    -M [name]   : defines a master site where all the bytecode will be 
 *                  fetched (by default, classes are loaded from the local
 *                  file system).
 *    -R dir      : specify jac_root directory 
 *    -G          : tells JAC to launch the administration GUI
 *    -A name     : launches an administration GUI on a remote server called 'name'
 *                  and do not start JAC.
 *    -W          : tells JAC to start the Web GUI server.
 *    -h (help)   : display the help.
 * </pre>
 * 
 * @see JacLoader
 * @author Renaud Pawlak
 * @author Lionel Seinturier
 * @author Laurent Martelli */

public class Jac {
   /** Verbose flag. */
   private static boolean verbose = false;

   /** Debug flag (more verbose). */
   private static boolean debug = false;

   /** Generate flag. */
   private static boolean gen = false;

   /** Clean flag. JAC cleans the temporary directory of translated
       class files. */
   private static boolean clean = false;

   /** Write flag. JAC writes the translated class files into a
       temporary directory. */
   private static boolean write = false;

   /** Server flag. JAC is launched in server mode. It waits for
       remote call and instantiations. */
   private static boolean server = false;

   /** Client flag. JAC binds to the namespace and the aspect-space. */
   private static boolean client = false;

   /** Stores the client host name in client or bind mode */
   private static String clientHost = "";

   /** Stores the server host name in bind mode */
   private static String serverHost = "";

   /** Stores the master server host name */
   private static String master = "";

   /** Distributed flag. JAC bootstraps a namespace and an
       aspect-space on the topology (a set of JAC servers). Then, it
       launches a deployment program. */
   private static boolean distributed = false;

   /** Store the class repository container name. */
   private static String classRepository = "";
   
   /** Internally used flag, starts a program if true. Set to false if
       an error occurs when parsing the options. */
   private static boolean start = true;

   /** The Javassist class loader. */
   public static javassist.Loader classLoader;

   /** The JAC root directory. */
   private static String jac_root = "";

   /** The class that is launched when JAC is started. */
   private static String launchedClass = "";

   /** A flag that tell if the GUI must be started. */
   private static boolean startGUI = false;

   /** A flag that tell if the GUI must be started. */
   private static boolean startWebGUI = false;

   private static Hashtable logLevels = new Hashtable();

   private static String logFileName;

   /**
    * The entry point of the JAC system.
    *
    * <p>Creates a JAC loader that will load the classes and translate them.
    *
    * @param args the command-line options
    * @see JacLoader */

   public static void main(String[] args) throws Throwable {

      args = parseOptions(args);
             
      if(clean) {
         if(debug) System.out.println("Cleaning temporary classes directory...");
         File f = new File(getClassesTmp());
         if (f.exists()) {
            cleanDirectory(f);
         }
      } else {
         //if(debug) System.out.println("Cleaning temporary classes directory...");
         //autoCleanDirectory("");
      }

      if( distributed || server || client || startWebGUI() ) {
         // launch RMI
         //Process p = Runtime.getRuntime().exec("rmiregistry");
         //p.waitFor();//Thread.currentThread().sleep(3000);
      }

      Class deamon = null;

      /** Try to launch a Jac server on the local host */
      if ( !master.equals("") ) {
         try {
            ClassLoader cl = (ClassLoader) Class.forName ( "jac.core.dist.DistdClassLoader" )
               .newInstance ();
            Class jac = cl.loadClass ( "jac.core.Jac" );
            jac.getMethod("setJacRoot", new Class[] {String.class}).invoke(
               null,new Object[] {jac_root});

            Class log = cl.loadClass ( "jac.util.Log" );
            Set logEntries = logLevels.entrySet();
            Iterator i = logEntries.iterator();
            while(i.hasNext())
            {
               Map.Entry entry = (Map.Entry)i.next();
               log.getMethod("setLevel", new Class[] {String.class,Integer.class}).invoke(
                  null,new Object[] {entry.getKey(),entry.getValue()});
            }

            deamon = cl.loadClass ( "jac.core.dist.rmi.RMIDistd" );
            System.out.println("-r "+master+" "+serverHost);
            deamon.getConstructor ( new Class[] { String[].class } )
               .newInstance ( new Object [] { 
                  new String[] {"-r",master,serverHost} } );

            Class bootstrap = cl.loadClass ( "jac.aspects.distribution.bootstrap.DistBootstrap" );
            bootstrap.getMethod("main", new Class[] {String[].class}).invoke(
               null,new Object[] {new String[0]});



            if(start) {

               ///** start the Aspect Component manager */
               //classLoader.run ( "jac.core.ACManager", null );
               Class acm = cl.loadClass ( "jac.core.ACManager" );
               acm.getMethod("main", new Class[] {String[].class}).invoke(
                  null,new Object[] {new String[0]});
               
               /** runs the application if any */
               if( args.length > 0 ) {
                  Class appRep = cl.loadClass ( "jac.core.ApplicationRepository" );
                  appRep.getMethod("main", new Class[] {String[].class}).invoke(
                     null,new Object[] {args});
               }
               Class jacObject = cl.loadClass ( "jac.core.JacObject" );
               jacObject.getMethod("main", new Class[] {String[].class}).invoke(
                  null,new Object[] {new String[0]});
               //Semaphore waitingSemaphore = new Semaphore();
               //waitingSemaphore.acquire();
            }

         } catch ( Exception e ) {
            System.out.println ("Error: cannot launch Jac server.");
            e.printStackTrace();
            while (true) {}
         }
      } else {
         classLoader = new javassist.Loader();
         classLoader.delegateLoadingOfPackage("org.");
         JacLoader loader = new JacLoader(verbose, debug, write, clean);
         classLoader.getClassPath().addPath(loader);
         classLoader.getClassPath().addPath(getClassesTmp());
         if ( distributed || startWebGUI() ) {
            deamon = classLoader.loadClass ( "jac.core.dist.rmi.RMIDistd" );
            deamon.getConstructor ( new Class[] { String[].class } )
               .newInstance ( new Object [] { new String[] {serverHost} } );
         }
         if ( client ) {
            deamon = classLoader.loadClass ( "jac.core.dist.rmi.RMIDistd" );
            deamon.getConstructor ( new Class[] { String[].class } )
               .newInstance ( new Object [] { new String[] { clientHost } } );
         }

         Class jac = classLoader.loadClass ( "jac.core.Jac" );

         jac.getMethod( "setJacRoot", new Class[] { String.class } ).invoke(
            null, new Object[] { jac_root } );
         if ( startGUI() ) {
            jac.getMethod( "setStartGUI", new Class[] {boolean.class} ).invoke(
               null, new Object[] {new Boolean(true)} );
         }
         if ( startWebGUI() ) {
            jac.getMethod( "setStartWebGUI", new Class[] {boolean.class} ).invoke(
               null, new Object[] {new Boolean(true)} );
         }

         if(start) {

            Class log = classLoader.loadClass ( "jac.util.Log" );
            Set logEntries = logLevels.entrySet();
            Iterator i = logEntries.iterator();
            while(i.hasNext())
            {
               Map.Entry entry = (Map.Entry)i.next();
               log.getMethod("setLevel", new Class[] {String.class,Integer.class}).invoke(
                  null,new Object[] {entry.getKey(),entry.getValue()});
            }
            if( logFileName != null && deamon != null ) {
               log.getMethod("setFileName", new Class[] {String.class,String.class}).invoke(
                  null,new Object[] {logFileName,
                                     ((String)deamon.getMethod("getLocalContainerName",new Class[0])
                                     .invoke(null, new Object[0]))+" - "});
            }
            
            classLoader.loadClass( "jac.util.Repository" );
            if ( distributed || client || server ) {
               /** bootstrap the distributed JAC system */
               classLoader.run("jac.aspects.distribution.bootstrap.DistBootstrap", null);
            }
            //if ( client ) {
            //   /** bind to the distributed JAC system */
            //   classLoader.run("jac.aspects.distribution.bootstrap.BindClient", 
            //                   new String[] {clientHost});
            //}
            /** start the Aspect Component manager */
            classLoader.run ( "jac.core.ACManager", null );

            /** runs the application if any */
            if( args.length > 0 ) {
               classLoader.run ( "jac.core.ApplicationRepository", args );
            }
            classLoader.run ( "jac.core.JacObject", null );
            //Semaphore waitingSemaphore = new Semaphore();
            //waitingSemaphore.acquire();
         }
      }
   }
   
   /**
    * Parse the command line and extract the options to set the values
    * of the appropriate fields.
    *
    * @param args the command-line options */
 
   private static String[] parseOptions(String[] args) {
      int i = 0;
      String[] ret = args;
      if (args.length == 0) {
         start = false;
         displayLaunchingHelp();
         return ret;
      }
      try {
      parse:
         while((i < args.length) && args[i].startsWith("-")) {
            switch (args[i].charAt(1)) {
               case '-':
                  i++;
                  break parse;
               case 'r':
                  System.out.println( "- JAC version 0.8.1 - " +
                                      "Get new release at http://jac.aopsys.com/ -" );
                  break;
               case 'v':
                  verbose = true;
                  break;
               case 'd':
                  debug = true;
                  break;
               case 'g':
                  gen = true;
                  start = false;
                  break;
               case 'w':
                  write = true;
                  break;
               case 'c':
                  clean = true;
                  break;
               case 'D':
                  distributed = true;
                  if( (i+1>=args.length) || args[i+1].startsWith("-") ||
                      args[i+1].endsWith(".jac")) {
                     serverHost = "s0";
                  } else {
                     i++;
                     serverHost = args[i];
                  }
                  break;
               case 'M':
                  i++;
                  master = args[i];
                  break;
               case 'V':
                  i++;
                  String category = args[i];
                  int equalIndex = args[i].indexOf('=');
                  Integer level = new Integer(1);
                  if (equalIndex!=-1) {
                     category = args[i].substring(0,equalIndex);
                     level = new Integer(args[i].substring(equalIndex+1));
                  }
                  logLevels.put(category,level);
                  break;
               case 'L':
                  i++;
                  logFileName = args[i];
                  break;
               case 'h':
                  start = false;
                  displayLaunchingHelp();
                  break;
               case 'R':
                  i++;
                  Jac.setJacRoot( args[i] );
                  break;
               case 'G':
                  Jac.setStartGUI(true);
                  break;
               case 'x':
                  PrintStream stream = new PrintStream(new FileOutputStream("out.txt"));
                  System.setOut(stream);
                  System.setErr(stream);
                  break;
               case 'A':
                  i++;
                  String serverName = args[i];
                  try {
                     // Argl!! I hate reflection in Java!
                     // this is a very simple way to launch a remote GUI
                     Object rc = Class.forName("jac.core.dist.RemoteContainer").
                        getMethod("resolve",new Class[] {String.class}).
                        invoke(null,new Object[] {serverName} );
                     if (rc==null) {
                        Log.error("Could not resolve "+serverName);
                        System.exit(1);
                     }
                     Object top = Class.forName("jac.core.dist.RemoteContainer").
                        getMethod("bindTo",new Class[] {String.class}).
                        invoke(rc,new Object[] {"JAC_topology"} );
                     Class.forName("jac.core.dist.RemoteRef")
                        .getMethod("invoke",new Class[]{String.class,Object[].class})
                        .invoke(top, new Object[] { "launchGUI", new Object[] {} });
                  } catch( Exception e ) {
                     e.printStackTrace();
                  }
                  start = false;
                  break;
               case 'W':
                  Jac.setStartWebGUI(true);
                  if ( client || server ) {
                     System.out.println("Incompatible options (-W, -S, and -C are incompatible)");
                     displayLaunchingHelp();
                  }
                  break;
               default:
                  System.out.println("Wrong option: " + args[i]);
                  displayLaunchingHelp();
            }
            if (args[i].startsWith("-S")) { i++; break; }
            if (args[i].equals( "-g" ) ) { break; }
            i++;
         }
      } catch (Exception e) {
         e.printStackTrace();
         
         System.out.println("Error while arguments parsing");
         displayLaunchingHelp();
      }

      //if(i == args.length) {
         //start = false;
      //   return ret;
      //}
      if (i>0) {
         ret = new String[args.length - i];
         System.arraycopy(args, i, ret, 0, args.length - i);
      }
      return ret;
   }

   /**
    * Print the help (displayed when the -h option is set). */

   public static void displayLaunchingHelp() {
      System.err.println("Launch a JAC application with :\n"+
                         "% java jac.core.Jac [options] app.jac arg1, ...\n" + 
                         "  where 'app.jac' is a JAC application descriptor\n" + 
                         "  where options are:\n" + 
                         "    -r (release): JAC displays the release number\n" +
                         "    -v (verbose): set JAC class loader to verbose\n" +
                         "    -d (debug)  : the JAC class loader is set to debugging mode\n" +
                         "    -c (clean)  : JAC cleans the tempory classes directory (previously created\n"+
                         "                  with 'write' option)\n" +
                         "    -w (write)  : the JAC class loader writes on disk all the translated classes\n"+
                         "                  (and use them for the next run)\n" +
                         "    -V log      : JAC activates the corresponding log\n"+
                         "    -L file     : redirect all the logs to the given file\n"+
                         "    -D [name] (dist): JAC runs in distributed mode (with a listening deamon\n"+
                         "                  name. If no name is given, default is s0.\n"+
                         "    -M [name]   : defines a master site where all the bytecode will be\n"+ 
                         "                  fetched (by default, classes are loaded from the local\n"+
                         "                  file system).\n"+
                         "    -R dir      : specify jac_root directory\n"+
                         "    -G          : tells JAC to launch the administration GUI\n"+
                         "    -A <name>   : launches an administration GUI on a remote server called 'name'\n"+
                         "                  and do not start JAC.\n"+
                         "    -W          : tells JAC to launch the GUI web server\n"+
                         "    -h (help)   : display the help\n");
      System.exit( 0 );
   }

   /**
    * This method is internally used to clean the cache directory of
    * the translated classes (to use the cache, use the -w option and
    * clean it with the -c option).
    *
    * @param d the directory to clean */

   protected static void cleanDirectory(File dir) {

      if (debug) System.out.println("Deleting dir " + dir.getPath() + "...");
      
      File[] in_d = dir.listFiles();

      for (int i=0; i<in_d.length; i++) {
         if (in_d[i].isDirectory()) {
            cleanDirectory(in_d[i]);
         } else {
            if (debug) System.out.println("Deleting file " + dir.getPath() + "...");
            in_d[i].delete();
         }
      }
      dir.delete();
   }

   /* like cleanDirectory(), but only remove files 
      if the original class file is newer */
   protected static void autoCleanDirectory(String dirName) {

      
      File dirFile = new File(getClassesTmp()+dirName);
      if (!dirFile.exists())
         return;
      if (debug) System.out.println("Auto cleaning dir " + dirFile.getPath() + "...");
      File[] in_d = dirFile.listFiles();

      if (!dirName.equals("")) 
         dirName += "/";
      for (int i=0; i<in_d.length; i++) {
         if (in_d[i].isDirectory()) {
            autoCleanDirectory(dirName+in_d[i].getName());
         } else {
            File orig = new File(getJacRoot()+"classes/"+dirName+in_d[i].getName());
            if (debug) System.out.println(in_d[i]+" : "+orig.lastModified()+" / "+in_d[i].lastModified());
            if (orig.lastModified()>in_d[i].lastModified()) {
               if (debug) System.out.println("Deleting file " + in_d[i] + "...");
               in_d[i].delete();
            }
         }
      }
      dirFile.delete();
   }
   
   /**
    * Returns the JAC root directory.<p>
    *
    * The root directory is usually defined by an environment variable
    * that is set to <code>$HOME/jac</code> where $HOME is the home
    * directory of the current user.<p>
    *
    * It can also be defined at launching time with the
    * <code>-R</code> option.<p>
    *
    * @return the JAC root directory value
    * @see #setJacRoot(String) */

   public static String getJacRoot() {
      if (jac_root==null) {
         System.err.println("No JAC_ROOT. Cannot continue.");
         System.exit(-1);
      }
      return jac_root;
   }

   /**
    * Returns the name of the directory where translated classes are
    * stored */

   public static String getClassesTmp() {
      return getJacRoot()+"classes_tmp/";
   }

   /**
    * Sets the JAC root directory.<p>
    *
    * @param dir the path of the JAC root directory
    * @see #getJacRoot()
    */

   public static void setJacRoot(String dir) {
      //      System.out.println("setJacRoot("+dir+")");
      if (! dir.endsWith("/")) 
         dir += "/";
      jac_root = dir;
   }

   /**
    * Gets the name of the class that has been launched when JAC was
    * started (in an autonomous or a distributed mode).<p>
    *
    * The launched class is the class that contains the static
    * <code>run</code> method.<p>
    *
    * @return the launched class name
    * @see #setLaunchedClass(String) */

   public static String getLaunchedClass() {
      return launchedClass;
   }

   /**
    * Sets the launched class.<p>
    *
    * @param className the launched class name
    * @see #getLaunchedClass() */

   public static void setLaunchedClass( String className ) {
      launchedClass = className;
   }
   
   /**
    * Returns the logical name of the program that was launched when
    * the JAC system started.<p>
    *
    * The logical program name is by convention the package name of
    * the class that runs the launched program.<p>
    *
    * For instance, if the command line is:<p>
    *
    * <ul><pre>jac -c -D jac.samples.agenda.Run</pre></ul><p>
    *
    * Then the launched program is <code>jac.samples.agenda</code><p>
    *
    * @return the launched program name
    * @see #getLaunchedClass() */

   public static String getLaunchedProgram() {
      String lc = getLaunchedClass();
      if ( lc == null || lc.equals( "" ) ) {
         return "";
      }
      String lp = lc;
      int ndx = 0;
      if ( ( ndx = lp.indexOf( ' ' ) ) != -1 ) {
         lp = lc.substring( 0, ndx );
      }
      if ( ( ndx = lp.lastIndexOf( '.' ) ) != -1 ) {
         lp = lc.substring( 0, ndx );
      }

      return lp;
   }

   /**
    * Tells if the JAC system must try to launch the GUI of the
    * program is exist (see the <code>-G</code> option).<p>
    *
    * If the program does not provides a GUI aspect, then the default
    * GUI is launched.<p>
    *
    * @return true if the GUI must be launched
    * @see #setStartGUI(boolean) */

   public static boolean startGUI() {
      return startGUI;
   }

   /**
    * Tells if the JAC system must try to launch the Web GUI of the
    * program is exist (see the <code>-G</code> option).<p>
    *
    * If the program does not provides a Web GUI aspect, then the default
    * GUI is launched.<p>
    *
    * @return true if the Web GUI must be launched
    * @see #setStartWebGUI(boolean) */

   public static boolean startWebGUI() {
      return startWebGUI;
   }

   /**
    * Tells the JAC system to start the GUI.<p>
    *
    * @see #startGUI() */

   public static void setStartGUI(boolean start) {
      startGUI = start;
   }

   /**
    * Tells the JAC system to start the GUI.<p>
    *
    * @see #startWebGUI() */

   public static void setStartWebGUI(boolean start) {
      startWebGUI = start;
   }

}

