/*
  Copyright (C) 2001 Renaud Pawlak, Lionel Seinturier.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import javassist.*;
import java.io.*;
import java.lang.reflect.Modifier;
import java.util.Hashtable;

import jac.util.*;

/**
 * A Javassist class loader path for JAC objects.
 *
 * <p>This loader path delegates the translation of JAC objects to
 * <code>WrappeeTranslator</code>.
 *
 * @see WrappeeTranslator
 * @author Renaud Pawlak
 * @author Lionel Seinturier
 */

public class JacLoader implements UserClassPath {

   /** 
    * Stores the bytecode of the loaded and translated classes.
    * 
    * @see isLoaded(String)
    * @see getLoadedBytecode(String) */

   private Hashtable loadedBytecodes = new Hashtable();

   /** Verbose flag. */
   private boolean verbose = false;

   /** Debug flag (more verbose). */
   private boolean debug = false;

   /** If true, caches the tranlated classes on disk. */
   private boolean write = false;

   /** If true, clears the disk cache. */
   private boolean clean = false;

   /**
    * Create a JacLoader.
    *
    * @param verbose prints out informations about the loaded classes
    * @param debug prints out more informations (also about
    * translation)
    * @param write if true, caches the tranlated classes on disk
    * @param clean if true, clears the disk cache */

   public JacLoader(boolean verbose, boolean debug, boolean write, boolean clean) {
      this.verbose = verbose;
      this.debug = debug;
      this.write = write;
      this.clean = clean;
   }

   /**
    * Returns <code>null</code>. Needed by Javassist. */

   public InputStream openClassfile(String classname) {
      return null;
   }

   /**
    * Returns true if the class is already loaded.
    * 
    * @param classname the class name to test
    * @return true if already loaded
    */

   public boolean isLoaded (String classname) {
      return loadedBytecodes.containsKey (classname);
   }

   /**
    * Returns a loaded class bytecode.
    *
    * @param classname the class name
    * @return an array of byte that correspond to the classname
    * bytecode */

   public byte[] getLoadedBytecode (String classname) {
      return (byte[])loadedBytecodes.get (classname);
   }

   /**
    * Load and translate the bytecode if needed. It does not modify
    * the class file on disk.
    *
    * <p>This method delegates the translation work to
    * <code>WrappeeTranslator.translateClass()</code>.
    *
    * @param name the full name of the class
    * @param in the input stream from which the class file is readed 
    * @return a new input stream that contains the translated bytecode
    * @see WrappeeTranslator#translateClass(CtClass,boolean,boolean,boolean,boolean)
    * */

   public InputStream filterClassfile(String name, InputStream in)
      throws IOException, CannotCompileException
   {
      if (loadedBytecodes.containsKey (name)) {
         return new ByteArrayInputStream((byte[])loadedBytecodes.get (name));
      }
      InputStream ret;
      //      if(verbose || debug) System.out.println("Creating CtClass...");
      CtClass c = new CtClass(in);
      String classname = c.getName();

      try {
         c = WrappeeTranslator.translateClass( c, verbose, debug, write, clean);
      } catch (Exception e) {
         e.printStackTrace();
         System.exit( 1 );
      }
      ret = c.getInputStream();
      if ( WrappeeTranslator.classIsAdapted(classname) ) {
         if (!loadedBytecodes.containsKey (classname)) {
            Log.trace( "loading", 1, "memoizing the translated class bytecode" );
            loadedBytecodes.put( classname, c.toBytecode() );
         }
      }
      //      if(verbose || debug) System.out.println("-- END");
      return ret;
   }
}












