/*
  Copyright (C) 2001 Renaud Pawlak, Lionel Seinturier.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import jac.core.rtti.*;
import jac.util.*;

/**
 * This class is the root class of all JAC objects.
 *
 * <p>It allows the user (the aspect component programmer) to access
 * and manipulate the objects that have been created into JAC thanks
 * to a set of useful static primitives.
 *
 * <p>This class also implements all the translation-independant
 * instance-methods of a wrappable object in the JAC system (defined by
 * the Wrappee interface). A Jac object subclasses this class at
 * load-time and is mixed with the <code>WrappeTemplate</code>
 * bytecode by calling the <code>translateClass()</code> method of the
 * <code>WrappeeTranslator</code> class.
 *
 * @version 0.8.1
 * @author <a href="mailto:pawlak@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:Lionel.Seinturier@lip6.fr">Lionel Seinturier</a>
 *
 * @see Wrappee
 * @see WrappeeTemplate 
 * @see WrappeeTranslator */

public abstract class JacObject implements Wrappee {

   /**************************************************************
    * Static members.
    **************************************************************/

   /** The JAC property file. */
   transient private static String prop_file_name = "jac.prop";

   /** Stores properties read from the JAC property file. */
   transient public static Properties props = null;

   /** Memorize the JAC object instances of the system. */
   transient public static ArrayList instances = new ArrayList();

   /** Fast access for the JAC objects. */
   transient public static Hashtable reverseInstances = new Hashtable();

   /** The number of created JAC objects. */
   transient public static int instancecount = 0;

   /** Memorize when the JAC system was launched. */
   transient static Date start_time = new Date();

   /** Bootstrapping flag. */ 
   transient public static boolean callBootstrap = false;

   /** Stores collaborations for all threads. */
   transient static Hashtable collaborations = new Hashtable();

   /** The local Aspect Component manager for this cointainer. */
   transient public static ACManager acm = null;

   /** A flag that indicates it a remote instantiation is currenlty
       performed. */
   transient public static boolean remoteInstantiation = false;

   /** Static initialization of the collaborations. */

   static {
      Thread current = Thread.currentThread();
      collaborations.put ( current, new Collaboration() );
   }

   /** Static initialization of the JAC local properties. */

   static {

      FileInputStream fis = null;
      Properties defaults = new Properties();
      
      try {
         props = new Properties();
         // Try to read a jac.prop in JAC_ROOT
         try {
            fis = new FileInputStream( Jac.getJacRoot() + prop_file_name );
            Log.trace("jac.prop","reading file "+Jac.getJacRoot() + prop_file_name);
            props.load( fis );
            Log.trace("jac.prop","Props:"+props);
         } catch(FileNotFoundException e) {
            Log.warning("No "+prop_file_name+
                        " found in JAC_ROOT ("+Jac.getJacRoot()+")");
         }

         // Try to read a jac.prop in $HOME
         try {
            String homedir = System.getProperty("user.home",null);
            if (homedir!=null) {
               fis = new FileInputStream(homedir +"/"+ prop_file_name);
               Log.trace("jac.prop","reading file "+homedir +"/"+ prop_file_name);
               defaults.putAll(props);
               props = new Properties(defaults);
               props.load( fis );
               Log.trace("jac.prop","Props:"+props);
            }
         } catch(FileNotFoundException e) {
         }
         
      // Try to read a jac.prop file from the current directory
         try {
            fis = new FileInputStream( prop_file_name );
            Log.trace("jac.prop","reading file "+"./"+ prop_file_name);
            defaults.putAll(props);
            props = new Properties(defaults);
            props.load( fis );
            Log.trace("jac.prop","Props:"+defaults);
         } catch(FileNotFoundException e) {
         }
      } catch (Exception e) {
         e.printStackTrace();
      }

      if (fis==null)
         System.out.println( "-- WARNING: no jac.prop found");
   }

   static {
      GarbageCollector.forceGC();
   }

   /**
    * Called by JAC when the program returns.
    *
    * <p>It upcalls the ACManager.afterApplicationStarted()
    * method.
    */
   public static void main(String[] args) {
      ((ACManager)ACManager.get()).afterApplicationStarted();
   }

   /**
    * Returns the date when the JAC system was launched.
    *
    * @return the date when JAC was launched */ 
   
   public static Date getStartTime() {
      return start_time;
   }

   /**
    * This method returns the nth object that has been created
    * in the Jac system.
    *
    * @param nth  the object index
    * @return     the requested object
    */

   public static Object getObject( int nth ) {
      return instances.get( nth );
   }


   /**
    * This method returns the index of anobject that has been created
    * in the Jac system.
    *
    * @param obj  the object
    * @return     the index of the object
    */

   public static int getObjectIndex( Object obj ) {
      int i = ((Integer) reverseInstances.get(obj)) . intValue();
      return i;
   }


   /**
    * This method returns the object counter of the Jac system (last
    * created object is <code>getObject(objectCount()-1)</code>).
    *
    * @return  the number of created objects
    */

   public static int objectCount() {
      return instances.size(); //instancecount;
   }
    
    
   /**
    * This method returns all the created objects in the JAC system.
    *
    * @return  all the JAC objects
    */
   
   public static List getObjects() {
      return instances;
   }


   /**
    * This method returns all the Jac objects of a given type as an
    * array.
    *
    * @param type the type to get
    * @return all the Jac objects of a given type */
   
   public static Object[] getObjects(String type) {
      Vector objects = new Vector();

      try {
         Class cl = Class.forName(type);
         
         for (int i = 0; i < JacObject.objectCount(); i++) {
            if (cl.isAssignableFrom(JacObject.getObject(i).getClass())) {
               objects.add(JacObject.getObject(i));
            }
         }
      } catch( Exception e ) { 
         e.printStackTrace(); 
      }
      return objects.toArray();
   }


   /**
    * This method returns all the JAC objects of a given type as an
    * array.
    *
    * @param type the type to get
    * @return all the JAC objects of a given type */

   public static Object[] getObjects(Class type) {
      Vector objects = new Vector();

      Iterator it = instances.iterator();
      while ( it.hasNext() ) {
         Object cur = it.next();
         if ( type.isAssignableFrom(cur.getClass()) ) {
            objects.add( cur );
         }
      }
      return objects.toArray();
   }

   /**
    * This method returns all the JAC objects that match any of the
    * given types.
    *
    * @param types the types to get
    * @return all the JAC objects that match any of the given types */
   
   public static Object[] getObjects(String[] types) {
      Vector objects = new Vector();
      for (int i = 0; i < types.length; i++) {
         try {
            objects.addAll( Arrays.asList(
               getObjects(Class.forName(types[i]))) );
         } catch( Exception e ) { e.printStackTrace(); }
      }       
      return objects.toArray();
   }


   /**
    * This method returns all the Jac objects that match any of the
    * given types.
    *
    * @param types the types to get
    * @return all the Jac objects that match any of the given types */
   
   public static Object[] getObjects(Class[] types) {
      Vector objects = new Vector();
      for (int i = 0; i < types.length; i++) {
         objects.addAll( Arrays.asList(getObjects(types[i])) );
      }       
      return objects.toArray();
   }

   /**
    * Get the collaboration for the current thread.
    *
    * @return the current collaboration
    * @see Collaboration */

   public static Collaboration getCollaboration () {
      Thread current = Thread.currentThread();
      Collaboration ret = (Collaboration) collaborations.get(current);
      if ( ret == null ) {
         collaborations.put ( current, ret = new Collaboration() );
      }
      return ret;
   }

   /**
    * Set a new collaboration for the current thread.
    *
    * @param collaboration the collaboration
    * @see Collaboration
    * @see #setCollaboration(Collaboration) */

   public static void setCollaboration ( Collaboration collaboration ) {
      Thread current = Thread.currentThread();
      collaborations.put(current, collaboration );
   }

   /**
    * Serialize an object into an array of bytes.
    * 
    * <p>If the programmer uses this method, the aspect components are
    * upcalled (<code>whenSerialized()</code>) to parametrize the
    * serialization by filling a <code>SerializedJacObject</code> instance.
    *
    * <p>A symetric process is implemented by <code>deserialize()</code>
    *
    * @param src the object to serialize
    * @return an array of bytes
    * 
    * @see AspectComponent#whenSerialized(SerializedJacObject)
    * @see SerializedJacObject
    * @see #deserialize(byte[])
    * @see JacObjectOutputStream */
    
   public static byte[] serialize( Object src ) {

      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      
      ObjectOutputStream oos = null;
      byte[] ret = null;
      
      try {
      
         oos = (ObjectOutputStream) new JacObjectOutputStream ( baos );
   
         oos.writeObject( src );
         oos.close();
      
         ret = baos.toByteArray();
         baos.close();
      
      }
      catch( Exception e ) { e.printStackTrace(); }
      
      return ret;
   }   
   
   /**
    * Deserialize an object from an array of bytes.
    *
    * <p>If the programmer uses this method, the aspect components are
    * upcalled (<code>whenDeserialized()</code> method) to parametrize the
    * deserialization by using a corresponding <code>SerializedJacObject</code>
    * instance.
    *
    * <p>A symetric process is implemented by <code>serialize()</code>
    *
    * @param data      the array of bytes
    * @return the object or null if a error has been encountered
    * 
    * @see AspectComponent#whenDeserialized(SerializedJacObject)
    * @see SerializedJacObject
    * @see #serialize(Object)
    * @see JacObjectInputStream */
    
   public static Object deserialize( byte[] data ) {
   
      ByteArrayInputStream bais = new ByteArrayInputStream(data);
      
      ObjectInputStream ois = null;
      Object ret = null;

      try {

         ois = (ObjectInputStream) new JacObjectInputStream ( bais );
         
         ret = ois.readObject();
         ois.close();
         bais.close();
      
      }
      catch( Exception e ) { e.printStackTrace(); }
      
      return ret;
   }

   /**************************************************************
    * Instance-level members.
    **************************************************************/

   /** A flag that memorize if the JAC objact was locally or remotly
       created. */
   transient private boolean remotelyCreated = false;

   /** Tells if the object was created by a cloning process. */
   transient private boolean isCloned = false;

   /** Store all the wrappers that wrap this JAC object. */
   private transient Vector wrappers;
   
   /** The corresponding wrapper classes of the <code>wrappers</code>
       field. */
   protected transient Vector wrapperClasses;

   /** The wrapping chains for the role methods. */
   protected transient Hashtable roleWrappingChains;

   /** The wrapping chains for the base methods. */
   protected transient Hashtable wrappingChains;

   /** The exceptions handlers for the base methods. */
   protected transient Hashtable catchingChains;

   /** The aspect components that have not treated this object yet. */
   protected transient Vector toNotifyACs = new Vector();

   /** The application this object's belongs to. */
   protected String applicationName;

   public String getApplication() {
      return applicationName;
   }

   /**
    * The default contructor for JAC objects.
    *
    * <p>It is called by any class that has been defined to be
    * wrappable. It adds the new instance to the system objects list
    * (see <code>getObjects</code>) and upcalls
    * <code>BaseProgramListener.whenNewInstance()</code> on the aspect
    * component manager so that this new instance will be handdle by
    * the aspect components when needed.
    *
    * @see #getObjects()
    * @see BaseProgramListener#whenNewInstance(Wrappee)
    * @see ACManager
    * @see AspectComponent */

   public JacObject() {
      if ( acm != null ) {
         // I should really use the collaboration for this stuff...
         if ( ! remoteInstantiation ) {
            String className = getClass().getName();
            if ( ! ( className.equals("jac.core.NameRepository") ||
                     className.equals("jac.core.ACManager") || 
                     className.equals("jac.core.dist.Topology") ) ) {
               applicationName = (String)getCollaboration().getAttribute("cur_App");
               acm.whenNewInstance ( this );
               Log.trace("application","new object "+this.getClass().getName()+
                         " for application "+applicationName);
            }
         } else {
            applicationName = (String)getCollaboration().getAttribute("cur_App");
            remotelyCreated = true;
         }
      }
      //boolean oldReg = false;
      //if ( acm != null ) {
      //   oldReg = acm.registering;
      //   acm.registering = true;
      //}
      
      Integer inst = new Integer(instancecount++);
      instances.add( this );
      reverseInstances.put( this, inst );

      //if ( acm != null ) {
      //   acm.registering = oldReg;
      //}
   }

   /**
    * This method deletes the current JAC object by removing it from
    * the different collections it has been inserted into by the
    * system.
    *
    * <p>The JAVA GC, will then be able to free it from memory on its
    * next run. */

   public void delete() {
      // remove from system list
      instances.remove(this);
      reverseInstances.remove(this);
      // remove from name repository
      NameRepository.get().unregisterObject(this);
      acm.whenDeleted(this);
   }

   /**
    * This method extends the default clone method to notify all the
    * active aspect components that a JacObject is beeing cloned.<p>
    * 
    * The reaction of an aspect component is then defined by the
    * <code>AspectComponent.whenClone()</code> method.
    *
    * @see AspectComponent#whenClone(Wrappee,Wrappee)
    * @see ACManager#whenClone(Wrappee,Wrappee) */

   public Object clone () {
      Object cloned = null;
      try {
         cloned = super.clone();
      } catch ( Exception e ) {
         return null;
      }
      // Re-init the aspect's state of the object
      ((JacObject)cloned).remotelyCreated = false;
      ((JacObject)cloned).isCloned = true;
      ((JacObject)cloned).wrappers = new Vector();
      ((JacObject)cloned).wrapperClasses = new Vector();
      ((JacObject)cloned).roleWrappingChains = new Hashtable();
      ((JacObject)cloned).wrappingChains = new Hashtable();
      ((JacObject)cloned).catchingChains = new Hashtable();
      ((JacObject)cloned).toNotifyACs = new Vector();
      // Upcalls the ACManager
      acm.whenClone( this, (Wrappee)cloned );
      return cloned;
   }

   public final boolean isCloned() {
      return isCloned;
   }

   /**
    * Tells if the object has been instantiated by a remote container
    * or by a local program.<p>
    *
    * @return true if remotly created */

   public final boolean isRemotlyCreated () {
      return remotelyCreated;
   }

   public final void wrap( Wrapper wrapper ) {
      if ( isWrappedBy( wrapper ) ) return;

      if ( wrappers == null ) {
         wrappers = new Vector();
         wrapperClasses = new Vector();
      }
      wrappers.add( wrapper );
      wrapperClasses.add( wrapper.getClass() );
      
      String ac = (String) getCollaboration().getAttribute( "cur_AC" );
      if ( ac != null ) {
         AspectComponent temp_ac = (AspectComponent) 
            ACManager.get().getObject( ac );
         if ( temp_ac != null ) { 
            temp_ac.addWrapper( wrapper );
            wrapper.setAspectComponent( ac );
         }
      }
   }

   static private HashMap wrappableMethods = new HashMap();
   
   public boolean isWrappable( String methodName ) {
      // We use a cache because it is too slow otherwise
      String fullmethodName = getClass().getName()+"."+methodName;
      Boolean ret = (Boolean)wrappableMethods.get(fullmethodName);
      if (ret == null) {
         try {
            getClass().getField( "_" + methodName + "_method_name" );
         } catch (Exception e) {
            Log.trace("wrappable","* isWrappable("+fullmethodName+") -> false");
            wrappableMethods.put(fullmethodName,new Boolean(false));
            return false;
         }
         Log.trace("wrappable","* isWrappable("+fullmethodName+") -> true");
         wrappableMethods.put(fullmethodName,new Boolean(true));
         return true;
      } else {
         Log.trace("wrappable","isWrappable("+fullmethodName+") -> "+ret);
         return ret.booleanValue();
      }
   }
   
   public final void wrap( Wrapper wrapper, String wrappingMethod,
                           String wrappedMethod ) {
      
      if ( isWrappable( wrappedMethod ) ) {

         if ( wrappingChains == null ) wrappingChains = new Hashtable();
         Vector wrappingChain = (Vector)wrappingChains.get( wrappedMethod );
         if ( wrappingChain == null ) {
            wrappingChains.put( wrappedMethod, wrappingChain = new Vector() );
         }
         
         int rank = 0;
         
         if ( acm != null ) {
            CompositionAspect ca = (CompositionAspect) 
               acm.objects.get( "JAC_composition_aspect" );
            if ( ca != null) {
               rank = ca.getWeaveTimeRank ( wrappingChain, wrapper, wrappingMethod );
            } else {
               if( wrappingChain.size() > 0 ) {
                  Log.warning( "no composition aspect found when wrapping "+
                               this.getClass().getName() + " with " + 
                               wrapper.getClass().getName() );
               }
            }
         }
         
         Object[] wrapping_method = new Object[2];
         wrapping_method[0] = wrapper;
         wrapping_method[1] = ClassRepository.getDirectMethodAccess(
            wrapper.getClass(),wrappingMethod)[0];

         if ( ((Method)wrapping_method[1]).getReturnType() == void.class && 
              ClassRepository.getDirectMethodAccess(
                 getClass(),wrappedMethod)[0].getReturnType() != void.class ) {
            throw new Error("Wrapping method return type mismatch "+
                            getClass().getName()+" : "+
                            wrappingMethod+","+wrappedMethod);
         }         
         wrappingChain.add( rank, wrapping_method );
         
         String ac = (String) getCollaboration().getAttribute( "cur_AC" );
         if ( ac != null ) {
            AspectComponent temp_ac = (AspectComponent) acm.objects.get( ac );
            if ( temp_ac != null ) { 
               temp_ac.addWrapper( wrapper );
               wrapper.setAspectComponent( ac );
            }
         }

         /*Class[] types = {
            Class.forName("jac.core.Wrapper") ,
            Class.forName("java.lang.String"),
            int.class
         };

         Object[] values = { wrapper, wrappingMethod, new Integer(-1) };

         getClass().getMethod("_wrap_" + wrappedMethod, types)
         . invoke(this, values);*/
        
         wrap ( wrapper );

      } else {
         
         /** Try to wrap the role methods */
         if ( roleWrappingChains == null ) {
            roleWrappingChains = new Hashtable();
         }
         Object[] roleWrappers = getWrappers();
         for ( int i = 0; i < roleWrappers.length; i++ ) {
            Method[] meths = ClassRepository.getDirectMethodAccess( 
               roleWrappers[i].getClass(), wrappedMethod );
            if ( meths != null && meths.length > 0 ) {
               Vector wrapping_chain = (Vector) roleWrappingChains.get( 
                  wrappedMethod );
               if ( wrapping_chain == null ) {
                  roleWrappingChains.put( wrappedMethod, wrapping_chain = new Vector() );
               }
               wrapping_chain.add( new Object[] { wrapper, meths[0] } );
               return;
            }
         }
         System.out.println ("Warning: cannot wrap '" + wrappedMethod + 
                             "' on " + this + " with " + wrappingMethod + ".");
      }

   }

   public final void wrap( Wrapper wrapper, String wrappingMethod,
                           String[] wrappedMethods ) {

      for (int i = 0; i < wrappedMethods.length; i++) {
         if (wrappedMethods[i] != null) {
            wrap ( wrapper, wrappingMethod, wrappedMethods[i] );
         }
      }
   }

   public final void wrapSameName( Wrapper wrapper, String wrappingMethods[] ) {
      for (int i = 0; i < wrappingMethods.length; i++) {
         if (wrappingMethods[i] != null) {
            wrap ( wrapper, wrappingMethods[i], wrappingMethods[i] );
         }
      }
   }

   public final void wrap( Wrapper wrapper, String[] wrappingMethods,
                           String[][] wrappedMethods ) {

      for ( int i = 0; i < wrappingMethods.length; i++ ) {
         if ( wrappedMethods[i] != null ) {
            wrap ( wrapper, wrappingMethods[i], wrappedMethods[i] );
         }
      }
   }

   public final void wrapAll( Wrapper wrapper, String wrapperMethod ) {

      try {
	    
         Method[] methods = getClass().getMethods();
	    
         for ( int i=0 ; i < methods.length ; i++ ) {
            if ( isWrappable( methods[i].getName() ) )
               wrap( wrapper, wrapperMethod, methods[i].getName() );
         }
         
      } catch ( Exception e ) { 
         e.printStackTrace(); 
         System.exit(1); 
      }
   }

   public final void wrapAll( Wrapper wrapper, String wrapperMethod, 
                              Class[] paramTypes ) 
   {
      Method[] methods = getClass().getMethods();
      
      for ( int i=0 ; i < methods.length ; i++ ) {
         if ( isWrappable( methods[i].getName() ) && 
              Arrays.equals(paramTypes, methods[i].getParameterTypes()))
            wrap( wrapper, wrapperMethod, methods[i].getName() );
      }
      
   }

   public final void wrapAllEx( Wrapper wrapper, String wrapperMethod, 
                                Class[] paramTypes ) 
   {
      Method[] methods = getClass().getMethods();
      
      for ( int i=0 ; i < methods.length ; i++ ) {
         if ( isWrappable( methods[i].getName() ) && 
              ! Arrays.equals(paramTypes, methods[i].getParameterTypes()))
            wrap( wrapper, wrapperMethod, methods[i].getName() );
      }
      
   }

   public final void wrapModifiers( Wrapper wrapper, String wrapperMethod ) {

      try {
         Vector methods = ClassRepository.get().getClass(this).getAllModifiers();
         for ( int i=0 ; i < methods.size() ; i++ ) {
            MethodItem method = (MethodItem)methods.get(i);
            if ( isWrappable( method.getName() ) )
               wrap( wrapper, wrapperMethod, method.getName() );
         }
      } catch ( Exception e ) { 
         e.printStackTrace(); System.exit(1); 
      }
   }

   public final void wrapAllEx( Wrapper wrapper, String wrapperMethod,
                                String[] excludedMethods ) {

      try {
	    
         Method[] methods = getClass().getMethods();
         //         Object[] values = { wrapper, wrapperMethod, new Integer(-1) };
         boolean excluded;
	    
         for ( int i=0 ; i < methods.length ; i++ ) {
            //            if ( methods[i].getName().startsWith("_wrap_") ) {
            if ( isWrappable( methods[i].getName() ) ) {
               excluded = false;
               for ( int j=0; j<excludedMethods.length; j++ ) {
                  if ( methods[i].getName().substring(6).equals(excludedMethods[j]) ) {
                     excluded = true;
                  }
               }

            if (!excluded) wrap( wrapper, wrapperMethod, methods[i].getName() );

                  //methods[i].invoke(this,values);
            }
         }

      //wrap ( wrapper );

      } catch ( Exception e ) { e.printStackTrace(); System.exit(1); }
   }

   public final void unwrap( Wrapper wrapper ) {
      if ( ! isWrappedBy( wrapper ) ) return;
      wrappers.remove( wrapper );
      wrapperClasses.remove( wrapper.getClass() );
      unwrapAll( wrapper );
   }

   public final void unwrap( Wrapper wrapper, String wrappingMethod,
                             String wrappedMethod ) {

      if ( ! isWrappable( wrappedMethod ) ) return;
      
      if ( wrappingChains == null ) return;
      Vector wrappingChain = (Vector)wrappingChains.get( wrappedMethod );
      if ( wrappingChain == null ) return;
      Iterator it = wrappingChain.iterator();
      while ( it.hasNext() ) {
         Object[] wrappingMeth = (Object[])it.next();
         if ( wrappingMeth[0] == wrapper && 
              ((Method)wrappingMeth[1]).getName().equals( wrappingMethod ) ) {
            it.remove();
         }
      }

   }

   public final void unwrapAll( Wrapper wrapper, String wrappingMethod ) {
      Method[] methods = getClass().getMethods();
         
      for ( int i=0 ; i < methods.length ; i++ ) {
         if ( isWrappable( methods[i].getName() ) )
            unwrap( wrapper, wrappingMethod, methods[i].getName() );
      }
   }

   public final void unwrapAll( Wrapper wrapper ) {
      Method[] wrapper_meths = wrapper.getClass().getMethods();

      for ( int i = 0; i < wrapper_meths.length; i++ ) {
         try {
            unwrapAll( wrapper, wrapper_meths[i].getName() );
         } catch ( Exception e ) {}
      }
   }

   public final Object[] getWrappers() {
      if ( wrappers == null ) {
         wrappers = new Vector();
         wrapperClasses = new Vector();
      }
      return wrappers.toArray();
   }

   public final Object[] getWrappers( Class wrapperType ) {
      if ( wrappers == null ) {
         wrappers = new Vector();
         wrapperClasses = new Vector();
      }
      if (wrapperType == null) {
         return wrappers.toArray();
      }
      Vector res = new Vector();
      for (int i = 0; i < wrappers.size(); i++) {
         if ( wrapperType.isAssignableFrom(wrappers.get(i).getClass()) ) {
            res.add( wrappers.get(i) );
         }
      }         
      return res.toArray();
   }

   public final boolean isWrappedBy( Wrapper wrapper ) {
      if ( wrappers == null ) {
         wrappers = new Vector();
         wrapperClasses = new Vector();
      }
      return wrappers.contains( wrapper );
   }

   public final Object[] getWrapperClasses() {
      if ( wrappers == null ) {
         wrappers = new Vector();
         wrapperClasses = new Vector();
      }
      return wrapperClasses.toArray();
   }

   public final boolean isExtendedBy( Class wrapperClass ) {
      if ( wrapperClass == null ) return false;
      if ( getWrappers( wrapperClass ).length > 0 ) {
         return true;
      } else {
         return false;
      }
   }

   public final Wrapper[] getWrappersForMethod(String methodName) {
      //Vector wrappingChain = wrappingChains.get( methodName );
      //     wrappers

      Field f1 = null;
      Vector mwrappers = null;
      try {
         f1 = getClass().getField ( "_" + methodName + "_wrappers" );
         mwrappers = (Vector) f1.get( this );
      } catch ( Exception e ) {
         e.printStackTrace();
         return null;
      }
      return null;
   }

   public final Vector getWrappingChain( String methodName ) {
      if ( wrappingChains == null ) return null;
      return (Vector) wrappingChains.get( methodName );
   }   

   public final void printWrappersForMethod(String methodName) {
      Field f1 = null;
      Field f2 = null;
      Vector mwrappers = null;
      Vector mwms = null;
      try {
         f1 = getClass().getField ( "_" + methodName + "_wrappers" );
         mwrappers = (Vector) f1.get( this );
         f2 = getClass().getField ( "_" + methodName + "_wrappingMethods" );
         mwms = (Vector) f2.get( this );
      } catch ( Exception e ) {
         e.printStackTrace();
         return;
      }
      System.out.println( "Method " + methodName + " is currently wrapped by: " );
      if ( mwrappers.size() == 0 ) {
         System.out.println( " <nothing>" );
      } else {
         for ( int i = 0; i < mwrappers.size(); i++ ) {
            System.out.println( " - " + mwrappers.get(i) + ":" + 
                                ((Method)mwms.get(i)).getName() );
         }
      }
   }

   public final Method getWrappingMethod( Wrapper wrapper, 
                                          String methodName ) {
      return null;
   }    
    
   public final Object nextWrapper() {
	
      Object ret = null;
        
      Collaboration c = getCollaboration();
      c.setRank( c.getRank() + 1 );
      Object[] interaction = c.getCurrentInteraction();

      Method m = null;
      try {
         m = getClass().getMethod(
            "_nextWrapper_" + interaction[1], 
            new Class[] { Object[].class, int.class } );
      } catch ( Exception e ) {}

      if ( m != null ) {
	  
         Object[] args = new Object[] { interaction[2], interaction[3] };

         try { ret = m.invoke( this, args ); }
         catch ( InvocationTargetException e ) {
      
            Throwable t = e.getTargetException();
            if(t instanceof WrappedThrowableException) {
               throw (RuntimeException)t;
            } else {
               throw new WrappedThrowableException ( t );
            }
         }
         catch ( IllegalAccessException e ) {
            e.printStackTrace();
            System.exit( 1 );
         }

      } else {
         System.out.println( "next role wrapper : " + (String) interaction[1] );
         ret = invokeRoleMethod( (String) interaction[1], (Object[]) interaction[2], false );
      }
	
      return ret;
   }

   public final void addExceptionHandler( Wrapper wrapper,
                                          String method,
                                          String listenedMethod ) {
      
      if ( ! isWrappable( listenedMethod ) ) {
         System.out.println ("Warning: cannot add catcher '" + method + "' on " + 
                             this + "." + listenedMethod + ".");
         return;
      }

      if ( catchingChains == null ) catchingChains = new Hashtable();
      Vector catchingChain = (Vector) catchingChains.get( listenedMethod );
      if ( catchingChain == null ) {
         catchingChains.put( listenedMethod, catchingChain = new Vector() );
      }
      
      Object[] catchingMethod = new Object[2];
      catchingMethod[0] = wrapper;
      catchingMethod[1] = ClassRepository.getDirectMethodAccess(wrapper.getClass(),method)[0];
      catchingChain.add( catchingMethod );
      
      String ac = (String) getCollaboration().getAttribute( "cur_AC" );
      if ( ac != null ) {
         AspectComponent temp_ac = (AspectComponent) ACManager.get().objects.get( ac );
         if ( temp_ac != null ) { 
            temp_ac.addWrapper( (Wrapper) wrapper );
            wrapper.setAspectComponent( ac );
         }
      }
   }

   public final void addExceptionHandler( Wrapper wrapper,
                                          String method ) {
      
      Method[] meths = getClass().getMethods();
      
      for ( int i = 0; i < meths.length; i++ ) {
         if ( isWrappable( meths[i].getName() ) ) {
            addExceptionHandler( wrapper, method, meths[i].getName() );
         }
      }
   }

   public Object[][] getExceptionHandlers( String listenedMethod ) {
      Object[][] ret = null;

      try {      
         Class[] types = {};
         Object[] values = {};
         ret = (Object[][])getClass().getMethod("_getExceptionHandler_" + listenedMethod, types)
            . invoke(this,values);
         
      } catch ( Exception e ) {
         e.printStackTrace();
         System.exit( 1 );
      }
      return ret;
   }

   public Object invokeRoleMethod( String name,
                                   Object[] parameters ) {
      return invokeRoleMethod( name, parameters, true );
   }
   
   /**
    * Internally used to process a role method invocation */

   private Object invokeRoleMethod( String name,
                                    Object[] parameters,
                                    boolean start ) {

      if ( start ) {

         /** Starts a new interaction */
         getCollaboration().newInteraction( 
            this, name, parameters );

         /** Upcall the Aspect Component manager. */
         try {
            acm.whenUsingNewInstance();
         } catch ( Exception e) {
            e.printStackTrace();
            System.exit( -1 );
         }
      }

      Object[] wrappers = getWrappers();
      Object ret = null;
      Method[] methods = null;
      boolean invoked = true;

      Vector wrapping_chain = null;
      if ( roleWrappingChains != null ) {
         wrapping_chain = (Vector) roleWrappingChains.get( name );
      }
      int rank = getCollaboration().getRank();
      if ( wrapping_chain != null && 
           rank < wrapping_chain.size() ) {
         Object[] wrappingMethod = (Object[])wrapping_chain.get( rank );
         Method m = (Method) wrappingMethod[1];
         try {
            ret = m.invoke( wrappingMethod[0], new Object[0] );
         } catch ( Exception e ) {
            e.printStackTrace();
         }
      } else {

         for (int i = 0; i < wrappers.length; i++) {
         
            methods = ClassRepository.getDirectMethodAccess( wrappers[i].getClass(), name );

            if ( methods[0] == null ) continue;
            
            for (int j = 0; j < methods.length; j++) {
               try {
                  invoked = false;
                  ret = methods[j].invoke( wrappers[i], parameters );
                  invoked = true;
               }
               catch ( IllegalArgumentException e ) {}                     
               catch ( InvocationTargetException e ) {
                  Throwable t = e.getTargetException();
                  JacObject.getCollaboration().endOfInteraction();            
                  if(t instanceof RuntimeException) {
                     throw (RuntimeException)t;
                  }
                  e.printStackTrace();
                  System.exit( 1 );         
               }
               catch ( IllegalAccessException e ) {
                  e.printStackTrace();
                  System.exit( 1 );
               }
            }
         }

         if ( ! invoked ) {
            JacObject.getCollaboration().endOfInteraction();
            throw new IllegalArgumentException(
               "No such method has been found for " + this + ": " + 
               name);
         }
      }
      if ( start ) {
         JacObject.getCollaboration().endOfInteraction(); 
      }
      return ret;
   }
   
   public final Object invoke (String name, Object[] parameters) {
	
      Object ret = null;

      Method[] methods = ClassRepository.getDirectMethodAccess( getClass(), name );
      boolean ok = false;
      for ( int i=0 ; i < methods.length ; i++) {
         try {
            ok = false;
            ret = methods[i].invoke( this, parameters );
            ok = true;
         }
         catch ( InvocationTargetException e ) {
            Throwable t = e.getTargetException();
            if(t instanceof RuntimeException) {
               throw (RuntimeException)t;
            }
            e.printStackTrace();
            System.exit( 1 );         
         }
         catch ( IllegalAccessException e ) {
            e.printStackTrace();
            System.exit( 1 );
         }
         catch ( Exception e ) {}                  
         if (ok) break;
      }
      if (!ok) {
         /** try to invoke a role method */
         try {
            ret = invokeRoleMethod( name, parameters );
         } catch ( IllegalArgumentException e ) {
            throw new IllegalArgumentException(
               "No such method has been found for " + this + ": " + name);
         }
      }
      
      return ret;
      
   }
   
   public final Object invokeOrg (String name, Object[] parameters) {
	
      Object ret = null;        

      Method[] methods = getClass().getMethods();
      boolean ok = false;
      boolean found = false;
      for ( int i=0 ; i < methods.length ; i++) {
         if ( methods[i].getName().equals("_org_" + name) && 
              methods[i].getParameterTypes().length == parameters.length) {
            found = true;
            try {
               ok = false;
               ret = methods[i].invoke( this, parameters );
               ok = true;
            }
            catch ( IllegalArgumentException e ) {}                     
            catch ( InvocationTargetException e ) {
               Throwable t = e.getTargetException();
               if(t instanceof RuntimeException) {
                  throw (RuntimeException)t;
               }
               e.printStackTrace();
               System.exit( 1 );         
            }
            catch ( IllegalAccessException e ) {
               e.printStackTrace();
               System.exit( 1 );
            }
            if (ok) break;
         }
      }
      // if there was no "_org_" prefixed method, try to call the method directly
      if (!found) {
         for ( int i=0 ; i < methods.length ; i++) {
            if ( methods[i].getName().equals(name) && 
                 methods[i].getParameterTypes().length == parameters.length) 
            {
               try {
                  ok = false;
                  ret = methods[i].invoke( this, parameters );
                  ok = true;
               }
               catch ( IllegalArgumentException e ) {}                     
               catch ( InvocationTargetException e ) {
                  Throwable t = e.getTargetException();
                  if(t instanceof RuntimeException) {
                     throw (RuntimeException)t;
                  }
                  e.printStackTrace();
                  System.exit( 1 );         
               }
               catch ( IllegalAccessException e ) {
                  e.printStackTrace();
                  System.exit( 1 );
               }
               if (ok) break;
            }
         }
      }
      if (!ok) {
         throw new IllegalArgumentException(
            "No such original method has been found: " +getClass().getName()+"."+name);
      }
      
      return ret;   
   }

   public void forceTreatment() {
      if (acm != null ) {
         acm.forceTreatment( this );
      }
   }

   /**************************************************************
    * State introspection primitives.
    **************************************************************/

   public final Field[] getFields() {

      Hashtable fields = ClassRepository.getDirectFieldAccess( getClass() );
      Object[] res = fields.values().toArray();
      Field[] ares = new Field[res.length];
      System.arraycopy ( res, 0, ares, 0, ares.length );
      return ares;

   }         

   public final Field getField( String name ) {

      Hashtable fields = ClassRepository.getDirectFieldAccess( getClass() );
      return (Field) fields.get( name );

   }

   public final Object getFieldValue ( String fieldName ) {      
      Hashtable ht = ClassRepository.getDirectFieldAccess( getClass() );
      Field field = (Field) ht.get( fieldName );
      if ( field == null ) return null;
      Object ret = null;
      try {
         ret = field.get( this );
      } catch ( Exception e ) {}
      return ret;
   }

   public final Object[] getState () {

      Hashtable fields = ClassRepository.getDirectFieldAccess( getClass() );
      Object[] tmp = fields.keySet().toArray();
      String[] names = new String[tmp.length];
      System.arraycopy( tmp, 0, names, 0, tmp.length );
      Object[] values = new Object[tmp.length];
      for ( int i = 0; i < names.length; i ++ ) {
         Field f = (Field) fields.get ( names[i] );
         try {
            values[i] = f.get( this );
         } catch ( Exception e ) {}
      }

      return new Object[] { names, values };
   }

   public final Object[] getState ( String[] fieldNames ) {

      Hashtable tmpfields = ClassRepository.getDirectFieldAccess( getClass() );
      Hashtable fields = new Hashtable();
      for ( int i = 0; i < fieldNames.length; i ++ ) {
         Field f = (Field) tmpfields.get( fieldNames[i] );
         if ( f != null ) {
            fields.put( fieldNames[i], f );
         }
      }
      Object[] tmp = fields.keySet().toArray();
      String[] names = new String[tmp.length];
      System.arraycopy( tmp, 0, names, 0, tmp.length );
      Object[] values = new Object[tmp.length];
      for ( int i = 0; i < names.length; i ++ ) {
         Field f = (Field) fields.get ( names[i] );
         try {
            values[i] = f.get( this );
         } catch ( Exception e ) {}
      }

      return new Object[] { names, values };
   }

   public final void setState ( Object[] state ) {

      Hashtable fields = ClassRepository.getDirectFieldAccess( getClass() );
      String[] names = (String[])state[0];
      Object[] values = (Object[])state[1];
      for ( int i = 0; i < names.length; i ++ ) {
         Field f = (Field) fields.get ( names[i] );
         try {
            f.set( this, values[i] );
         } catch ( Exception e ) {}
      }
   }

   public final void setFieldValue (String fieldName, Object value) {
      Hashtable ht = ClassRepository.getDirectFieldAccess( getClass() );
      Field field = (Field) ht.get( fieldName );
      if ( field == null ) return;
      try {
         field.set( this, value );
      } catch ( Exception e ) {}
   }
   
   /**
    * Default to string method for JAC objects.
    *
    * @return the name of the object as registered into the name
    * repository */

   public String toString() {
      Repository nr = NameRepository.get();
      String name = null;
      if( nr != null ) {
         name = nr.getName(this);
      }
      if( name == null ) {
         name = "<anonymous> "+super.toString();
      }
      return name;
   }

   /**
    * This method factorizes the common code that is used when the
    * next wrapper is called.
    *
    * @param wrappee the current wrappe
    * @param methodName the wrapped method that is called
    * @param params the parameters of the invocation
    * @param _rank the rank within the wrapping chain */

   public Object myNextWrapper( Wrappee _wrappee, String methodName, 
                                Object[] params, Integer _rank ) 
      throws Throwable {
      Object ret = null;
          
      JacObject wrappee = (JacObject)_wrappee;
      Method m = null;
      Object to_invoke = null;
      boolean direct_call = false;
      boolean start_rank = false;
      int rank = _rank.intValue();
      String oldApplicationName=null;

      if ( rank == 0) {
         start_rank = true;
      }

      try {

         //         Log.trace("jac", "calling " + methodName + " : " + rank + ":" + wrappee.getClass() );

         if ( start_rank ) {
            //            Log.trace("jac", "new interaction for " + getCollaboration() );
            getCollaboration().newInteraction(
               wrappee, methodName, params );
            // here, the attribute is persistent for optimization purpose
            // we save it to restore it at the end
            oldApplicationName = (String)getCollaboration().getAttribute("cur_App");
            getCollaboration().addPersistentAttribute("cur_App",applicationName);
            if ( acm != null ) {
               //               Log.trace("jac", "calling WUNI" );
               acm.whenUsingNewInstance ();
            }
         }

         String old_ac = (String)JacObject.getCollaboration().getAttribute("cur_AC");
         try {

            //            Log.trace("jac", 2, "calling, step 0" );

            Vector wrappingChain = null;
            if ( wrappee.wrappingChains != null ) {
               wrappingChain = (Vector)wrappee.wrappingChains.get( methodName );
            }

            //            Log.trace("jac", 2, "calling, step 1" );

            if (! start_rank ) {
               if ( acm != null ) {
                  acm.afterRunningWrapper( 
                     (Wrapper)((Object[])wrappingChain.get(rank - 1))[0],
                     ((Method)((Object[])wrappingChain.get(rank - 1))[1]).getName() );
               }
            }

            //            Log.trace("jac", 2, "calling, step 2" );

            if ( wrappingChain != null && (acm != null) && (! direct_call) ) {
               while ( rank < wrappingChain.size() && 
                       ! acm.beforeRunningWrapper( 
                          (Wrapper)((Object[])wrappingChain.get(rank))[0], 
                          ((Method)((Object[])wrappingChain.get(rank))[1]).getName() ) ) {
                  acm.afterRunningWrapper(
                     (Wrapper)((Object[])wrappingChain.get(rank))[0],
                     ((Method)((Object[])wrappingChain.get(rank))[1]).getName() );
                  //                  Log.trace("jac", "SKIPPING RANK " + rank );
                  rank++;
               }
            }

            //Log.trace("jac", 2, "calling, step 3" );

            if ( wrappingChain != null && rank < wrappingChain.size() && !direct_call ) {

               to_invoke = ((Object[])wrappingChain.get(rank))[0];
               m = (Method)((Object[])wrappingChain.get(rank))[1];
               JacObject.getCollaboration().addTransientAttribute(
                  "cur_AC", ((Wrapper)to_invoke).getAspectComponentName() );
               /*if( wrappee.getApplication()!=null && wrappee.getApplication().equals("bank") ) {
                  Log.trace("aspect-introspection", ((Wrapper)to_invoke).getAspectComponentName()
                            +": "+to_invoke.getClass()+"."+m.getName());
                            }*/
               //Log.trace("cur_AC","invoking next wrapper "+
               //          to_invoke.getClass().getName()+
               //          "("+methodName+")"+
               //         " with cur_AC = "+
               //          ((Wrapper)to_invoke).getAspectComponentName()+
               //          " on "+wrappee.getClass().getName());
               ret = m.invoke( to_invoke , new Object[0] );            
            } else {

               //Log.trace("jac", 2, "calling org " + methodName );
      
               JacObject.getCollaboration().addTransientAttribute(
                  "cur_AC", null );
               //Log.trace("cur_AC2","invoking base method "+
               //          methodName+" with cur_AC = null");
               try {
                  Method m2 = ClassRepository.getDirectMethodAccess( 
                     wrappee.getClass(),"_org_" + methodName )[0];
//                  System.out.println("thread "+Thread.currentThread().getName());
                  //                  System.out.println("Invoke "+(Object)m2+" = "+m2+"("+m2.isAccessible()+")"+" on "+wrappee+" with "+Arrays.asList(params));
                  ret = m2.invoke ( wrappee, params );
                  //                  System.out.println("org method _org_"+methodName+" called successfully");
               } catch ( IllegalArgumentException ill) {
                  try {
                     ret = ClassRepository.invokeDirect( wrappee.getClass(), 
                                                         "_org_" + methodName,
                                                         wrappee, params );
                  } catch ( NoSuchMethodException e ) {
                     e.printStackTrace();
                     System.exit(-1);
                  }
               } 
            }
   
            //Log.trace("jac", "end of calling org " + methodName );

         } finally {
            //Log.trace("cur_AC","restoring cur_AC = "+old_ac);
            JacObject.getCollaboration().addTransientAttribute(
               "cur_AC", old_ac );
         }


         if ( start_rank ) {
            getCollaboration().endOfInteraction();
            // we restore the calling application
            getCollaboration().addPersistentAttribute("cur_App",oldApplicationName);            
         }

      }
      catch ( InvocationTargetException e ) {
         Throwable t = e.getTargetException();
         
         Log.trace("jac", wrappee + "." + methodName + " catching " + t );

         if(t instanceof WrappedThrowableException) {
            t = ((WrappedThrowableException)t).getWrappedThrowable();
         } 

         if (start_rank) {
            JacObject.getCollaboration().endOfInteraction();            
         }            

         boolean caught = false;

         Vector catchingChain = null;
         if ( wrappee.catchingChains != null ) {
            catchingChain = (Vector)wrappee.catchingChains.get( methodName );
         } else {
            //Log.trace("jac", "Catching chains are null for " + wrappee );
         }
         
         if ( catchingChain != null ) { 
            Iterator it = catchingChain.iterator();
            Log.trace("jac", "Trying to find catcher..." );            
            while ( it.hasNext() ) {
               try {
                  Object[] wm = (Object[])it.next();
                  Log.trace("jac", "trying "+wm[1]+" on "+wm[0]+" ("+t+")"+" ->"+t.getClass() );
                  ((Method)wm[1]).invoke( wm[0], new Object[] { t } );
                  Log.trace("jac", "caught." );
                  caught = true;
                  // several handlers on the same exception can cause trouble!!!
                  // e.g.: InputWrapper.inputCanceled + TransactionWrapper.rollback!
                  // => handlers should be correctly ordered!
                  break;
               }
               catch ( Exception e1 ) {
                  if ( ( e1 instanceof InvocationTargetException ) ) {
                     throw new WrappedThrowableException ( 
                        ((InvocationTargetException)e1).getTargetException() );
                  }
               }
            }
         } else {
            //Log.trace("jac", 2, "Catching chain is null for " + wrappee );
         }

         if (!caught) {
            Log.trace("jac", "forwarding exception" );
            throw new WrappedThrowableException ( t );
         } else {
            Log.trace("jac", "silent exception" );
         }
      }
      catch ( IllegalAccessException e ) {
         e.printStackTrace();
         System.exit( 1 );
      }

      return ret;
   }

}

