/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.Serializable;

import jac.util.*;
import jac.core.rtti.*;
import java.lang.reflect.*;
import gnu.regexp.*;
import java.util.*;

/**
 * This class can be used by JAC aspect components to easily define a
 * set of method points on the base program that the aspects will use to
 * modify its behavior.
 *
 * <p>A method pointcut is defined through four pointcut
 * expressions. For the moment, these pointcut expressions are a
 * simple extension of regular expressions -- in an EMACS_LISP syntax
 * (see the GNU-regexp tutorial) that can be combined with the
 * <code>&&</code> operator (in this case all the regexps must match)
 * or the <code>||</code> operator (in this case, only one regexp must
 * match). Before a regexp, you can use the <code>!</code> (not)
 * operator to inverse the matching.
 *
 * <p>Depending on the pointcut expression, you can also
 * use keywords that will be dynamically interpreted.  We intend to
 * provide a more complete and suited desciption language. Note that
 * for the moment, the tricky cases that cannot be handled by these
 * expressions can be treated by programming the
 * <code>AspectComponent.whenUsingNewInstance</code> method.
 *
 * <ul><li><i>The host expression</i>: filters the components that
 * belong or not to the pointcut on a location basis. To be activated,
 * the name of the host where the pointcut is currently applied must
 * match this expression.</li>
 *
 * <li><i>The wrappee expression</i>: filters the components that
 * belong or not to the pointcut on a per-component basis. This
 * expression assumes that the component is named (that the naming
 * aspect is woven). If you need to modify all the components of the
 * same class equaly, then this expression should be <code>".*"</code>
 * and the <i>wrappee-class expression</i> should be used.</li>
 *
 * <li><i>The wrappee-class expression</i>: filters the components
 * that belong or not to the pointcut on a per-class basis. For
 * instance, <code>"A || B"</code> says that only the classes that are
 * named A or B belong to the pointcut; <code>packagename.* &&
 * classname</code> matches the class that belong to
 * <code>packagename</code> and that are named
 * <code>classname</code>.</li>
 *
 * <li><i>The wrappee-method expression</i>: for all the components
 * that match the two previous expressions, defines which methods must
 * be modified by the pointcut. For instance,
 * <code>"set.*(int):void"</code> matches all the methods that have
 * name that begins with "set", that take only one integer argument,
 * and that return nothing. The regular expression can contain several
 * builtin keywords to automatically match methods with special
 * semantics (<code>ALL</code>: all the methods of the class,
 * <code>MODIFIERS</code>: all the state modifiers,
 * <code>ACCESSORS</code>: all the state setters,
 * <code>GETTER(fielname)</code>: the getter for the given field,
 * <code>SETTER(fieldname)</code>: the setter for the given
 * field). For instance, <code>SETTERS && !.*(int).*</code> matches
 * all the setters, excluding the ones for the integer
 * fields.</li></ul>
 *
 * <p>A pointcut is also related to a wrapping class and a wrapping
 * method (and a precise wrapper, instance of the wrapping class, can
 * be specified if it is known). This couple implements the aspect
 * code in all the points that matches the over-depicted pointcut
 * expression.
 *
 * <p>Finally, when a pointcut is applied to the matching elements,
 * wrappers (instances of the wrapping class) are created (except if
 * the intialization wrapper is not null). In this case, the
 * <code>one2one</code> flag tells if one new wrapper instance nust be
 * created for each base component or if one unique wrapper instance
 * must be used for all the components.
 *
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 * @see AspectComponent */

public class MethodPointcut extends Pointcut {

   Vector wrappeeExprs = new Vector();
   Vector wrappeeClassExprs = new Vector();
   Vector wrappeeMethodExprs = new Vector();
   Vector hostExprs = new Vector();
   Vector iwrappeeExprs = new Vector();
   Vector iwrappeeClassExprs = new Vector();
   Vector iwrappeeMethodExprs = new Vector();
   Vector ihostExprs = new Vector();
   String wrappeeExpr;
   String wrappeeClassExpr;
   String wrappeeMethodExpr;
   String hostExpr;
   String wrappingClassName;
   String wrappingMethodName;
   String methodName;
   Object[] methodArgs;
   boolean one2one = true;
   AspectComponent aspectComponent = null;

   static String[] wrappeeKeywords = new String[] {
      "ALL"
   };
   static String[] classKeywords = new String[] {
      "ALL"
   };
   static String[] methodKeywords = new String[] {
      "ALL",
      "MODIFIERS",
      "ACCESSORS",
      "SETTER",
      "GETTER",
      "ADDERS",
      "ADDER",
      "REMOVERS",
      "REMOVER"
   };
   static String[] hostKeywords = new String[] {
      "ALL"
   };

   /**
    * Returns a readable description of the pointcut. */

   public String toString() {
      return "pointcut {"+wrappingClassName+","+wrappingMethodName+","+methodName+
         "}->{"+wrappeeExpr+","+wrappeeClassExpr+","+wrappeeMethodExpr+","+hostExpr+"}";
   }

   Wrapper commonWrapper = null;
   Wrapper initWrapper = null;

   /**
    * Instantiates a new pointcut with the given characterisitics.
    *
    * @param aspectComponent the aspect component this pointcut
    * belongs to
    * @param wrappeeExpr the wrappee definition that matches the names
    * as defined by the naming aspect
    * @param wrappeeClassExpr the wrappee class expression (matches
    * the fully qualified class name)
    * @param wrappeeMethodExpr the wrappee method expression (matches
    * the full method name as defined by the rtti)
    * @param initWrapper the instance of the wrapper used by this
    * pointcut, if null a new one is automatically created depending
    * on the one2one flag
    * @param wrappingClassName the name of the wrapper class
    * @param wrappingMethodName the name of the wrapping method
    * (within the wrapping class)
    * @param methodName the name of the aspect component
    * method to upcall when a mathing object is used
    * @param methodArgs the argument values for this method
    * method to upcall when a mathing object is used
    * @param hostExpr a regular expression that macthes the hosts
    * where the pointcut has to be applied (if null or empty string,
    * default is ".*" which means that the pointcut will be applied on
    * all the hosts)
    * @param one2One true if each new wrapped instance corresponds to
    * one different wrapper (it has no effect if
    * <code>initWrapper</code> is not null */

   public MethodPointcut ( AspectComponent aspectComponent, 
                           String wrappeeExpr, 
                           String wrappeeClassExpr, 
                           String wrappeeMethodExpr,
                           Wrapper initWrapper,
                           String wrappingClassName,
                           String wrappingMethodName,
                           String methodName,
                           Object[] methodArgs,
                           String hostExpr,
                           boolean one2One ) {

      this.aspectComponent = aspectComponent;
      this.wrappeeExpr = wrappeeExpr;
      this.wrappeeClassExpr = wrappeeClassExpr;
      this.wrappeeMethodExpr = wrappeeMethodExpr;
      this.hostExpr = hostExpr;
      this.initWrapper = initWrapper;
      this.wrappingClassName = wrappingClassName;
      this.wrappingMethodName = wrappingMethodName;
      this.methodName = methodName;
      this.methodArgs = methodArgs;
      this.one2one = one2one;
      Log.trace("wrap","new "+this);
   }

   Wrapper buildWrapper() {
      try {
         Class wrapperClass=Class.forName(wrappingClassName);
         if( methodArgs != null ) {
            Constructor[] cs = wrapperClass.getConstructors();
            for( int i=0; i<cs.length; i++ ) {
               try {
                  return (Wrapper)cs[i].newInstance( methodArgs );
               } catch( IllegalArgumentException e ) {}
            }
         }
         return (Wrapper)wrapperClass.newInstance();
      } catch(Exception e) {
         e.printStackTrace();
      }
      return null;
   }

   /**
    * Applies this pointcut to the given wrappee.
    *
    * <p>The pointcut application consists in wrapping the wrappee
    * accordingly to the pointcut description. Note that in JAC, the
    * pointcut is usually applied on a per-component basis, and when
    * the component (wrappee) is used for the first time.
    *
    * @param wrappee the component the current pointcut is applied to */
  
   public synchronized void applyTo( Wrappee wrappee ) {

      Log.trace("pointcut","checking "+this+" on "+wrappee); 

      if( wrappee.isCloned() ) {
         Log.trace("pointcut","do not apply aspects on clones");
         return;
      }

      parseExpr( "host expression", wrappee,
                 hostExpr, hostKeywords,
                 hostExprs, ihostExprs );
      if( ! isHostMatching( wrappee ) ) { return; }
      Log.trace("pointcut","host is matching");       

      parseExpr( "wrappee expression", wrappee,
                 wrappeeExpr, wrappeeKeywords,
                 wrappeeExprs, iwrappeeExprs );
      parseExpr( "wrappee class expression", wrappee,
                 wrappeeClassExpr, classKeywords, 
                 wrappeeClassExprs, iwrappeeClassExprs );

      if( ! isClassMatching( wrappee ) ) return;
      Log.trace("pointcut","class is matching"); 
      if( ! isNameMatching( wrappee ) ) return;
      Log.trace("pointcut","name is matching"); 

      // upcalls the method if exist
      if( methodName != null ) {
         try {
            ClassRepository.get().getClass( aspectComponent.getClass() )
               .invoke( aspectComponent, methodName, methodArgs ); 
         } catch( Exception e ) {
            e.printStackTrace();
         }
      }

      // stops if no wrapping infos if given
      if( initWrapper==null && wrappingClassName==null &&
          wrappingMethodName==null ) return;

      parseExpr( "wrappee method expression", wrappee,
                 wrappeeMethodExpr, methodKeywords, 
                 wrappeeMethodExprs, iwrappeeMethodExprs );

      Log.trace("pointcut","applying "+wrappeeMethodExpr+
                " --- "+this+" to "+wrappee); 
      
      Vector methodsToWrap = getMatchingMethodsFor( wrappee );
      Wrapper wrapper = null;
      if( initWrapper != null ) {
         wrapper = commonWrapper = initWrapper;
      } else {
         try {
            if( one2one ) {
               wrapper = buildWrapper();
            } else {
               if( commonWrapper == null )
                  commonWrapper = buildWrapper();
               wrapper = commonWrapper;
            }
         } catch( Exception e ) {
            e.printStackTrace();
         }
      }
      
      Log.trace("pointcut.wrap","methods to wrap: "+methodsToWrap); 

      // wrap the methods
      Vector eh = wrapper.getExceptionHandlers(wrapper.getClass());
      Iterator it = methodsToWrap.iterator();
      while( it.hasNext() ) {
         MethodItem method = (MethodItem) it.next();
         wrappee.wrap( wrapper, wrappingMethodName, method.getName() );
         
         // install exeption handler if exist
         Iterator it2 = eh.iterator();
         while( it2.hasNext() ) {
            String handler = (String) it2.next();
            wrappee.addExceptionHandler(wrapper,handler,method.getName());
         }
      }
   }

   /**
    * Gets the methods of the wrappee that are modified by this
    * pointcut.
    *
    * @param wrappee the component to test
    * @return a vector containing the matching method items */

   protected Vector getMatchingMethodsFor( Wrappee wrappee ) {
      Log.trace("wrap","getting matching methods for "+wrappee); 
      Vector result = new Vector();
      ClassItem cli = ClassRepository.get().getClass( wrappee.getClass() );
      Vector methods = cli.getAllMethods();
      Iterator it = methods.iterator();
      while ( it.hasNext() ) {
         Object o = it.next();
         MethodItem mi = (MethodItem)o;
         if( isMethodMatching( mi ) ) {
            Log.trace("wrap",mi+" matches"); 
            result.add( mi );
         }
      }
      return result;
   }

   /**
    * Tests if the given component class is modified (in a way or
    * another) by this pointcut.
    *
    * @param wrappee the component to test
    * @return true if the class matches */

   protected boolean isClassMatching( Wrappee wrappee ) {
      
      Iterator it = wrappeeClassExprs.iterator();
      Iterator iti = iwrappeeClassExprs.iterator();
      String name = wrappee.getClass().getName();
      try {
         while( it.hasNext() ) {
            RE regexp = new RE((String)it.next(), 0, RESyntax.RE_SYNTAX_EMACS);
            boolean inv = ((Boolean) iti.next()).booleanValue();
            Log.trace("wrap","isClassMatching: comparing "+name+" with "+
                      regexp+" (inv="+inv);
            if( regexp.isMatch( name ) ^ 
                inv ) {
               return false;
            }
         }
      } catch( Exception e ) {
         e.printStackTrace();
      }
      return true;
   }

   /**
    * Tests if the given component is modified (in a way or
    * another) by this pointcut.
    *
    * <p>Contrary to the class-based matching, this matching works on
    * a per-component basis and not on a per-class basis. This is
    * posible by using the naming aspect of the system (assuming it is
    * there). If the naming aspect appears not to be woven, then all
    * the components should mathes here.
    *
    * @param wrappee the component to test
    * @return true if the name matches */

   protected boolean isNameMatching( Wrappee wrappee ) {
      Iterator it = wrappeeExprs.iterator();
      Iterator iti = iwrappeeExprs.iterator();
      String name = NameRepository.get().getName(wrappee);
      if( name == null ) return false;
      try {
         while( it.hasNext() ) {
            String s = (String)it.next();
            if( isPathExpression( s ) ) {
               Log.trace("wrap","path expression found: "+s);
               return isInPath( wrappee, s );
            }
            RE regexp = new RE( s, 0, RESyntax.RE_SYNTAX_EMACS);
            boolean inv = ((Boolean) iti.next()).booleanValue();
            Log.trace("wrap","isNameMatching: comparing "+name+" with "+
                      regexp+" (inv="+inv);
            if( regexp.isMatch( name ) ^ 
                inv ) {
               return false;
            }
         }
      } catch( Exception e ) {
         e.printStackTrace();
      }
      return true;
   }

   /**
    * Tells if this expression is a path expression (of the form o/r/o...).
    *
    * @param expr the expression to check
    * @return true is a path expression */

   public static boolean isPathExpression( String expr ) {
      if( expr.indexOf('/') == -1 ) {
         return false;
      } else {
         return true;
      }
   }

   /**
    * Tells if this object is reachable for the given object path
    * expression.
    *
    * @param candidate the candidate object
    * @param pathExpr the path expression
    * @return boolean true if reachable */

   public static boolean isInPath( Object candidate, String pathExpr ) {
      StringTokenizer st = new StringTokenizer(pathExpr,"/");
      String root = st.nextToken();
      Collection accessible = ObjectRepository.getObjects( root );
      try {
         while (st.hasMoreTokens()) {
            Log.trace("wrap","intermediate accessibles are "+accessible);
            String relExpr = st.nextToken();
            accessible = getRelatedObjects( accessible, relExpr );
            String filterExpr = st.nextToken();
            accessible = filterObjects( accessible, filterExpr );
         }
      } catch( Exception e ) {
         Log.error("malformed path expression: "+pathExpr);
         e.printStackTrace();
         return false;
      }
      Log.trace("wrap","checking if "+candidate+
                " matches the path expression "+pathExpr+
                ", accessible objects are: "+accessible);
      return accessible.contains( candidate );
   }

   /**
    * Gets all the objects that are related to a set of objects through
    * relations that match the given expression.
    *
    * @param initial the initial objects
    * @param relationExpr the expression that matches the relations
    * @return a set of accessible objects */

   public static Collection getRelatedObjects( Collection initial,
                                               String relationExpr ) {
      Log.trace("wrap","getRelatedObjects "+relationExpr);
      Iterator it = initial.iterator();
      ClassRepository cr = ClassRepository.get();
      Vector res = new Vector();
      while( it.hasNext() ) {
         Object o = it.next();
         ClassItem cli = cr.getClass( o.getClass() );
         Log.trace("wrap","getting matching relations");
         Collection rels = cli.getMatchingRelations( relationExpr );
         Log.trace("wrap","matching relations are "+rels);
         Iterator itRels = rels.iterator();
         while( itRels.hasNext() ) {
            FieldItem fi = (FieldItem) itRels.next();
            if( fi.isReference() ) {
               Object ref = fi.get( o );
               if( ref != null ) {
                  Log.trace("wrap","adding referenced object "+ref);
                  res.add( ref );
               }
            } else if ( fi instanceof CollectionItem ) {
               Collection cref = ((CollectionItem)fi).getActualCollection( o );
               if( cref != null ) {
                  Log.trace("wrap","adding objects in collection");
                  res.addAll( cref );
               }
            }
            Log.trace("wrap","performing next relation");
         }
      }
      return res;
   }

   /**
    * Filters a collection of objects regarding an expression.
    *
    * @param initial the collection to filter
    * @param filter the filtering expression
    * @return a filtered collection that contains only the objects
    * that match the expression */

   public static Collection filterObjects( Collection initial,
                                           String filter ) {
      Log.trace("wrap","filterObjects "+filter);
      NameRepository nr = (NameRepository)NameRepository.get();
      Vector res = new Vector();
      try {
         // path expression allows the user to denote objects
         // with their index in the collection
         Integer index = new Integer(filter);
         res.add(((Vector)initial).get(index.intValue()));
         return res;
      } catch (Exception e) {}

      Iterator it = initial.iterator();
      try {
         RE regexp = new RE( filter, 0, RESyntax.RE_SYNTAX_EMACS);
         while( it.hasNext() ) {
            Object o = it.next();
            String name = nr.getName(o);
            if( name!=null && regexp.isMatch( name ) ) {
              res.add(o);
            }
         }
      } catch( Exception e ) {
         e.printStackTrace();
      }
      return res;
   }

   /**
    * Tests if the given component is modified (in a way or
    * another) by this pointcut.
    *
    * <p>Contrary to the class-based matching, this matching works on
    * a per-component basis and not on a per-class basis. This is
    * posible by using the naming aspect of the system (assuming it is
    * there). If the naming aspect appears not to be woven, then all
    * the components should mathes here.
    *
    * @param wrappee the component to test
    * @return true if the name matches */

   protected boolean isHostMatching( Wrappee wrappee ) {

      String name = "";
      try {
         Class distd = Class.forName("jac.core.dist.Distd");
         name = (String)distd.getMethod("getLocalContainerName",new Class[0])
            .invoke(null,new Object[0]);
      } catch (Exception e) {
         e.printStackTrace();
         return false;
      }
      if (name == null) return false;

      Iterator it = hostExprs.iterator();
      Iterator iti = ihostExprs.iterator();
      try {
         while( it.hasNext() ) {
            RE regexp = new RE((String)it.next(), 0, RESyntax.RE_SYNTAX_EMACS);
            boolean inv = ((Boolean) iti.next()).booleanValue();
            //Log.trace("pointcut","isHostMatching: comparing "+name+" with "+
            //          regexp+" (inv="+inv);
            if( regexp.isMatch( name ) ^ 
                inv ) {
               return false;
            }
         }
      } catch( Exception e ) {
         e.printStackTrace();
      }
      //Log.trace("pointcut","Host "+name+" is matching "+hostExprs);      
      return true;
   }

   /**
    * Tests if the given method item is modified (in a way or
    * another) by this pointcut.
    *
    * @param method the method to test
    * @return true if the method matches */

   protected boolean isMethodMatching( MethodItem method ) {
      Iterator it = wrappeeMethodExprs.iterator();
      Iterator iti = iwrappeeMethodExprs.iterator();
      String name = method.toString();
      try {
         while( it.hasNext() ) {
            RE regexp = new RE((String)it.next(), 0, RESyntax.RE_SYNTAX_EMACS);
            boolean inv = ((Boolean) iti.next()).booleanValue();
            Log.trace("wrap","isMethodMatching: comparing "+name+" with "+
                      regexp+" (inv="+inv);
            if( regexp.isMatch( name ) ^ 
                inv ) {
               return false;
            }
         }
      } catch( Exception e ) {
         e.printStackTrace();
      }

      return true;
   }



   protected String parseKeyword( Wrappee wrappee, String keywordExpr, 
                                  String[] parameters ) {

      String result = "";

      if( keywordExpr.equals("ALL") ) {
         result = ".*";

      } else if( keywordExpr.equals("MODIFIERS") ) {

         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
     
         Vector mods = cli.getAllModifiers();
         Iterator it = mods.iterator();
         while( it.hasNext() ) {
            MethodItem m = (MethodItem) it.next();
            String s = replace(m.toString(),"[]","\\[\\]");
            result = result+s;
            if( it.hasNext() ) {
               result = result + "\\|";
            }
         }
      } else if( keywordExpr.equals("ACCESSORS") ) {

         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
     
         Vector mods = cli.getAllAccessors();
         Iterator it = mods.iterator();
         while( it.hasNext() ) {
            MethodItem m = (MethodItem) it.next();
            String s = replace(m.toString(),"[]","\\[\\]");
            result = result+s;
            if( it.hasNext() ) {
               result = result + "\\|";
            }
         }
      } else if( keywordExpr.equals("REMOVERS") ) {

         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
     
         Vector mods = cli.getAllRemovers();
         Iterator it = mods.iterator();
         while( it.hasNext() ) {
            MethodItem m = (MethodItem) it.next();
            String s = replace(m.toString(),"[]","\\[\\]");
            result = result+s;
            if( it.hasNext() ) {
               result = result + "\\|";
            }
         }
      } else if( keywordExpr.equals("ADDERS") ) {

         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
     
         Vector mods = cli.getAllAdders();
         Iterator it = mods.iterator();
         while( it.hasNext() ) {
            MethodItem m = (MethodItem) it.next();
            String s = replace(m.toString(),"[]","\\[\\]");
            result = result+s;
            if( it.hasNext() ) {
               result = result + "\\|";
            }
         }
      } else if( keywordExpr.equals("SETTER") ) {

         if( parameters==null || parameters.length>1 || parameters.length==0 ) {
            Log.error("pointcut parsing: wrong number of arguments for the SETTER keyword");
            return "#ERROR#";
         }
         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
         FieldItem fi = cli.getField(parameters[0]);
         if( fi==null ) {
            Log.warning("pointcut", "field "+parameters[0]+" does not exist in class "+cli);
            return "#ERROR#";
         }
         MethodItem[] writers = fi.getWritingMethods();
         if (writers!=null && writers.length>0) {
            result = writers[0].toString();
         } else {
            Log.warning("pointcut", "field "+parameters[0]+" does not exist in class "+cli);
            return "#ERROR#";
         }

      } else if( keywordExpr.equals("GETTER") ) {

         if( parameters==null || parameters.length>1 || parameters.length==0 ) {
            Log.error("pointcut parsing: wrong number of arguments for the GETTER keyword");
            return "#ERROR#";
         }
         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
         FieldItem fi = cli.getField(parameters[0]);
         if( fi==null ) {
            Log.warning("pointcut", "field "+parameters[0]+" does not exist in class "+cli);
            return "#ERROR#";
         }
         MethodItem[] accessors = fi.getAccessingMethods();
         if (accessors!=null && accessors.length>0) {
            result = accessors[0].toString();
         } else {
            Log.warning("pointcut", "field "+parameters[0]+
                        " does not have any setter in class "+cli);
            return "#ERROR#";
         }
      } else if( keywordExpr.equals("REMOVER") ) {

         if( parameters==null || parameters.length>1 || parameters.length==0 ) {
            Log.error("pointcut parsing: wrong number of arguments for the REMOVER keyword");
            return "#ERROR#";
         }
         ClassItem cli = ClassRepository.get().getClass(wrappee.getClass());
         FieldItem fi = cli.getField(parameters[0]);
         if( fi==null ) {
            Log.warning("pointcut", "field "+parameters[0]+" does not exist in class "+cli);
            return "#ERROR#";
         }
         if ( ! (fi instanceof CollectionItem) ) {
            Log.warning("pointcut", "field "+parameters[0]+" is not a collection in class "+cli);
            return "#ERROR#";
         }
         CollectionItem ci = (CollectionItem)fi;
         MethodItem[] removers = ci.getRemovingMethods();
         if (removers!=null && removers.length>0) {
            result = removers[0].toString();
         } else {
            Log.warning("pointcut", "collection "+parameters[0]+
                        " does not have any remover in class "+cli);
            return "#ERROR#";
         }
      }


      Log.trace("wrap","parsed keyword "+keywordExpr+" => "+result);

      return result;

   }

}
