/*
  Copyright (C) 2001 Renaud Pawlak.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.util.*;
import jac.core.rtti.*;
import gnu.regexp.*;


/**
 * This class allows the JAC applications to access objects that are
 * managed within the current VM or that can be defined outside of it
 * by the currently woven aspects.
 *
 * <p>By default, the object repository methods only return the JAC
 * objects that are accessible through the
 * <code>JacObject.getObject</code> methods. However, since it upcalls
 * the <code>ACManager.whenGetObjects()</code> method, all the aspect
 * components can change the returned object set.
 *
 * <p>As typical examples:
 *
 * <ul><li>the persistence aspects may load objects that correspond to
 * the get filter so that the application can see the objects stored
 * in the database
 *
 * <li>the owning aspect can remove some objects from the returned
 * collection if the current user is not allowed to access them
 *
 * <li>the distribution aspect can initiate a distributed seek over
 * all the containers of the topology to add some objects that
 * correspond to the filter but that may have not been declared yet in
 * this VM</ul>
 *
 * @see JacObject#getObjects()
 * @see JacObject#getObjects(Class)
 * @see ACManager#whenGetObjects(List,ClassItem)
 * @see AspectComponent#whenGetObjects(List,ClassItem)
 */

public class ObjectRepository {

   /**
    * Gets all the instantiated JAC objects on the current VM and on
    * all the external objects sources known by the aspects (maybe
    * resticed by some aspects).
    *
    * @return a collection of acessible objects
    * @see #getObjects(ClassItem) */

   public static Collection getObjects() {
      ArrayList objects = JacObject.instances;
      ((ACManager)ACManager.get()).whenGetObjects(objects,null);
      return objects;
   }

   /**
    * Gets all the JAC objects instances of a given class on the
    * current VM and on all the external objects sources known by the
    * aspects (maybe resticed by some aspects).
    *
    * @return a collection of acessible instances of <code>cl</code>
    * @see #getObjects(ClassItem) */

   public static Collection getObjects(ClassItem cl) {
      List objects = new ArrayList(Arrays.asList(
         JacObject.getObjects(cl.getActualClass())));
      ((ACManager)ACManager.get()).whenGetObjects(objects,cl);
      return objects;
   }

   /**
    * Gets the set of JAC objects that match the expression.
    *
    * @param expr a regular expression
    * @return the objects set */

   public static Collection getObjects(String expr) {
      Collection all =getObjects();
      Collection res = new Vector();
      RE re;
      try {
         re = new RE( expr );
      } catch ( Exception e ) {
         e.printStackTrace();
         return null;
      }
      NameRepository nr = (NameRepository)NameRepository.get();
      Iterator it = all.iterator();
      while ( it.hasNext() ) {
         Object o = it.next();
         String name = nr.getName( o );
         if( name == null ) continue;
         if( re.isMatch( name ) ) {
            res.add( o );
         }
      }
      return res;
   }

}
