/*
  Copyright (C) 2001 Renaud Pawlak

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.io.Serializable;

import jac.util.*;
import jac.core.rtti.*;
import java.lang.reflect.*;
import gnu.regexp.*;
import java.util.*;

/**
 * This abstract class is the definition of a pointcut in JAC.
 *
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 * @see AspectComponent
 * @see MethodPointcut */

public abstract class Pointcut implements Serializable {

   /**
    * Applies this pointcut to the given wrappee.
    *
    * @param wrappee the component the current pointcut is applied to */
  
   public abstract void applyTo( Wrappee wrappee );

   /**
    * Parses a keyword expression and returns its actual value as a
    * regular expression regarding the context. */

   protected abstract String parseKeyword( Wrappee wrappee, 
                                           String keywordExpr, 
                                           String[] parameters );

   /**
    * A generic method that parses a pointcut expression and stores
    * the result within a vector.
    *
    * @param descr a humain readable desciption of the pointcut
    * expression type (used to make logs clearer)
    * @param expr a pointcut expression
    * @param result the parsing result */

   protected void parseExpr( String descr, Wrappee wrappee, String expr, 
                             String[] keywords, Vector result, Vector inv ) {

      result.clear();
      inv.clear();
      expr = replace( expr, " || ", "\\|" );
      int pos = 0;
      boolean end = false;

      if( expr.charAt(0) == '!' ) {
         pos=1;
         inv.add( new Boolean(false) );
      } else {
         inv.add( new Boolean(true) );
      }
      while( ! end ) {
         int newpos = expr.indexOf( " && ", pos );
         try {
            if( newpos != -1 ) {
               result.add( replaceKeywords( wrappee,
                     expr.substring( pos, newpos ), keywords ) );
               pos = newpos + 4;
               if( expr.charAt(pos) == '!' ) {
                  pos++;
                  inv.add( new Boolean(false) );
               } else {
                  inv.add( new Boolean(true) );
               }
            } else {
               result.add( replaceKeywords( wrappee, 
                  expr.substring( pos ), keywords ) );
               end = true;
            }
         } catch (Exception e) {
            Log.error("Invalid pointcut definition, "+descr+
                      " construction failed at position "+pos+": "+e);
         }
      }
   }                           

   /** Replace the keywords within an expression. */
   String replaceKeywords( Wrappee wrappee, String expr, String[] keywords ) {
      String newExpr=expr; 
      for(int i=0; i<keywords.length; i++) {
         newExpr=replaceKeyword( wrappee, newExpr, keywords[i] ); 
      }
      return newExpr;
   }

   /** Parse the parammeters of a keyword */
   String[] parseParameters(String params) {
      if( params.equals("") ) return null;
      if( params.charAt(0) != '(' ) return null;
      params = params.substring(1, params.indexOf(')',0));
      StringTokenizer st = new StringTokenizer( params, "," );
      String[] result = new String[st.countTokens()];
      int i=0;
      while (st.hasMoreTokens()) {
         result[i++]=(String)st.nextToken();
      }
      return result;
   }

   int parametersLength(String params) {
      if( params.equals( "" ) ) return 0;
      if( params.charAt(0) != '(' ) return 0;
      return params.indexOf(')',0)+1;
   }

   /** Replace a keyword within an expression. */
   String replaceKeyword( Wrappee wrappee, String expr, String keyword ) { 

      int pos = 0;
      boolean end = false;
      String newExpr = "";

      while( ! end ) {
         int newpos = expr.indexOf( keyword, pos );
         try {
            if( newpos != -1 ) {
               newExpr = newExpr + expr.substring( pos, newpos - pos ) +
                  parseKeyword( wrappee, keyword, 
                                parseParameters(expr.substring(newpos+keyword.length())) );
               pos = newpos + keyword.length() + 
                  parametersLength(expr.substring(newpos+keyword.length()));
            } else {
               newExpr = newExpr + expr.substring( pos );
               end = true;
            }
         } catch (Exception e) {
            Log.error("Invalid keyword '"+keyword+"'");
         }
      }
      return newExpr;
   }

   /**
    * A useful method that replaces all the occurences of a string.
    *
    * @param orgString the original string
    * @param oldString the string to replace (if found) in the
    * original string
    * @param newString the string that replaces all the occurences of
    * old string
    * @return a new string with the occurences replaced */

   public static String replace(String orgString, 
                                String oldString, String newString ) {

      int pos = 0;
      boolean end = false;

      String result = "";

      while( ! end ) {
         int newpos = orgString.indexOf( oldString, pos );
         try {
            if( newpos != -1 ) {
               result = result + orgString.substring( pos, newpos ) +
                  newString;
            } else {
               result = result + orgString.substring( pos );
            }
         } catch (Exception e) {
            e.printStackTrace();
            return orgString;
         }
         pos = newpos + oldString.length();
         if( newpos == -1 ) end = true;
      }
      //      System.out.println("--- replaced "+orgString+" with "+result);
      return result;
   }                           

}




















