/*

  Copyright (C) 2001 Renaud Pawlak, Lionel Seinturier.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import java.util.Vector;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.io.Serializable;

/**
 * This interface describe all the methods that are useful for a
 * wrappee and that are implemented by the JacObject class.
 *
 * @version 0.8.1
 * @author <a href="http://cedric.cnam.fr/~pawlak/index-english.html">Renaud Pawlak</a>
 * @author <a href="http://www-src.lip6.fr/homepages/Lionel.Seinturier/index-eng.html">Lionel Seinturier</a> */

public interface Wrappee extends Cloneable, Serializable {

   /**
    * Add a wrapper to the current wrappee.
    *
    * <p>Any method of this wrapper can then be used on the wrappee
    * with <code>invokeRoleMethod</code>.
    *
    * <p>To precise which method of this wrapper should actually wrap
    * the current wrappee methods, use the <code>wrap</code> methods.
    *
    * @param wrapper the new wrapper.
    *
    * @see #wrap(Wrapper,String,String)
    * @see #wrap(Wrapper,String,String[])
    * @see #wrap(Wrapper,String[],String[][])
    * @see #wrapAll(Wrapper,String)
    * @see #wrapAllEx(Wrapper,String,String[])
    * @see #invokeRoleMethod(String,Object[]) */

   void wrap( Wrapper wrapper );

   /**
    * Wrap a given method of the current JAC object with a given
    * method of a given wrapper.
    *
    * <p>All the other methods of the wrapper will be automatically
    * considered as potential role methods (this method calls the
    * <code>wrap(Wrapper)</code> method.
    *
    * @param wrapper the wrapper where the wrapping method is
    * implemented
    * @param wrappingMethod  the name of the wrapping method
    * @param wrappedMethod the name of the wrapped method
    *
    * @see Wrapper
    * @see #wrap(Wrapper) */

   void wrap( Wrapper wrapper, String wrappingMethod,
              String wrappedMethod );

   /**
    * A nicer way to write <code>wrap</code> when several base methods
    * need to be wrapped.
    *
    * @param wrapper the wrapper where the wrapping method is
    * implemented
    * @param wrappingMethod the name of the wrapping method in the
    * wrapper
    * @param wrappedMethods the names of the wrapped methods.
    *
    * @see #wrap(Wrapper,String,String) */

   void wrap( Wrapper wrapper, String wrappingMethod,
                     String[] wrappedMethods );

   /**
    * A nicer way to write <code>wrap</code> when several wrapping
    * methods need to wrap several base methods.
    *
    * @param wrapper the wrapper
    * @param wrappingMethods the names of the wrapping methods in the
    * wrapper
    * @param wrappedMethods an array of array containing the names of
    * the wrapped methods for each wrapping method
    *
    * @see #wrap(Wrapper,String,String) */

   void wrap( Wrapper wrapper, String[] wrappingMethods,
                     String[][] wrappedMethods );

   /**
    * A convenient way to wrap methods when the wrapped methods have
    * the same name as the wrapping methods.
    *
    * @param wrapper the wrapper
    * @param wrappingMethods the names of the wrapping methods in the
    * wrapper
    *
    * @see #wrap(Wrapper,String,String) */
   void wrapSameName( Wrapper wrapper, String wrappingMethods[] );

   /**
    * Wrap all the wrappable (public) methods of the current wrappee.
    *
    * @param wrapper the wrapper
    * @param wrapperMethod the name of the wrapping method
    *
    * @see #wrap(Wrapper,String,String)
    * @see #wrapModifiers(Wrapper,String)
    * @see #wrapAllEx(Wrapper,String,String[]) */

   void wrapAll( Wrapper wrapper, String wrapperMethod );

   /**
    * Wrap all the methods with the given parameter types.
    */
   void wrapAll( Wrapper wrapper, String wrapperMethod, 
                 Class[] paramTypes ) ;

   /**
    * Wrap all the methods but those with the given parameter types.  
    */
   void wrapAllEx( Wrapper wrapper, String wrapperMethod, 
                   Class[] paramTypes ) ;

   /**
    * Wrap all the writer wrappable (public) methods of the current
    * wrappee.
    *
    * @param wrapper the wrapper
    * @param wrapperMethod the name of the wrapping method
    *
    * @see #wrap(Wrapper,String,String)
    * @see #wrapAll(Wrapper,String)
    * @see #wrapAllEx(Wrapper,String,String[]) */

   void wrapModifiers( Wrapper wrapper, String wrapperMethod );

   /**
    * Wrap all the wrappable (public) methods of the current wrappee
    * except the one given in the <code>excludedMethods</code>
    * parameter.
    *
    * @param wrapper the wrapper object
    * @param wrapperMethod the name of the wrapping method
    * @param exludedMethods some methods that do not need to be
    * wrapped
    *
    * @see #wrapAll(Wrapper,String)
    * @see #wrapModifiers(Wrapper,String)
    * @see #wrap(Wrapper,String,String) */

   void wrapAllEx( Wrapper wrapper, String wrapperMethod,
                          String[] excludedMethods );

   /**
    * Definitively removes a wrapper.
    *
    * <p>When this method is called, none of the methods of the
    * wrapper can be used as role methods anymore.
    *
    * <p>This method before calls the <code>unwrapAll</code> to ensure
    * that none of the current wrappee methods are yet wrapped.
    *
    * @param wrapper the wrapper to remove
    *
    * @see #unwrapAll(Wrapper)
    * @see #invokeRoleMethod(String,Object[]) */

   void unwrap( Wrapper wrapper );

   /**
    * Unwrap a single method.
    *
    * <p>The wrapper must implement <code>wrappingMethod</code>. If
    * the wrapped method was not actually wrapped at the time this
    * method is called, then this code has no effect.
    *
    * <p>To definitively remove the wrapper so that none of its method
    * will not be considered as role methods anymore, use the
    * <code>unwrap(Wrapper)</code> method.
    *
    * @param wrapper the wrapper
    * @param wrapperMethod the name of the method that must unwrap
    * <code>wrappeeMethod</code>
    * @param wrappeeMethod the name of the method to unwrap
    *
    * @see #unwrap(Wrapper) */

   void unwrap( Wrapper wrapper, String wrappingMethod,
                       String wrappedMethod );

   /**
    * Unwrap all the methods that are currently wrapped by the given
    * wrapping method.
    *
    * <p>To definitively remove the wrapper so that none of its method
    * will not be considered as role methods anymore, use the
    * <code>unwrap(Wrapper)</code> method.
    *
    * @param wrapper the wrapper
    * @param wrappingMethod the name of the method that must unwrap
    * all the wrappee methods
    *
    * @see #unwrap(Wrapper)
    * @see #unwrap(Wrapper,String,String) */

   void unwrapAll( Wrapper wrapper, String wrappingMethod );

   /**
    * Unwrap all the methods that are currently wrapped by any
    * wrapping method of the given wrapper.
    *
    * <p>To definitively remove the wrapper so that none of its method
    * will not be considered as role methods anymore, use the
    * <code>unwrap(Wrapper)</code> method. 
    *
    * @param wrapper the wrapper to remove
    *
    * @see #unwrap(Wrapper)
    * @see #unwrap(Wrapper,String,String) */

   void unwrapAll( Wrapper wrapper );

   /**
    * This method returns the wrapper objects that currently wrap
    * wrappee.
    *
    * @return the wrappers array for the wrappee
    *
    * @see #wrap(Wrapper)
    * @see #isWrappedBy(Wrapper) */

   Object[] getWrappers();

   /**
    * Returns a set of wrappers of a given type that wrap the current
    * wrappee.
    *
    * <p>If the <code>wrapperType</code> parameter is null,
    * returns all wrappers.
    * 
    * @param wrapperType the type of the wrappers
    * @return an array containing the matching wrappers 
    *
    * @see #wrap(Wrapper)
    * @see #wrap(Wrapper,String,String)
    * @see #getWrappers() */

   Object[] getWrappers( Class wrapperType );

   /**
    * Test if the given wrapper wraps the current wrappee.
    *
    * @param wrapper the candidate wrapper
    * @return true if wrapped
    * 
    * @see #wrap(Wrapper)
    * @see #getWrappers() */

   boolean isWrappedBy( Wrapper wrapper );

   /**
    * This method returns the classes of all the wrappers that
    * currently wrap the current wrappee.
    *
    * @return the wrapper classes array
    *
    * @see #wrap(Wrapper)
    * @see #isExtendedBy(Class) */

   Object[] getWrapperClasses();

   /**
    * Returns true if the wrappee is wrapped by an instance of wrapper
    * class (the wrapper type is then extended). If the given wrapper
    * class is null, return false.
    *
    * @param wrapperClass the class to test
    * @return true if wrapped
    *
    * @see #getWrapperClasses() */

   boolean isExtendedBy( Class wrapperClass );

   /**
    * This method returns the wrappers that wrap the given
    * method.
    *
    * @param methodName the name of the method that is wrapped by
    * the wrapper(s)
    * @return the wrapper(s) array for the given method */

   Wrapper[] getWrappersForMethod(String methodName);

   /**
    * This method returns the wrapping chain of a given method.
    *
    * @param methodName the name of the wrapped method
    * @return the wrapper(s) array for the given method */
   
   

   /**
    * This method prints out the wrappers and wrapping methods for a
    * given methods.
    *
    * @param methodName the name of the method that is wrapped by
    * the wrapper(s) */

   void printWrappersForMethod(String methodName);

   /**
    * For a given wrapper (retrieved by
    * <code>getWrappersForMethod</code>, this method returns the
    * method that wraps the <code>methodName</code> method.
    *
    * <p>Not available yet.
    *
    * @param wrapper the wrapper.
    * @param methodName the name of the method that is supposed to
    * wrap a method of the wrappee.
    * @return the wrapper method that wraps <code>methodName</code>
    * @see Wrappee#getWrappersForMethod(String) */

   Method getWrappingMethod( Wrapper wrapper, 
                                    String methodName );

   /**
    * Returns the wrapping chain for a given method of the current
    * wrappee.
    *
    * <p>The vector contains 2 elements array where the first element is
    * a reference on the wrapper and the second one is the wrapping
    * method in this wrapper. If none wrapper wraps the given method,
    * then the result maybe null or an empty vector.<p>
    *
    * @param methodName the name of the method in the wrappe
    * @return the wrapping chain, null or an empty vector if not
    * wrapped, always null if the given method is not supported by the
    * current wrappee class */

   Vector getWrappingChain( String methodName );
    
   /**
    * This method calls the next wrapper of the current collaboration.
    * 
    * <p>The programmer should not use this method. S/he should use
    * the <code>Wrapper.proceed()</code> method instead, which is
    * simplier to use.
    *
    * @return the return value of the subsequent collaboration
    *
    * @see Collaboration
    * @see Wrapper#proceed() */

   Object nextWrapper();

   /**
    * Adds an exception handler to the specified method.
    *
    * <p>An exception handling method must at least present one method
    * with the following prototype:
    *
    * <ul><pre>
    * void listeningMethodName (ExceptionType toBeCautch);
    * </pre></ul>
    *
    * <p>An exception handling method can handle any throwable object,
    * thus including errors, runtime and non-runtime exceptions.
    *
    * @param wrapper the wrapper that implements <code>method</code>
    * @param method the name of the exception handling method in the
    * wrapper class
    * @param listenedMethod the wrapper method name the handler is
    * added to */
   
   void addExceptionHandler( Wrapper wrapper,
                             String method,
                             String listenedMethod );

   /**
    * Adds an exception handling method to all the methods of the
    * wrappee.
    *
    * @param wrapper the wrapper that implements the method
    * @param method the exception handling method
    *
    * @see #addExceptionHandler(Wrapper,String,String) */

   void addExceptionHandler( Wrapper wrapper,
                             String method );

   /**
    * Get the exception handlers for the specified method.
    *
    * <p>Each element of the returned array is composed of 2 elements
    * arrays containing the wrapper and the method that handles the
    * exception in this wrapper.

    * @param listenedMethod the name of the method we want to know
    * about.
    * @return an array of exception handlers */   
   
   Object[][] getExceptionHandlers( String listenedMethod );

   /**
    * Invoke a role method on the wrappee.
    *
    * <p>The programmer can also use the <code>invoke</code>
    * method. However, whith the <code>invoke</code> method, if a
    * method of the wrappee matches the invoked method, then the role
    * method is not called. Thus, the programmer should use this
    * method to ensure that the role method is actually called.
    * 
    * @param name the name of the role method to invoke.
    * @param parameters the parameters.
    * @return the returned object
    * 
    * @see #invoke(String,Object[]) */

   Object invokeRoleMethod( String name,
                                   Object[] parameters );
   
   /**
    * This method calls a method on the wrappee by knowing its name and
    * arguments.
    * 
    * <p>If no matching method is found, <code>invoke</code> tries to
    * invoke the method as a role method.
    *
    * @param name the name of the method to call
    * @param parameters the argument values
    * @return the called method return value as an object
    *
    * @see #invokeRoleMethod(String,Object[]) */

   Object invoke (String name, Object[] parameters);
   
   /**
    * This method can be used to shortcut the wrapping chain and
    * directly call the original method.
    * 
    * <p><b>NOTE</b>: avoid doing this. It is not clean to shortcut
    * the whole wrapping chain. Instead, use the <code>invoke</code>
    * method or, if you really need to,
    * <code>AspectComponent.before/afterRunningWrapper()</code> to
    * skip some wrappers.
    *
    * @param name the name of the method to call
    * @param parameters the argument values
    * @return the called method return value as an object
    *
    * @see AspectComponent#beforeRunningWrapper(Wrapper,String)
    * @see AspectComponent#afterRunningWrapper(Wrapper,String)
    * @see #invoke(String,Object[])
    **/

   Object invokeOrg (String name, Object[] parameters);

   /**
    * Force the treatment of this object by all the currently woven
    * aspect components.
    *
    * @see ACManager#forceTreatment(Wrappee)
    * @see AspectComponent */

   void forceTreatment();

   /**
    * Return all the fields of the wrappee.<p>
    *
    * This includes private and protected fields of the wrappee class
    * and of all its superclasses. They are set to be accessible so
    * that the wrappers can perform full introspection and
    * intercession of the wrappees states.<p>
    *
    * @return the fields array */

   Field[] getFields();

   /**
    * Return one field of the wrappee.<p>
    *
    * This method allows access to private and protected fields so
    * that the wrappers can perform full introspection and
    * intercession of the wrappees states.<p>
    *
    * It seeks in all superclasses if the field is not found in the
    * current one.<p>
    *
    * @param name the field name
    * @return the matching field if found, null otherwise */

   Field getField( String name );

   /**
    * Return the value of a given field of the wrappee.<p>
    *
    * This method works on public, protected, and private fields.<p>
    *
    * If the field does not exist in the wrappee,
    * <code>getFieldValue</code> tries to read the field from the
    * wrappers that currently wrap the wrappee. Thus, a wrappee state
    * can be considered as an extended state.<p>
    * 
    * This method must can be used within the wrappers to
    * instrospect their wrappee state. The programmer should not use
    * the Java standard introspection features since it does not allow
    * seamless access to the wrapper states. If the field does not
    * exist in the wrapper or any of the wrapper, return null.<p>
    *
    * @param fieldName the name of the field
    * @return the field value 
    * 
    * @see Wrapper
    * @see #setFieldValue(String,Object) */

   Object getFieldValue ( String fieldName );

   /**
    * Return the state of the wrappee.<p>
    *
    * This method works on public, protected, and private fields.<p>
    *
    * The returned two-dim array contains an string array for the
    * field names and an object array for their values.<p>
    *
    * <ul><pre>
    * - getState()[0] = String[] { field names }
    * - getState()[1] = Object[] { corresponding values }
    * - getState()[0][n] = nth field name
    * - getState()[1][n] = nth field value
    * </pre></ul>
    *
    * @return the wrappee state
    * @see #setState(Object[]) */

   Object[] getState ();

   /**
    * Return a partial state of the wrappee.<p>
    *
    * The state only contains the fields named by the
    * <code>fieldNames</code> array. To get all the fields, use the
    * <code>getState()</code> method.<p>
    *
    * This method works on public, protected, and private fields.<p>
    *
    * @return the wrappee state
    * @see #getState() */

   Object[] getState ( String[] fieldNames );

   /**
    * Set the state of the wrappee.<p>
    *
    * This method works on public, protected, and private fields.<p>
    *
    * The state format is the one defined in <code>getState()</code>
    *
    * @return the wrappee state
    * @see #getState() */

   void setState ( Object[] state );   

   /**
    * Set the value of a given field of the current wrappee.<p>
    *
    * This method works on public, protected, and private fields.<p>
    *
    * If the field does not exist in the wrappee,
    * <code>setFieldValue</code> tries to set the field value of the
    * wrappers that currently wrap the wrappee. Thus, a wrappee state
    * can be considered as an extended state.
    * 
    * <p>This method must can be used within the wrappers to intercess
    * (modify) their wrappee state. The programmer should not use the
    * Java standard introspection features since it does not allow
    * seamless access to the wrapper states. If the field does not
    * exist in the wrapper or any of the wrapper, nothing happens.
    *
    * @param fieldName the name of the field
    * @return the field value 
    * 
    * @see Wrapper
    * @see #getFieldValue(String) */

   void setFieldValue (String fieldName, Object value);

   /**
    * Returns the application the object belongs to.
    *
    * @return the application's name */

   String getApplication();

   /**
    * This method extends the default clone method to notify all the
    * active aspect components that a JacObject is beeing cloned.<p>
    * 
    * The reaction of an aspect component is then defined by the
    * <code>AspectComponent.whenClone()</code> method.
    *
    * @see AspectComponent#whenClone(Wrappee,Wrappee)
    * @see ACManager#whenClone(Wrappee,Wrappee) */

   Object clone ();

   /**
    * This method tells if the current wrappee is a clone from an
    * original one.
    *
    * @return true if cloned */

   boolean isCloned();

}






