/*
  Copyright (C) 2001 Renaud Pawlak, Lionel Seinturier.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package jac.core;

import javassist.*;
import java.lang.reflect.*;
import java.util.*;
import java.io.*;

import java.io.FileInputStream;
import java.io.FileNotFoundException;

import jac.util.Log;

/**
 * <code>WrappeeTranslator</code> implements a
 * <code>translateClass()</code> method that translates a regular java
 * class into a wrappable one.
 * 
 * <p>This translation is usually done at
 * <i>load time</i>. After the translation, the class subclasses
 * the <code>JacObject</code> class.
 *
 * <p>To optimize and simplify the translation, this method uses
 * templated bytecodes pieces extracted from
 * <code>WrappeeTemplate</code>.
 *
 * @see JacObject
 * @see WrappeeTemplate
 *
 * @version 0.8.1
 * @author Renaud Pawlak
 * @author Lionel Seinturier
 */

public class WrappeeTranslator {

   /** The wrappee template class name. */
   protected static String templateString = "jac.core.WrappeeTemplate";

   /** The JAC object class name. */
   protected static String jacObjectString = "jac.core.JacObject";

   /** The name of a peculiar vector needed for the translation. */
   protected static String jacVectorString = "jac.core.utils.JacVector";

   /**
    * The property file defines the classes that need to be translated
    * with the following property:
    * 
    * <ul><pre>
    * jac.toAdapt: classToAdapt1 classToAdapt2 ...
    * </pre></ul>
    * 
    * <p>Even if they are not in the jac.prop file, the following
    * system-used classes are translated by default:
    * 
    * <ul><pre>
    * - jac.util.Repository
    * - jac.core.NameRepository
    * - jac.core.dist.Topology
    * </pre></ul>
    */
    
   /** The name of the property file. */
   protected static String propFileName = "jac.prop";

   /** The name of the property in the the jac.prop file. */
   protected static String toNotAdaptProp = "jac.toNotAdapt";

   /** The name of the property in the the jac.prop file. */
   protected static String toAdaptProp = "jac.toAdapt";

   /** The properties loaded from the jac.prop file. */
   protected static Properties props;

   /** Store the classes translated to be not wrappable. */
   public static HashSet classesToNotAdapt = new HashSet();
   /** Store the packages (set of classes) translated to be not
       wrappable. */
   public static HashSet packagesToNotAdapt = new HashSet();

   /** Store the classes translated to be wrappable. */
   public static HashSet classesToAdapt = new HashSet();
   /** Store the packages (set of classes) translated to be
       wrappable. */
   public static HashSet packagesToAdapt = new HashSet();

   /** Store the count of homonyms methods. */
   protected static Hashtable methodCounts;
    
   static {
      
      FileInputStream fis = null;
      Properties defaults = new Properties();
      
      try {
         // Try to read a jac.prop in JAC_ROOT
         try {
            fis = new FileInputStream( Jac.getJacRoot() + propFileName );
            Log.trace("jac.prop","reading file "+Jac.getJacRoot() + propFileName);
            props = new Properties();
            props.load( fis );
            Log.trace("jac.prop","Props:"+props);
         } catch(FileNotFoundException e) {
         }
         
         // Try to read a jac.prop in $HOME
         try {
            String homedir = System.getProperty("user.home",null);
            if (homedir!=null) {
               fis = new FileInputStream(homedir +"/"+ propFileName);
               Log.trace("jac.prop","reading file "+homedir +"/"+ propFileName);
               defaults.putAll(props);
               props = new Properties(defaults);
               props.load( fis );
               Log.trace("jac.prop","Props:"+props);
            }
         } catch(FileNotFoundException e) {
         }
         
         // Try to read a jac.prop file from the current directory
         try {
            fis = new FileInputStream( propFileName );
            Log.trace("jac.prop","reading file "+"./"+ propFileName);
            defaults.putAll(props);
            props = new Properties(defaults);
            props.load( fis );
            Log.trace("jac.prop","Props:"+defaults);
         } catch(FileNotFoundException e) {
         }
      } catch (Exception e) {
         e.printStackTrace();
      }
   

      try {

         String prop = props.getProperty(toNotAdaptProp);
         
         if ( prop == null ) {
            System.out.println( "-- WARNING: no property jac.toNotAdapt found in property file jac.prop" );
         }
         else {
            
            StringTokenizer st = new StringTokenizer( prop );
            //System.out.println( prop );
            
            while ( st.hasMoreElements() ) {
               String element = (String)st.nextElement();
               if (element.endsWith(".*")) {
                  packagesToNotAdapt.add(element.substring(0,element.length()-2));
               } else {
                  classesToNotAdapt.add(element);
               }
            }
         }
         
         prop = props.getProperty(toAdaptProp);
         
         if ( prop == null ) {
            System.out.println( "-- WARNING: no property jac.toAdapt found in property file jac.prop" );
         }
         else {
            
            StringTokenizer st = new StringTokenizer( prop );
            //System.out.println( prop );
            
            while ( st.hasMoreElements() ) {
               String element = (String)st.nextElement();
               if (element.endsWith(".*")) {
                  packagesToAdapt.add(element.substring(0,element.length()-2));
               } else {
                  classesToAdapt.add(element);
               }
            }
         }
         
      } catch( Exception e ) { e.printStackTrace(); }

   }

   /**
    * Tells if the given class has to be adapted.<p>
    *
    * To know this, it loads 
    
   /**
    * Do the translation and update the class file on the file system.
    *
    * @param to_translate the name of the class file to be
    * translated.  */

   public static void translateClassFile(String to_translate)
      throws Exception {
      CtClass to_adapt = new CtClass(to_translate);
      to_adapt = translateClass(to_adapt, false, false, false, false);
      to_adapt.compile();
   }

   /**
    * Tell wether a class is to be adapted
    *
    * @param className a <code>String</code> value
    * @return a <code>boolean</code> value
    */
   public static boolean classIsAdapted(String className) {
      int index = className.lastIndexOf(".");
      if ( (! classesToAdapt.contains(className) && 
            ! (index!=-1 && packagesToAdapt.contains(className.substring(0,index))) ) && 
           ( className.endsWith( ".Run" ) || 
             className.endsWith( ".Main" ) ||
             className.endsWith( "AC" ) ||
             className.endsWith( "Wrapper" ) ||
             (className.indexOf('$') != -1) ||
             className.endsWith( "Exception" ) ||
             className.startsWith( "jac.core" ) ||
             className.startsWith( "jac.util" ) ||
             className.startsWith( "jac.aspects" ) ||
             className.startsWith( "java." ) ||
             className.startsWith( "javax." ) ||
             className.startsWith( "org." ) ||
             classesToNotAdapt.contains(className) ||
             (index!=-1 && packagesToNotAdapt.contains(className.substring(0,index)))
           ) ) return false; 
      else {
         return true;
      }
   }

   /**
    * Wrapper for the bytecode translation with <code>force=false</code>.
    *
    * @param to_adapt  the class to be translated
    * @param verbose verbose translation if true
    * @param debug more verbose if true
    * @param write write the translated class on disk if true
    * @param clean clean the set of writen classes
    * @return the translated Javassist class.
    * @see #translateClass(CtClass,boolean,boolean,boolean,boolean)
    */  
   public static CtClass translateClass(CtClass to_adapt,
                                        boolean verbose,
                                        boolean debug, 
                                        boolean write,
                                        boolean clean ) throws Exception {

      return translateClass( to_adapt, verbose, debug, write, clean, false );
   }   

   /**
    * Do the bytecode translation.
    *
    * <p>This method modifies specified class files so that they
    * represent wrappable classes.  All the objects that are loaded
    * implement the JacObject class and their instances can be
    * wrapped by a wrapper instance.
    * 
    * <p>If the original class file looks like:
    *
    * <ul><pre>
    * class Person {
    *   public int f(int i) { return i + 1; }
    * }
    * </pre></ul>
    *
    * <p>It will be translated into a class like:
    *
    * <ul><pre>
    * class Person extends JacObject {
    *   protected Vector _f_method_name;
    *   public int _org_f(int i) { return i + 1; }
    *   // used to start the wrapping chain
    *   // the templated code of f is defined in WrappeeTemplate
    *   public Object _nextWrapper_f( Object[] args, int rank ) {...}
    *   public int f(int i) { <i>wraps _nextWrapper_f</i> }
    * }
    * </pre></ul>
    *
    * <p>To optimize and simplify the translation, this method uses
    * templated bytecodes pieces extracted from
    * <code>WrappeeTemplate</code>.
    *
    * @param to_adapt  the class to be translated
    * @param verbose verbose translation if true
    * @param debug more verbose if true
    * @param write write the translated class on disk if true
    * @param clean clean the set of writen classes
    * @param force force the translation despite the jac.prop rules
    * @return the translated Javassist class.
    * @see WrappeeTemplate
    * @see WrappeeTemplate#_nextWrapper_FNAME(Object[],int)  */

   public static CtClass translateClass(CtClass to_adapt,
                                        boolean verbose,
                                        boolean debug, 
                                        boolean write,
                                        boolean clean, 
                                        boolean force ) throws Exception {

      if( debug ) System.out.println( "loading: " + to_adapt.getName() );

      String className = to_adapt.getName();

      /**
       * First, check whether the class is to be adapted,
       * next, whether it has been already adapted.
       */

      if ( to_adapt.isInterface() ) {
         if( debug ) System.out.println( 
            "translating: interface "+to_adapt.getName()+" not adapted" );
         return to_adapt;
      }

      String name = to_adapt.getName();

      if ( ! force ) {
         if (!classIsAdapted(name)) {
            if( debug ) 
               System.out.println("translating: "+className+" not adapted");
            return to_adapt;
         } else {
            if( debug ) 
               System.out.println("translating: "+className+" is to be adapted");
         }
      }
      
      File fclass = new File(getTempFile(to_adapt));
      File fOrigClass = new File(Jac.getJacRoot()+"classes/"+
                                 to_adapt.getName().replace('.','/')+".class");
      if (fclass.exists() && fclass.lastModified()>fOrigClass.lastModified()) 
         return to_adapt;
	
      /** None of the above mentionned cases are verified. Adapt the class. */

      if( verbose || debug) System.out.println( "translating: class " + to_adapt.getName() +
                                                " -> " + fclass);
      methodCounts = new Hashtable();
      
      try {
         to_adapt.getDeclaredField("__JAC_TRANSLATED");
         if( verbose ) System.out.println( "translating: class " + to_adapt.getName() +" already adapted");
         return to_adapt;
      } catch (Exception e) {}

      CtClass meta_wrappee = new CtClass(templateString);
      CtClass rootclass = new CtClass(jacObjectString);
      CodeConverter cc;
      CtNewMethod m;
      String orgname=null;
      boolean multiple_prototypes = false;
      String cur_method_name;

      CtClass objectclass = new CtClass("java.lang.Object");
      CtMethod inobject=null;

      CtClass superclass = to_adapt.getSuperclass();
      CtClass prev = to_adapt;
      while (!superclass.getName().equals("java.lang.Object") &&
             !superclass.getName().equals(jacObjectString)) {
         prev = superclass;
         superclass = superclass.getSuperclass();
         if( !prev.getName().equals(to_adapt.getName()) ) {
            if( debug ) System.out.println( "loading: superclass translation of " + prev.getName() );
            translateClass( prev, verbose, debug, true, clean );
         }
      }
      if(superclass.getName().equals("java.lang.Object")) {
         prev.setSuperclass(rootclass);
      }

      to_adapt = new CtClass(to_adapt.getInputStream());    // intermediate class file

      CtNewField f;
	
      CtMethod[] methods = to_adapt.getDeclaredMethods();

      methodCounts.clear();
      
      for(int i=0; i<methods.length; i++) {

         int mod = methods[i].getModifiers();
         if( (!Modifier.isPublic(mod)) || Modifier.isAbstract(mod) ||
             Modifier.isInterface(mod) || Modifier.isStatic(mod) ) continue;

         if(methods[i].getName().equals("toString")||
            methods[i].getName().equals("clone")) continue;

         if(getMethodCount(methods[i].getName()) == 0) {
            multiple_prototypes = false;
            cur_method_name = methods[i].getName();
         } else {
            multiple_prototypes = true;
            cur_method_name = methods[i].getName() + getMethodCount(methods[i].getName());
            if(debug) System.out.println("translating: multiple prototypes found for " + cur_method_name );
         }
         incrMethodCount(methods[i].getName());

         // Creates some new fields
         if (!multiple_prototypes) {
         
            f = new CtNewField(Modifier.STATIC,
                               CtClass.forName("java.lang.String"),
                               "_" + cur_method_name + "_method_name");
            f.setInitializer(FieldInitializer.constant(methods[i].getName()));
            f.bePublic();
            to_adapt.addField(f);
           
            // Creates one _nextWrapper method per method
            m = CtNewMethod.copy(
               meta_wrappee.getDeclaredMethod("_nextWrapper_FNAME"),
               "_nextWrapper_" + cur_method_name,
               null
            );
            to_adapt.addMethod(m);

         }

         if(debug) System.out.println("translating: method " + methods[i].getName() );

         methods[i].setName("_org_" + methods[i].getName());

      }
      /** Constructor translation does not work... waiting for BCEL version...
      CtConstructor[] constructors = to_adapt.getDeclaredConstructors();

      for(int i=0; i<constructors.length; i++) {

         int mod = constructors[i].getModifiers();
         if( Modifier.isPrivate(mod) || Modifier.isAbstract(mod) ||
             Modifier.isInterface(mod) || Modifier.isProtected(mod) || Modifier.isStatic(mod) ) continue;

         if(debug) System.out.println("translating: constructor " + constructors[i].getName() );

         CtNewConstructor newConstructor = CtNewConstructor.make( 
            constructors[i].getParameterTypes(),
            constructors[i].getExceptionTypes(),
            CtNewConstructor.PASS_ARRAY,
            constructors[i],
            null );                                

         constructors[i].setName("_orgc_" + constructors[i].getName());

         int paramsCount = constructors[i].getParameterTypes().length;
         
         constructors[i].setBody( meta_wrappee.getDeclaredMethod("emptyMethod"+paramsCount), null );

         to_adapt.addConstructor( newConstructor );

      }
      */

      to_adapt = new CtClass(to_adapt.getInputStream());    // intermediate class file
      methods = to_adapt.getDeclaredMethods();

	
      for( int i=0 ; i < methods.length ; i++ ) {

         int mod = methods[i].getModifiers();
         if( Modifier.isPrivate(mod) || Modifier.isAbstract(mod) ||
             Modifier.isInterface(mod) || Modifier.isProtected(mod) ||
             Modifier.isStatic(mod) ) continue;

         // Convert the _nextWrapper_ methods to work on the right fields...

         if (methods[i].getName().startsWith("_nextWrapper_",0)) {
            if(debug) System.out.println( "translating: converting the body of " +
                                          methods[i].getName() );
            cc = new CodeConverter();
            orgname = methods[i].getName().substring(13);

            cc.redirectFieldAccess(
               to_adapt, "_FNAME_method_name",
               to_adapt, "_" + orgname + "_method_name"
            );
            /*
            cc.redirectMethodCall(
               to_adapt, "myNextWrapper",
               rootclass, "myNextWrapper"
            );
            */
            methods[i].instrument(cc);
         }

         // Create the stubs...

         if (methods[i].getName().startsWith("_org_",0)) {

            if(debug) System.out.println( "translating: creating the stub for " +
                                          methods[i].getName() );

            orgname = methods[i].getName().substring(5);

            m = CtNewMethod.makeWrapped(
               methods[i].getReturnType(),
               orgname,
               methods[i].getParameterTypes(),
               methods[i].getExceptionTypes(),
               to_adapt.getDeclaredMethod("_nextWrapper_" + orgname),
               ConstParameter.integer(0)
            );

            to_adapt.addMethod(m);
         }
      }

      f = new CtNewField(CtClass.booleanType, "__JAC_TRANSLATED");
      to_adapt.addField(f);

      CtConstructor[] cs = to_adapt.getConstructors();
      boolean exist = false;
      for (int i = 0; i < cs.length; i++) {
         if (cs[i].getParameterTypes().length == 0) {
            exist = true;
         }
      }
      if (!exist) {
         if(debug||verbose) System.out.println( "translating: creating default constructor" );
         to_adapt.addConstructor( CtNewConstructor.makeDefault() );
         f = new CtNewField(CtClass.booleanType, "__JAC_ADDED_DEFAULT_CONSTRUCTOR");
         f.beStatic();
         f.bePublic();
         to_adapt.addField(f);
      }

      if(write) {
         if(debug||verbose) System.out.println( "translating: writing class on disk" );
         to_adapt.compile( getTempDirectory( to_adapt ) );
      }

      /** Register the adapted class */
      
      if(debug) System.out.println("translating: <<< " + to_adapt.getName() + " successfully translated >>>");

      return to_adapt;
   }
   

   /**
    * When translating a class, this method is used to count the
    * methods of a given name and know if it is a multiprototype
    * method (homonyms).
    * 
    * @param name the name of the method to increment the occurence */

   static void incrMethodCount( String name ) {

      if(methodCounts.get(name) != null) {
         Integer count = (Integer)methodCounts.get(name);
         methodCounts.remove(name);
         methodCounts.put ( name, new Integer (count.intValue() + 1) );
      } else {
         methodCounts.put ( name, new Integer (1) );
      }

   }


   /**
    * Returns the current count for a given method name.
    *
    * @param name the name of the method
    * @return the number of occurences it has been encountered for the
    * moment */

   static int getMethodCount( String name ) {

      if(methodCounts.containsKey(name)) {
         return ((Integer)methodCounts.get(name)).intValue();
      }
      return 0;
   }

   /**
    * This method returns the directory where the .class file of the
    * translated <code>c</code> class is stored when the
    * <code>-w</code> option is activated.
    *
    * @param c the Javassist class
    * @return its relative location in the disk cache */

   static String getTempDirectory (CtClass c) {
       String pn = c.getPackagename();
       if (pn == null) {
	   pn = "";
       } else {
	   pn = pn.replace('.','/');
       }
       return Jac.getClassesTmp() + pn;
   }
   
   static String getTempFile(CtClass c) {
      return getTempDirectory(c)+"/"+c.getSimpleName()+".class";
   }
}
