package jac.core.parsers.acc;

import java_cup.runtime.Scanner;
import java_cup.runtime.Symbol;
import java.io.*;
import java.net.*;

import jac.util.Log;

public class AccScanner implements Scanner {
   PushbackReader input;
   String streamName;
   int line;
   boolean bol; // beginning of line
   boolean previousBol; // beginning of line

   AccScanner include = null;

   public AccScanner(Reader input, String streamName) {
      this.input = new PushbackReader(input,2);
      this.streamName = streamName;
      this.line = 1;
      this.bol = true;
      //System.out.println(streamName+":INIT");
   }

   public AccScanner(String streamName) throws IOException
   {
      InputStream inputStream = null;
      try {
         Log.trace("AccScanner","FileInputStream("+streamName+")");
         inputStream = new FileInputStream(streamName);
      } catch ( FileNotFoundException e ) {
         try {
            Log.trace("AccScanner","getResourceAsStream("+streamName+")");
            inputStream = ClassLoader.getSystemClassLoader().getResourceAsStream(streamName);
         } catch (Exception e2 ) {
            Log.trace("AccScanner","new URL("+streamName+")");
            try {
               URL url = new URL(streamName);
               inputStream = url.openStream();
            } catch (MalformedURLException e3) {
               e3.printStackTrace();
            }
         }
      }

      if (inputStream==null)
         throw new FileNotFoundException("file not found : "+streamName);

      this.input = new PushbackReader(new InputStreamReader(inputStream),2);
      this.streamName = streamName;
      this.line = 1;
      this.bol = true;

      //      System.out.println(streamName+":INIT");
   }

   public void printState() {
      Log.error(getLine());
      if (include!=null)
         include.printState();
   }

   public Symbol next_token() throws java.lang.Exception
   {
      if (include!=null) {
         Symbol token = include.next_token();
         if (token.sym != AccSymbols.EOF) {
            return token;
         }
         include = null;
      }

      boolean quoted = false;
      int c = input.read();
      while (true) {
         // skip white spaces and CR
         while (c==' ' || c=='\t' || c=='\n' || c=='\r') {
            //System.out.println(streamName+":"+line+":WHITESPACE");
            if (c=='\n' || c=='\r') {
               //System.out.println(streamName+":"+line+":NEWLINE1");
               line++;
               bol = true;
               if (c=='\r') {
                  c = input.read();
                  if (c != '\n')
                     input.unread(c);
               }
            }
            c = input.read();
         }
         // skip comments
         if (c=='/') {
            int ahead = input.read();
            if (ahead=='/') {
               // c++-style comments
               //System.out.println(streamName+":"+line+":C++ COMMENT");
               c = input.read();
               while (c!='\n' && c!='\r' && c!=-1) {
                  c = input.read();
               }
               continue;
            } if (ahead=='*') {
               // c-style comments
               //System.out.println(streamName+":"+line+":C COMMENT");
               int prev = c;
               c = input.read();
               boolean state = false;
               while (!(state && c=='/') && c!=-1) {
                  if ((c=='\n' && prev!='\r') || c=='\r') line++;
                  state = (c=='*');
                  prev = c;
                  c = input.read();
               }
               if (c==-1) {
                  Log.error("unclosed comment");
                  printState();
                  break;
               }
               c = input.read();
               continue;
               
            } else {
               input.unread(ahead);
               break;
            }
         }
         break;
      }

      Symbol token = new Symbol(AccSymbols.error);

      //System.out.println("Testing char `"+c+"'");
      switch (c) {
         case -1: 
         case 65535:
            token = new Symbol(AccSymbols.EOF);
            //System.out.println(streamName+":TOKEN "+token.sym+"["+token.value+"]");
            break;
         case '{':
            token = new Symbol(AccSymbols.LBRACE,"{");
            break;
         case '}':
            token = new Symbol(AccSymbols.RBRACE,"}");
            break;
         case ',':
            token = new Symbol(AccSymbols.COMMA,",");
            break;
            // when -1 is unread(), it becomes \uFFFF
         case ';':
            token = new Symbol(AccSymbols.EOL,";");
            break;
         case '"':
            // quoted string
            {
               quoted = true;
               StringWriter value = new StringWriter();
               c = input.read();
               boolean escaped = false;
               if (c=='\\') {
                  escaped = true;
                  c=readEscaped(input.read());
               }
               while(c!='"' || escaped) {
                  value.write(c);
                  c = input.read();
                  if (c=='\n')
                     line++;
                  escaped = false;
                  if (c=='\\') {
                     c=readEscaped(input.read());
                     escaped = true;
                  }
               }
               //System.out.println("value="+value);
               token = new Symbol(AccSymbols.ATOMIC_VALUE,value.toString());
            }
            break;
         default:
            {
               StringWriter value = new StringWriter();
               String delim="{},; \t\n\r";
               while(delim.indexOf(c)==-1 && c!=-1) {
                  value.write(c);
                  c = input.read();
               }
               input.unread(c);
               //System.out.println("value="+value);
               token = new Symbol(AccSymbols.ATOMIC_VALUE,value.toString());
            }
      }

      previousBol = bol;

      if (bol && token.sym == AccSymbols.ATOMIC_VALUE && token.value.equals("include")) {
         bol = false;
         String filename = (String)next_token().value;
         //System.out.println(streamName+":INCLUDE("+next.value+")");
         include = new AccScanner(filename);
         return next_token();
      }
      if (!quoted && token.sym == AccSymbols.ATOMIC_VALUE && 
          ( token.value.equals("class") || 
            token.value.equals("method") || 
            token.value.equals("attribute") || 
            token.value.equals("block") ) )
      {
         token.sym = AccSymbols.CLASS;
         //System.out.println(streamName+":TOKEN block");
      }
      //System.out.println(streamName+":TOKEN "+token.sym+"["+token.value+"]");
      Log.trace("parser","read token "+token+"  = "+token.value);
      bol = false;
      return token;
   }

   /**
    * Read a char from input, possibly escaped by '\'
    */
   int readEscaped(int c) throws IOException 
   {
      switch (c) {
         case 'n':
            return '\n';
         case 'r':
            return '\r';
         case 't':
            return '\t';
         default:
            return c;
      }
   }

   /**
    * Return the current line number
    */
   public String getLine() {
      if (include!=null)
         return include.getLine();
      else
         return streamName+":"+(previousBol?line-1:line);
   }

}

