/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;

/**
 * This class defines a meta item that corresponds to the
 * <code>java.lang.reflect.Method</code> and to the
 * <code>java.lang.reflect.Constructor</code> meta elements.
 *
 * <p>It appears that methods and contructors can be seen as
 * semantically similar (a constructor is nothing else that a special
 * static method that returns an object that is of the owner class
 * type). Thus, we decide to wrap those two meta item into a unique
 * one (on contrary to the Java choice)<p>
 *
 * @author Laurent Martelli
 * @author Renaud Pawlak
 * @version 0.8.1 */

public abstract class AbstractMethodItem extends MetaItemDelegate {
   
   /**
    * Default contructor to create a new abstract method item.<p>
    *
    * @param delegate the <code>java.lang.reflect</code> actual
    * meta item */

   public AbstractMethodItem ( Object delegate ) throws InvalidDelegateException {
      super(delegate);
   }

   /**
    * Gets the parameter types of this abstract method item.<p>
    *
    * @return the actual method parameter types
    */

   public abstract Class[] getParameterTypes();

   /**
    * Gets the number of parameters
    *
    * @return the number of parameters.
    */
   public int getParameterCount() {
      return getParameterTypes().length;
   }
   
   /**
    * Return the full method name, ie with parameter types.
    *
    * @return The full method name
    */

   public String getFullName() {
      //.substring(0, getName().lastIndexOf('.'))
      String ret = getName() + "(";
      Class[] pts = getParameterTypes();
      for ( int i = 0; i < pts.length; i++ ) {
         ret = ret + NamingConventions.getShortClassName( pts[i] );
         if ( i < pts.length - 1 ) ret = ret + ",";
      }
      ret = ret + ")";
      return ret;
   }

}// MethodItem
