/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */


package jac.core.rtti;

import java.util.*;
import java.lang.reflect.*;
import jac.util.*;
import gnu.regexp.*;

/**
 * This class defines a meta item that corresponds to the
 * <code>java.lang.reflect.Class</code> meta element.<p>
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1
 */

public class ClassItem extends MetaItemDelegate {

   static Class jacObjectClass = ClassRepository.jacObjectClass;
   static Class wrappeeClass = ClassRepository.wrappeeClass;

   /**
    * Default contructor to create a new class item object.<p>
    *
    * @param delegate the <code>java.lang.reflect.Class</code> actual
    * meta item */

   public ClassItem ( Class delegate ) throws InvalidDelegateException {
      super( delegate );
   }

   private int collsCount = 0;
   private int refsCount = 0;
   private int primsCount = 0;
   private int methodsCount = 0;
   private int constructorsCount = 0;

   /**
    * Stores the fields of the class.<p> */

   protected Hashtable fields = new Hashtable();

   /**
    * Stores the methods of the class. */

   protected Hashtable methods = new Hashtable();

   /**
    * Gets a field from its name. Throw an exception if the field does
    * not exist.<p>
    *
    * @param name the field name
    * @return a <code>FieldItem</code> instance 
    * @see #getFieldNoError
    */

   public FieldItem getField( String name ) {
      FieldItem ret = (FieldItem) fields.get(name);
      if (ret==null)
         throw new NoSuchFieldException(this,name);
      return ret;
   }

   /**
    * Gets a field from its name. Don't throw an exception if the
    * field does not exist.<p>
    *
    * @param name the field name
    * @return a <code>FieldItem</code> instance, or null if no such
    * field exists in the class 
    * @see #getField(String)
    */
   public FieldItem getFieldNoError(String name) {
      return (FieldItem) fields.get(name);
   }

   /**
    * Gets a collection from its name.<p>
    *
    * @param name the collection name
    * @return a <code>CollectionItem</code> instance
    */

   public CollectionItem getCollection( String name ) {
      return (CollectionItem) fields.get(name);
   }

   /**
    * Gets all the fields for this class item.<p>
    *
    * @return an array of field items
    */

   public FieldItem[] getFields() {
      Collection c = fields.values();
      FieldItem[] res = new FieldItem[c.size()];
      Iterator it = c.iterator();
      int i = 0;
      while ( it.hasNext() ) {
         res[i] = (FieldItem) it.next();
         i++;
      }
      return res;
   }

   /**
    * Gets all the primitive fields for this class item.<p>
    *
    * @return an array of field items
    */

   public FieldItem[] getPrimitiveFields() {
      Collection c = fields.values();
      FieldItem[] res = new FieldItem[primsCount];
      Iterator it = c.iterator();
      int i = 0;
      while ( it.hasNext() ) {
         FieldItem field = (FieldItem) it.next();
         if (field.isPrimitive()) {
            res[i] = field;
            i++;
         }
      }
      return res;
   }

   /**
    * Gets all the references for this class item.<p>
    *
    * @return an array of field items
    */

   public FieldItem[] getReferences() {
      Collection c = fields.values();
      FieldItem[] res = new FieldItem[refsCount];
      Iterator it = c.iterator();
      int i = 0;
      while ( it.hasNext() ) {
         FieldItem field = (FieldItem) it.next();
         if (field.isReference()) {
            res[i] = field;
            i++;
         }
      }
      return res;
   }

   /**
    * Gets all the references and collections that match the
    * expression for this class item.<p>
    *
    * @return a vector of field items */

   public Collection getMatchingRelations( String expr ) {
      Vector res = new Vector();
      Collection c = fields.values();
      Iterator it = c.iterator();
      RE re;
      try {
         re = new RE( expr );
      } catch ( Exception e ) {
         e.printStackTrace();
         return null;
      }
      while ( it.hasNext() ) {
         FieldItem field = (FieldItem) it.next();
         if (field.isReference() || (field instanceof CollectionItem)) {
            if( re.isMatch( field.getName() ) ) {
               res.add( field );
            }
         }
      }
      return res;
   }

   /**
    * Gets all the collections for this class item.<p>
    *
    * @return an array of field items
    */

   public CollectionItem[] getCollections() {
      Collection c = fields.values();
      CollectionItem[] res = new CollectionItem[collsCount];
      Iterator it = c.iterator();
      int i = 0;
      while ( it.hasNext() ) {
         FieldItem field = (FieldItem) it.next();
         if ( field instanceof CollectionItem ) {
            res[i] = (CollectionItem) field;
            i++;
         }
      }
      return res;
   }


   /**
    * Add a field item to this class item.<p>
    *
    * @param field the new field
    */
 
   void addField( FieldItem field ) {
      if ( ! fields.containsKey( field.getName() ) ) {
         try {
            field.setParent( this );
         } catch( Exception e ) {
            e.printStackTrace();
            System.exit( -1 );
         }
         fields.put( field.getName(), field );
         if ( field instanceof CollectionItem ) {
            collsCount++;
         } else if ( wrappeeClass.isAssignableFrom( field.getActualField().getType() ) ) {
            refsCount++;
         } else {
            primsCount++;
         }  
      }
   }
   
   /**
    * Gets a set of homonym methods (including constructors) from
    * their names.<p>
    *
    * @param name the name of the methods
    * @return a <code>MethodItem</code> instance */

   public AbstractMethodItem[] getAbstractMethods(String name) 
      throws NoSuchMethodException
   {
      Log.trace("rtti.method","getAbstractMethods("+name+")");
      AbstractMethodItem[] res=null;
      if (name.endsWith(")")) {
         String name2 = name.substring(0,name.indexOf("("));
         Log.trace("rtti.method","name2 : "+name2);
         AbstractMethodItem[] meths = 
            (AbstractMethodItem[])methods.get(name2);
         if (meths!=null) {
            for (int i=0; i<meths.length; i++) {
               Log.trace("rtti.method","trying "+meths[i].getFullName());
               if (meths[i].getFullName().endsWith(name)) {
                  res = new AbstractMethodItem[] {meths[i]};
               }
            }
         }
      } else {
         res = (AbstractMethodItem[])methods.get(name);
      }
      if (res == null) {
         throw new NoSuchMethodException(
            "ClassItem.getAbstractMethods: no such method "+name+" in "+this);
      }
      if (res.length>1) {
         Log.warning("rtti","Ambiguous method name "+this+"."+name);
         new AmbiguousMethodNameException().printStackTrace();
      }
      return res;
   }

   /**
    * Gets an abstract method from its name.
    *
    * <p>If this method has homonym(s), then an
    * <code>AmbiguousMethodNameException</code> is thrown.
    *
    * <p>An abstract method can be a static, an instance method or a
    * constructor.
    *
    * @param name the name of the method to search
    * @return the corresponding method if found
    * @see #getMethod(String)
    * @see #getAbstractMethods(String) */


   public AbstractMethodItem getAbstractMethod(String name) 
      throws NoSuchMethodException, AmbiguousMethodNameException
   {
      AbstractMethodItem[] res=getAbstractMethods(name);      
      if (res.length>1) {
         throw new NoSuchMethodException("Ambiguous method name "+this+"."+name);
      }
      return res[0];
   }

   /**
    * Strip the arguments part from a full method name.
    * aMethod(aType) -> aMethod
    * aMethod -> aMethod
    */
   String stripArgs(String name) {
      int index = name.indexOf("(");
      if (index!=-1)
         return name.substring(0,index);
      else
         return name;
   }

   /**
    * Gets a set of homonym methods (excluding constructors) from
    * their names.<p>
    *
    * @param name the name of the methods
    * @return a <code>MethodItem</code> instance */

   public MethodItem[] getMethods(String name) 
      throws NoSuchMethodException
   {
      MethodItem[] res = null;

      Log.trace("rtti.method","getMethods("+name+")");
      if (name.endsWith(")")) {
         String name2 = name.substring(0,name.indexOf("("));
         Log.trace("rtti.method","name2 : "+name2);
         MethodItem[] meths = (MethodItem[])methods.get(name2);
         for (int i=0; i<meths.length; i++) {
            Log.trace("rtti.method","trying "+meths[i].getFullName());
            if (meths[i].getFullName().endsWith(name)) {
               res = new MethodItem[] {meths[i]};
            }
         }
      } else {
         res = (MethodItem[])methods.get(name);
      }

      if (res == null) {
         Log.trace("rtti.method",2,"methods = "+methods);
         throw new NoSuchMethodException(
            "ClassItem.getMethods: no such method "+name+" in class "+this);
      }
      return res;
   }

   /**
    * Gets a method from its name.
    *
    * <p>If this method has homonym(s), then an
    * <code>AmbiguousMethodNameException</code> is thrown.
    *
    * <p>A method can be a static or instance method but not a
    * constructor.
    *
    * @param name the name of the method to search
    * @return the corresponding method if found
    * @see #getAbstractMethod(String)
    * @see #getMethods(String) */

   public MethodItem getMethod(String name) 
      throws NoSuchMethodException, AmbiguousMethodNameException
   {
      MethodItem[] res= getMethods(name);
      if (res.length>1) {
         throw new AmbiguousMethodNameException(
            "Ambiguous method name "+this+"."+name+" choice is:"
            +Arrays.asList(res));
      }
      return res[0];
   }

   /**
    * Gets a set of arrays containing all the method items for this
    * class item.
    *
    * <p>Each arrays contains the methods of the same name.
    *
    * @return a collection containing the methods */

   public Vector getMethods() {
      Collection c = methods.values();
      Vector res = new Vector();
      Iterator it = c.iterator();
      while ( it.hasNext() ) {
         Object[] methods = (Object[]) it.next();
         if ( methods[0] instanceof MethodItem ) {
            res.add(methods);
         }
      }
      return res;
   }

   /**
    * Gets all the method items for this class item.<p>
    *
    * @return a collection containing the methods
    */

   public Vector getAllMethods() {
      Collection c = methods.values();
      Vector res = new Vector();
      Iterator it = c.iterator();
      while ( it.hasNext() ) {
         Object[] methods = (Object[]) it.next();
         if ( methods[0] instanceof MethodItem && 
              !ClassRepository.isJacMethod(
                 ((MethodItem)methods[0]).getName()) ) {
            res.addAll(Arrays.asList(methods));
         }
      }
      return res;
   }

   /**
    * Gets all the method items that modify the state of the instances
    * for this class item.
    *
    * @return a collection containing the modifiers */

   public Vector getAllModifiers() {

      Vector methods = getAllMethods();
      Iterator it = methods.iterator();
      Vector modifiers = new Vector();

      while(it.hasNext()) {
         MethodItem method = (MethodItem)it.next();

         if ( method.isStateModifier() ) {
            modifiers.add(method);
            continue;
         }

         FieldItem[] writtenFields = method.getWrittenFields();
         if ( writtenFields != null ) {
            modifiers.add(method);
            continue;
         }

         CollectionItem[] addedCollections = method.getAddedCollections();
         if ( addedCollections != null ) {
            modifiers.add(method);
            continue;
         } 

         CollectionItem[] removedCollections = method.getRemovedCollections();
         if ( removedCollections != null ) {
            modifiers.add(method);
            continue;
         }
      }
      return modifiers;
   }

   /**
    * Gets all the method items that access the state of the instances
    * for this class item.
    *
    * @return a collection containing the accessors */

   public Vector getAllAccessors() {

      Vector methods = getAllMethods();
      Iterator it = methods.iterator();
      Vector accessors = new Vector();

      while(it.hasNext()) {
         MethodItem method = (MethodItem)it.next();

         if ( method.hasAccessedReferences() ) {
            accessors.add(method);
            continue;
         }

         FieldItem[] accessedFields = method.getAccessedFields();
         if ( accessedFields != null ) {
            accessors.add(method);
            continue;
         }
      }
      return accessors;
   }


   /**
    * Gets all the method items that removes an object from a
    * collection of this class item.
    *
    * @return a collection containing the modifiers */

   public Vector getAllRemovers() {

      Vector methods = getAllMethods();
      Iterator it = methods.iterator();
      Vector removers = new Vector();

      while(it.hasNext()) {
         MethodItem method = (MethodItem)it.next();

         if ( method.getRemovedCollections() != null ) {
            removers.add(method);
            continue;
         }
      }
      return removers;
   }

   /**
    * Gets all the method items that adds an object from a
    * collection of this class item.
    *
    * @return a collection containing the modifiers */

   public Vector getAllAdders() {

      Vector methods = getAllMethods();
      Iterator it = methods.iterator();
      Vector adders = new Vector();

      while(it.hasNext()) {
         MethodItem method = (MethodItem)it.next();

         if ( method.getAddedCollections() != null ) {
            adders.add(method);
            continue;
         }
      }
      return adders;
   }

   /**
    * Gets all the constructors items for this class item.<p>
    *
    * @return a collection containing the constructors */

   public Vector getConstructors() {
      Collection c = methods.values();
      Vector res = new Vector();
      Iterator it = c.iterator();
      while ( it.hasNext() ) {
         Object[] methods = (Object[]) it.next();
         if ( methods[0] instanceof ConstructorItem ) {
            res.addAll(Arrays.asList(methods));
         }
      }
      return res;
   }

   public ConstructorItem getConstructor(Class[] parameterTypes) 
      throws NoSuchMethodException
   {
      Log.trace("rtti.constructor","getConstructor "+Arrays.asList(parameterTypes));
      Vector constructors = getConstructors();
      for (int i=0; i<constructors.size(); i++) {
         ConstructorItem constructor = (ConstructorItem)constructors.get(i);
         Log.trace("rtti.constructor","comparing with "+Arrays.asList(constructor.getParameterTypes()));
         if (Arrays.equals(constructor.getParameterTypes(),parameterTypes)) {
            return constructor;
         }
      }
      return null;
      /*
      throw new NoSuchMethodException("No such constructor for class "+getName()+
                                      " with parameters "+Arrays.asList(parameterTypes));
      */
   }

   /**
    * Gets the constructor item of this class item that matches the
    * given <code>java.lang.reflect.Constructor</code>.
    *
    * @return the corresponding constructor item */

   public ConstructorItem getConstructor(Constructor constructor) {
      Log.trace("rtti","getConstructor("+constructor+")");
      Collection constructors = getConstructors();
      Iterator it = constructors.iterator();
      while ( it.hasNext() ) {
         ConstructorItem[] current = (ConstructorItem[])it.next();
         for (int i=0; i<current.length; i++) {
            Log.trace("rtti","compare with "+current[i]);
            if (current[i].getActualConstructor().toString().equals(constructor.toString())) {
               return current[i];
            }
         }
      }
      return null;
   }

   /**
    * Creates a new instance of this class item by using the default
    * constructor (the one with no parameters).
    *
    * @return the newly created instance */

   public Object newInstance() 
      throws InstantiationException, IllegalAccessException {
      return getActualClass().newInstance();
   }

   /**
    * Creates a new instance of this class item by using the 
    * constructor that matches the given parameter types.
    *
    * @param types the types of the constructor arguments
    * @param values the arguments values
    * @return the newly created instance */

   public Object newInstance(Class[] types,Object[] values) 
      throws InstantiationException, IllegalAccessException, 
             java.lang.NoSuchMethodException, InvocationTargetException {
      Constructor c = getActualClass().getConstructor(types);
      return c.newInstance(values);
   }

   /**
    * Gets the class represented by this class item.<p>
    *
    * @return the actual class
    * @see #getType()
    */

   public Class getActualClass() {
      return (Class) delegate;
   }

   /**
    * Gets the superclass of this class item.<p>
    *
    * @return the superclass, null if the superclass is Object (or
    * JacObject) */

   public ClassItem getSuperclass() {
      if(getActualClass().getSuperclass()==Object.class
         || getActualClass().getSuperclass()==ClassRepository.jacObjectClass) 
         return null;
      return ClassRepository.get().getClass(getActualClass().getSuperclass());
   }

   public int getModifiers() {
      return ((Class)delegate).getModifiers();
   }

   /**
    * Synonym of <code>getActualClass</code>.<p>
    *
    * @return the actual class
    * @see #getActualClass()
    */

   public Class getType() {
      return getActualClass();
   }

   /**
    * Add a method item to this class item.<p>
    *
    * @param method the new method
    */
 
   public void addMethod( MethodItem method ) {
      String name = method.getName();
      Log.trace("rtti.method","addMethod: "+name+" -> "+getName()+"."+method);
      if ( ! methods.containsKey( name ) ) {
         methods.put( method.getName(), new MethodItem[] { method } );
      } else {
         MethodItem[] meths = (MethodItem[]) methods.get( name );
         MethodItem[] newMeths = new MethodItem[meths.length + 1];
         System.arraycopy( meths, 0, newMeths, 0, meths.length);
         newMeths[meths.length] = method;
         methods.remove( name );
         methods.put( name, newMeths );
      }
      methodsCount++;
      try {
         method.setParent( this );
      } catch( Exception e ) {
         e.printStackTrace();
         System.exit( -1 );
      }
   }   
   
   /**
    * Add a constructor item to this class item.<p>
    *
    * @param constructor the new constructor
    */
 
   public void addConstructor( ConstructorItem constructor ) {
      String name = NamingConventions.getShortConstructorName(constructor.getActualConstructor());
      Log.trace("rtti.method","addConstructor: "+name+" -> "+constructor);
      if ( ! methods.containsKey( name )) {
         methods.put( name, new ConstructorItem[] { constructor } );
      } else {
         ConstructorItem[] constructors = 
            (ConstructorItem[]) methods.get( name );
         ConstructorItem[] newConstructors = 
            new ConstructorItem[constructors.length + 1];
         System.arraycopy( constructors, 0, newConstructors, 0, constructors.length);
         newConstructors[constructors.length] = constructor;
         methods.remove( constructor.getName() );
         methods.put( name, newConstructors );
      }
      constructorsCount++;
      try {
         constructor.setParent( this );
      } catch( Exception e ) {
         e.printStackTrace();
         System.exit( -1 );
      }
   }   
   
   /**
    * Tests if a method exist in this class item.<p>
    * 
    * @return true if exist
    */

   public boolean hasMethod( String name ) {
      return methods.containsKey( stripArgs(name) );
   }

   /**
    * Tests if a field exist in this class item.<p>
    * 
    * @return true if exist
    */

   public boolean hasField( String name ) {
      return (name!=null) && fields.containsKey( name );
   }

   public String getName() {
      return ((Class)delegate).getName();
   }

   /**
    * A shortcut to invoke a method (avoid to get the method item).
    *
    * @param object the object to perform the invocation on (may be
    * null for a static method)
    * @param methodName the name of the method to invoke
    * @param parameters the parameters passed to the method
    * @see MethodItem#invoke(Object,Object[]) */

   public Object invoke( Object object, String methodName, Object[] parameters )
      throws IllegalAccessException, InvocationTargetException {
      return getMethod( methodName ).invoke( object, parameters );
   }

   /** 
    * Tests if this class item corresponds to a Jac system class.
    *
    * <p>It simply checks that the package name does not start with
    * "jac.core" or "jac.aspects".
    *
    * @return true if this is a Jac system class */

   public boolean isSystemClass() {
      return getName().startsWith("jac.core") || getName().startsWith("jac.aspects");
   }

   /**
    * The exception that is thrown when the accessed method does not
    * exist. */
   public class NoSuchMethodException extends RuntimeException {
      public NoSuchMethodException(String msg) { super(msg); }
      //      public NoSuchMethodException() { super(); }
   }

   /**
    * The exception that is thrown when the accessed method has some
    * synonymes (methods with same names but different parameter
    * types). */
   public class AmbiguousMethodNameException extends RuntimeException {
      public AmbiguousMethodNameException(String msg) { super(msg); }
      public AmbiguousMethodNameException() { super(); }
   }
}
