/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import jac.util.*;
import java.util.*;
import java.lang.reflect.*;
import java.security.*;

/**
 * This class defines the class repository of the rtti aspect.<p>
 *
 * It contains class items that are field and method items
 * agregates.<p>
 *
 * @see ClassItem
 * @see MethodItem
 * @see FieldItem
 * @see CollectionItem
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1
 */

public class ClassRepository {

   /**
    * Get the sole instance of class repository.<p>
    * 
    * @return the class repository */
   
   public static ClassRepository get () {
      if ( classRepository == null ) 
         classRepository = new ClassRepository();
      return classRepository;
   }
   
   /**
    * Store the sole instance of class repository. */

   protected static transient ClassRepository classRepository = null;

   /** Stores the Jac objects root class. */
   public static Class jacObjectClass;
   /** Stores the Wrappee class. */
   public static Class wrappeeClass;
   
   static {
      try {
         jacObjectClass = Class.forName( "jac.core.JacObject" );
         wrappeeClass = Class.forName( "jac.core.Wrappee" );
      } catch( Exception e ) {
         e.printStackTrace();
         System.exit( -1 );
      }
   }  

   /**
    * Store direct access to the methods of all the classes by name on
    * optimization purpose. */
   transient static Hashtable directMethodAccess = new Hashtable();

   /**
    * Store direct access to the fields of all the classes by name on
    * optimization purpose. */
   transient static Hashtable directFieldAccess = new Hashtable();

   /**
    * Fill the direct access to methods hashtable for a given class.
    *
    * <p>This is used to optimize the method calls when you only got
    * the method name.
    *
    * @param cl the class to treat.
    *
    * @see #getDirectMethodAccess(Class,String) */

   public static void fillDirectMethodAccess( final Class cl ) {
      if( cl == null ) return;
      if ( directMethodAccess == null ) directMethodAccess = new Hashtable();
      if ( directMethodAccess.containsKey( cl ) ) return;
      final Hashtable methods = new Hashtable();
      AccessController.doPrivileged(
         new PrivilegedAction() {
               public Object run() {
                  Method[] meths = cl.getMethods();
                  AccessibleObject.setAccessible( meths, true );
                  for ( int i = meths.length - 1; i >= 0; i-- ) {
                     if ( methods.containsKey( meths[i].getName() ) ) {
                        Method[] oldms = (Method[]) methods.get( meths[i].getName() );
                        Method[] newms = new Method[oldms.length+1];
                        System.arraycopy( oldms, 0, newms, 0, oldms.length );
                        newms[oldms.length] = meths[i];
                        methods.remove( meths[i].getName() );
                        methods.put( meths[i].getName(), newms );
                     } else {
                        methods.put( meths[i].getName(), new Method[] { meths[i] } );
                     }
                  }
                  return null;
               }
            }
      );
      directMethodAccess.put( cl, methods );
   }

   /**
    * Fill the direct access to fields hashtable for a given class.
    *
    * <p>This is used to optimize the access of protected and private
    * fields.
    *
    * @param cl the class to treat. */

   public static void fillDirectFieldAccess( final Class cl ) {
      if( cl == null ) return;
      if ( directFieldAccess == null ) directFieldAccess = new Hashtable();
      if ( directFieldAccess.containsKey( cl ) ) return;
      final Hashtable fields = new Hashtable();
      AccessController.doPrivileged(
         new PrivilegedAction() {
               public Object run() {
                  Class curcl = cl;
                  while ( curcl != jacObjectClass && curcl != Object.class ) {
                     Field[] fs = curcl.getDeclaredFields();
                     AccessibleObject.setAccessible( fs, true );
                     for ( int i = 0; i < fs.length; i++ ) {
                        if ( (! Modifier.isStatic ( fs[i].getModifiers() )) &&
                             (! ( cl.isAssignableFrom( wrappeeClass ) && 
                                  isJacField ( fs[i].getName() ) )) &&
                             (! fs[i].getName().equals("__JAC_TRANSLATED"))  &&
                             (! fs[i].getName().equals("__JAC_ADDED_DEFAULT_CONSTRUCTOR")) ) {
                           fields.put( fs[i].getName(), fs[i] );
                        }
                     }
                     curcl = curcl.getSuperclass();
                  }
                  return null;
               }
            }
      );   
      directFieldAccess.put( cl, fields );
   }

   /**
    * Returns a Hashtable that maps fields with their names.
    *
    * <p>For efficiency, the programmer should use this method instead
    * of using the Java refection API to retrives values of protected
    * or private fields within a class and all its superclasses.
    * 
    * @param cl the class where the field is supposed to be
    * @param name the field name
    * @return an array containing the matching methods (homonyms)
    * 
    * @see #fillDirectFieldAccess(Class) */

   public static Hashtable getDirectFieldAccess( Class cl ) {
      fillDirectFieldAccess(cl);
      return ((Hashtable)directFieldAccess.get(cl));
   }

   /**
    * Call a directly acceded method.
    *
    * @param cl the class that supports the method
    * @param methodName the method to call
    * @param o the object to call the method on
    * @param params the arguments
    *
    * @see #getDirectMethodAccess(Class,String)
    */

   public static Object invokeDirect( Class cl, 
                                      String methodName, 
                                      Object o, 
                                      Object[] params ) 
      throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
      Method[] directMethods = getDirectMethodAccess( 
         cl, methodName );
      boolean ok = false;
      Object ret = null;
      for ( byte i = 0; i < directMethods.length; i++ ) {
         try {
            ret = directMethods[i].invoke( o, params );
            ok = true;
         } catch ( IllegalArgumentException e1 ) {
         }
      }
      if ( ! ok ) {
         throw new NoSuchMethodException();
      }
      return ret;
   }

   /**
    * Return an array of methods that correspond to the given class
    * and given method name.
    *
    * <p>For efficiency, the programmer should use this method instead
    * of using the Java refection API.
    * 
    * @param cl the class where the method is supposed to be
    * @param name the method name
    * @return an array containing the matching methods (homonyms), a
    * one array element with <code>null</code> inside if none method
    * matched (this ugly result was introduced for backward
    * compatibility)
    * 
    * @see #fillDirectMethodAccess(Class) */

   public static Method[] getDirectMethodAccess(Class cl, String name) {
      fillDirectMethodAccess(cl);
      Method[] ret = (Method[]) ((Hashtable)directMethodAccess.get(cl)).get(name);
      if (ret == null ) {
         return new Method[] { null };
      } else {
         return ret;
      }
   }

   /**
    * Returns true is this class defautlt constructor (the one with no
    * parameter) has been added by JAC at class load-time.
    *
    * @param cl the class to test
    * @return true if added, false if programmer defined */

   public static boolean isDefaultConstructorAdded( Class cl ) {
      try {
         cl.getField("__JAC_ADDED_DEFAULT_CONSTRUCTOR");
      } catch ( Exception e ) {
         return false;
      }
      return true;
   }

   /**
    * Return true if the method is added by JAC at class load-time.
    *
    * @param methodName the method to check
    * @return true if added at load-time */

   public static boolean isJacMethod (String methodName) {
      if ( ( !methodName.equals("toString") &&
             getDirectMethodAccess(jacObjectClass, methodName)[0] != null )|| 
           methodName.startsWith("_") )
         return true;
      return false;
   }

   /**
    * Return true if the field is added by JAC at class load-time. */

   public static boolean isJacField (String fieldName) {
      try {
         jacObjectClass.getDeclaredField (fieldName);
      } catch ( Exception e ) {
         if ( fieldName.startsWith("_") ) {
            return true;
         }
         return false;
      }
      return true;
   }

   /**
    * Return true if the field part of the JacObject class.
    *
    * @param fieldName the field to test */

   public static boolean isJacObjectField (String fieldName) {
      try {
         jacObjectClass.getDeclaredField (fieldName);
      } catch ( Exception e ) {
         return false;
      }
      return true;
   }

   /**
    * Get method names.
    *
    * @param src  the source object containing the methods
    * @return     the method names as an array of strings
    */
   
   public static String[] getMethodsName( Class cl ) {
   
      String[] methodNames = null;
      Vector tmp = new Vector();
      
      try {
      
         Method[] methods = cl.getMethods();

         for ( int i=0 ; i < methods.length ; i++ ) {
            if ( (! Modifier.isStatic ( methods[i].getModifiers() )) &&
                 (! isJacMethod ( methods[i].getName() )) ) {
               tmp.add ( methods[i].getName() );
            }
         }

         methodNames = new String[ tmp.size() ];

         for ( int i=0 ; i < tmp.size() ; i++ ) {
            methodNames[i] = (String) tmp.get ( i );
         }

      }
      catch( Exception e ) { e.printStackTrace(); }
      
      return methodNames;
   }

   /**
    * Get modifiers names.
    *
    * @param src  the source object containing the methods
    * @return     the modifiers names as an array of strings
    */
   
   public static String[] getModifiersNames( Class cl ) {
   
      ClassItem cli = ClassRepository.get().getClass( cl );
      Vector modifiers = cli.getAllModifiers();
      
      String[] methodNames = new String[modifiers.size()];

      for( int i=0; i<methodNames.length; i++ ) {
         methodNames[i]=((MethodItem)modifiers.get(i)).getName();
      }
      return methodNames;
   }

   /**
    * Adds a set of written fields to a method.<p>
    *
    * This method shortcuts the writting of RTTI aspects for
    * programs. Equivalent effects can be achieved by using the RTTI
    * aspect API.<p>
    *
    * @param cl the involved class
    * @param methodName the method name
    * @param fieldNames the set of fields that are written by this
    * method
    * @see MethodItem#addWrittenField(FieldItem)
    */

   public static void addWrittenFields( Class cl, 
                                        String methodName, 
                                        String[] fieldNames ) 
   {
      ClassItem cli = ((ClassRepository)ClassRepository.get()).getClass( cl.getName() );
      MethodItem[] mis = cli.getMethods( methodName );

      if ( mis != null ) {
         for ( int i = 0; i < mis.length; i++ ) {
            for ( int j = 0; j < fieldNames.length; j++ ) {
               FieldItem fi = cli.getField( fieldNames[j] );
               if ( fi != null ) {
                  mis[i].addWrittenField( fi );
               }
            }
         }
      }
      
   }  

   /**
    * Adds a set of accessed fields to a method.<p>
    *
    * This method shortcuts the writting of RTTI aspects for
    * programs. Equivalent effects can be achieved by using the RTTI
    * aspect API.<p>
    *
    * @param cl the involved class
    * @param methodName the method name
    * @param fieldNames the set of fields that are accessed by this
    * method
    * @see MethodItem#addAccessedField(FieldItem)
    */

   public static void addAccessedFields( Class cl, 
                                         String methodName, 
                                         String[] fieldNames ) {
      
      ClassItem cli = ((ClassRepository)ClassRepository.get()).getClass( cl.getName() );
      MethodItem[] mis = cli.getMethods( methodName );

      if ( mis != null ) {
         for ( int i = 0; i < mis.length; i++ ) {
            for ( int j = 0; j < fieldNames.length; j++ ) {
               FieldItem fi = cli.getField( fieldNames[j] );
               if ( fi != null ) {
                  mis[i].addAccessedField( fi );
               }
            }
         }
      }
      
   }  


   /**
    * Adds a set of added collections to a method.<p>
    *
    * This method shortcuts the writting of RTTI aspects for
    * programs. Equivalent effects can be achieved by using the RTTI
    * aspect API.<p>
    *
    * @param cl the involved class
    * @param methodName the method name
    * @param fieldNames the set of collections that are added by this
    * method
    * @see MethodItem#addAddedCollection(CollectionItem) */

   public static void addAddedCollections( Class cl, 
                                           String methodName, 
                                           String[] collectionNames ) {
      
      ClassItem cli = ((ClassRepository)ClassRepository.get()).getClass( cl.getName() );
      MethodItem[] mis = cli.getMethods( methodName );

      if ( mis != null ) {
         for ( int i = 0; i < mis.length; i++ ) {
            for ( int j = 0; j < collectionNames.length; j++ ) {
               CollectionItem ci = (CollectionItem) cli.getField( collectionNames[j] );
               if ( ci != null ) {
                  mis[i].addAddedCollection( ci );
               }
            }
         }
      }
      
   }  

   /**
    * Adds a set of removed collections to a method.<p>
    *
    * This method shortcuts the writting of RTTI aspects for
    * programs. Equivalent effects can be achieved by using the RTTI
    * aspect API.<p>
    *
    * @param cl the involved class
    * @param methodName the method name
    * @param fieldNames the set of collections that are removed by this
    * method
    * @see MethodItem#addRemovedCollection(CollectionItem) */

   public static void addRemovedCollections( Class cl, 
                                             String methodName, 
                                             String[] collectionNames ) {
      
      ClassItem cli = ((ClassRepository)ClassRepository.get()).getClass( cl.getName() );
      MethodItem[] mis = cli.getMethods( methodName );

      if ( mis != null ) {
         for ( int i = 0; i < mis.length; i++ ) {
            for ( int j = 0; j < collectionNames.length; j++ ) {
               CollectionItem ci = (CollectionItem) cli.getField( collectionNames[j] );
               if ( ci != null ) {
                  mis[i].addRemovedCollection( ci );
               }
            }
         }
      }
      
   }

   /**
    * The default constructor will set the classRepository field to
    * the right value (singleton pattern). */

   public ClassRepository () {
      classRepository = this;
   }

   /**
    * This method returns an existing class from its name.
    *
    * <p>If the class is not registered yet, the class repository
    * automatically builds default runtime informations (using naming
    * conventions), and seamlessly registers the new
    * <code>ClassItem</code> instance into the class repository.
    *
    * <p>Note: in case of manual class registrations, a class must be
    * registered with its full name to avoid name conflicts.<p>
    *
    * @param name the name of the class to get
    * @return the class if exist, null otherwise */

   public ClassItem getClass(String name)
   {
      String wrappedName = Classes.getPrimitiveTypeWrapper(name);
      MetaItem res = (MetaItem) getObject( wrappedName );
      if ( res == null ) {
         res = buildDefaultRTTI( wrappedName );
         if( res != null ) 
            register( wrappedName, res );
         else
            Log.trace("rtti","*** buildDefaultRTTI("+wrappedName+") returned null");
      }
      if (res!=null && res instanceof ClassItem)
         return (ClassItem)res;
      else
         throw new NoSuchClassException(name);
   }

   /**
    * Returns a ClassItem or a VirtualClassItem from its name. It
    * first tries to find a ClassItem, and if it fails, it returns a
    * VirtualClassItem.
    */

   public MetaItem getVirtualClass(String name) 
   {
      MetaItem ret;
      try {
         ret = getClass(name);
      } catch (NoSuchClassException e) {
         ret = (MetaItem)getObject(name);
         if (ret == null)
            throw e;
      }
      return ret;
   }

   /**
    * This method returns the class item that corresponds to the given
    * object class.<p>
    *
    * @param object the object to get the class item
    * @return the class item
    * @see #getClass(String) */

   public ClassItem getClass( Object object ) {
      return getClass(object.getClass().getName());
   }

   /**
    * This method returns the class item that corresponds to the given
    * class.<p>
    *
    * @param class the class to get the class item
    * @return the class item
    * @see #getClass(String) */

   public ClassItem getClass( Class cl ) {
      return getClass(cl.getName());
   }

   /**
    * This method builds the default RTTI for a given class name and
    * returns a corresponding class item.<p>
    *
    * @param name the name of the class to build
    * @return the corresponding class item, null if the given name is
    * not a class name */

   protected ClassItem buildDefaultRTTI( String name ) {

      Class cl = null;   
      ClassItem cli = null;
      Log.trace("rtti", ">>> Constructing RTTI infos for class " + name );

      try {
         
         try {
            cl = Class.forName( name );
         } catch ( Exception e ) {
            return null;
         }
         
         cli = new ClassItem( cl );

         // methods
         Method[] meths = cl.getMethods();
         Log.trace("rtti.methods","adding methods for "+cli);
         for( int i = 0; i < meths.length; i++ ) {
            if( meths[i].getDeclaringClass() == jacObjectClass ||
                meths[i].getDeclaringClass() == Object.class || 
                meths[i].getName().startsWith("_") ) continue;
            Log.trace("rtti.methods","adding method "+meths[i]);
            MethodItem mi = new MethodItem( meths[i]);
            cli.addMethod( mi );
            if ( NamingConventions.isModifier( meths[i].getName() ) ) {
               mi.setStateModifier( true );
            }
         }

         // fields and collections
         Collection fields = getDirectFieldAccess( cl ).values();
         Iterator it = fields.iterator();
         while ( it.hasNext() ) {
            Field f = (Field) it.next();
            Log.trace("rtti", "constructing " + f );
            Class t = f.getType();
            FieldItem fi = null;
            MethodItem mis[] = null;
            Method m = null;
            if ( Collection.class.isAssignableFrom( t ) || 
                 Map.class.isAssignableFrom( t ) || 
                 t.isArray() ) 
            {
               Log.trace("rtti", t.getName()+" -> "+f.getName()+ " is a collection");
               fi = new CollectionItem( f );
               m = NamingConventions.defaultAdderForField( cl, f.getName() );
               if ( m != null ) {
                  mis = cli.getMethods( m.getName() );
                  for (int i=0; i<mis.length; i++) {
                     mis[i].addAddedCollection( (CollectionItem)fi );
                     ((CollectionItem)fi).addAddingMethod( mis[i] );
                  }
               }
               m = NamingConventions.defaultRemoverForField( cl, f.getName() );
               if ( m != null ) {
                  mis = cli.getMethods( m.getName() );
                  for (int i=0; i<mis.length; i++) {
                     mis[i].addRemovedCollection( (CollectionItem)fi );
                     ((CollectionItem)fi).addRemovingMethod( mis[i] );
                  }
               }
            } else {
               Log.trace("rtti", f.getName()+ " is a field");
               fi = new FieldItem( f );
            }
            m = NamingConventions.defaultSetterForField( cl, f.getName() );
            if ( m != null ) {
               mis = cli.getMethods( m.getName() );
               for (int i=0; i<mis.length; i++) {
                  mis[i].addWrittenField( fi );
                  fi.addWritingMethod( mis[i] );
               }
             }
            m = NamingConventions.defaultGetterForField( cl, f.getName() );
            if ( m != null ) {
               mis = cli.getMethods( m.getName() );
               for (int i=0; i<mis.length; i++) {
                  mis[i].addAccessedField( fi );
                  fi.addAccessingMethod( mis[i] );
               }
             }
            cli.addField( fi );
         }

         // constructors
         Constructor[] constructors = cl.getConstructors();
         for( int i = 0; i < constructors.length; i++ ) {
            if ( ! ( isDefaultConstructorAdded( cl ) && 
                 constructors[i].getParameterTypes().length == 0 ) ) {
               cli.addConstructor(new ConstructorItem(constructors[i]));
            }
         }
      } catch ( InvalidDelegateException e ) {
         e.printStackTrace();
         System.exit( -1 );
      }
      return cli;

   }

   /**
    * Link class items to their names */
   public Hashtable objects = new Hashtable();

   /**
    * Reverse hashtable to find an class item from its name */
   public Hashtable names = new Hashtable();

   /**
    * Register a new class into the class repository.
    *
    * @param logicalName the key that allows to find the class
    * @param object the class to register
    * @return true if the class registered, false if already
    * registered
    *
    * @see #unregister(String) */

   public boolean register ( String logicalName, Object object ) {
      Object o = objects.get( logicalName );
      if ( o != null ) {
         objects.remove( logicalName );
         names.remove( o );
      }
      objects.put( logicalName, object );      
      names.put ( object, logicalName );
      return true;
   }

   /**
    * Unregister a class from the repository.
    *
    * @param logicalName the class name
    *
    * @see #register(String,Object) */

   public void unregister ( String logicalName ) {
      if ( objects.get( logicalName ) == null ) {
         return;
      }
      names.remove( objects.get( logicalName ) );
      objects.remove( logicalName );
   }

   /**
    * Returns true if a class is registered with this name.
    *
    * @param logicalName the key that allows to find the class
    *
    * @see #register(String,Object) */

   public boolean isRegistered ( String logicalName ) {
      if ( names.contains( logicalName ) ) {
         return true;
      }
      return false;
   }

   /**
    * Return all the registered classes as an array.
    *
    * <p>Reverse operation is <code>getNames()</code>.
    *
    * @return the registered classes in this repository
    *
    * @see #register(String,Object)
    * @see #getNames() */

   public Object[] getClasses () {
      return objects.values().toArray();
   }

   /**
    * Return the names of the registered classes as an array.    
    *
    * @return the registered classes names in this repository
    *
    * @see #register(String,Object) */

   public Object[] getNames () {
      return names.values().toArray();
   }

   /**
    * Return a registered classes for a given logical name.
    * 
    * <p>Return <code>null</code> if the name does not correspond to
    * any registered class or if <code>logicalName</code> is null.
    *
    * <p>Reverse operation is <code>getName(Object)</code>.
    *
    * @param logicalName the key that allows to find the class
    * @return the corresponding object, null if not registered
    *
    * @see #register(String,Object)
    * @see #getName(Object) */

   public Object getObject ( String logicalName ) {
      if ( logicalName == null ) return null;
      return objects.get ( logicalName );
   }

   /**
    * Returns the name of a registered class. Null if not
    * registered.
    *
    * <p>Reverse operation is <code>getObject(String)</code>.
    *
    * @param object the class to get the name of
    * @return the class name, null if not registered
    *
    * @see #register(String,Object) 
    * @see #getObject(String)
    */

   public String getName ( Object object ) {
      if ( object == null ) {
         return null;
      }
      return (String) names.get ( object );
   }

   /**
    * Dump all the registered classes in this class repository. */

   public void dump () {
      Enumeration keys = objects.keys();
      
      System.out.println ( getClass().getName() + " dump:");
      while (keys.hasMoreElements()) {
         String key = (String) keys.nextElement();
         System.out.println(" - " + key + " : " + objects.get( key ) );
      }
   }

   /* instantiate an object from a string */
   public static Object instantiate(Class type, String value) {
      if (type==int.class) {
         type = Integer.class;
      } else if (type==float.class) {
         type = Float.class;
      } else if (type==double.class) {
         type = Double.class;
      } else if (type==boolean.class) {
         type = Boolean.class;
      } else if (type==byte.class) {
         type = Byte.class;
      } else if (type==char.class) {
         type = Character.class;
      } else if (type==long.class) {
         type = Long.class;
      } else if (type==short.class) {
         type = Short.class;
      }
      try {
         Constructor constructor = type.getConstructor(new Class[] {String.class});
         return constructor.newInstance(new Object[] {value});
      } catch (Exception e) {
         e.printStackTrace();
         return null;
      }
   }
}

class NoSuchClassException extends RuntimeException {
   public NoSuchClassException(String name) {super("No such class in repository : "+name);}
}
