/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;
import java.util.*;
import jac.util.*;

/**
 * This class defines a meta item that corresponds to a
 * <code>java.lang.reflect.Field</code> meta element that is of an
 * array, a collection, or a map type.<p>
 *
 * @see java.util.Collection
 * @see java.util.Map
 * @see java.util.Arrays
 * @see java.lang.reflect.Field
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1
 */

public class CollectionItem extends FieldItem {

   private Class type = null;

   /**
    * Default contructor to create a new collection item object.<p>
    *
    * @param delegate the <code>java.lang.reflect.Field</code> actual
    * meta item */

   public CollectionItem ( Field delegate ) throws InvalidDelegateException {
      super( delegate );
      type = delegate.getType();
      if ( ! ( Collection.class.isAssignableFrom( type ) || 
               Map.class.isAssignableFrom( type ) || 
               type.isArray() ) ) {
         throw new InvalidDelegateException();
      }  
   }

   MethodItem[] addingMethods;
   
   /**
    * Gets the method items that access this collection for adding.<p>
    *
    * @return value of addingMethods.  */

   public MethodItem[] getAddingMethods() {
      return addingMethods;
   }
   
   /**
    * Sets the methods that access this collection for adding.<p>
    *
    * @param addingMethods value to assign to addingMethods.
    */

   public void setAddingMethods( MethodItem[] addingMethods ) {
      this.addingMethods = addingMethods;
   }

   /**
    * Adds a new adding method for this field.<p>
    *
    * @param addingMethod the method to add
    */

   public void addAddingMethod( MethodItem addingMethod ) {
      if ( addingMethods == null ) {
         addingMethods = new MethodItem[] { addingMethod };
      }
   }

   MethodItem[] removingMethods;
   
   /**
    * Gets the methods that access this collection for removing.<p>
    *
    * @return value of removingMethods.
    */

   public MethodItem[] getRemovingMethods() {
      return removingMethods;
   }
   
   /**
    * Sets the methods that access this collection for removing.<p>
    *
    * @param removingMethods value to assign to removingMethods.
    */

   public void setRemovingMethods( MethodItem[] removingMethods ) {
      this.removingMethods = removingMethods;
   }

   /**
    * Adds a new removing method for this field.<p>
    *
    * @param removingMethod the method to add
    */

   public void addRemovingMethod( MethodItem removingMethod ) {
      if ( removingMethods == null ) {
         removingMethods = new MethodItem[] { removingMethod };
      }
   }

   /**
    * Clears all the methods that has been set to be removers or
    * adders for this collection item.
    *
    * @see #addAddingMethod(MethodItem)
    * @see #setAddingMethods(MethodItem[])
    * @see MethodItem#removeAddedCollection(CollectionItem)
    * @see #addRemovingMethod(MethodItem)
    * @see #setRemovingMethods(MethodItem[])
    * @see MethodItem#removeRemovedCollection(CollectionItem) */

   public void clearMethods() {
      super.clearMethods();
      if (removingMethods != null) {
         for (int i=0; i<removingMethods.length; i++) {
            removingMethods[i].removeRemovedCollection(this);
         }
         removingMethods = null;
      }
      if (addingMethods != null) {
         for (int i=0; i<addingMethods.length; i++) {
            addingMethods[i].removeAddedCollection(this);
         }
         addingMethods = null;
      }
   }

   /**
    * Clears the collection on the given object.<p>
    *
    * For maps and collections, it delegates to the <code>clear</code>
    * method. For array, it resets the adding index so that added
    * elements will crush existing ones.<p>
    *
    * @param substance the object where to clean the collection
    */

   public void clear( Object substance ) {

      Field f = (Field) delegate;

      try {
         if (f.get(substance)!=null) {
            if ( Collection.class.isAssignableFrom( type ) ) { 
               ((Collection)f.get( substance )).clear();
            }
            if ( ( Map.class.isAssignableFrom( type ) ) ) { 
               ((Map)f.get( substance )).clear();
            }
            if ( type.isArray() )
               f.set( substance, Array.newInstance( 
                  type.getComponentType(), 0 ) );
         }
      } catch ( Exception e ) { 
         e.printStackTrace(); 
      }
   }

   /**
    * Tells if this collection is actually an array.<p>
    *
    * @return true if an array
    */

   public boolean isArray() {
      return type.isArray();
   }

   /**
    * Gets the collection object represented by this collection
    * item.
    *
    * <p>It returns a <code>java.util.Collection</code> representation of
    * the actual collection (i.e.  either a <code>Collection</code>, a
    * <code>Map</code>, or an array).
    *
    * <p>The programmer should rather use the methods that gets the
    * actual collection by using the accessor for this collection if
    * any since it allows possible aspects applications if needed.
    *
    * @return the actual object collection representation
    * @see #getActualCollectionThroughAccessor(Object)
    * @see #toCollection(Object) */

   public Collection getActualCollection( Object substance ) {
      Object value = null;

      try {
         value = ((Field)delegate).get( substance );
      } catch ( Exception e ) {
         Log.error("getActualCollection("+getName()+") failed");
         e.printStackTrace();
         return null;
      }
      return toCollection(value);
   }

   /**
    * Gets the collection object represented by this collection item
    * by using the accessor defined in the class containing this
    * collection if any.
    *
    * <p>Use this method to be sure that all the aspects will be
    * applied to the substance when the collection is retrieved.
    *
    * <p>It returns a <code>java.util.Collection</code> representation of
    * the actual collection (i.e.  either a <code>Collection</code>, a
    * <code>Map</code>, or an array).
    *
    * @return the actual object collection representation
    * @see #getActualCollection(Object)
    * @see #toCollection(Object) */

   public Collection getActualCollectionThroughAccessor(Object substance) {
      Object value = getThroughAccessor(substance);
      return toCollection(value);
   }

   /**
    * A useful method to convert any kind of collection to a
    * <code>java.util.Collection</code> compatible instance.
    *
    * <p>Supported converted value types are
    * <code>java.util.Collection</code>, <code>java.util.Map</code>,
    * and Java arrays.
    *
    * @param value an object of one of the supported types
    * @return the collection compliant converted value */

   public static Collection toCollection(Object value) {
      if ( value == null )  return null;

      if ( value instanceof Collection ) { 
         return (Collection) value;
      } else if ( value instanceof Map ) { 
         return ((Map) value).values();
      } else if ( value.getClass().isArray() ) {
         return Arrays.asList( (Object[]) value );
      }
      return null;
   }

   /**
    * Adds an item to the collection.<p>
    *
    * The <code>extraInfos</code> param must represent a key if the
    * collection is an hastable. If the collection is a
    * <code>java.util</code> collection or if it is an array, the
    * extra informations can be null (in this case, the new item is
    * added at the end of the collection), and they can be not
    * null. In this case, it must be an <code>Integer</code> instance
    * that represents the index of the new item within the
    * collection (must be a list in this case).<p>
    *
    * @param substance the collection where to add the new item
    * @param newItem the item to add
    * @param extraInfos this optional parameter must be used when the
    * collection need some other information when adding a item to it
    * as, for instance, the key for a map */

   public void add( Object substance, Object newItem, Object extraInfos ) {

      Field f = (Field) delegate;

      try {      
      
         if ( ( Collection.class.isAssignableFrom( type ) ) ) { 
            // Set and List
            if ( extraInfos == null ) {
               ((Collection)f.get( substance )).add( newItem );
            } else {
               ((List)f.get( substance )).add( 
                  ((Integer)extraInfos).intValue(), newItem );
            }
         } else if ( ( Map.class.isAssignableFrom( type ) ) ) { 
            // Map
            ((Map)f.get( substance )).put( extraInfos, newItem );
         } else if ( type.isArray() ) {
            // Arrays
            // THIS IS BUGGED !!!
            if ( extraInfos == null ) {
               Object oldArray = f.get( substance );
               int index = 0;
               if( oldArray != null ) {
                  index = Array.getLength( oldArray );
               }
               Object newArray;
               f.set( substance, 
                      newArray = Array.newInstance( 
                         type.getComponentType(), index+1 ) );

               if( oldArray != null ) {
                  for( int i=0; i<index; i++) {
                     Array.set( newArray, i, Array.get( oldArray, i ) );
                  }
               }
               
               Array.set( newArray, index, newItem );

            } else {
               ((Object[])f.get( substance ))[((Integer)extraInfos).intValue()] = newItem;
               //arrayAddingIndex = ((Integer)extraInfos).intValue() + 1;
            }
         }
      
      } catch ( Exception e ) { e.printStackTrace(); }
      
   }

   /**
    * Removes an item from the collection.<p>
    *
    * The <code>extraInfos</code> param must represent a key if the
    * collection is an hastable. If the collection is a
    * <code>java.util</code> collection or if it is an array, the
    * extra informations can be null (in this case, the new item is
    * added at the end of the collection), and they can be not
    * null. In this case, it must be an <code>Integer</code> instance
    * that represents the index of the new item within the
    * collection (must be a list in this case).<p>
    *
    * @param substance the collection where to remove the new item
    * @param item the item to remove (can be null if extra infos are
    * not null)
    * @param extraInfos this optional parameter must be used when the
    * collection need some other information when adding a item to it
    * as, for instance, the key for a map */

   public void remove( Object substance, Object item, Object extraInfos ) {

      Field f = (Field) delegate;

      try {      
      
         if ( ( Collection.class.isAssignableFrom( type ) ) ) { 
            // Set and List
            if ( extraInfos == null ) {
               ((Collection)f.get( substance )).remove( item );
            } else {
               ((List)f.get( substance )).remove( item );
            }
         } else if ( ( Map.class.isAssignableFrom( type ) ) ) { 
            // Map
            ((Map)f.get( substance )).remove( extraInfos );
         } else if ( type.isArray() ) {
            // Arrays
            // THIS IS BUGGED !!!
            if ( extraInfos == null ) {
               Object oldArray = f.get( substance );
               int index = 0;
               if( oldArray != null ) {
                  index = Array.getLength( oldArray );
               }
               Object newArray;
               f.set( substance, 
                      newArray = Array.newInstance( 
                         type.getComponentType(), index-1 ) );

               if( oldArray != null ) {
                  for( int i=0; i<index; i++) {
                     Array.set( newArray, i, Array.get( oldArray, i ) );
                  }
               }
               
               //Array.set( newArray, index, newItem );

            } else {
               //((Object[])f.get( substance ))[((Integer)extraInfos).intValue()] = item;
               //arrayAddingIndex = ((Integer)extraInfos).intValue() + 1;
            }
         }
      
      } catch ( Exception e ) { e.printStackTrace(); }
      
   }

   /**
    * Tells if this collection item is compliant with a
    * <code>java.util.List</code> interface.
    *
    * @return true if compliant */

   public boolean isList() {
      return java.util.List.class.isAssignableFrom(getType());
   }
   
   /**
    * Tells if this collection item is compliant with a
    * <code>java.util.Map</code> interface.
    *
    * @return true if compliant */

   public boolean isMap() {
      return java.util.Map.class.isAssignableFrom(getType());
   }
   
   /**
    * Tells if this collection item is compliant with a
    * <code>java.util.Set</code> interface.
    *
    * @return true if compliant */

   public boolean isSet() {
      return java.util.Set.class.isAssignableFrom(getType());
   }
   
   /**
    * Always returns false for collections (maybe not very
    * semantically clear).
    *
    * @return false */

   public boolean isPrimitive() {
      return false;
   }

   /**
    * Always returns false for collections (maybe not very
    * semantically clear).
    *
    * @return false */

   public boolean isReference() {
      return false;
   }
}
