/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;

/**
 * This class defines a meta item that corresponds to the
 * <code>java.lang.reflect.Constructor</code> meta element.<p>
 *
 * @author Laurent Martelli
 * @author Renaud Pawlak
 * @version 0.8.1
 */

public class ConstructorItem extends AbstractMethodItem {

   /**
    * Transforms a constructor items array into a constructors array
    * containing the <code>java.lang.reflect</code> constructors
    * wrapped by the constructor items.<p>
    *
    * @param constructorItems the Constructor items
    * @return the actual Constructors in <code>java.lang.reflect</code> */

   public static Constructor[] toConstructors( ConstructorItem[] constructorItems ) {
      Constructor[] res = new Constructor[constructorItems.length];
      for ( int i = 0; i < constructorItems.length; i++ ) {
         if ( constructorItems[i] == null ) {
            res[i] = null;
         } else {
            res[i] = constructorItems[i].getActualConstructor();
         }
      }
      return res;
   }

   
   /**
    * Default contructor to create a new constructor item object.<p>
    *
    * @param delegate the <code>java.lang.reflect.Constructor</code> actual
    * meta item */

   public ConstructorItem ( Constructor delegate ) throws InvalidDelegateException {
      super( delegate );
   }

   /**
    * Get the constructor represented by this constructor item.<p>
    *
    * @return the actual constructor
    */

   public Constructor getActualConstructor() {
      return (Constructor) delegate;
   }

   public String getName() {
      return ((Constructor)delegate).getName();
   }

   public Class getType() {
      return getParent().getType();
   }

   /**
    * Creates a new instance of the class item that is parent of this
    * constructor item.
    *
    * @param params the parameters needed by this constructor (see the
    * types)
    * @see #getParameterTypes() */

   public Object newInstance(Object [] params) 
      throws InstantiationException, IllegalAccessException, InvocationTargetException {
      return getActualConstructor().newInstance(params);
   }

   /**
    * A nice way to construct a new instance when the constructor does
    * not take any arguements (it throws an exception if it is not the
    * case).
    *
    * @see #getParameterTypes() */

   public Object newInstance() 
      throws InstantiationException, IllegalAccessException, InvocationTargetException {
      return getType().newInstance();
   }

   public Class[] getParameterTypes() {
      return ((Constructor)delegate).getParameterTypes();
   }

   public String toString() {
      if (delegate!=null) {
         Constructor m = (Constructor)delegate;
         String ret = m.getName() + "(";
         Class[] pts = m.getParameterTypes();
         for ( int i = 0; i < pts.length; i++ ) {
            ret = ret + NamingConventions.getShortClassName( pts[i] );
            if ( i < pts.length - 1 ) ret = ret + ",";
         }
         ret = ret + ")";
         return ret;
      }
      else 
         return "???";
   }

   public String getFullName() {
      String ret = NamingConventions.getShortConstructorName(this) + "(";
      Class[] pts = getParameterTypes();
      for ( int i = 0; i < pts.length; i++ ) {
         ret = ret + NamingConventions.getShortClassName( pts[i] );
         if ( i < pts.length - 1 ) ret = ret + ",";
      }
      ret = ret + ")";
      return ret;
   }

   
}// ConstructorItem
