/*
  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;
import jac.util.*;

/**
 * This class defines a meta item that corresponds to the
 * <code>java.lang.reflect.Field</code> meta element.<p>
 *
 * <p>In addition to the <code>java.lang.reflect</code> classical
 * features, this RTTI method element is able to tell if a field is
 * accessed for reading or writting by a given method.
 *
 * <p>It also provides some modification methods that are aspect compliant.
 *
 * <p>For the moment, default meta informations are setted by the
 * <code>ClassRepository</code> class using some naming
 * conventions. In a close future, these informations will be deduced
 * from the class bytecodes analysis at load-time.
 *
 * @see java.lang.reflect.Field
 * @see #getWritingMethods()
 * @see #getAccessingMethods()
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1 */

public class FieldItem extends MetaItemDelegate {

   static Class wrappeeClass = ClassRepository.wrappeeClass;

   /**
    * Transforms a field items array into a fields array containing
    * the <code>java.lang.reflect</code> fields wrapped by the method
    * items.<p>
    *
    * @param fieldItems the field items
    * @return the actual fields in <code>java.lang.reflect</code>
    */

   public static Field[] toFields( FieldItem[] fieldItems ) {
      Field[] res = new Field[fieldItems.length];
      for ( int i = 0; i < fieldItems.length; i++ ) {
         if ( fieldItems[i] == null ) {
            res[i] = null;
         } else {
            res[i] = fieldItems[i].getActualField();
         }
      }
      return res;
   }

   /**
    * Default contructor to create a new field item object.<p>
    *
    * @param delegate the <code>java.lang.reflect.Field</code> actual
    * meta item */

   public FieldItem ( Field delegate ) throws InvalidDelegateException {
      super( delegate );
   }


   public final Object getAttribute(String name) {
      Object value = super.getAttribute(name);
      if(value==null) {
         ClassItem parent=(ClassItem)getParent();
         parent=parent.getSuperclass();
         if(parent!=null) {
            if( parent.hasField(getName()) ) {
               FieldItem parentField=parent.getField(getName());
               value=parentField.getAttribute(name);
            }
         }
      }
      return value;
   }


   MethodItem[] accessingMethods;
   
   /**
    * Get the methods that access this field for reading.<p>
    *
    * @return value of accessingMethods.
    */
   public final MethodItem[] getAccessingMethods() {
      return accessingMethods;
   }
   
   /**
    * Set the methods that access this field for reading.<p>
    *
    * @param accessingMethods value to assign to accessingMethods.
    */

   public final void setAccessingMethods( MethodItem[] accessingMethods ) {
      this.accessingMethods = accessingMethods;
   }

   /**
    * Add a new accessing method for this field.<p>
    *
    * @param accessingMethod the method to add
    */

   public final void addAccessingMethod( MethodItem accessingMethod ) {
      if ( accessingMethods == null ) {
         accessingMethods = new MethodItem[] { accessingMethod };
      }
   }
   
   MethodItem[] writingMethods;
   
   /**
    * Get the methods that access this field for writing.<p>
    *
    * @return value of writingMethods.
    */
   public final MethodItem[] getWritingMethods() {
      return writingMethods;
   }
   
   /**
    * Set the methods that access this field for writing.<p>
    *
    * @param writingMethods value to assign to writingMethods.
    */

   public final void setWritingMethods( MethodItem[] writingMethods ) {
      this.writingMethods = writingMethods;
   }

   /**
    * Add a new writing method for this field.<p>
    *
    * @param writingMethod the method to add
    */

   public final void addWritingMethod( MethodItem writingMethod ) {
      if ( writingMethods == null ) {
         writingMethods = new MethodItem[] { writingMethod };
      }
   }
   
   /**
    * Remove accessing and writing methods
    */
   public void clearMethods() {
      if (writingMethods != null) {
         for (int i=0; i<writingMethods.length; i++) {
            writingMethods[i].removeWrittenField(this);
         }
         writingMethods = null;
      }
      if (accessingMethods != null) {
         for (int i=0; i<accessingMethods.length; i++) {
            accessingMethods[i].removeAccessedField(this);
         }
         accessingMethods = null;
      }
   }

   /**
    * Get the field represented by this field item.<p>
    *
    * @return the actual field
    */

   public final Field getActualField() {
      return (Field) delegate;
   }

   /**
    * Get the value of this field item for a given object.<p>
    *
    * @param object the object that supports the field
    * @return the field value in the given object
    * @see #set(Object,Object) */

   public final Object get( Object object ) {
      Object ret = null;
      try {
         ret = ((Field)delegate).get( object );
      } catch ( Exception e ) {
         e.printStackTrace();
      }
      return ret;
   }

   /**
    * Get a field value through accessor if it exists.<p>
    *
    * @param object the object that supports the field
    */
   public final Object getThroughAccessor(Object substance) {
      Object value = null;
      String name;
      boolean accessorFound = false;
      MethodItem[] accessors = getAccessingMethods();
      if (accessors!=null && accessors.length>0) {
         Log.trace("rtti.field","Found "+accessors.length+
                   " accessors for field " + getName());
         for (int i=0; i<accessors.length;i++) {
            Log.trace("rtti.field","Accessor "+accessors[i]);
            if (NamingConventions.isGetter(accessors[i].getName())) {
               try {
                  Log.trace("rtti.field",this+": invoking "+accessors[i]);
                  value = (accessors[i].invoke(substance,new Object[0]));
                  accessorFound = true;
               } catch (IllegalArgumentException e) {
                  accessorFound=false;
                  //e.printStackTrace();
               } catch (IllegalAccessException e) {
                  accessorFound=false;
                  //e.printStackTrace();
               } catch ( InvocationTargetException e ) {
                  throw (RuntimeException)e.getTargetException();
               }
            }
         } 
      }

      if( ! accessorFound ) {
         Log.warning("rtti.field","No accessor found for field " + this);
         value = get(substance);
      }
      return value;
   }
   
   /**
    * Sets the value of this field item for a given object.<p>
    *
    * @param object the object that supports the field
    * @param value the value to set
    * @see #get(Object) */

   public final void set(Object object, Object value) {
      try {
         if ( value==null && getType().isPrimitive() ) {
            Log.warning("Cannot set primitive field "+this+" to null");
         } else {
            ((Field)delegate).set( object, value );
         }
      } catch ( Exception e ) {
         e.printStackTrace();
      }
   }

   /**
    * Sets the value of this field item by using its setter method if
    * any (else use the <code>set</code> method.
    *
    * @param subtance the object to set the field of 
    * @param value the new value of the field
    * @see #set(Object,Object) */

   public final void setThroughWriter(Object substance, Object value) {
      Log.trace("rtti.field","setThroughWriter "+substance+"."+getName()+","+value);
      String name;
      boolean writerFound = false;
      MethodItem[] writers = getWritingMethods();
      if (writers!=null && writers.length>0) {
         Log.trace("rtti.field","Found "+writers.length+
                   " writers for field " + getName());
         for (int i=0; i<writers.length;i++) {
            Log.trace("rtti.field","Writer "+writers[i]);
            if (NamingConventions.isSetter(writers[i].getName())) {
               try {
                  Log.trace("rtti.field",this+": invoking "+writers[i]);
                  writers[i].invoke(substance,new Object[] { value });
                  writerFound = true;
                  return;
               } catch (IllegalArgumentException e) {
                  writerFound=false;
                  //e.printStackTrace();
               } catch (IllegalAccessException e) {
                  writerFound=false;
                  //e.printStackTrace();
               } catch ( InvocationTargetException e ) {
                  throw (RuntimeException)e.getTargetException();
               }
            }
         } 
      }
      if ( ! writerFound ) {
         Log.warning("rtti.field","No writer found for field " + this);
         set(substance,value);
      }
   }

   public MethodItem getSetter() {
      MethodItem[] writers = getWritingMethods();
      if (writers!=null && writers.length>0) {
         Log.trace("rtti.field","Found "+writers.length+
                   " writers for field " + getName());
         for (int i=0; i<writers.length;i++) {
            Log.trace("rtti.field","Writer "+writers[i]);
            if (NamingConventions.isSetter(writers[i].getName())) {
               return writers[i];
            }
         }
      }
      throw new Error("Field "+getName()+" does not have a setter"); 
   }

   public final String getName() {
      return ((Field)delegate).getName();
   }

   public final Class getType() {
      return ((Field)delegate).getType();
   }

   public final ClassItem getTypeItem() {
      return ClassRepository.get().getClass(getType());
   }

   /**
    * Tells if this field item represents a primitive type (that is to
    * say it is of a type that is not a reference towards a Jac
    * object).
    *
    * <p>Allways returns false on collections (use
    * <code>isWrappable</code> to know if the field is wrappable).
    *
    * @return true if primitive
    * @see #isReference() */
 
   public boolean isPrimitive() {
      return ! wrappeeClass.isAssignableFrom(getType());
   }

   /**
    * Tells if this field item represents a reference type (that is to
    * say it is of a type that is a reference towards a Jac
    * object).
    *
    * <p>Allways returns false on collections (use
    * <code>isWrappable</code> to know if the field is wrappable).
    *
    * @return true if reference
    * @see #isPrimitive()
    * @see #isWrappable() */

   public boolean isReference() {
      return wrappeeClass.isAssignableFrom(getType());
   }

   /**
    * Tells if the field item represents a wrappable type.
    *
    * <p>This method can be used on all kinds of field item including
    * collections (on contrary to <code>isPrimitive</code> and
    * <code>isReference</code>).
    *
    * @return true if wrappable
    * @see #isPrimitive()
    * @see #isReference() */

   public boolean isWrappable() {
      return wrappeeClass.isAssignableFrom(getType());
   }
   
}
