/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;
import java.util.*;

/**
 * This class defines the super class for all the meta items whithin
 * the rtti aspect.<p>
 *
 * A meta item encapsulates a <code>java.lang.reflect</code> item so
 * that the user of this item can add extra informations
 * (attributes). Typically this feature can be used by an aspect to
 * tag an element of the model to react to this tag later on.<p>
 *
 * Examples:<p> <ul> 
 *
 * <li>A persistence aspect can tag some field persistent, add methods
 * that change the object states even if they do not fit naming
 * conventions...<p>
 *
 * <li>A GUI can tag a given field to be invisible or a class to be
 * displayed by a special view (eg a given Swing component)...
 *
 * </ul>
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1
 */

public abstract class MetaItem {

   /** A correspondance table between the RRTI attributes and the
       aspect components that set them. */
   protected static Hashtable attrACs = new Hashtable();

   /**
    * Unsets all the attributes of all the RTTI items that have been
    * set by a given aspect component.<p>
    *
    * @param acName the aspect component name
    * @see #unsetAttribute(String) */
   
   public static void unsetAttributesFor( String acName ) {}

   /* A <code>name -> value</code> container. */
   private Hashtable attributes = new Hashtable();

   boolean isRegisteredAC( String name ) {
      Object ac = null;
      try {
         Class acmc = Class.forName("jac.core.ACManager");
         Object acm = acmc
            .getMethod("get",new Class[0])
            .invoke(null,new Object[0]);
         ac = acmc
            .getMethod("getObject",new Class[]{String.class})
            .invoke(acm,new Object[]{name});
      } catch (Exception e) {
         e.printStackTrace();
      }
      if (ac==null) return false;
      else return true;
   }

   String getCurAC() {
      Object acName = null;
      try {
         Class jacObj = Class.forName("jac.core.JacObject");
         Object coll = jacObj
            .getMethod("getCollaboration",new Class[0])
            .invoke(null,new Object[0]);
         acName = coll.getClass()
            .getMethod("getAttribute",new Class[]{String.class})
            .invoke(coll,new Object[]{"cur_AC"});
      } catch (Exception e) {
         e.printStackTrace();
      }
      return (String)acName;
   }

   /**
    * Gets the value of an attribute.
    *
    * @param name the name of the attribute
    * @return the value of the attribute
    */

   public Object getAttribute(String name) {
      String acName = (String) attrACs.get( name );
      if( acName != null && (! isRegisteredAC( acName )) ) {
         return null;
      }
      return attributes.get(name);
   }

   /**
    * Sets the value of an attribute.
    *
    * @param name the name of the attribute
    * @param value the value of the attribute
    */

   public final void setAttribute(String name, Object value) {

      if ( name == null ) return;
      String acName = (String) getCurAC();
      if ( acName != null ) {
         if ( ! attrACs.containsKey( name ) ) {
            attrACs.put( name, acName );
         }
      }
      attributes.put(name,value);
   }

   /**
    * Unsets the value of an attribute.
    *
    * @param name the name of the attribute
    * @param value the value of the attribute
    */

   public void unsetAttribute(String name) {
      attributes.remove(name);
   }

   /**
    * This method gets the name of the meta item by delegating to the
    * actual <code>java.lang.reflect</code> meta item.<p>
    *
    * @return the item name */

   public abstract String getName();
}


