/*

  Copyright (C) 2001 Renaud Pawlak, Laurent Martelli
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.core.rtti;

import java.lang.reflect.*;
import jac.util.*;

import java.util.*;

/**
 * This class defines a meta item that corresponds to the
 * <code>java.lang.reflect.Method</code> meta element.
 *
 * <p>In addition to the <code>java.lang.reflect</code> classical
 * features, this RTTI method element is able to tell if a method is a
 * setter, a getter, or more generally speaking, a state modifier for
 * a given field of the object it belongs to. And, if this field is a
 * collection (an array, a list or a map), then this meta element is
 * able to tell if the method it represents adds or removes elements
 * to or from this collection.
 *
 * <p>It also provides to introspection features for references on Jac
 * objects.
 *
 * <p>For the moment, default meta informations are setted by the
 * <code>ClassRepository</code> class using some naming
 * conventions. In a close future, these informations will be deduced
 * from the class bytecodes analysis at load-time.
 *
 * @see java.lang.reflect.Method
 * @see #getWrittenFields()
 * @see #getAccessedFields()
 * @see #getAddedCollections()
 * @see #getRemovedCollections()
 * @see #isStateModifier()
 * @see #hasAccessedReferences()
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1 */

public class MethodItem extends AbstractMethodItem {

   /**
    * Transforms a method items array into a methods array containing
    * the <code>java.lang.reflect</code> methods wrapped by the method
    * items.<p>
    *
    * @param methodItems the method items
    * @return the actual methods in <code>java.lang.reflect</code>
    */

   public static Method[] toMethods( MethodItem[] methodItems ) {
      Method[] res = new Method[methodItems.length];
      for ( int i = 0; i < methodItems.length; i++ ) {
         if ( methodItems[i] == null ) {
            res[i] = null;
         } else {
            res[i] = methodItems[i].getActualMethod();
         }
      }
      return res;
   }
   
   /**
    * Default contructor to create a new method item object.<p>
    *
    * @param delegate the <code>java.lang.reflect.Method</code> actual
    * meta item */

   public MethodItem ( Method delegate ) throws InvalidDelegateException {
      super( delegate );
   }

   /** Stores the fields that are written by this method.<p> */
   FieldItem[] writtenFields = null;
   
   /**
    * Gets the value of the fields that are written by this method.<p>
    *
    * @return value of writtenFields.  */

   public final FieldItem[] getWrittenFields() {
      return writtenFields;
   }
   
   /**
    * Sets the value of the fields that are written by this method.<p>
    *
    * @param writtenFields value to assign to writtenFields
    * @see #addWrittenField(FieldItem)
    */

   public final void setWrittenFields( FieldItem[] writtenFields ) {
      this.writtenFields = writtenFields;
   }

   /**
    * Adds a new written field for this method.<p>
    *
    * @param writtenField the field to add
    * @see #setWrittenFields(FieldItem[])
    * @see #removeWrittenField(FieldItem)
    */

   public final void addWrittenField( FieldItem writtenField ) {
      if ( writtenFields == null ) {
         writtenFields = new FieldItem[] { writtenField };
      } else {
         FieldItem[] tmp = new FieldItem[writtenFields.length + 1];
         System.arraycopy( writtenFields, 0, tmp, 0, writtenFields.length );
         tmp[writtenFields.length] = writtenField;
         writtenFields = tmp;
      }
   }

   /**
    * Removes a new written field for this method.<p>
    *
    * @param writtenField the field to remove
    * @see #addWrittenField(FieldItem)
    */

   public final void removeWrittenField( FieldItem field ) {
      if ( writtenFields != null ) {
         Vector v = new Vector(Arrays.asList(writtenFields));
         v.remove(field);
         writtenFields = new FieldItem[v.size()];
         System.arraycopy(v.toArray(),0,writtenFields,0,v.size());
      }
   }

   /** Stores the fields that are read by this method.<p> */
   FieldItem[] accessedFields = null;

   /** Stores the number of references read by this method */
   int numAccessedReferences = 0;

   /**
    * Tells if this method accesses any references of the class it
    * belongs to.
    *
    * @return true if one or more references are accessed */
   
   public final boolean hasAccessedReferences() {
      return numAccessedReferences>0;
   }

   /**
    * Gets the value of the fields that are written by this method.<p>
    *
    * @return value of accessedFields.  */

   public final FieldItem[] getAccessedFields() {
      return accessedFields;
   }
   
   /**
    * Sets the value of the fields that are read by this method.<p>
    *
    * @param accessedFields value to assign to accessedFields
    * @see #addAccessedField(FieldItem)
    */

   public final void setAccessedFields(FieldItem[]  accessedFields) {
      this.accessedFields = accessedFields;
      numAccessedReferences = 0;
      if (accessedFields!=null) {
         for(int i=0; i<accessedFields.length; i++) {
            if (accessedFields[i].isReference())
               numAccessedReferences++;
         }
      }
   }

   /**
    * Adds a new accessed field for this method.<p>
    *
    * @param accessedField the field to add
    * @see #setAccessedFields(FieldItem[])
    * @see #removeAccessedField(FieldItem)
    */

   public final void addAccessedField( FieldItem accessedField ) {
      if (accessedField.isReference())
         numAccessedReferences++;
      if ( accessedFields == null ) {
         accessedFields = new FieldItem[] { accessedField };
      } else {
         FieldItem[] tmp = new FieldItem[accessedFields.length + 1];
         System.arraycopy( accessedFields, 0, tmp, 0, accessedFields.length );
         tmp[accessedFields.length] = accessedField;
         accessedFields = tmp;
      }
   }

   /**
    * Removes an accessed field for this method.<p>
    *
    * @param accessedField the field to remove
    * @see #addAccessedField(FieldItem)
    */

   public final void removeAccessedField( FieldItem field ) {
      if ( accessedFields != null ) {
         Vector v = new Vector(Arrays.asList(accessedFields));
         if (v.remove(field)) {
            if (field.isReference())
               numAccessedReferences--;
            accessedFields = new FieldItem[v.size()];
            System.arraycopy(v.toArray(),0,accessedFields,0,v.size());
         }
      }
   }

   /** Store the collections that are added by this method.<p> */
   CollectionItem[] addedCollections = null;
   
   /**
    * Get the value of the collections that are added by this
    * method.<p>
    *
    * @return value of addedCollections.  */

   public final CollectionItem[] getAddedCollections() {
      return addedCollections;
   }
   
   /**
    * Sets the value of the collections that are added by this method.<p>
    *
    * @param addedCollections value to assign to addedCollections
    * @see #addAddedCollection(CollectionItem)
    */

   public final void setAddedCollections( CollectionItem[] addedCollections ) {
      this.addedCollections = addedCollections;
   }

   /**
    * Adds a new added collection for this method.<p>
    *
    * @param addedCollection the collection to add
    * @see #removeAddedCollection(CollectionItem)
    */

   public final void addAddedCollection( CollectionItem addedCollection ) {
      if ( addedCollections == null ) {
         addedCollections = new CollectionItem[] { addedCollection };
      } else {
         CollectionItem[] tmp = new CollectionItem[addedCollections.length + 1];
         System.arraycopy( addedCollections, 0, tmp, 0, addedCollections.length );
         tmp[addedCollections.length] = addedCollection;
         addedCollections = tmp;
      }
   }

   /**
    * Removes an existing added collection for this method.<p>
    *
    * @param addedCollection the collection to add
    * @see #addAddedCollection(CollectionItem) */

   public final void removeAddedCollection( CollectionItem collection ) {
      if ( addedCollections != null ) {
         Vector v = new Vector(Arrays.asList(addedCollections));
         v.remove(collection);
         addedCollections = new CollectionItem[v.size()];
         System.arraycopy(v.toArray(),0,addedCollections,0,v.size());
      }
   }

   /** Stores the collections that are removed by this method.<p> */
   CollectionItem[] removedCollections = null;
   
   /**
    * Gets the value of the collections that are removed by this
    * method.<p>
    *
    * @return value of removedCollections.  */

   public final CollectionItem[] getRemovedCollections() {
      return removedCollections;
   }
   
   /**
    * Sets the value of the collections that are removed by this method.<p>
    *
    * @param removedCollections value to assign to removedCollections
    * @see #addRemovedCollection(CollectionItem)
    */

   public final void setRemovedCollections( CollectionItem[] removedCollections ) {
      this.removedCollections = removedCollections;
   }

   /**
    * Adds a new removed collection for this method.<p>
    *
    * @param removedCollection the collection to add
    * @see #setRemovedCollections(CollectionItem[])
    * @see #removeRemovedCollection(CollectionItem)
    */

   public final void addRemovedCollection( CollectionItem removedCollection ) {
      if ( removedCollections == null ) {
         removedCollections = new CollectionItem[] { removedCollection };
      } else {
         CollectionItem[] tmp = new CollectionItem[removedCollections.length + 1];
         System.arraycopy( removedCollections, 0, tmp, 0, removedCollections.length );
         tmp[removedCollections.length] = removedCollection;
         removedCollections = tmp;
      }
   }

   /**
    * Removes an existing removed collection for this method.<p>
    *
    * @param removedCollection the collection to remove
    * @see #addRemovedCollection(CollectionItem)
    */

   public final void removeRemovedCollection( CollectionItem collection ) {
      if ( removedCollections != null ) {
         Vector v = new Vector(Arrays.asList(removedCollections));
         v.remove(collection);
         removedCollections = new CollectionItem[v.size()];
         System.arraycopy(v.toArray(),0,removedCollections,0,v.size());
      }
   }

   /** Stores a flag that tells if this method is a state modifier. */
   protected boolean stateModifier = false;   

   /**
    * Sets this method to be a state modifier (ie when it is called,
    * the state of the object changes).
    *
    * @param bool true if modifier
    * @see #isStateModifier() */

   public final void setStateModifier( boolean bool ) {
      stateModifier = bool;
   }

   /**
    * Tells if this method is a state modifier (ie when it is called,
    * the state of the object changes).
    *
    * <p>By default, is returns false unless the method fowllows the
    * naming conventions of a mofifier or if someone configured it to
    * be true with the <code>setStateModifier</code> method.
    *
    * @param bool true if modifying
    * @see #setStateModifier(boolean)
    * @see NamingConventions#isModifier(String) */

   public final boolean isStateModifier() {
      if( writtenFields != null && writtenFields.length>0) 
         return true;
      return stateModifier;
   }     

   /**
    * Gets the method represented by this method item.<p>
    *
    * @return the actual method
    */

   public final Method getActualMethod() {
      return (Method) delegate;
   }

   public final String getName() {
      return ((Method)delegate).getName();
   }

   public final Class getType() {
      return ((Method)delegate).getReturnType();
   }

   public final Class[] getParameterTypes() {
      return ((Method)delegate).getParameterTypes();
   }

   public final String toString() {
      if (delegate!=null) {
         return getFullName()+":"+
            NamingConventions.getShortClassName( 
               ((Method)delegate).getReturnType() );
      }
      else 
         return "???";
   }

   /**
    * Invokes this method on the given object and with the given
    * parameters values.
    *
    * @param object a class this method belongs to intance
    * @param parameters the values of the parameters to invoke this
    * method with */

   public final Object invoke(Object object, Object[] parameters) 
      throws IllegalAccessException, InvocationTargetException 
   {
      //      Log.trace("invoke",toString()+" invoke("+object+","+Arrays.asList(parameters)+")");
      return ((Method)delegate).invoke(object,parameters);
   }

   /**
    * Invokes this static method with the given parameters values.
    *
    * @param parameters the values of the parameters to invoke this
    * method with */

   public final Object invokeStatic(Object[] parameters) 
      throws IllegalAccessException, InvocationTargetException 
   {
      //      Log.trace("invoke",toString()+" invoke("+object+","+Arrays.asList(parameters)+")");
      return ((Method)delegate).invoke(null,parameters);
   }

   /**
    * Invokes a method with uninitialized parameters.
    *
    * <p>The parameters array is initialized before the invocation
    * with default values.
    *
    * @param object the target object
    * @param parameters the maybe initialized values of the parameters
    * to invoke this method with (primitive parameters can be null) */

   public final Object invokeWithInit(Object object,
                                      Object[] parameters)
      throws IllegalAccessException, InvocationTargetException
   {
      
      Class[] paramTypes = getParameterTypes();
      for (int i=0; i< parameters.length; i++) {
         if (parameters[i]==null) {
            if (paramTypes[i] == float.class) {
               parameters[i] = new Float(0.0);
            } else if (paramTypes[i] == long.class) {
               parameters[i] = new Long(0);
            } else if (paramTypes[i] == double.class) {
               parameters[i] = new Double(0.0);
            } else if (paramTypes[i] == byte.class) {
               parameters[i] = new Byte((byte)0);
            } else if (paramTypes[i] == char.class) {
               parameters[i] = new Character(' ');
            } else if (paramTypes[i] == short.class) {
               parameters[i] = new Short((short)0);
            } else if (paramTypes[i] == int.class) {
               parameters[i] = new Integer(0);
            } else if (paramTypes[i] == boolean.class) {
               parameters[i] = new Boolean(false);
            }
         }
      }
      return ((Method)delegate).invoke(object,parameters);
   } 

   public final boolean isSetter() {
      return NamingConventions.isSetter(getName());
   }
   public final boolean isAdder() {
      return NamingConventions.isAdder(getName());
   }
   public final boolean isRemover() {
      return NamingConventions.isRemover(getName());
   }
}// MethodItem
