/*
  JAC-Core version 0.8.1

  Renaud Pawlak, pawlak@cnam.fr, CEDRIC Laboratory, Paris, France.
  Lionel Seinturier, Lionel.Seinturier@lip6.fr, LIP6, Paris, France.

  JAC-Core is free software. You can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as
  published by the Free Software Foundation.
  
  JAC-Core is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  This work uses the Javassist system - Copyright (c) 1999-2000
  Shigeru Chiba, University of Tsukuba, Japan.  All Rights Reserved.  */


package jac.core.rtti;

import jac.core.AspectComponent;
import jac.core.Wrappee;
import jac.core.JacObject;
import jac.util.Log;

import java.util.Vector;

/**
 * This class defines the rtti aspect.
 *
 * <p>It allows the programmer to add some runtime type informations on
 * the classes of its applications.
 *
 * @see ClassItem
 * @see MethodItem
 * @see FieldItem
 * @see CollectionItem
 *
 * @author Renaud Pawlak
 * @author Laurent Martelli
 * @version 0.8.1 */

public interface RttiConf {

   /**
    * Adds some written fields to a given method.
    *
    * <p>This configuration method must be used when a method changes
    * some field values and does not follow the naming conventions (it
    * is not a setter or an adder for instance).
    *
    * @param className the class the method belongs to
    * @param methodName the method
    * @param writtenFields some new fields the method writes */ 

   void addWrittenFields( String className, String methodName, 
                                 String[] writtenFields );

   /**
    * Adds some accessed fields to a given method.
    *
    * <p>This configuration method must be used when a method reads
    * some field values and does not follow the naming conventions (it
    * is not a getter for instance).
    *
    * @param className the class the method belongs to
    * @param methodName the method
    * @param writtenFields some new fields the method reads */ 

   void addAccessedFields( String className, String methodName, 
                           String[] accessedFields );

   /**
    * Declare a field to be of a given type
    * 
    * @param className the class the field belongs to
    * @param fieldName the field
    * @param type the type
    */
   void setFieldType( String className, String fieldName, String type);

   /**
    * Declare a method's parameters to be of a given type
    * 
    * @param className the class the field belongs to
    * @param methodName the field
    * @param type the type
    */
   void setParametersType( String className, String methodName, String[] types) ;

   /**
    * Create a new virtual class. Virtual classes allow you extend and
    * refine the types used by the application, so that aspects can
    * behave differently. 
    */
   void newVirtualClass( String className );

   /**
    * Declare an association made of two roles.
    *
    * @see jac.aspects.associations.AssociationsAC
    */

   void declareAssociation( ClassItem classA, String fieldNameA, 
                            ClassItem classB, String fieldNameB );

   /**
    * Declare a repository to which all instances of type will be
    * automatically added.
    *
    * @param type the type of objects to add in the repository
    * @param repositoryName the name of the object holding the repository
    * @param repositoryCollection the collection to add the objects to
    */

      void defineRepository( ClassItem type, 
                             String repositoryName,
                             CollectionItem repositoryCollection ); 

   /**
    * This configuration method defines root objects.
    *
    * <p>Root objects are never garbaged since they are considered as
    * an entry point to retrieve the application objects that are
    * related to them.
    *
    * @param name the name of the root object */

   void addRoot(String name);

   /**
    * This configuration method tells that the fields (references or
    * collections) must be cloned when the class is cloned.
    *
    * @param className the class name
    * @param fields the names of the fields that are cloned */

   void setClonedFields(String className, String[] fields);

}

