/*

  Copyright (C) 2001 Laurent Martelli

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package jac.util;

import java.util.Hashtable;
import java.util.Map;
import java.io.PrintStream;
import java.io.FileOutputStream;

/**
 * The <code>Log</code> class provides a means to send some
 * informational, debugging or error messages like
 * <code>System.out.println</code> but traces can be enabled or
 * disabled at runtime.
 *
 * <p>JAC supports a -V launching option that allows the user to
 * enable a given trace category (for instance, by using <code>-V
 * jac</code>, the user can see what happens in the JAC core system.
 *
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> */

public final class Log {

   // enabled traces
   static Hashtable levels = new Hashtable();

   /** This flag tells if the Log traces into a file or on the
       screen. */
   public static boolean toFile = false;
   static PrintStream out;

   static String logHeader = "";

   /**
    * Sets the logging file.
    *
    * @param name the file name */

   public static void setFileName(String name, String header) {
      try {
         logHeader = header;
         out = new PrintStream( new FileOutputStream(name) );
         toFile = true;
      } catch( Exception e ) {
         e.printStackTrace();
      }
   }
   
   /**
    * Traces with a level lower than <code>levels[category]</code> are
    * discarded.
    *
    * @param category a string representing the category to trace
    * @param level the level of the trace (1 is always printed)
    * @param message the message to print out */

   public static void trace(String category, int level, String message) {
      if (category!=null && levels.containsKey(category)) {
         if (level<=((Integer)levels.get(category)).intValue()) {
            if( toFile ) {
               out.println(logHeader+category+": "+message);
            } else {
               System.out.println(category+": "+message);
            }
         }
      } else {
         levels.put(category,new Integer(0));
      }
   }

   /* this causes too much dependency troubles
   public static void trace(int level, String message) {
      String category = (String)JacObject.getCollaboration().getAttribute("cur_AC");
      Log.trace(category,level,message);
   }
   public static void trace(String message) {
      Log.trace(1,message);
   }
   */

   /**
    * Traces with a level equals to 1.
    *
    * @param category a string representing the category to trace
    * @param message the message to print out
    * @see #trace(String,int,String) */

   public static void trace(String category, String message) {
      Log.trace(category,1,message);
   }

   /**
    * Traces an error into the <code>System.err</code> stream.
    *
    * @param message the error message to print out */

   public static void error(String message) {
      if( toFile ) {
         out.println(logHeader+"ERROR: "+message);
      } else {
         System.err.println("ERROR: "+message);
      }
   }

   /**
    * Traces a warning into the <code>System.err</code> stream.
    *
    * @param message the warning message to print out */

   public static void warning(String message) {
      if( toFile ) {
         out.println(logHeader+"WARNING: "+message);
      } else {
         System.err.println("WARNING: "+message);
      //      System.out.println("WARNING: "+message);
      }
   }

   /**
    * Traces a warning into the <code>System.err</code> stream only if
    * the given category is enable to trace.
    *
    * @param message the warning message to print out */
 
   public static void warning(String category, String message) {
      warning(category+": "+message);
   }

   /**
    * Sets the verbose level of a given category.
    *
    * <p>The higher, the more traces are printed out.
    *
    * @param category the category to set the level of
    * @param level the category verbose level */
 
   public static void setLevel(String category, Integer level) {
      levels.put(category,level);
   }

   /**
    * Returns a Map category -> enabled saying which traces are enables
    */
   public static Map getLevels() {
      return levels;
   }
}
